# Author:  A M Chavan 
# Created: 2006-03-25, from orangLaSillaPriority.tcl
# 
# $Id: orangLaSillaPriorityMPG.tcl,v 1.2 2006/03/26 03:46:58 amchavan Exp $

## True/false: is this an MPG run?
## MPG run ID are like "077.A-9123(B)", in the 9000 range
##-------------------------------------------------------
proc isMPG {ob} {
	set id [getProgId $ob]
	set number [lindex [split $id "-"] 1]
	set mpg [string range $number 0 0]
	if { $mpg == "9" } {
		return 1
	} else {
		return 0
	}
}

## MPG run: yes=1, no=2 -- we privilege MPI runs
##-----------------------------------------------
proc getMpgScore {ob} {
	set isMpg [isMPG $ob]
	if {$isMpg} {
		set mpg "1"
		set desc "MPG run\n"
	} else {
		set mpg "2"
		set desc "Non-MPG run\n"
	}
    
    return [list $mpg $desc]
}

## run started: yes=1, no=2 
##-------------------------------------------
proc getRunStartedScore {isRunStarted} {
    if {$isRunStarted} {
        set started "1"
        set desc "Started run\n"
    } else {
        set started "2"
        set desc "Non-started run\n"
    }                           
    
    return [list $started $desc]
}

## time critical: yes=1, no=2  
##---------------------------------------------------
proc getTimeCriticalScore {isTimeCritical} {
    if {$isTimeCritical} {
        return [list "1" "Time critical\n"]
    } else {
        return [list "2" "Not time critical\n"]
    }                           
}

## Run property
##
## Taken from "Service Mode Execution Priorities" at
## http://www.pl.eso.org/paranal/sciops/team_only/procedures/fyi_sm_reminder.html
##
##    * 0. RRM               [RRM is not available, see Bugzilla item 5838]
##    * 1. ToO               [ToO is not available, see Bugzilla item 5839]
##    * 2. Pre-imaging
##    * 3. Carry over
##    * 4. Guaranteed time   [NOT RELEVANT, said Olivier]
##    * 5. Large programmes
##    * 6. Chilean programmes
##    * A. Normal A queue
##    * B. Normal B queue
##    * C. Normal C queues 
##--------------------------------------------------
proc getRunPropertyScore {isPreImaging isLarge isChilean isCarryOver rankClass} {
    
    if {$isPreImaging} {
        set runProperty "2"
        set desc "Pre-imaging"
    } elseif {$isCarryOver} {
        set runProperty "3"
        set desc "Carry-over"      
    } elseif {$isLarge} {
        set runProperty "5"
        set desc "Large program"       
    } elseif {$isChilean} {
        set runProperty "6"
        set desc "Chilean"             
    } else {
        set runProperty $rankClass
        set desc "Class $rankClass" 
    }
    
    append desc " run\n"
    return [list $runProperty $desc]
} 

## Seeing class: internal problem=0, 
##               close to constraint=1,
##               any other value=2
##--------------------------------------------------
proc getSeeingClass {curSeeing reqSeeing}  {

    set desc "Current seeing ($curSeeing) "
    set desc2 "close to OB constraint ($reqSeeing): class = "

    if { $reqSeeing >= $curSeeing && $reqSeeing < $curSeeing * 1.5} {

        set seeingClass "1"
        append desc $desc2
        append desc "$seeingClass\n"

    } elseif {$reqSeeing >= $curSeeing * 1.5} {

        set seeingClass "2"
        append desc "not "
        append desc $desc2
        append desc "$seeingClass\n"
        
    } else {
    
        set seeingClass "0"
        set desc "Inconsistent seeing value: $curSeeing\n"
    }
    
    return [list $seeingClass $desc]
}

## Sky transparency class: internal problem=0, 
##                         close to constraint=1,
##                         any other value=2
##--------------------------------------------------
proc getSkyTranspClass {curTransp reqTransp} {

    set desc "Current sky transparency ($curTransp) "
    set desc2 "close to OB constraint ($reqTransp): class = "
   
    set currentTranspValue [string range $curTransp 0 0]
    set transpValue [string range $reqTransp 0 0]
    set transpLimit [expr $currentTranspValue + 1]
    
    if { $transpValue >= $currentTranspValue && $transpValue <= $transpLimit} {

        set transpClass "1"
        append desc $desc2
        append desc "$transpClass\n"

    }  elseif {$transpValue > $transpLimit} {

        set transpClass "2" 
        append desc "not "
        append desc $desc2
        append desc "$transpClass\n"
        
    } else {
        set transpClass "0"
        set desc "Inconsistent sky transparency value: $curTransp\n"
    }
    return [list $transpClass $desc]
}

## FLI class: internal problem=0, 
##            close to constraint=1,
##            any other value=2
##--------------------------------------------------
proc getFliClass {curFLI reqFLI} {
    
    set desc "Current FLI ($curFLI) "
    set desc2 "close to OB constraint ($reqFLI): class = "
    
    set fliLimit [expr $curFLI + 0.5]
    if { $reqFLI >= $curFLI && $reqFLI <= $fliLimit } {

        set fliClass "1"
        append desc $desc2
        append desc "$fliClass\n"

    } elseif {$reqFLI > $fliLimit } {

        set fliClass "2" 
        append desc "not "
        append desc $desc2
        append desc "$fliClass\n"
        
    } else {
    
    	# Issue: sometimes the moon is too brigth but it's
    	# still below the horizon. This proc doesn't know it.
    	# See Bugzilla item 5844
    	# Patched by having this proc returning "1" here
        # (should be "0")
        set fliClass "1"
        set desc "Possibly inconsistent FLI value: $curFLI, class = $fliClass\n"
    }
    return [list $fliClass $desc]
}


## Rank the input OB
##--------------------------------------------------
proc rank {ob} {

    # Retrieve OB information and ranking parameters
    #--------------------------------------------------
    set userPriority [getUserPriority $ob]
    set airmass [getAirmass $ob]
    set dec [getDec $ob]
    set ra [getRa $ob]
    set diffDec [getDiffDec $ob]
    set diffRA [getDiffRA $ob]  
    set fli [format "%4.2f" [getFLI $ob]]
    set moonAngularDistance [getMoonAngularDistance $ob]
    set propDec [getPropDec $ob]
    set propRA [getPropRA $ob]
    set seeing [format "%3.1f" [getSeeing $ob]]
    set skyTransparency [getSkyTransparency $ob]        
    set strehlRatio [getStrehlRatio $ob]
    set executionTime [getExecutionTime $ob]
    set period [getPeriod $ob]
    set progId [getProgId $ob]
    set obsMode [getObsMode $ob]
    set telescope [getTelescope $ob]
    set rankClass [getRankClass $ob]
    set rank [getRank $ob]
    set timeCritical [getTimeCritical $ob]
    set chilean [getChilean $ob]
    set preImaging [getPreImaging $ob]
    set carryOver [getCarryOver $ob]
    set large [getLarge $ob]
    set secondHalfTI [getSecondHalfTI $ob]
    set runStarted [getRunStarted $ob]
    set lst [getLST]
    set ha [format "%06.2f" [getHourAngle $ob]]
    set currentSeeing [getCurrentSeeing]
    set currWindDirect [getCurrentWindDirection]
    set currentSky [getCurrentSkyTransparency]
    set currentFli [format "%4.2f" [getCurrentFLI]]
    
    # Compute some preliminary values and 
    # their justification
    #--------------------------------------------------
    set vjSeeing    [getSeeingClass $currentSeeing $seeing]
    set vjSkyTransp [getSkyTranspClass $currentSky $skyTransparency]
    set vjFLI       [getFliClass $currentFli $fli]
    set vjTimeCrit  [getTimeCriticalScore $timeCritical]
    set vjRunStarted [getRunStartedScore $runStarted]
    set vjRunProps [getRunPropertyScore $preImaging $large $chilean $carryOver $rankClass]
    set vjMPG      [getMpgScore $ob]

    # Here we accumulate the ranking score and the
    # justification for the ranking itself
    #--------------------------------------------------
    set score ""
    set description ""
    
    # These will be useful later
    #--------------------------------------------------
    set seeingClass    [lindex $vjSeeing 0]
    set skyTranspClass [lindex $vjSkyTransp 0]
    set fliClass       [lindex $vjFLI 0]
    set weatherTotalValue [expr $seeingClass + $skyTranspClass + $fliClass]
    
    # Most important criterium is whether this is an MPG run
    #-------------------------------------------------------
    append score       [lindex $vjMPG 0]
    append description [lindex $vjMPG 1]
    
    # Most important criterium is the rank class
    #--------------------------------------------------
    if { [string first $rankClass  "ABC"] < 0 }  {
		set rankClass "Z"
    }

    ## rank class
    append score $rankClass 
    append description "Rank class = $rankClass\n"

	## time critical or not
	append score [lindex $vjTimeCrit 0]
	append description [lindex $vjTimeCrit 1]
        
	## seeing, sky transparency and FLI  
	append score $weatherTotalValue 
	append description "Observing conditions index = $weatherTotalValue\n"
        
	## seeing class    
	append score $seeingClass
	append description [lindex $vjSeeing 1]
        
 	## sky transparency
	append score $skyTranspClass
	append description [lindex $vjSkyTransp 1]
        
	## fli
	append score $fliClass
	append description [lindex $vjFLI 1]    


    # For B class OBs consider whether the run was started
    #------------------------------------------------------
    if {$rankClass == "B" } {                           
        append score [lindex $vjRunStarted 0]
        append description [lindex $vjRunStarted 1]
    } else {
        append score "X"
        append description "Don't care whether run was started (rank class $rankClass)\n"
    }
    
    # Required seeing
    #----------------------------------------
    set reqs [format "%3.1f" $seeing]
    append score $reqs
    append description "Required seeing: $reqs\n"
    
    # For A OBs: pre-imaging, large, chilean, carryover...
    #-----------------------------------------------------
    if {$rankClass == "A"}  {
        append score       [lindex $vjRunProps 0]
        append description [lindex $vjRunProps 1]
    } else {
        append score "X"
    }

    # Hour angle
    # To simplify sorting we add 12, so that 
    # -12..+12 becomes 0..24. 
    # Smaller HA values will rank higher 
    #------------------------------------------
    append score [format "%05.2f" [expr $ha + 12.0]]
    append description "Hour angle = [format "%.2f" $ha]\n"
    
    # FLI and Internal priority
    #----------------------------------------
    append score $fli
    append description "FLI constraint = $fli\n"

    append score $userPriority
    append description "User priority = $userPriority\n"
    
    ## returns the result
    return [list $score $description]
}



## this have to be always called
init
rankall

# end of file orangRank.tcl
