package org.eso.ohs.phase2.apps.p2pp;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Hashtable;

import javax.swing.JOptionPane;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.FileIOUtils;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.persistence.AppConfig;

/**
 * One of a pair of methods to prevent concurrent access to local storage.
 * We use a strange convention (ie the absence of a file) to indicate
 * that a user subdir is locked.  This is because File.delete() is the
 * nearest things we have (in JDK 1.1) to an atomic test-and-set file
 * operation.  JDK 1.2 and beyond have createNewFile().
 *
 * <p>
 * This kind of locking is not atomic and is unsatisfactory in many ways
 * but it's better than nothing and we give the user the option to
 * override it.  If you can do better, please do!
 */
public  class LockMgr {

    private AppConfig appcfg = AppConfig.getAppConfig();
    private File cacheDirFile = new File(appcfg.getCacheDir());
    private String ppid;
    private static String lck_prefix;

    /** amchavan, 20-Dec-2001 */
    private static Logger stdlog_ = Logger.getLogger(LockMgr.class);

    public LockMgr() {
        ppid = System.getProperty("runtime.ppid");        
    }
    /*
            LockMgr(String l_ppid)
            {
                ppid = l_ppid;
            }
    */

    /**
     * A hashtable to keep track of the local storage areas we have
     * locked. Normally we only have one locked at a time but during
     * ReloginAction, for a short period, we have both old and
     * new cache dirs locked.
         *
         * This method does nothing upon every attempt to re-login, using the
         * "Login as" option from the main menu, with the same userID.
     */
    private Hashtable<Long, Long> locks_ = new Hashtable<Long, Long>();
    
    /** remember the userID of the current session */
	private int loc_userID;

    public void lockLocalStorage(int userId) throws ObjectIOException {
        lck_prefix = "lockinfo." + userId;

        if (userId == loc_userID) {
            stdlog_.debug("lockLocalStorage: same userID");
            return;
        } else {
            stdlog_.debug("lockLocalStorage: different userID");
            loc_userID = userId;
        }

        File unlockedFile = new File(cacheDirFile, "unlocked." + userId);
        // File infoFile = new File(cacheDirFile, lck_prefix + "." + ppid);
        /*
         * For users, who migrate from V2.1 to V2.2 there need to be a check,
         * for a missing unlocked file. Otherwise p2pp would assume, that
         * the user was logged in just because the unlocked-file is missing.
         */
        FilenameFilter lck_filter = new FilenameFilter() {
            public boolean accept(File dir, String name) {
                if (name.startsWith(lck_prefix))
                    return true;
                return false;
            }
        };
        /*
         * IF no unlockedFile exists but a lockinfoFile, then assume, that
         * somebody is already logged in with this userId
         * IF neither an unlockedFile nor a lockinfo file exists, that assume
         * that that the user is migrating from V2.1 to V2.2 or lost the
         * lockinfo File for some other reason.
         */
        if (!unlockedFile.delete()
            && cacheDirFile.list(lck_filter).length != 0) {
            String lockedBy = "";
            String[] lockFiles = cacheDirFile.list(lck_filter);
            if (lockFiles.length > 0) {
                lockedBy =
                    new File(cacheDirFile + File.separator + lockFiles[0])
                        .getAbsolutePath();
            }

            JOptionPane.showMessageDialog(
                MainWindow.getPrincipalWindow(),
                "The system is apparently already in use for this user ID"
                    + "\nConcurrent use for a user ID is not possible.\n"
                    + "\nIf you are sure this system and user ID are not in-use"
                    + "\nat this time, please remove the lock file:"
                    + "\n"
                    + lockedBy
                    + "\nand restart the application."
                    + "\n\nCannot continue!",
                "Warning message",
                JOptionPane.ERROR_MESSAGE);

            stdlog_.fatal("UserID " + userId + " is already in use - exiting");
            System.exit(0);
        }

        markAsLocked(userId);
        createInfoFile(userId);
    }

    private void markAsLocked(int userId) {
        Long idLong = new Long(userId);
        locks_.put(idLong, idLong);
    }

    private void markAsUnlocked(int userId) {
        Long idLong = new Long(userId);
        locks_.remove(idLong);
    }

    private boolean haveLocked(int userId) {
        Long idLong = new Long(userId);
        return locks_.containsKey(idLong);
    }

    private void createInfoFile(int userId) {
        File infoFile = new File(cacheDirFile, "lockinfo." + userId + "." + ppid);
        try {
            infoFile.delete();
            FileIOUtils.stringToFile(
                System.getProperty("user.name"),
                infoFile.getAbsolutePath());
        } catch (IOException e) {
            stdlog_.error("", e);
        }
    }

    public void unlockLocalStorage(int userId) {
        File unlockedFile = new File(cacheDirFile, "unlocked." + userId);
        File infoFile = new File(cacheDirFile, "lockinfo." + userId + "." + ppid);

        if (haveLocked(userId)) {
            infoFile.renameTo(unlockedFile);
            markAsUnlocked(userId);
        }
    }

    public void createLockForNewStorage(int userId) {
        createInfoFile(userId);
        markAsLocked(userId);
    }
}
