package org.eso.ohs.phase2.apps.p2pp;

import java.io.File;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentException;
import org.eso.ohs.instruments.InstrumentList;

/** 
 * This class instanciates a Singleton Reference to a List of
 * Instrument objects. The class extends InstrumentList, and 
 * adds the ability to download IPs from the appropriate
 * storage manager if the IP is not found locally.
 * 
 * This class uses p2pp.Persistence to download IPs, and is therefore
 * specific to the P2PP application.
 * 
 * @version $id$
 * @author Nick Kornweibel for ESO. 
 */

public class P2PPInstrumentList extends InstrumentList {

    private String ipCacheDir_ = null;
    
    private boolean ipMode_;
    
    private String instDir_;

    /**
     * Standard log channel 
     */
    private static Logger stdlog_ =
        Logger.getLogger(P2PPInstrumentList.class);

    protected P2PPInstrumentList(boolean ipMode,String instDir,String ipc) {
        super();
        ipCacheDir_ = ipc;
        ipMode_ = ipMode;
        instDir_ = instDir;
        if (ipCacheDir_ != null && ipCacheDir_.length() > 0) {
            File f = new File(ipCacheDir_);
            ipCacheDir_ = f.getAbsolutePath();
        } else {
            ipCacheDir_ = null;
        }
    }

    /**
     * @return the instrument requested, with the version requested.
     * If the instrument is not found in the local cache, persistence
     * will be queried for the IP, which will be downloaded and extracted
     * if located. If not found, or on error, return null.
     *
     * NOTE:
     * Unfortunately it is unclear what to do in the case of error...if an
     * instrument is required from the database, but fails to be downloaded,
     * then it is handled as if the IP simply did not exist and null is returned.
     */
    public Instrument getInstrument(String inst, float version) {
        Instrument rc = super.getInstrument(inst, version);
        if (rc == null) {
            stdlog_.debug(
                "Downloading IP "+inst+" v"+version);
            // query persistemce and download the IP if it exists.
            try {
                if (Persistence.getInstance().downloadInstrumentPackage(
                    inst, version)) {
                        reloadInstrumentList(ipMode_,instDir_,ipCacheDir_);
                        rc = InstrumentList.getInstance().getInstrument(inst, version);
                } else {
                    stdlog_.debug("IP not downloaded: "+inst+" v"+version);
                }
                
            } catch (InstrumentException e1) {
                // failed to reload IPs
                stdlog_.error("Failed to reload IPs", e1);
                rc = null;
            } catch (ObjectIOException e2) {
                // failed to download IP
                stdlog_.error("Failed to download IP", e2);
                rc = null;
            }
        }
        return rc;
    }
}