package org.eso.ohs.phase2.apps.p2pp.views;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;

import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;

import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.dfs.ObservationDescription;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentList;
import org.eso.ohs.instruments.Parameter;
import org.eso.ohs.instruments.TemplateSignature;
import org.eso.ohs.instruments.TemplateSignatureFactory;
import org.eso.ohs.persistence.Config;
import org.eso.ohs.persistence.Media;
import org.eso.ohs.phase2.apps.p2pp.Persistence;


/**
 * This is the data model that contains the OBView. 
 * The datamodel delegates to the ODGridModel for the majority
 * of the functionality except in the situation where there 
 * is an acquisition template
 * @see ODTemplateTable
 * @see ODGridModel
 * @see OBView
 * @see ODView   
 *
 * Add call to calculate execution time 
 * @author cmavros change made 07/Sept/2001
 */

@SuppressWarnings("serial")
public class AcquisitionModel
	extends AbstractTableModel
	implements TemplateModel, PropertyChangeListener {

	/**
	 * Model for linked od
	 */

	private ODGridModel model_;

	/**
	 * Current ObservationBlock selected
	 */

	private ObservationBlock ob_;

	/**
	 * The acqTemplate_ associated with the OD
	 */
	private TemplateSignature acqTemplate_;

	/**
	 * Listens for changes in the data model
	 */
	private AcquisitionModelListener listener_;

	/**
	 * Static reference to the currently selected Acquisition Data model
	 */
	private static AcquisitionModel acqModel_;

	/**
	 * This flag is set when data is changed
	 */
	private boolean dirty_ = false;

	private boolean editModeOps_ = false;

	/**
	 * Id for this model
	 */
	static private long id_ = 0;

	/**
	 * Number of columns in model
	 */

	private int columnCount_ = 0;

	/**
	 * Number of rows in model
	 */
	private int rowCount_ = 0;

	/** 
	 * Static version of the AcqModel
	 * @param ob ObservationBlock to be displayed
	 */
	
	public static AcquisitionModel getAcqModel(ObservationBlock ob) {
		return getAcqModel(ob, false);
	}
	
	/*
	 * @param ob the ob for which and ACQ model is required
	 * @param asNew true: force reload of the ACQ model even if the ob is tha
	 * same.
	 * @return AcquisitionModel
	 */
	public static AcquisitionModel getAcqModel(ObservationBlock ob, boolean asNew) {
		if (asNew || id_ != ob.getId()) {
			acqModel_ = new AcquisitionModel(ob);
			AcquisitionModel.id_ = ob.getId();
		}
		return acqModel_;
	}

	public void setEditModeOps(boolean edit) {
		editModeOps_ = edit;
	}

	/**
	 * @param ob Currently displated OB
	 */

	public AcquisitionModel(ObservationBlock ob) {
		ob_ = ob;
		acqTemplate_ = ob_.getTemplateSignature();
		ob_.addPropertyChangeListener(this);
		model_ = ODGridModel.getCachedODModel(this, ob_.getOd());
		if (listener_ == null)
			listener_ = new AcquisitionModelListener();
		model_.addTableModelListener(listener_);
		if (Config.getCfg().isInEngineeringMode() && acqTemplate_ != null) {
			try {
				// System.out.println("@@@@@@@@@@@@@ Refreshing Acquisition Template");
				TemplateSignatureFactory.refreshTemplate(acqTemplate_);
			} catch (IOException e) {
			}
		}
	}

	/**
	 * Currently selected instrument associated with 
	 * this ob
	 *
	 */
	public Instrument getInstrument() {
		ObservationDescription od = ob_.getOd();
		return InstrumentList.getInstance().getInstrument(
			od.getInstrument(),
			od.getIPVersion());

	}

	/** 
	 * Column is a label or not
	 * @param column column
	 */

	public boolean isLabel(int column) {
		boolean label = false;
		if (acqTemplate_ != null) {
			if (column == 0)
				label = true;
			if (column >= 2)
				label = model_.isLabel(column - 2);
		} else if (model_ != null) {
			label = model_.isLabel(column);
		}
		return label;
	}

	/**
	 * Observation Description
	 * @return Observation Description
	 */
	public ObservationDescription getOD() {
		return ob_.getOd();
	}

	/**
	 * Save changes to the undelying business objects
	 *
	 * <br>ChangeLog: cmavros 21-11-01 Add execution calculation time 
	 */
	public void saveChanges() {
		if (model_ != null && model_.isDirty()) {
			model_.saveChanges();
		}
		if (dirty_)
			ob_.setTemplateSignature(acqTemplate_);
		dirty_ = false;
	}

	public ObservationBlock getOB() {
		return ob_;
	}

	/**
	 * Whether or not a cell is editable in the grid
	 * @param row row
	 * @param column column
	 */

	public boolean isCellEditable(int row, int column) {
		boolean edit = true;
		if (!Config.getCfg().isInOperationsMode()) {
			//return true;
			if (ob_ != null && ob_.isCheckedIn())
				return false;
			if (ob_ != null) {
				Persistence objMgr = Persistence.getInstance();

				if (!objMgr.isInRegistry(ob_.getId())
					|| !Media.LOCAL.equals(objMgr.getMediaFor(ob_))) {
					return false;
				}
			}
		}
		if (acqTemplate_ != null) {
			Parameter[] plist = acqTemplate_.getDisplayParamList();
			if (column == 0)
				edit = false;
			else if (column == 1 && row >= plist.length) {
				edit = false;
			} else if (column >= 2) {
				edit = model_.isCellEditable(row, column - 2);
			}
		} else {
			if (model_ != null)
				edit = model_.isCellEditable(row, column);
		}
		return edit;
	}

	/**
	 * Get Acquisition Template Signature
	 */

	public TemplateSignature getAcquisition() {
		return ob_.getTemplateSignature();
	}

	/**
	 * If a new od is connected to the ob
	 * a propertyChange event occurs
	 * @param ev PropertyChangeEvent    
	 */

	public void propertyChange(PropertyChangeEvent ev) {
		if (acqTemplate_ != ob_.getTemplateSignature()) {
			acqTemplate_ = ob_.getTemplateSignature();
			fireTableStructureChanged();
		}
	}

	public void insertColumn(int column, TemplateSignature ts) {
		if (acqTemplate_ != null) {
			if (!ts.getTemplateType().equalsIgnoreCase("acquisition")) {
				if (column <= 1)
					model_.insertColumn(0, ts);
				else
					model_.insertColumn(column - 2, ts);
			} else
				ob_.setTemplateSignature(ts);
		} else {
			if (!ts.getTemplateType().equalsIgnoreCase("acquisition"))
				model_.insertColumn(column, ts);
			else
				ob_.setTemplateSignature(ts);
		}

	}

	/**
	 * Remove a column from the associated od model
	 * @param sourceColumn source column
	 */

	public void removeColumn(int column) {
		// System.out.println("CLASS AcquisitionModel - removeColumn");
		if (acqTemplate_ == null)
			model_.removeColumn(column);
		else {
			if (column <= 1) {
				ob_.setTemplateSignature(null);
				fireTableStructureChanged();
			} else
				model_.removeColumn(column - 2);
		}
	}

	/**
	 * Get the value at a particular row or column
	 *
	 * @param row row
	 * @param column column
	 * @return value at row,column  
	 */
	public Object getValueAt(int row, int col) {
		if (acqTemplate_ != null) {
			Parameter pList[] = acqTemplate_.getDisplayParamList();
			if (col == 0) {
				if (row < pList.length)
					return pList[row].getLabel();
			} else if (col == 1) {
				if (row < pList.length) {
					return pList[row].getDisplayValue();
				}
			} else {
				return model_.getValueAt(row, col - 2);
			}
		} else {
			if (model_ != null)
				return model_.getValueAt(row, col);
		}
		return null;
	}

	/**
	 * Set the value at a particular row or column
	 *
	 * @param row row
	 * @param column column
	 */

	public void setValueAt(Object value, int row, int column) {
		if (acqTemplate_ != null) {
			if (column == 1) {
				Parameter[] plist = acqTemplate_.getDisplayParamList();
				if (plist != null) {
					if (row < plist.length) {
						if (!plist[row].displayValueDifferent())
							plist[row].setValue((String) value);
						dirty_ = true;
						saveChanges();
						fireTableChanged(
							new TableModelEvent(this, row, row, column));
					}
				}
			} else if (column > 2) {
				model_.setValueAt(value, row, column - 2);
			}
		} else {
			if (model_ != null) {
				model_.setValueAt(value, row, column);
			}
		}
	}

	/**
	 * Return the name of the column in the Table
	 * @param column column
	 * @return name of column
	 */

	public String getColumnName(int column) {
		if (acqTemplate_ != null) {
			if (column == 0)
				return acqTemplate_.getTemplateName();
			else if (column == 1)
				return "1";
			else if (column >= 2)
				return model_.getColumnName(column - 2);
		} else {
			if (model_ != null)
				return model_.getColumnName(column);
		}
		return null;
	}

	/**
	 * Set the columncount for the datamodel
	 * @param columnCount column count
	 */

	public void setColumnCount(int columnCount) {
		columnCount_ = columnCount;
	}

	/**
	 * Calculate the column count for the data model
	 */

	public void calcColumnCount() {
		if (acqTemplate_ != null && model_ != null)
			columnCount_ = model_.getColumnCount() + 2;
		else if (model_ != null)
			columnCount_ = model_.getColumnCount();
		else if (acqTemplate_ != null)
			columnCount_ = 2;
		else
			columnCount_ = 0;
	}

	/**
	 * Column count
	 * @return column count
	 */

	public int getColumnCount() {
		calcColumnCount();
		return columnCount_;
	}

	/**
	 * Return a class for the column in the grid
	 * @return column class
	 */

	public Class getColumnClass() {
		return Object.class;
	}

	/**
	 * Get the row count
	 * @return row count
	 */

	public int getRowCount() {
		if (acqTemplate_ != null) {
			int acqCount = acqTemplate_.getDisplayParamList().length;
			if (model_ != null) {
				if (acqCount > model_.getRowCount())
					rowCount_ = acqCount;
				else
					rowCount_ = model_.getRowCount();
			} else
				rowCount_ = acqCount;
		} else if (model_ != null) {
			rowCount_ = model_.getRowCount();
		}
		return rowCount_;
	}

	/**
	 * Currently selected column
	  * @return current column
	 */

	public int getCurrColumn() {
		if (acqTemplate_ != null
			&& model_ != null
			&& model_.getColumnCount() > 0) {
			return model_.getCurrColumn() + 2;
		} else if (acqTemplate_ != null) {
			return 1;
		} else if (model_ != null) {
			return model_.getCurrColumn();
		}
		return -1;
	}

	/**
	 * Return the array of labels or a Template 
	 * Signature for a givewn column
	 * return array of strings or refernce to Template Signature
	 */

	public Object getDataItem(int column) {
		if (acqTemplate_ != null) {
			if (column == 1)
				return acqTemplate_;
			else if (column == 0) {
				Parameter[] plist = acqTemplate_.getDisplayParamList();
				String[] labels = new String[plist.length];
				return labels;
			} else {
				return model_.getDataItem(column - 2);
			}
		} else {
			if (model_ != null)
				return model_.getDataItem(column);
		}
		return null;
	}

	/**
	 * Get the template Signatures associated with 
	 * the linked od from its data model
	 * @return array of Template Signatures
	 */

	public TemplateSignature[] getTemplateSignatures() {
		if (model_ != null)
			return model_.getTemplateSignatures();
		return null;
	}

	/**
	 * Move column in grid and change the associated data model
	 * by firing a fireTableStructureChanged event
	 * @param sourceColumn source column
	 * @param destColumn destination column
	 */

	public void moveColumn(int sourceColumn, int destColumn) {
		if (acqTemplate_ != null) {
			if (sourceColumn <= 1 || destColumn <= 1) {
				fireTableStructureChanged();
				return;
			} else {
				model_.moveColumn(sourceColumn - 2, destColumn - 2);
			}
		} else {
			model_.moveColumn(sourceColumn, destColumn);
		}
	}
    
    
   

	/**
	 * This class listens for changed in data in the table model
	 * and fires the appropriate events
	 * 
	 * Calculate execution time ob_.getExecTime()
	 * then convert value to hr:mm:ss
	 * if a change was made in table.
	 * <br>ChangeLog: cmavros change made 07/Sept/2001
	 *
	 */

	private class AcquisitionModelListener implements TableModelListener {
		public void tableChanged(TableModelEvent e) {
			if (e.getFirstRow() == TableModelEvent.HEADER_ROW) {
				fireTableStructureChanged();
				return;
			}
			if (e.getType() == TableModelEvent.UPDATE) {
				
				fireTableDataChanged();
				return;
			}
		}
	}
    /**
     * Returns the dirty_.
     * @return boolean
     */
    public boolean isDirty() {
        return dirty_;
    }

    /**
     * Sets the dirty_.
     * @param dirty_ The dirty_ to set
     */
    public void setDirty(boolean dirty) {
        dirty_ = dirty;
    }

}
