/* $Id: ODGridModel.java,v 1.9 2006/04/27 13:25:00 pnunes Exp $ */

package org.eso.ohs.phase2.apps.p2pp.views;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.swing.JOptionPane;
import javax.swing.event.TableModelEvent;
import javax.swing.table.AbstractTableModel;

import org.eso.ohs.dfs.ObservationDescription;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentList;
import org.eso.ohs.instruments.Parameter;
import org.eso.ohs.instruments.TemplateSignature;
import org.eso.ohs.instruments.TemplateSignatureFactory;
import org.eso.ohs.persistence.Config;
import org.eso.ohs.persistence.Media;
import org.eso.ohs.phase2.apps.p2pp.Persistence;

/**
* This class extends the AbstractTableModel. The data model is represented by
* a Vector class which contains TemplateSignature objects and an array of strings
* which represnt the Template Signature labels. Each column in the data model 
* is an element of the Vector. The JTable ODGrid which uses this model uses the
* add,remove and insert column methods to implement dragging and droping of columns. 
 */

public class ODGridModel
    extends AbstractTableModel
    implements TemplateModel, PropertyChangeListener {

    /**
     * Dirty flag set when data changes
     */
    private boolean dirty_ = false;

    /**
     * Grid is read only or not
     */
    private boolean readOnly_ = false;

    /**
     *  Observation Desription represented in datamodel
     */
    private ObservationDescription od_;

    /**
     * Vector that holds the data for the grid
     */
    private Vector<Object> data_ = new Vector<Object>();

    /**
     * Vector that holds data that represents which cells 
     * are invalid
     */
    private Vector errors_ = new Vector();

    /**
     * Number of columns in the grid. This is updated as templates 
     * are added and removed.
     */
    private int numColumns = 0;

    /**
     * The number of rows in the grid. This will correspond to 
     * the number of parameters per template signature
     */
    private int numRows = 0;

    /**
     * Id which identifier this model
     */
    private long id_;

    /**
     * The currently selectedcolumn
     */
    private int currColumn = 0;

    /**
     * Hashtable of ODGridModel instances currently in use.
     * This is needed so that if an object is displayed in two
     * or more JTables, the same model is used in both and thus
     * everything remains in sync.
     */
    private static Hashtable<Object, ODGridModel> cachedModels_ = new Hashtable<Object, ODGridModel>();

    /**
     * Current model for the od
     * @param od Observation Description
     * @param asNew force reloading of the model even if cached.
     * <br>ChangeLog: T Schlichting, 14-Feb-01: fixed 1246
     */
    public static ODGridModel getCachedODModel(
        Object key,
        ObservationDescription od) {
        return getCachedODModel(key, od, false);
    }

    public static ODGridModel getCachedODModel(
        Object key,
        ObservationDescription od,
        boolean asNew) {

        Enumeration<Object> enum0 = cachedModels_.keys();
        ODGridModel model = null;

        // Do we already have a cached model for this OD?
        if (!asNew) {
            while (enum0.hasMoreElements()) {
                ODGridModel m =
                    cachedModels_.get(enum0.nextElement());

                if (m.getId() == od.getId()) {
                    model = m;
                    break;
                }
            }
        }

        // No cached model -- create a new one.
        if (model == null) {
            model = new ODGridModel(od);
            model.setId(od.getId());
        }

        // If model is empty (eg new) fill it
        // We always reset contents if we're in Engineering Mode.
        // T Schlichting, 14-Feb-01: fixed 1246
        model.setData(od.getSignatures());

        // Store a reference to this model in the cache.
        // We must do this regardless of whether it's new or old!
        cachedModels_.put(key, model);

        return model;
    }

    /**
     * Set the od for the model
     * @param od_ Observation Description
     */
    public void setOD(ObservationDescription od) {
        od_ = od;
    }

    /**
     * Observation Description
     * @return Observation Description
     */
    public ObservationDescription getOD() {
        return od_;
    }

    /**
      * <br>ChangeLog: T Schlichting, 14-Feb-01: fixed 1246
      */
    public void propertyChange(PropertyChangeEvent ev) {
        boolean tableStructureChanged = false;
        TemplateSignature[] newTsfs = od_.getSignatures();
        TemplateSignature[] oldTsfs = getTemplateSignatures();
        /*
         * IF    an initial situation
         * OR    new cells, rows or columns in the table
         * THEN  set flag that setData get's called
         */
        if (oldTsfs == null || newTsfs.length != oldTsfs.length)
            tableStructureChanged = true;
        else {
            /*
             * walk through the set of templates in this grid (not through every
             * cell) and see, whether a change has happened there
             */
            for (int i = 0; i < newTsfs.length; i++) {
                if (oldTsfs[i] != newTsfs[i])
                    tableStructureChanged = true;
            }
        }
        if (tableStructureChanged) {
            /*
             * setData fires a TableStructureChangeEvent anyway, so we don't
             * need to do it here.
             */
            setData(newTsfs);
        }
    }

    /**
     * Construct an ODGridModel. Only called from
     * within class
     */
    private ODGridModel(ObservationDescription od) {
        od_ = od;
        od_.addPropertyChangeListener(this);
    }

    /**
     * Set id to uniquely identify this model
     * @param id Id
     */
    public void setId(long id) {
        id_ = id;
    }

    /**
     * Get id
     * @return id
     */
    public long getId() {
        return id_;
    }

    /**
     * Check to see if this a given column is a label
     * @param column column
     * @return boolean value 
     */
    public boolean isLabel(int column) {
        //System.out.println("Data size " + data_.size());
        if (data_.elementAt(column) instanceof String[])
            return true;
        return false;
    }

    /**
     * Returns the column name associated with a particular column index.
     *
     * @param columnIndex index to identify the column name.
     * @return Name of column     
     */
    public String getColumnName(int columnIndex) {
        if (data_.elementAt(columnIndex) instanceof String[]) {
            TemplateSignature ts =
                (TemplateSignature) data_.elementAt(columnIndex + 1);
            return ts.getTemplateName();
        } else {
            int j = 0;
            while (data_.elementAt(columnIndex) instanceof TemplateSignature) {
                j++;
                columnIndex--;
            }
            return (Integer.toString(j));

        }
    }

    /**
     * Returns the currently selected column
     * @return currently selected column
     */
    public int getCurrColumn() {
        // ensure that the cuurColumn is in the bounds of the available
        // columns - otherwise, move it into the bounds
        if (currColumn >= getColumnCount()) {
            currColumn = getColumnCount() - 1;
        }
        if (currColumn < 0) {
            currColumn = 0;
        }
        return currColumn;
    }

    /**
     * Returns the instrument name associate with the od
     * @return instrument name
     */
    public Instrument getInstrument() {
        return InstrumentList.getInstance().getInstrument(
            od_.getInstrument(),
            od_.getIPVersion());
    }

    /**
     * Complete the interface of the TemplateModel interface
     * @return Template Signature
     */
    public TemplateSignature getAcquisition() {
        return null;
    }

    /**  
     * This method sets the data_ vector so the table can be re-initialized.
     * after the data is saved.
     *
     * @param templateSignature array of templateSignatures
     */
    void setData(TemplateSignature[] templateSignature) {
        // System.out.println("@@@@@@@@@@@@ Start Setting data");
        Vector<String> labels = new Vector<String>();
        data_.removeAllElements();
        numColumns = 0;
        TemplateSignature ts;
        if (Config.getCfg().isInEngineeringMode()) {
            for (int i = 0; i < templateSignature.length; i++) {
                try {
                    TemplateSignatureFactory.refreshTemplate(
                        templateSignature[i]);

                } catch (IOException e) {
                }
            }

        }
        for (int i = 0; i < templateSignature.length; i++) {
            Parameter[] pList = templateSignature[i].getDisplayParamList();
            if (pList.length > numRows)
                numRows = pList.length;
            if (i - 1 >= 0
                && templateSignature[i - 1] instanceof TemplateSignature
                && templateSignature[i].getTemplateName().equals(
                    templateSignature[i - 1].getTemplateName())) {
                data_.addElement(templateSignature[i]);
                numColumns += 1;
            } else {
                labels.removeAllElements();
                for (int j = 0; j < pList.length; j++) {
                    Parameter p = pList[j];
                    if (p.getLabel() != null)
                        labels.addElement( p.getLabel());
                    else
                        labels.addElement( p.getParameterName());
                }
                String labelarray[] = new String[labels.size()];
                labels.copyInto(labelarray);
                data_.addElement(labelarray);
                data_.addElement(templateSignature[i]);
                numColumns += 2;
            }
        }
        // System.out.println("@@@@@@@@@ End Set Dsta");
        fireTableStructureChanged();
        // System.out.println("Affter firse Structure changed");
    }

    /**
     * Column count for the table
     * @return Number of columns
     */
    public int getColumnCount() {
        return numColumns;
    }

    /**
     * Set the column count for the table
     * @param colCount column count 
     */
    public void setColumnCount(int colCount) {
        numColumns = colCount;
    }

    /**
     * Returns the number of rows in the grid
     * @return number of rows
     */
    public int getRowCount() {
        return numRows;
    }

    /**
     * @return int the data for a particular columns. 
     * This will either be a template signature or a parameter list,
     * or null if the column is out of range. 
     */
    public Object getDataItem(int column) {
        Object rc = null;
        if (column >= 0 && column < data_.size()) {
            rc = data_.elementAt(column);
        }
        return rc;
    }

    /**
     * Returns the data for a particular columns. This will either 
     * be a template signature or a parameter list.
     */
    public void setDataItem(int column, TemplateSignature ts) {
        data_.removeElementAt(column);
        data_.insertElementAt(ts, column);
    }

    /**
     * Return the template Signatures from the data Vector.
     * @return list of templateSignatues in the grid
     */
    public TemplateSignature[] getTemplateSignatures() {
        if (numColumns > 1) {
            Vector<TemplateSignature> tsfVector = new Vector<TemplateSignature>();
            for (int i = 0; i < data_.size(); i++) {
                if (data_.elementAt(i) instanceof TemplateSignature) {
                    tsfVector.addElement(
                        (TemplateSignature) data_.elementAt(i));
                }
            }
            TemplateSignature[] tsfs = new TemplateSignature[tsfVector.size()];
            tsfVector.copyInto(tsfs);
            return tsfs;
        }
        return null;
    }

    /**
     * Set the read only flag
     */
    public void setReadOnly(boolean readOnly) {
        readOnly_ = readOnly;
    }

    /**
     * Read only flag
     * @return boolean value
     */
    public boolean getReadOnly() {
        return readOnly_;
    }

    /**
     * Set the row count for the data model
     * @param rowCount Row count
     */
    public void setRowCount(int rowCount) {
        numRows = rowCount;
    }

    /**
     * Return a class for the column in the grid
     * @return column class
     */
    public Class<Object> getColumnClass() {
        return Object.class;
    }

    /**
     * Set whether a cell is editable or not. All cells 
     * in a label are read only and any values in a parameter
     * that have the fixed flag set
     *
     * @param row row in datamodel 
     * @param column column in datamodel
     */
    public boolean isCellEditable(int row, int column) {
        boolean edit = false;
        if (!Config.getCfg().isInOperationsMode()) {
            if (od_ != null && od_.isCheckedIn())
                return false;
            if (od_ != null) {
                Persistence objMgr = Persistence.getInstance();
                if (!objMgr.isInRegistry(od_.getId())) {
                    return false;
                }

                Media media = objMgr.getMediaFor(od_.getId());

                if (!Media.LOCAL.equals(media))
                    return false;
            }
        }
        if (!readOnly_
            && data_.elementAt(column) instanceof TemplateSignature) {
            TemplateSignature ts = (TemplateSignature) data_.elementAt(column);
            Parameter[] plist = ts.getDisplayParamList();
            if (plist != null && row < plist.length) {
                if (!plist[row].isFixed())
                    edit = true;
            }
        }
        return edit;
    }

    /**
     * Whether data has changed or not
     * @return boolean value
     */
    public boolean isDirty() {
        return dirty_;
    }

    /**
     * Save changed to the underlying business object
     */
    public void saveChanges() {
        if (dirty_) {
            TemplateSignature[] tsfs = getTemplateSignatures();
            if (tsfs != null)
                od_.setSignatures(tsfs);
        }
        dirty_ = false;
    }

    /**
     * Set the value for an element in the grid. This will update an indivdal parameter with 
     * the value entered
     * 
     * @param value Value to enter into parameter
     * @param rowIndex row index
     * @param columnIndex column index
     */
    public void setValueAt(Object value, int rowIndex, int columnIndex) {
        if (columnIndex >= data_.size()) {
            JOptionPane.showMessageDialog(
                null,
                "SETVALUEAT HAS MESSED UP",
                "SETVALUEAT ERROR",
                JOptionPane.WARNING_MESSAGE);
            return;
        }

        if (data_.elementAt(columnIndex) instanceof TemplateSignature) {
            TemplateSignature ts =
                (TemplateSignature) data_.elementAt(columnIndex);
            Parameter[] plist = ts.getDisplayParamList();
            if (plist != null) {
                Parameter p = plist[rowIndex];
                if (!p.displayValueDifferent())
                    p.setValue((String) value);
                dirty_ = true;
                saveChanges();
                fireTableChanged(
                    new TableModelEvent(this, rowIndex, rowIndex, columnIndex));
            }
        }

    }

    /**
     * Returns the data for a particular cell.
     * @param aRow row index 
     * @param aColumn column index
     */
    public Object getValueAt(int aRow, int aColumn) {
        if (data_.size() > 0 && data_.size() > aColumn) {
            if (data_.elementAt(aColumn) instanceof TemplateSignature) {
                TemplateSignature ts =
                    (TemplateSignature) data_.elementAt(aColumn);
                Parameter[] pList = ts.getDisplayParamList();
                if (aRow < pList.length) {
                    return pList[aRow].getDisplayValue();
                }
            } else {
                String[] labels = (String[]) data_.elementAt(aColumn);
                if (aRow < labels.length)
                    return labels[aRow];
            }
        }
        return null;
    }

    /**
     * Insert a column into the datamodel at a given position.
     * @param column column
     * @param ts TemplateSignature
     */
    public void insertColumn(int column, TemplateSignature ts) {
        insertColumnInternal(column, ts);
        fireTableStructureChanged();
    }

    /**
     * Insert a column into the datamodel at a particualar offset
     * @param column index into datamodel where ts should be inserted   
     * @param ts Template Signature
     */
    public void insertColumnInternal(int column, TemplateSignature ts) {
        currColumn = column;
        boolean insert = false;
        boolean islabel = false;
        Vector pLabels = new Vector();
        TemplateSignature ts1 = null;
        String[] labels = ts.getLabels();
        if (column == 0 && data_.size() == 0) {
            data_.addElement(labels);
            data_.addElement(ts);
            numColumns += 2;
            currColumn += 1;
            insert = true;
        } else if (data_.elementAt(column) instanceof TemplateSignature) {
            ts1 = (TemplateSignature) data_.elementAt(column);
            if (ts1.getTemplateName().equals(ts.getTemplateName())) {
                data_.insertElementAt(ts, column + 1);
                numColumns++;
                currColumn++;
                insert = true;
            } else if (
                column + 1 < getColumnCount()
                    && data_.elementAt(column + 1) instanceof TemplateSignature) {
                for (int i = column; i >= 0; i--) {
                    if (data_.elementAt(i) instanceof String[]) {
                        String[] plabels = (String[]) data_.elementAt(i);
                        data_.insertElementAt(plabels, column + 1);
                        numColumns += 1;
                        break;
                    }
                }
            } else if (
                column + 2 < getColumnCount()
                    && data_.elementAt(column + 2) instanceof TemplateSignature) {
                TemplateSignature ts2 =
                    (TemplateSignature) data_.elementAt(column + 2);
                if (ts.getTemplateName().equals(ts2.getTemplateName())) {
                    data_.insertElementAt(ts, column + 2);
                    numColumns++;
                    currColumn += 2;
                    insert = true;
                }
            }

        } else {
            islabel = true;
            ts1 = (TemplateSignature) data_.elementAt(column + 1);
            if (ts1.getTemplateName().equals(ts.getTemplateName())) {
                numColumns++;
                currColumn++;
                data_.insertElementAt(ts, column + 1);
                insert = true;
            } else if (
                column - 1 > 0
                    && data_.elementAt(column - 1) instanceof TemplateSignature) {
                ts1 = (TemplateSignature) data_.elementAt(column - 1);
                if (ts1.getTemplateName().equals(ts.getTemplateName())) {
                    data_.insertElementAt(ts, column - 1);
                    numColumns++;
                    insert = true;
                }
            }
        }
        if (!insert) {
            if (!islabel) {
                data_.insertElementAt(ts, column + 1);
                data_.insertElementAt(labels, column + 1);
                currColumn += 2;
            } else {
                data_.insertElementAt(ts, column);
                data_.insertElementAt(labels, column);
                if (column == 0)
                    currColumn = 1;
                else
                    currColumn++;
            }
            insert = true;
            numColumns += 2;
        }
        if (labels.length > getRowCount())
            setRowCount(labels.length);
        if (currColumn <= 0)
            currColumn = 1;
    }

    /**
     * Remove a column and fire a tableStructureChangedEvent
     * @param column column to be removed
     * 
     */
    public void removeColumn(int column) {
        removeColumnInternal(column);
        fireTableStructureChanged();
    }

    /**
     * Delete a column into the datamodel at a particualar index.
     * @param column remove column at this index
     *
     */
    public void removeColumnInternal(int column) {
        currColumn = column;
        boolean remove = false;
        if (column < data_.size() && column >= 0) {
            if (data_.elementAt(column) instanceof TemplateSignature) {
                if (column + 1 == numColumns
                    && data_.elementAt(column - 1) instanceof String[]) {
                    data_.removeElementAt(column);
                    data_.removeElementAt(column - 1);
                    numColumns -= 2;
                    currColumn -= 2;
                    remove = true;
                } else if (
                    column + 1 < numColumns
                        && data_.elementAt(column + 1) instanceof String[]
                        && data_.elementAt(column - 1) instanceof String[]) {
                    if (column - 2 >= 0 && column + 2 < numColumns) {
                        TemplateSignature ts2 =
                            (TemplateSignature) data_.elementAt(column + 2);
                        TemplateSignature ts1 =
                            (TemplateSignature) data_.elementAt(column - 2);
                        if (ts2
                            .getTemplateName()
                            .equals(ts1.getTemplateName())) {
                            data_.removeElementAt(column + 1);
                            numColumns--;
                        }
                    }
                    data_.removeElementAt(column);
                    data_.removeElementAt(column - 1);
                    numColumns -= 2;
                    currColumn -= 2;
                    remove = true;
                } else {
                    if (column - 1 >= 0
                        && data_.elementAt(column - 1)
                            instanceof TemplateSignature)
                        currColumn--;
                    data_.removeElementAt(column);
                    numColumns--;
                    remove = true;
                }
            } else {
                if (column + 2 < numColumns
                    && data_.elementAt(column + 1) instanceof TemplateSignature
                    && data_.elementAt(column + 2) instanceof TemplateSignature) {
                    TemplateSignature ts1 =
                        (TemplateSignature) data_.elementAt(column + 1);
                    TemplateSignature ts2 =
                        (TemplateSignature) data_.elementAt(column + 2);
                    if (ts1.getTemplateName().equals(ts2.getTemplateName())) {
                        data_.removeElementAt(column + 1);
                        currColumn++;
                        numColumns--;
                        remove = true;
                    }
                } else {
                    data_.removeElementAt(column + 1);
                    data_.removeElementAt(column);
                    numColumns -= 2;
                    currColumn--;
                    remove = true;
                }
            }
        }
        if (currColumn <= 0 && getColumnCount() > 0)
            currColumn = 1;
    }

    /**
     * Move a column into the datamodel at a particualar index.
     * This is a composite of doing a remove and an insert.
     * @param sourceColumn  Source Column 
     * @param destColumn Destination column
     */
    public void moveColumn(int sourceColumn, int destColumn) {
        boolean insert = false;
        int destinationColumn = destColumn;
        TemplateSignature ts = null;
        TemplateSignature ts1 = null;
        if (data_.elementAt(sourceColumn) instanceof TemplateSignature)
            ts = (TemplateSignature) data_.elementAt(sourceColumn);
        else
            return;
        if (data_.elementAt(destinationColumn) instanceof TemplateSignature) {
            ts1 = (TemplateSignature) data_.elementAt(destinationColumn);
        }
        int colcount = getColumnCount();
        removeColumnInternal(sourceColumn);
        if (sourceColumn < destinationColumn) {
            destinationColumn =
                destinationColumn - (colcount - getColumnCount());
        }
        if (data_.elementAt(destinationColumn) instanceof TemplateSignature) {
            if (ts1.getTemplateName().equals(ts.getTemplateName())) {
                if (sourceColumn > destColumn)
                    destinationColumn--;
            }
        }
        insertColumnInternal(destinationColumn, ts);
        fireTableStructureChanged();
    }

}
