// -*-c++-*-
#ifndef _ArcProdRep_h_
#define _ArcProdRep_h_

/*
 * E.S.O. - VLT project/ESO Archive
 *
 * $Id: ArcProdRep.h,v 1.3 2005/10/17 13:21:09 nferguso Exp $
 *
 * ArcProdRep.h - class definitions for class ArcProdRep, a class 
 * representing an application that read from the
 * observations database
 *
 * who                  when         what
 * -------------------  -----------  ----------------------------------------
 * Elisabetta Angeloni  25 May 1998  Created
 * Neil Ferguson        26 Nov 2004  Converted to DfsStdLib/SybConn
 * Neil Ferguson        20 Sep 2005  Removed obsolete methods
 */

#include "TimeStamp.h" 

#include <ArcDbGen.h>
#include <ArcProduct.h>
#include <ArcAstoFile.h>

/**
 * This class and its methods are meant to be used for reading from
 * observations database.
 * An instance consists of references to the database and the tables
 * where the data products information resides.
 * This class includes a set of member functions for retrieving
 * the data products catalog information.
 */

class ArcProdRep : public ArcDbGen {

private:
    
    DfsCString data_productsTab_;
    
    // Search a given product in the asto tables (mdfiles or cdfiles)
    int selectAstoFile(ArcAstoFile& astoFile, const DfsCString& fileId);
    int selectCDFile(ArcAstoFile& astoFile, const DfsCString& fileId);
    int selectMDFile(ArcAstoFile& astoFile, const DfsCString& fileId);
    int isFileAstoArchived(const DfsCString& fileId);
    int isFileNgasArchived(const DfsCString& fileId);

public:

    
/**@name Constructors
 */
//@{
    /// Default constructor: it supplies a connection to the observations repository.
    // The default values are read from a configuration file
    ArcProdRep(DfsCString dbAlias="DPREP");
    
    /// Constructor that requires username, password, server and database name (a default is supplied).
    // In case an empty server name is given, the value of $DSQUERY variable is taken
    ArcProdRep(DfsCString dbName, DfsCString dbPasswd, 
               DfsCString dbServer="", DfsCString dbName="observations");

    /// Constructor that requires a valid SybaseConnection object
    ArcProdRep(SybaseConnection& database);

//@}

/**@name  Methods for data retrieval
 * These methods retrieve product's information from the 
 * asto and observations databases
 */
//@{

/**
****************************************************************************
* Generic method for retrieving from the database a list of data products 
* given a generic where clause. 
* It execute the following SELECT statement:
SELECT dp_id, mjd_obs 
FROM data_products

* If no where clause is specified, all the products from the table
* data_products are retrieved.
* In order to build the where clause, the methods that return the column's names
* should be used.

@memo
Generic method for retrieving from the database a list of products 
given a generic where clause. 
@return
0 if OK, 1 if an error occurred
The number of retrieved items is given by dataProductList.entries()
NOTE: it user's responsability to free the list using dataProductList.clear()
@param
dataProductList (OUTPUT) list containing the retrieved products,
whereClause - (INPUT) generic where clause
@see
ArcProdList user-defined type, RWDBCriterion class
*/

    int getProdCat(ArcProdList& dataProductList,
		   const DfsCString& whereClause = "0 = 0");

/**
****************************************************************************
Update the column data_products.rel_date to a given value
for all the rows matching the where condition.
This method is used by PI dppacker for updating the rel_date
when the CD is completed
@memo
Update the column data_products.rel_date to a given value
for all the rows matching the where condition
@return
0 if OK, 1 if an error occurred
@param
relDate (INPUT) value to be put in rel_date,
whereClause - (INPUT) generic where clause
*/
    int updateReleaseDate(TimeStamp relDate, const DfsCString& whereClause);

/**
****************************************************************************
This method retrieves the corresponding location (url) of a list of products.
The output is put in a list sorted by media type (magnetic disk or CD-ROM)
and by diskname.
Other information returned are the file size in bytes and the type of 
storage media (see type enum MediaType).

@memo
This method retrieves the corresponding location (url) of a list of products
@return
0 if OK, 1 if an error occurred
The number of retrieved items is given by urlList.entries()
NOTE: it user's responsability to free the list using urlList.clearAndDestroy()
@param
urlList (OUTPUT) list containing the product urls,
prodList - (INPUT) array of pointers containg the product IDs
@see
ArcSortFileList user-defined type
*/

    int getProdUrl(ArcSortFileList& urlList, 
                   char** prodList);

/**
****************************************************************************
This method reads from the table data_products summary information for a 
given products.
The searching criteria is
- product Id (= data_products.dp_id)
This service is used by dppacker command for preparing DP-SUMMARY file.
@memo
This method reads from the table data_products summary information for a 
given products
@return
0 if OK, 1 if an error occurred , -1 if no data found
@param
dpId  (INPUT) product Id as searching criteria 
obsName    - (OUTPUT) OB name (= data_products.obs_name)
category   - (OUTPUT) product category (= data_products.dp_cat)
object     - (OUTPUT) object name (= data_products.object)
ra         - (OUTPUT) RA (= data_products.ra)
dec        - (OUTPUT) DEC (= data_products.dec)
exptime    - (OUTPUT) Exposure time (= data_products.exptime)
insId      - (OUTPUT) Instrument Id (= data_products.ins_id)
insMode    - (OUTPUT) Instrument Mode (= data_products.ins_mode)
filterPath - (OUTPUT) FILTER's list (= data_products.filter_path)
grisPath   - (OUTPUT) GRIS's list (= data_products.gris_path)
gratPath   - (OUTPUT) GRAT's list (= data_products.grat_path)
slitPath   - (OUTPUT) SLIT's list (= data_products.slit_path)
*/
    
    int getProdInfo(const char* dpId, DfsCString& obsName, DfsCString& category, 
                    DfsCString& object, int& ra, int& dec, double& exptime,
                    DfsCString& origFile, DfsCString& progId,
                    DfsCString& insId, DfsCString& insMode, DfsCString& filterPath, 
                    DfsCString& grisPath, DfsCString& grat_path, DfsCString& slitPath) const;
    
/**
****************************************************************************
This method extracts from the table mdfiles the list of files that could be 
deleted according to:
- time period (see data_products.mjd_obs)
- max size 
- data cache server (see mdfiles.diskname)
The files that can be deleted must be stored on CDs put online
This service is used by dpmkspace command.
@memo
This method extracts from the table mdfiles the list of files that could be 
deleted
@return
0 if OK, 1 if an error occurred
The number of retrieved items is given by fileList.entries()
NOTE: it user's responsability to free the list using urlList.clearAndDestroy()
@param
fileList (OUTPUT) list containing the product urls,
totalSize - (OUTPUT) actual size in MB of disk space that can be freed not greater than maxSize
diskName  - (INPUT) hostname where the file to be deleted are stored
maxSize   - (INPUT) max size in MB of disk space to be freed
startMjd  - (INPUT) initial time period to be considered (see data_products.mjd_obs)
endMjd    - (INPUT) final time period to be considered (see data_products.mjd_obs)
@see
ArcSortFileList user-defined type
*/

    int getDeletableFiles(ArcSortFileList& fileList, int& totalSize,
                          const DfsCString& diskName, const int& maxSize,
                          const double& startMjd, const double& endMjd);

/**
****************************************************************************
This method deletes an entry from mdfiles table using file_id as condition.
@memo
This method deletes an entry from mdfiles table using file_id as condition.
@return
0 if OK, 1 if an error occurred
NOTE: it user's responsability to rollback or commit the transaction
@param
fileId  (INPUT) id of the file entry to be deleted (see mdfiles.fileId)
session - (INPUT) database session
*/

    int deleteAstoFile(const DfsCString& fileId, SybaseConnection& session); 

//@}

/**@name Names of Columns
 * These methods get the column's name of the table data_products
 * of observations database.
 * These methods must be used for building a generic where clause
 */
//@{

    /// Return the named column
    DfsCString getCol(const char* name) const { return (data_productsTab_ + "." + name); } 
  
    /// Data product ID (column "data_products.dp_id"). Require an integer type
    DfsCString dpId() const { return (data_productsTab_ + ".dp_id"); }

    /// value of keyword MJD-OBS (column "data_products.mjd_obs"). Require a double type
    DfsCString MJD_OBS() const { return (data_productsTab_ + ".mjd_obs"); }

    /// Exposure Start date (column "data_products.exp_start"). Require a RWDBDateTime class
    DfsCString expStartDate() const { return (data_productsTab_ + ".exp_start"); }

    /// Release date (column "data_products.rel_date"). Require a RWDBDateTime class
    DfsCString releaseDate() const { return (data_productsTab_ + ".rel_date"); }

    /// value of keyword OBS ID (column "data_products.ob_id"). Require an integer type
    DfsCString OBS_ID() const { return (data_productsTab_ + ".ob_id"); }

    /// value of keyword TELESCOPE (column "data_products.tel_id"). Require a string type
    DfsCString TELESCOPE() const { return (data_productsTab_ + ".tel_id"); }

    /// value of keyword INSTRUME (column "data_products.ins_id"). Require a string type
    DfsCString INSTRUME() const { return (data_productsTab_ + ".ins_id"); }

    /// value of keyword OBS PROG ID (column "data_products.prog_id"). Require a string type
    DfsCString OBS_PROG_ID() const { return (data_productsTab_ + ".prog_id"); }

    /// value of keyword DPR CATG (column "data_products.dp_cat"). Require a string type
    DfsCString DPR_CATG() const { return (data_productsTab_ + ".dp_cat"); }

    /// value of keyword DPR TYPE (column "data_products.dp_type"). Require a string type
    DfsCString DPR_TYPE() const { return (data_productsTab_ + ".dp_type"); }

    /// value of keyword DPR TECH (column "data_products.dp_tech"). Require a string type
    DfsCString DPR_TECH() const { return (data_productsTab_ + ".dp_tech"); }

    /// value of keyword EXPTIME (column "data_products.exptime"). Require a double type
    DfsCString EXPTIME() const { return (data_productsTab_ + ".exptime"); }

    /// value of keyword OBS TPLNO (column "data_products.tpl_seqno"). Require an integer type
    DfsCString OBS_TPLNO() const { return (data_productsTab_ + ".tpl_seqno"); }

    /// value of keyword TPL EXPNO (column "data_products.tpl_expno"). Require an integer type
    DfsCString TPL_EXPNO() const { return (data_productsTab_ + ".tpl_expno"); }

    /// value of keyword DET ID (column "data_products.det_id"). Require a string type
    DfsCString DET_ID() const { return (data_productsTab_ + ".det_id"); }

    /// value of keyword DET EXP ID (column "data_products.det_expid"). Require an integer type
    DfsCString DET_EXP_ID() const { return (data_productsTab_ + ".det_expid"); }

    /// value of keyword RA (column "data_products.ra"). Require an integer type (milliarcsec)
    DfsCString RA() const { return (data_productsTab_ + ".ra"); }

    /// value of keyword DEC (column "data_products.dec"). Require an integer type (milliarcsec)
    DfsCString DEC() const { return (data_productsTab_ + ".dec"); }

    /// value of keyword OBJECT (column "data_products.object"). Require a string type
    DfsCString OBJECT() const { return (data_productsTab_ + ".object"); }

    /// value of keyword INS MODE (column "data_products.ins_mode"). Require a string type
    DfsCString INS_MODE() const { return (data_productsTab_ + ".ins_mode"); }

    /// value of keyword OBS NAME (column "data_products.obs_name"). Require a string type
    DfsCString OBS_NAME() const { return (data_productsTab_ + ".obs_name"); }

    /// value of keyword NAXIS1 (column "data_products.naxis1"). Require an integer type
    DfsCString NAXIS1() const { return (data_productsTab_ + ".naxis1"); }

    /// value of keyword NAXIS2 (column "data_products.naxis2"). Require an integer type
    DfsCString NAXIS2() const { return (data_productsTab_ + ".naxis2"); }

    /// value of keyword TEL AIRM START (column "data_products.tel_airm_start"). Require a double type
    DfsCString TEL_AIRM_START() const { return (data_productsTab_ + ".tel_airm_start"); }

    /// value of keyword ORIGFILE (column "data_products.origfile"). Require a string type
    DfsCString ORIGFILE()const  { return (data_productsTab_ + ".origfile"); }

    /// value of keyword DET CHIP1 ID (column "data_products.det_chip1id"). Require a string type
    DfsCString DET_CHIP1_ID() const { return (data_productsTab_ + ".det_chip1id"); }

    /// value of keywords INS*NAME (column "data_products.light_path"). Require a string type
    DfsCString LIGHT_PATH() const { return (data_productsTab_ + ".light_path"); }

    /// value of keywords INS*FILT*NAME (column "data_products.filter_path"). Require a string type
    DfsCString FILTER_PATH() const { return (data_productsTab_ + ".filter_path"); }

    /// value of keywords INS*GRIS*NAME (column "data_products.gris_path"). Require a string type
    DfsCString GRIS_PATH() const { return (data_productsTab_ + ".gris_path"); }

    /// value of keywords INS*GRAT*NAME (column "data_products.grat_path"). Require a string type
    DfsCString GRAT_PATH() const { return (data_productsTab_ + ".grat_path"); }

    /// value of keywords INS*SLIT*NAME (column "data_products.slit_path"). Require a string type
    DfsCString SLIT_PATH() const { return (data_productsTab_ + ".slit_path"); }

//@}
           
}; // end of definition of class ArcProdRep

#endif /* _ArcProdRep_h_ */
