// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observator
// LICENSE:   ESO internal use only
//
// PROJECT:   VLT Data Flow Syste
// AUTHOR:    Christophe Moins- ESO/SDD/DF
// SUBSYSTEM: arcsys
//
// $Revision: 1.7 $
// $Log: CalibProg.h,v $
// Revision 1.7  2009/06/15 12:30:10  vforchi
// Removed daemon mode, added reduced data handling
//
// Revision 1.6  2008/09/11 08:31:07  vforchi
// fixed compilation problems with gcc 4.2.4
//
// Revision 1.5  2007/07/10 08:53:42  cmoins
// removal of the list of previously used instrument
// when fetching a list of calibration files
//
// Revision 1.4  2007/05/24 11:37:12  cmoins
// previous retrieved files are taken into account when looking for package files
//
// Revision 1.3  2007/05/04 09:56:46  cmoins
// removal of storeData, updateProgDB and getMaxTime implementation
//
// Revision 1.2  2007/04/26 12:55:29  cmoins
// adding of a StorageService template parameter
//
// Revision 1.1  2007/04/25 15:54:05  cmoins
// implementation of a calibration program
//
// 
//------------------------------------------------------------------------------

#ifndef CALIBPROG_H
#define CALIBPROG_H

#include <string>
#include <DfsLog.h>
#include <boost/tokenizer.hpp>
#include "ObsProg.h"
#include "RawDataProduct.h"
#include "DPPackerDef.h"

using namespace std;

/// Calibration program template class.
/**
 * A calibration program contains only calibration files. It will use an
 * ObservationService and a StorageService template parameter to remain 
 * independent of the data retrieval layer and of the storage service layer
 */
template <class ObservationService,class StorageService>
class CalibProg : public ObsProg<StorageService>
{
public :

#if defined UNIT_TEST
  friend class TestCalibProg;
  friend class TestObsProgPackage;
#endif 
 
  /// \brief Constructor
  /**
   * \param progId program identifier
   * \param connAlias data repository connection parameter
   */
  CalibProg(string progId,
            string connAlias):
    ObsProg<StorageService>(progId,connAlias)
  {}

  /// \brief Destructor
  ~CalibProg()
  {}

  /// \brief fetches a list of files associated to the program
  /**
   * \param startTime start time of the files retrieval
   * \param startTimeAdjustment correction to apply on start time for 
   *                            calibration files retrieval
   * \param endTime end time of the files retrieval
   * \param endTimeAdjustment   correction to apply on end time for 
   *                            calibration files retrieval
   * \return 0 if successful, -1 otherwise 
   */
  virtual int fetchData(double  startTime,
			double  startTimeAdjustment,
			double  endTime,
			double  endTimeAdjustment,
			const MediaFile::ResultSet&  rpreviousFileSet,
			vector<string> reducedDataLocations);
  
  /// \brief instantiates a copy of the object
  /// returns the address of the copy
  virtual ObsProg<StorageService>* clone() const ;

};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Fetches a list of files associated to the program
//-----------------------------------------------------------------------------
template <class ObservationService,class StorageService>
int
CalibProg<ObservationService,StorageService>::
fetchData(double startTime,
	  double  startTimeAdjustment,
	  double endTime,
	  double  endTimeAdjustment,
	  const MediaFile::ResultSet&  rpreviousFileSet,
          vector<string> reducedDataLocations)
{
  typedef boost::tokenizer<boost::char_separator<char> > tokenizer;

  boost::char_separator<char> sep(",");
  int retStatus = 0;
  string connAlias = ObsProg<StorageService>::getAlias();
  ObservationService observationService(connAlias);  

  DFS_TRACE("CalibProg<ObservationService>::fetchData");

  RawDataProduct::ResultSet dataFileSet = ObsProg<StorageService>::getDataFileSet();
  string progId = ObsProg<StorageService>::getProgId();
  if (observationService.findCalibFiles(progId,
					startTime,
					endTime,
					dataFileSet) > 0)
    {
      DFS_INFO("Files retrieval succeeded for %s",progId.c_str());
      ObsProg<StorageService>::setDataFileSet(dataFileSet);
    }
  else
    {
      DFS_ERROR("Files retrieval failed for %s",progId.c_str());
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// instantiates a copy of the object
//------------------------------------------------------------------------------

template <class ObservationService,class StorageService>
ObsProg<StorageService>*
CalibProg<ObservationService,StorageService>::
clone()  const
{
  return new CalibProg<ObservationService,StorageService>(*this);
}
#endif /* CALIBPROG_H */
