//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//COPYRIGHT (c) 1996 European Southern Observatory
//LICENSE
//
//PROJECT:   VLT On-line Data Flow
//AUTHOR:    <Michael Wiedmer - ESO/DMD/DPG>
//SUBSYSTEM: <Core>
//
//$Revision: 1.1 $
//$Log: CorCheckValid.h,v $
//Revision 1.1  2004/09/06 14:35:21  cguirao
//*** empty log message ***
//
//Revision 1.1  2004/06/23 12:47:09  cguirao
//First check in
//
//Revision 1.2  1998/06/30 08:49:48  dataflow
//Removing $NAME
//
//Revision 1.1  1997/04/10 08:05:44  dfadmin
//Initial revision
//
//Revision 1.3  1996/09/09 13:01:42  mwiedmer
//standardise
//
//Revision 1.2  1996/09/03 14:41:27  mwiedmer
//added RCS Name tag
//
//Revision 1.1  1996/09/02 11:04:01  dfadmin
//First release
//
//Revision 1.3  1996/09/02 09:45:35  mwiedmer
//first release
//
//Revision 1.2  1996/09/02 08:25:43  mwiedmer
//*** empty log message ***
//
//-----------------------------------------------------------------------------

#ifndef COR_CHECKVALID_H
#define COR_CHECKVALID_H 

// includes from this library
#include <cor_assert.h>
// includes from other ESO libraries
// includes from other libraries
// system includes

const unsigned short corValidCheck = 12345;
const unsigned short corDeadCheck = 0xDEAD;

/// class for checking instance validity
class CorCheckValid {
/**
  {\footnotesize $Id: CorCheckValid.h,v 1.1 2004/09/06 14:35:21 cguirao Exp $ } \\

  {\large SYNOPSIS} \\
  \begin{verbatim}
    #include <CorCheckValid.h> 

    CorCheckValid myClass;
  \end{verbatim}

  {\large DESCRIPTION} \\
  This class serves three roles:

  o It asserts if member functions are called through a null pointer.

  o It asserts if an instance is destructed twice.

  o It asserts if member functions are called on a rubbish pointer

  \begin{verbatim}

    MyClass* c = 0;
    c->somefunc();     << Bang

  \end{verbatim}
  
  Note that classes check themselves. It is not up to the users of
  the class to do anything, or even to know about the checking.

  Sadly the overhead is permanent, it cannot be removed in release code
  as the danger of linking development and non-development libraries
  is too great. (i.e the classes would be different sizes)
  
  Luckily there are just two simple macros to handle all of the work.

  There is absolutely no code inserted into member functions not compiled
  with -DEBUG (except the constructos and destructors, where a tiny
  amount of code is inserted).

  {\large CAUTIONS} \\
  There is a two byte overhead in each instance that uses the
  checker so it is only sensible for larger classes such as containers.

  {\large EXAMPLES} \\
  In your classes definition add the line

  COR_CHECKVALID_CLASS;

  Typically this would go in the private section right at the bottom,
  it is of no interest to the end user.
  (If class is a base class then put it at the end of the
   protected section so derived class can check themselves.)

  \begin{verbatim}
  #include <CorCheckValid.h>

  class MyClass {
    public:
     Myclass();
     int doit();
    protected:
     int mydata;

     COR_CHECKVALID_CLASS;
   };
  \end{verbatim}
 
  In each memeber function add the macro call COR_CHECKVALID;
  right at the top of the function

  \begin{verbatim}
  int MyClass::doit() {
    COR_CHECKVALID;
  ..
  }
  \end{verbatim}
*/
  
public:

  /// c'tor
  CorCheckValid();

  /// d'tor
  ~CorCheckValid();

  /// checks if this instance is still OK
  void checkOk(
    const void* aThisPtr,
    const char* aFile,
    int aLine
  ) const;

protected:
  unsigned short mChecknum;  
};

// define the macros, ensure that no checks (what so ever) take place
// in non-development versions.

#ifdef  COR_USE_ASSERT
#define COR_CHECK_VALID_OK checkOk(this,__FILE__,__LINE__)
#define COR_CHECKVALID corValidChecker.COR_CHECK_VALID_OK
#else
#define COR_CHECKVALID
#define COR_CHECK_VALID_OK
#endif

// inline functions
#include <CorCheckValid.icc>

#define COR_CHECKVALID_CLASS \
  CorCheckValid corValidChecker

#endif

// Set mode for emacs.
// Local Variables:
// mode:C++
// End:

