//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//COPYRIGHT (c) 1996 European Southern Observatory
//LICENSE
//
//PROJECT:   VLT On-line Data Flow
//AUTHOR:    Michael Wiedmer - ESO/DMD/DPG
//SUBSYSTEM: System Core
//
//$Revision: 1.2 $
//$Log: CorDebug.h,v $
//Revision 1.2  2004/10/26 15:23:53  nferguso
//Temporary check-in for repository move.
//
//Revision 1.1  2004/09/06 14:35:21  cguirao
//*** empty log message ***
//
//Revision 1.1  2004/06/23 12:47:09  cguirao
//First check in
//
//Revision 1.2  1998/06/30 08:49:49  dataflow
//Removing $NAME
//
//Revision 1.1  1997/12/03 15:05:47  dfadmin
//Initial revision
//
//Revision 1.1  1997/11/27 18:13:43  dataflow
//*** empty log message ***
//
//-----------------------------------------------------------------------------

#ifndef COR_DEBUG_H
#define COR_DEBUG_H

// includes from this library
#include <CorBool.h>
#include <cor_debug.h>
// includes from other ESO libraries
#include "DfsCString.h"
// includes from other libraries
// system includes

// typedef, extern, enums
// variables: consts, statics, exported variables (declared extern elsewhere)
// local forward function declarations
class CorUnitTest;

#ifdef COR_USE_DEBUG

//@@Man: COR_L1(), COR_L2(), COR_L2(), COR_DEBUG()
//@@Type: MACROS
//@@See:  COR_ASSERT, COR_TRACE
//@@Memo: debugging functionality
/**      This macro is provided to faciliate the use of debugging functions
         in a standard and uniform way.
         In particular, it can be used to execute runtime debugging
         code.
*/

// - COR_L1(FUNC)    executes the function FUNC if the variable is 1

// - COR_L2(FUNC)    executes the function FUNC if the variable is 1 or 2

// - COR_L3(FUNC)    executes the function FUNC if the variable is 1, 2 or 3

//@@{
/// Debugging Function - run code at debugging level 1
#define COR_L1(CODE)\
{\
  const char* env = getenv(corDEBUG_ENV);\
  if (env &&\
      (atoi(env) < 2)) {\
    CODE;\
  }\
}

/// Debugging Function - level 2
#define COR_L2(CODE)\
{\
  const char* env = getenv(corDEBUG_ENV);\
  if (env &&\
      (atoi(env) < 3)) {\
    CODE;\
  }\
}

/// Debugging Function - level 3
#define COR_L3(CODE)\
{\
  const char* env = getenv(corDEBUG_ENV);\
  if (env &&\
      (atoi(env) < 4)) {\
    CODE;\
  }\
}
/**
  {\footnotesize $Id: CorDebug.h,v 1.2 2004/10/26 15:23:53 nferguso Exp $}

  {\large SYNOPSIS} \\
  \begin{verbatim}
    #include <cor_debug.h> 

    void foo();

    COR_L1(foo());
    COR_L2(foo());
    COR_L3(foo());
  \end{verbatim}

  {\large DESCRIPTION} \\
  This macro serves as an interface to using debugging code in the application.
  It accepts a function call as its only parameter and will execute that
  function if the user desires so (through setting an environment variable or
  some such).\\
  COR_L1(foo()); executes foo() if the environment variable is set to 1\\
  COR_L2(foo()); executes foo() if the environment variable is set to 1 or 2\\
  COR_L3(foo()); executes foo() if the environment variable is set to 1, 2 or
  3\\
  The environment variable is defined in "cor_debug.h".

  {\large EXAMPLE}\\
  \begin{verbatim}
    #include <cor_debug.h> 

    class A {
    public:
      A(){}
      ~A(){}
      void foo() {
        COR_L1(check_sanity());
      }
    protected:
      void check_sanity(){
        // check the validity of this instance
      }
    };

    int main() {
      COR_L2(cout << "This is some text to help me debug.")
      
      A a;
      a.foo();   //  this will also call check_sanity() in the debug version
                 //  if the environment variable defined in cor_debug.h is
                 //  set to "1"

      return 0;
    }
      
    
  \end{verbatim}

  {\large SEE ALSO}\\
  \Ref{cor_debug.h}
 
  */
//@@}


#else

#define COR_L1(CODE)
#define COR_L2(CODE)
#define COR_L3(CODE)

#endif  // COR_USE_DEBUG defined

//=============================================================================
// Debug mechanism
#ifdef COR_USE_DEBUG

//@@Man: COR_DEBUG()
//@@Type: MACROS
//@@See:  COR_ASSERT
//@@Memo: debugging functionality
/**      This macro is provided to faciliate the use of debugging functions
         in a standard and uniform way.
*/

//@@{
/// Debugging Function
#define COR_DEBUG(CODE)\
if (getenv(corDEBUG_ENV)) {\
  CODE;\
}
/**
  {\footnotesize $Id: CorDebug.h,v 1.2 2004/10/26 15:23:53 nferguso Exp $ } \\

  {\large SYNOPSIS} \\
  \begin{verbatim}
    #include <cor_assert.h.h> 

    void foo();
    
    COR_DEBUG(foo());
  \end{verbatim}

  {\large DESCRIPTION} \\
  This macro serves as an interface to using debugging code in the application.
  It accepts a function call as its only parameter and will execute that
  function if the user desires so (through setting an environment variable or
  some such).

  {\large EXAMPLE}
  \begin{verbatim}
    #include <cor_assert.h.h> 

    class A {
    public:
      A(){}
      ~A(){}
      void foo() {
        COR_DEBUG(check_sanity());
      }
    protected:
      void check_sanity(){
        // check the validity of this instance
      }
    };

    int main() {
      COR_DEBUG(cout << "This is some text to help me debug.")
      
      A a;
      a.foo();   //  this will also call check_sanity() in the debug version

      return 0;
    }
      
    
  \end{verbatim}

  

  */
//@@}


#else

#define COR_DEBUG(CODE) 

#endif  // COR_USE_DEBUG defined


class CorDebug {
public:

  ///
  CorDebug();
  /** {\large Purpose} \\
  The default c'tor.

  {\large Preconditions} \\
  \begin{itemize}
  \item Itemize this section if necessary. Remove itemization otherwise.
  \end{itemize}

  {\large Exceptions}\\
  \begin{itemize}
  \item Itemize this section if necessary. Remove itemization otherwise.
  \end{itemize}
  */


  ///
  ~CorDebug();
  /** {\large Purpose} \\
  The d'tor.
  
  {\large Preconditions} \\
  \begin{itemize}
  \item Itemize this section if necessary. Remove itemization otherwise.
  \end{itemize}
  
  {\large Exceptions}\\
  \begin{itemize}
  \item Itemize this section if necessary. Remove itemization otherwise.
  \end{itemize}
  */


  ///
  int getExecutionLevel() const;


  ///
  static int setExecutionLevel();


protected:

  ///
  CorDebug(
    /** {\large Purpose} \\
    The copy c'tor creates a new instance from a reference.

    {\large Preconditions} \\
    \begin{itemize}
    \item Itemize this section if necessary. Remove itemization otherwise.
    \end{itemize}

    {\large Exceptions}\\
    \begin{itemize}
    \item Itemize this section if necessary. Remove itemization otherwise.
    \end{itemize}

    {\large Arguments}\\
    */
    ///
    const CorDebug& aRef
    /** the reference instance */
  );


  ///
  CorDebug& operator=(
    /** {\large Purpose} \\
    The assignment operator makes this instance look like another one.

    {\large Preconditions} \\
    \begin{itemize}
    \item Itemize this section if necessary. Remove itemization otherwise.
    \end{itemize}

    {\large Returns}\\
    A reference to this instance.

    {\large Exceptions}\\
    \begin{itemize}
    \item Itemize this section if necessary. Remove itemization otherwise.
    \end{itemize}

    {\large Arguments}\\
    */
    ///
    const CorDebug& aRef
    /** the reference instance */
  );


private:
  static int mExecuteLevel;  // the execution level as defined
                             // in the environment variable
};

#endif // COR_DEBUG_H

