// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2003, 2004 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System - RogueWave->dfsStdLib port
// AUTHOR:    Neil Ferguson - ESO/DMD/DFS
// SUBSYSTEM: dfsStdLib RWCSubString replacement
//
// $Revision: 1.9 $
// $Log: DfsCSubString.h,v $
// Revision 1.9  2008/07/24 16:26:23  nferguso
// Added ISO 14882 compliant forward declarations, as required by GCC since
// version 4.1.0.
//
// Revision 1.8  2004/09/01 13:32:06  nferguso
// Avoid missing return warnings on HP.
//
// Revision 1.7  2004/08/17 15:54:22  nferguso
// Documentation updates.
//
// Revision 1.6  2004/08/10 15:56:38  nferguso
// Removed redundant #include directives.
//
// Revision 1.5  2004/07/29 14:17:39  nferguso
// DfsCSubstring now has a public (default) copy constructor, in order to
// comply with ISO 14882. Comments were changed appropriately.
//
// The copy-assignment operator now casts its parameter to void. This stops
// the compiler from giving unwanted warnings about unused parameters.
//
// Revision 1.4  2004/03/26 15:07:39  nferguso
// Copyright years updated.
//
// Revision 1.3  2004/03/11 13:16:21  nferguso
// Removed superfluous semicolons.
//
// Revision 1.2  2004/02/20 10:10:05  nferguso
// Added more comments.
//
// Revision 1.1.1.1  2003/11/28 14:21:16  nferguso
// First Check-In
//
// 
//------------------------------------------------------------------------------

#ifndef DFS_CSUBSTRING_H
#define DFS_CSUBSTRING_H

// Includes from this library. 
#include <DfsCString.h>

// C includes.
#include <cassert>

// Forward declaration. 
class DfsCString; 

/// The DfsCSubString class is a partial replacement of the RWCSubString class. 
/** Only methods required by existing DFS code have been reimplemented, as new 
 * DFS code should be implemented using STL capabilities. 
 *
 * NOTE: As with RogueWave, code with effects similar to the following: 
 * @code 
 * DfsCString parent("String contents.");
 * parent(0, 3) = parent(0, 3) = "Test str";
 * @endcode 
 * should be avoided, in order to prevent unexpected multi-assign side effects.
 *
 * Both implementations are designed on the basis that:
 *
 * - substrings may only exist as temporary objects. 
 * - substrings may be invalidated by changes to their parent strings. 
 * - a parent string must remain in existence until all of its substrings have
 *   been destroyed. 
 * 
 * As a result, no mechanism has been provided for updating a substring when 
 * its parent string changes. 
 *
 * ISO 14882 requires the DfsCSubString class to have a copy constructor before
 * instances of it can be returned from a DfsCString or DfsCTokenizer method. 
 * Application code <I>must not</I> use this capability to make further copies 
 * of a DfsCSubString object. 
 *
 * As DfsCSubString methods should correspond exactly to the original RogueWave
 * API, this documentation is sparse. Consult the RogueWave documentation for 
 * more details. 
 */ 
class DfsCSubString
{
    /// Only DfsCString and DfsCTokenizer may instantiate DfsCSubStrings. 
    friend class DfsCString;

    /// Only DfsCString and DfsCTokenizer may instantiate DfsCSubStrings. 
    friend class DfsCTokenizer;

public:
    // The default constructor for this class is private. The user may not 
    // create DfsCSubString objects directly. 

    // The default copy constructor suffices for this class. It is required 
    // for returning DfsCSubString values from DfsCString and DfsCTokenizer 
    // methods, but should not be used otherwise. 

    // The default destructor suffices for this class. 

    /// @brief Replaces this DfsCSubString with the @a sourceString, by 
    /// modifying the substring's parent string. 
    /** If the sourceString is a different length from the original substring,
     * the parent string length will change as necessary. 
     */
    DfsCSubString& operator=(const char* sourceString); 

    /// Returns true if this is the null substring, false otherwise. 
    DfsBoolean isNull() const;

private:
    bool null;                  ///< True if this is a null substring. 
    DfsCString* parentString;   ///< The DfsCString containing this substring. 
    size_t startIndex;          ///< The substring's starting index. 
    size_t subStringLength;     ///< The length of the substring. 

    /// The user may not create DfsCSubString objects directly. 
    DfsCSubString();

    /// The user may not create DfsCSubString objects directly. 
    DfsCSubString(const DfsCString& parent, size_t index, size_t length);

    /// The user may not copy-assign DfsCSubString objects. 
    /** RogueWave interpret this assignment operator to mean "Overwrite the 
     * characters of this substring with those of the @a sourceSubString. If 
     * the sourceSubString is a different length from the original substring, 
     * the parent string length will change as necessary." DfsCSubStrings do 
     * not provide that operation. 
     */
    DfsCSubString& operator=(const DfsCSubString& sourceSubString) 
    {
        // Avoid unused parameter warnings. 
        static_cast<void>(sourceSubString); 

        // This method should never be used. 
        assert(0); 

        // Avoid missing return warnings. 
        return(*this); 
    }

    /// Returns an STL string iterator set to the start of the substring. 
    /** This method must not be called on a null substring. */
    std::string::iterator begin() const;

    /// Returns an STL string iterator set to the end of the substring. 
    /** This method must not be called on a null substring. */ 
    std::string::iterator end() const;
};

#endif  // DFS_CSUBSTRING_H
