// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2003, 2004 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System - RogueWave->dfsStdLib port
// AUTHOR:    Neil Ferguson - ESO/DMD/DFS
// SUBSYSTEM: dfsStdLib RWCTokenizer replacement
//
// $Revision: 1.6 $
// $Log: DfsCTokenizer.h,v $
// Revision 1.6  2004/09/01 13:32:06  nferguso
// Avoid missing return warnings on HP.
//
// Revision 1.5  2004/08/17 15:54:22  nferguso
// Documentation updates.
//
// Revision 1.4  2004/08/10 15:56:38  nferguso
// Removed redundant #include directives.
//
// Revision 1.3  2004/03/26 15:07:39  nferguso
// Copyright years updated.
//
// Revision 1.2  2004/02/20 10:10:05  nferguso
// Added more comments.
//
// Revision 1.1.1.1  2003/11/28 14:21:16  nferguso
// First Check-In
//
// 
//------------------------------------------------------------------------------

#ifndef DFS_CTOKENIZER_H
#define DFS_CTOKENIZER_H

// Includes from this library
#include <DfsCString.h>

/// The DfsCTokenizer class is a partial replacement of the RWCTokenizer class.
/** Only methods required by existing DFS code have been reimplemented, as new 
 * DFS code should be implemented using STL capabilities. 
 *
 * NOTE: Users should not assign new values to a token substring supplied by a 
 * DfsCTokenizer. The assignment will have no effect on either the tokeniser, 
 * or the string it was constructed from. However, it will be safe, in contrast 
 * to the RogueWave version, which behaves very oddly after such an assignment. 
 *
 * As DfsCTokenizer methods should correspond exactly to the original RogueWave
 * API, this documentation is sparse. Consult the RogueWave documentation 
 * for more details. 
 */ 
class DfsCTokenizer
{
public:
    /// Constructs a tokeniser based on the given @a sourceString. 
    DfsCTokenizer(const DfsCString& sourceString);

    // The default destructor suffices for this class. 

    /// Advances to the next token, and returns it as a substring. 
    DfsCSubString operator()();

    /// Advances to the next token, and returns it as a substring. 
    /** Tokens are delimited by any of the characters in @a delimiterSet, 
     * or by an embedded null byte. */
    DfsCSubString operator()(const char* delimiterSet);

protected:
    
private:
    /// This STL string contains a copy of the string being tokenised. 
    /** The tokeniser protects itself from changes to the source string by 
     * taking a copy of its contents. This ensures that the tokeniser has 
     * full control of its state, and also prevents the user from changing 
     * the source string by assigning new values to token substrings. 
     */ 
    std::string masterTokenString;

    /// The current index within the masterTokenString. 
    size_t masterTokenIndex;

    /// This DfsCString acts as parent string to the current token substring. 
    /** Using a separate parent string from the masterTokenString means that 
     * the user cannot corrupt the tokeniser state by assigning new values 
     * to token substrings. 
     */
    DfsCString currentTokenParent;

    /// Users may not copy DfsCTokenizer objects. 
    DfsCTokenizer(const DfsCTokenizer&)
    {
        assert(0);      // This method should never be called. 
    }
    
    /// Users may not copy DfsCTokenizer objects. 
    DfsCTokenizer& operator=(const DfsCTokenizer&)
    {
        assert(0);      // This method should never be called. 
        return(*this);  // Avoid missing return warnings. 
    }
};

#endif  // DFS_CTOKENIZER_H
