// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2004 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System - RogueWave->dfsStdLib port
// AUTHOR:    Neil Ferguson - ESO/DMD/DFS
// SUBSYSTEM: dfsStdLib DfsListBase - base class for all DfsStdLib list classes
//
// $Revision: 1.1 $
// $Log: DfsListBase.h,v $
// Revision 1.1  2004/08/23 08:53:01  nferguso
// Extracted common list functionality from DfsTPtrDlist, DfsTPtrSlist,
// DfsTValDlist, and DfsTValSlist into a new DfsListBase base class,
// which is inherited by the four list classes.
//
//
//------------------------------------------------------------------------------

#ifndef DFS_LIST_BASE_H
#define DFS_LIST_BASE_H

// Standard C++ includes
#include <iterator>
#include <list>

// Includes from this library.
#include <DfsUtils.h>

/// DfsListBase is the base class for all DfsStdLib list classes. 
/** This class is inherited by DfsTPtrDlist, DfsTPtrSlist, DfsTValDlist, 
 * and DfsTValSlist. It provides common capabilities not already provided 
 * by the encapsulated STL list, such as indexed access to list elements. 
 *
 * <B>NOTES:</B> 
 *
 * - DfsListBase does not contain enough functionality to allow direct use, 
 * nor does it specify any virtual methods. As a result, it should only be 
 * used indirectly through its subclasses. 
 * 
 * - This class is implicitly unit tested by the unit tests for its four 
 * subclasses. 
 *
 * - Indexed element access has linear complexity for lists. Although 
 * DfsListBase provides sequential access optimisation, vector indices 
 * are still faster. Non-sequential access should be avoided altogether. 
 */
template<typename ELEMENT> 
class DfsListBase
{
public:
    /// Creates an empty list. 
    DfsListBase<ELEMENT>() : 
        stlList(),
        lastElementIndex(0), 
        lastElementValid(false),
        listLength(0)
    {
        lastElementPos = stlList.begin(); 
    }

    /// Copy constructor. 
    DfsListBase<ELEMENT>(const DfsListBase<ELEMENT>& source) : 
        stlList(source.stlList),
        lastElementIndex(0), 
        listLength(source.listLength)
    {
        lastElementPos = stlList.begin(); 
        lastElementValid = (listLength != 0);
    }
    
    // The default destructor suffices for this class. 

    /// @brief Replaces the contents of this list with a copy of the 
    /// @a source list's contents. 
    DfsListBase<ELEMENT>& operator=(const DfsListBase<ELEMENT>& source)
    {
        if(&source != this)
        {
            stlList = source.stlList;
            listLength = source.listLength;

            lastElementPos = stlList.begin(); 
            lastElementIndex = 0;
            lastElementValid = (listLength != 0);
        }

        return(*this);
    }

    /// Returns the number of entries in this list. 
    size_t entries() const
    {
        return(listLength);
    }

protected: 
    /// @brief This STL list stores the contents of a DfsStdLib list, and 
    /// provides part of its functionality. 
    std::list<ELEMENT> stlList; 

    /// @brief This STL list iterator marks the location of the last list 
    /// element accessed by an indexing operation. 
    /** Storing this position means that sequential access patterns can be 
     * accelerated by a substantial margin. 
     */
    mutable typename std::list<ELEMENT>::iterator lastElementPos;

    /// The last index requested by an indexing operation. 
    mutable size_t lastElementIndex;

    /// True if the last index is valid. 
    mutable bool lastElementValid;

    /// The current length of this list. 
    size_t listLength; 

    /// @brief Generates and returns an STL list iterator set to the element 
    /// with the given @a index. 
    /** This method uses offset calculations to find the quickest move to the 
     * desired index, which is from the closest of the following points: 
     * 
     * - the start of the list. 
     * - the last indexed element, if available. 
     * - the end of the list. 
     *
     * The need for this algorithm stems from the fact that STL lists do not 
     * provide indexed access to their elements. List indexing is always less 
     * efficient than vector indexing; if indexed access is the main use for 
     * a container, then one should seriously consider using a vector instead. 
     *
     * <B>WARNING:</B> @a index must be in range for the current list length. 
     * Calls to this method are expected to range-check it beforehand. 
     */ 
    typename std::list<ELEMENT>::iterator 
    locateNewElementPos(const size_t index) const
    {
        // Provides mutable access to the STL list without exposing it to all 
        // the other const methods. 
        std::list<ELEMENT>* mutableList = 
            const_cast<std::list<ELEMENT>*>(&stlList); 

        // This becomes the iterator that will be returned. 
        typename std::list<ELEMENT>::iterator locatedPos(lastElementPos); 

        // Offset of the given index relative to the last indexed element. 
        int lastOffset = 
            static_cast<int>(index) - static_cast<int>(lastElementIndex);

        if(!lastElementValid)
        {
            // The last indexed element is not available - offset from the 
            // start of the list instead. 
            locatedPos = mutableList->begin(); 
            lastOffset = static_cast<int>(index) - 0; 
        }

        // Consider the fastest cases first. 
        switch(lastOffset)
        {
          case -1:
            return(--locatedPos);       // One step backward. 
            break;
            
          case 0:
            return(locatedPos);         // Current position. 
            break;
            
          case 1:
            return(++locatedPos);       // One step forward. 
            break;
        }

        // This case requires a non-trivial move. 
        {
            int beginOffset = index;
            int endOffset = index - listLength;

            if(beginOffset < -lastOffset)
            {
                // Closest starting point is the start of the list. 
                locatedPos = mutableList->begin();
                std::advance(locatedPos, beginOffset);
            }
            else if(-endOffset < lastOffset)
            {
                // Closest starting point is the end of the list. 
                locatedPos = mutableList->end();
                std::advance(locatedPos, endOffset);
            }
            else
            {
                // Closest starting point is the last indexed element. 
                std::advance(locatedPos, lastOffset);
            }

            return(locatedPos);
        }
    }

    /// @brief Moves the last indexed element iterator to the given @a index, 
    /// stores the numeric index value, and marks them both as being valid. 
    /** <B>WARNING:</B> @a index must be in range for the current list length. 
     * Calls to this method are expected to range-check it beforehand. 
     */
    typename std::list<ELEMENT>::iterator 
    moveLastElementPos(const size_t index) const 
    {
        // Locate the required iterator point for this index and store it. 
        lastElementPos = locateNewElementPos(index);
        lastElementIndex = index; 
        lastElementValid = true;

        return(lastElementPos);
    }

private: 

};

#endif  // DFS_LIST_BASE_H
