// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2004 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System - RogueWave->dfsStdLib port
// AUTHOR:    Neil Ferguson - ESO/DMD/DFS
// SUBSYSTEM: dfsStdLib RWTPtrDlistIterator replacement
//
// $Revision: 1.7 $
// $Log: DfsTPtrDlistIterator.h,v $
// Revision 1.7  2004/08/17 15:54:22  nferguso
// Documentation updates.
//
// Revision 1.6  2004/08/10 15:56:38  nferguso
// Removed redundant #include directives.
//
// Revision 1.5  2004/04/05 12:19:32  nferguso
// Rationalised include hierarchy.
//
// Revision 1.4  2004/03/26 15:07:40  nferguso
// Copyright years updated.
//
// Revision 1.3  2004/03/14 18:38:05  nferguso
// Code in insertAfterPoint() reformatted to fit within 80 columns.
//
// Revision 1.2  2004/03/11 15:45:52  nferguso
// Revised a comment in operator++().
// Removed:
// - superfluous includes.
// - redundant private default constructor.
//
// Revision 1.1  2004/03/11 13:21:53  nferguso
// First Check-In
//
// 
//------------------------------------------------------------------------------

#ifndef DFS_TPTR_DLIST_ITERATOR_H
#define DFS_TPTR_DLIST_ITERATOR_H

// Standard C includes. 
#include <cassert>

// Includes from this library.
#include <DfsConstants.h>
#include <DfsTPtrDlist.h>

// Required to resolve adverse include file ordering. 
template<typename ELEMENT> class DfsTPtrDlist;

/// @brief The DfsTPtrDlistIterator class is a partial replacement of the 
/// RWTPtrDlistIterator class. 
/** Only methods required by existing DFS code have been reimplemented, as 
 * new DFS code should be implemented using STL capabilities. 
 * 
 * As DfsTPtrDlistIterator methods should correspond exactly to the original 
 * RogueWave API, this documentation is sparse. Consult the RogueWave 
 * documentation for more details. 
 */ 
template<typename ELEMENT>
class DfsTPtrDlistIterator
{
public:
    /// Creates a new DfsTPtrDlistIterator for the given @a parent list. 
    DfsTPtrDlistIterator<ELEMENT>(DfsTPtrDlist<ELEMENT>& parent) : 
        parentList(&parent), 
        currentPos(parent.stlList.begin()),
        preBegin(true)
    {
        // This constructor does nothing beyond member initialisation. 
    }
    
    // The default destructor suffices for this class. 

    /// Advances this DfsTPtrDlistIterator to the next element. 
    /** Returns @c true if the advance left the iterator referencing a 
     * defined value, or @c false otherwise. 
     *
     * NB: This is not strictly what the RogueWave documentation describes, 
     * but it matches their implementation. 
     */
    DfsBoolean operator++()
    {
        if(preBegin)
        {
            // Conceptual advance to the first element. 
            preBegin = false; 
        }
        else if(currentPos != parentList->stlList.end())
        {
            // At least one position follows the current one - advance to 
            // the next position. 
            ++currentPos;
        }

        return(currentPos != parentList->stlList.end());
    }
    
    /// Moves this DfsTPtrDlistIterator backward to the previous element. 
    /** Returns @c true if the decrement left the iterator referencing a 
     * defined value, or @c false otherwise. 
     *
     * NB: This is not strictly what the RogueWave documentation describes, 
     * but it matches their implementation. 
     */ 
    DfsBoolean operator--()
    {
        if(currentPos == parentList->stlList.begin())
        {
            // Cannot decrement past the start of the STL list - adopt a 
            // conceptual position just before the first element. 
            preBegin = true;
        }
        else
        {
            // At least one element precedes the current one - move back 
            // to the previous element. 
            --currentPos;
        }

        return(!preBegin);
    }

    /// @brief Inserts the given @a element immediately after the element 
    /// currently referenced by this DfsTPtrDlistIterator. 
    /** NOTES: 
     *
     * - If the iterator is (conceptually) set to the undefined position just 
     * before the first element, the inserted element will become the first 
     * element in the list. 
     * - If the iterator references a defined value, the inserted element will 
     * be placed immediately after that value in the list. 
     * - <B>WARNING:</B> If the iterator is set to the undefined position just 
     * after the last element, the behaviour of this method is undefined.
     *
     * Although these points are not documented by RogueWave, this behaviour 
     * matches their implementation. 
     */
    void insertAfterPoint(ELEMENT* element)
    {
        if(preBegin)
        {
            // Insert the element at the start of the list. 
            currentPos = parentList->stlList.insert(
                currentPos, 
                static_cast<void*>(element)); 
        }
        else
        {
            // Insert the element immediately after the current position. 
            parentList->stlList.insert(
                ++std::list<void*>::iterator(currentPos),
                static_cast<void*>(element));
        }

        // The parent list is now one element longer. 
        parentList->listLength += 1;

        // Reset the last indexed element position in the parent list. 
        parentList->lastElementPos = parentList->stlList.begin(); 
        parentList->lastElementIndex = 0; 
        parentList->lastElementValid = true;
    }

    /// @brief Returns the DfsTPtrDlist element currently referenced by this 
    /// DfsTPtrDlistIterator. 
    ELEMENT* key()
    {
        return(static_cast<ELEMENT*>(*currentPos));
    }

    /// @brief Removes the DfsTPtrDlist element currently referenced by this 
    /// DfsTPtrDlistIterator from the list and returns it. 
    /** If the element being removed is the first element, the iterator will 
     * be set to a conceptual position just before the new first element. 
     * Otherwise, it will be left referencing the element which preceded the 
     * removed element. This behaviour matches the RogueWave implementation. 
     */ 
    ELEMENT* remove()
    {
        // Make a copy of the current element. 
        ELEMENT* element = static_cast<ELEMENT*>(*currentPos);

        // Remove the current element. 
        currentPos = parentList->stlList.erase(currentPos);

        // The parent list is now one element shorter. 
        parentList->listLength -= 1;

        // Reset the last indexed element position in the parent list. 
        parentList->lastElementPos = parentList->stlList.begin(); 
        parentList->lastElementIndex = 0; 
        parentList->lastElementValid = (parentList->listLength != 0); 

        // Emulate RogueWave - decrement the new iterator position. 
        this->operator--();

        // Return the removed element. 
        return(element); 
    }
    
protected:
    
private:
    /// The DfsTPtrDlist that this DfsTPtrDlistIterator refers to. 
    DfsTPtrDlist<ELEMENT>* parentList;

    /// @brief This iterator refers to the parent DfsTPtrDlist's encapsulated 
    /// STL list; it already provides most of the required functionality. 
    std::list<void*>::iterator currentPos; 

    /// @brief True if this DfsTPtrDlistIterator is (conceptually) set to the 
    /// undefined position just before the first element. 
    bool preBegin; 

    /// @brief The copy constructor is not provided by RogueWave and should 
    /// never be used. 
    DfsTPtrDlistIterator<ELEMENT>(const DfsTPtrDlistIterator<ELEMENT>& source) 
    {
        assert(0);
    }

    /// @brief The assignment operator is not provided by RogueWave and should 
    /// never be used. 
    DfsTPtrDlistIterator<ELEMENT>& operator=(
        const DfsTPtrDlistIterator<ELEMENT>& source) 
    {
        assert(0);
    }
};

#endif  // DFS_TPTR_DLIST_ITERATOR_H
