// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2004 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System - RogueWave->dfsStdLib port
// AUTHOR:    Neil Ferguson - ESO/DMD/DFS
// SUBSYSTEM: dfsStdLib RWTPtrSlist replacement
//
// $Revision: 1.7 $
// $Log: DfsTPtrSlist.h,v $
// Revision 1.7  2004/08/23 08:53:01  nferguso
// Extracted common list functionality from DfsTPtrDlist, DfsTPtrSlist,
// DfsTValDlist, and DfsTValSlist into a new DfsListBase base class,
// which is inherited by the four list classes.
//
// Revision 1.6  2004/08/17 15:54:22  nferguso
// Documentation updates.
//
// Revision 1.5  2004/08/10 15:56:38  nferguso
// Removed redundant #include directives.
//
// Revision 1.4  2004/03/26 15:07:40  nferguso
// Copyright years updated.
//
// Revision 1.3  2004/03/11 15:41:23  nferguso
// Added DfsTPtrSlistIterator as a friend class.
//
// Revision 1.2  2004/03/11 13:20:28  nferguso
// Corrected an off-by-one error in locateNewElementPos().
//
// Revision 1.1  2004/02/27 16:17:15  nferguso
// First Check-In
//
// 
//------------------------------------------------------------------------------

#ifndef DFS_TPTR_SLIST_H
#define DFS_TPTR_SLIST_H

// Standard C++ includes
#include <algorithm>

// Includes from this library.
#include <DfsListBase.h>
#include <DfsTPtrSlistIterator.h>

// Required to resolve adverse include file ordering. 
template<typename ELEMENT> class DfsTPtrSlistIterator;

/// @brief The DfsTPtrSlist class is a partial replacement of the 
/// RWTPtrSlist class. 
/** Only methods required by existing DFS code have been reimplemented, 
 * as new DFS code should be implemented using STL capabilities. 
 * 
 * The STL does not provide a singly-linked list type, so DfsTPtrSlist 
 * is implemented using a doubly-linked list instead. 
 *
 * NB: Indexed element access has linear complexity for lists. Although 
 * DfsTPtrSlist provides sequential access optimisation, vector indices
 * are still faster. Non-sequential access should be avoided altogether. 
 * 
 * As DfsTPtrSlist methods should correspond exactly to the original 
 * RogueWave API, this documentation is sparse. Consult the RogueWave 
 * documentation for more details. 
 */ 
template<typename ELEMENT>
class DfsTPtrSlist : public DfsListBase<void*>
{
    /// DfsTPtrSlistIterators need access to the private data members. 
    friend class DfsTPtrSlistIterator<ELEMENT>;

public:
    /// Creates an empty DfsTPtrSlist. 
    DfsTPtrSlist<ELEMENT>() : 
        DfsListBase<void*>()
    {
        // This constructor does nothing beyond base class initialisation. 
    }

    /// Copy constructor. 
    DfsTPtrSlist<ELEMENT>(const DfsTPtrSlist<ELEMENT>& source) : 
        DfsListBase<void*>(source)
    {
        // This constructor does nothing beyond base class initialisation. 
    }
    
    // The default destructor suffices for this class. 

    /// @brief Replaces the contents of this DfsTPtrSlist with a copy 
    /// of the @a source list's contents. 
    DfsTPtrSlist<ELEMENT>& operator=(const DfsTPtrSlist<ELEMENT>& source)
    {
        // Delegate assignment to the base list class. 
        this->DfsListBase<void*>::operator=(source); 
        return(*this); 
    }

    /// Returns the element at the given @a index. 
    /** This method always performs bounds checking on the index, and throws 
     * a DfsBoundsErr exception for index values which are out of range. 
     */
    ELEMENT*& operator[](size_t index) throw(DfsBoundsErr)
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(index, listLength);

        // Move the last indexed element iterator to the new index, and 
        // return a reference to the pointer stored at that location. 
        return(*reinterpret_cast<ELEMENT**>(&(*moveLastElementPos(index))));
    }
    
    /// Returns the element at the given @a index. 
    /** This method always performs bounds checking on the index, and throws 
     * a DfsBoundsErr exception for index values which are out of range. 
     */
    ELEMENT* const& operator[](size_t index) const throw(DfsBoundsErr)
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(index, listLength);

        // Move the last indexed element iterator to the new index, and 
        // return a reference to the pointer stored at that location. 
        return(*reinterpret_cast<ELEMENT**>(&(*moveLastElementPos(index))));
    }

    /// Inserts the given @a element at the end of this DfsTPtrSlist. 
    void append(ELEMENT* element)
    {
        stlList.push_back(static_cast<void*>(element));
        ++listLength; 
    }
    
    /// Returns the element at the given @a index. 
    /** This method always performs bounds checking on the index, and throws 
     * a DfsBoundsErr exception for index values which are out of range. 
     */
    ELEMENT*& at(size_t index) throw(DfsBoundsErr)
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(index, listLength);

        // Move the last indexed element iterator to the new index, and 
        // return a reference to the pointer stored at that location. 
        return(*reinterpret_cast<ELEMENT**>(&(*moveLastElementPos(index))));
    }
    
    /// Returns the element at the given @a index. 
    /** This method always performs bounds checking on the index, and throws 
     * a DfsBoundsErr exception for index values which are out of range. 
     */
    ELEMENT* const& at(size_t index) const throw(DfsBoundsErr)
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(index, listLength);

        // Move the last indexed element iterator to the new index, and 
        // return a reference to the pointer stored at that location. 
        return(*reinterpret_cast<ELEMENT**>(&(*moveLastElementPos(index))));
    }

    /// @brief Clears all stored pointers, without affecting the objects they 
    /// point to. 
    void clear()
    {
        stlList.clear();                        // Clear the pointers. 
        lastElementPos = stlList.begin();       // Reset the last indexed
        lastElementIndex = 0;                   // element position. 
        lastElementValid = false;
        listLength = 0;                         // List length is zero. 
    }

    /// @brief Clears all stored pointers, destroying the objects they point 
    /// to by calling the @c delete operator. 
    /** <B>WARNING:</B> This method will <I>not</I> call @c delete[] for an 
     * array ELEMENT type. Array types must be deleted by the user. 
     */ 
    void clearAndDestroy()
    {
        // Loop through all stored pointers, deleting their objects. 
        std::list<void*>::iterator deletePos = stlList.begin();
        std::list<void*>::iterator endPoint = stlList.end();

        while(deletePos != endPoint)
        {
            delete static_cast<ELEMENT*>(*deletePos);
            ++deletePos;
        }

        stlList.clear();                        // Clear the pointers. 
        lastElementPos = stlList.begin();       // Reset the last indexed
        lastElementIndex = 0;                   // element position. 
        lastElementValid = false;
        listLength = 0;                         // List length is zero. 
    }
    
    /// @brief Returns @c true if this DfsTPtrSlist contains a pointer to an 
    /// object that compares equal to the object pointed to by @a element, 
    /// @c false otherwise. 
    /** NB: This method performs a linear search through the list. */
    bool contains(const ELEMENT* element) const
    {
        // Find the appropriate pointer, if one exists. 
        std::list<void*>::const_iterator foundPos(
            find_if(stlList.begin(), stlList.end(), PtrEquals(element))); 

        return(foundPos != stlList.end());
    }

    /// @brief If this DfsTPtrSlist contains a pointer to an object that 
    /// compares equal to the object pointed to by @a element, @c find() 
    /// returns that pointer. Otherwise, a null pointer will be returned. 
    /** NB: This method performs a linear search through the list. */
    ELEMENT* find(const ELEMENT* element) const
    {
        // Find the appropriate pointer, if one exists. 
        std::list<void*>::const_iterator foundPos(
            find_if(stlList.begin(), stlList.end(), PtrEquals(element))); 

        if(foundPos != stlList.end())
        {
            // Found a match - return the corresponding pointer. 
            return(static_cast<ELEMENT*>(*foundPos));
        }
        else
        {
            // No match found - return a null pointer. 
            return(0);
        }
    }

    /// Inserts the given @a element at the end of this DfsTPtrSlist. 
    /** @return Always returns @c true. */
    bool insert(ELEMENT* element)
    {
        // Insert the element. 
        stlList.push_back(static_cast<void*>(element));
        ++listLength; 
        return(true);
    }

    /// @brief If this DfsTPtrSlist contains a pointer to an object that 
    /// compares equal to the object pointed to by @a element, removes that 
    /// pointer from the list and returns it. Otherwise, a null pointer will 
    /// be returned. 
    /** NB: This method performs a linear search through the list. */ 
    ELEMENT* remove(const ELEMENT* element)
    {
        // Find the appropriate pointer, if one exists. 
        std::list<void*>::iterator foundPos(stlList.begin());
        unsigned int foundIndex = 0;
        PtrEquals ptrEquals(element); 

        while((foundPos != stlList.end()) && (!ptrEquals(*foundPos))) 
        {
            ++foundPos;
            ++foundIndex;
        }

        if(foundPos != stlList.end())
        {
            // Found a match - store the pointer. 
            ELEMENT* foundPtr = static_cast<ELEMENT*>(*foundPos); 

            // Remove the list element. 
            stlList.erase(foundPos);
            --listLength;

            if(foundIndex == lastElementIndex)
            {
                // The last indexed element was removed - the stored position 
                // must be marked invalid. 
                lastElementPos = stlList.begin();
                lastElementIndex = 0;
                lastElementValid = false;
            }
            else if(foundIndex < lastElementIndex)
            {
                // Removed an element before the last indexed element, so the 
                // last index must be decremented. The iterator is unaffected. 
                --lastElementIndex;
            }
            
            return(foundPtr);
        }
        else
        {
            // No match found - return a null pointer. 
            return(0);
        }
    }

    /// @brief Removes and returns the last element, or a null pointer if the 
    /// list is empty. 
    ELEMENT* removeLast()
    {
        if(listLength)
        {
            // List contains elements - get the last element. 
            std::list<void*>::iterator lastPos = --stlList.end(); 

            // Store the pointer. 
            ELEMENT* lastPtr = static_cast<ELEMENT*>(*lastPos);
            
            // Erase the last element. 
            stlList.erase(lastPos);
            --listLength;

            if(lastElementIndex == listLength)
            {
                // The last element was also the last indexed element, so the 
                // stored position must be marked invalid. 
                lastElementPos = stlList.begin();
                lastElementIndex = 0;
                lastElementValid = false;
            }

            return(lastPtr);
        }
        else
        {
            // This list is empty. 
            return(0);
        }
    }

protected:
    
private:
    /// @brief This class permits equality comparisons on objects for which 
    /// only pointers are available. 
    class PtrEquals : std::unary_function<const void* const, bool> 
    {
    public:
        /// @brief Creates a new PtrEquals instance to make comparisons with 
        /// the object referenced by @a compare. 
        PtrEquals(const ELEMENT* compare) :
            element(compare)
        {
            // This constructor does nothing beyond member initialisation. 
        }

        /// @brief This operator returns the result of an equality comparison 
        /// on the object pointed to by @a param and the object provided at 
        /// construction time. 
        bool operator()(const void* const param)
        {
            return(*static_cast<const ELEMENT*>(param) == *element); 
        }

    private:
        /// A pointer to the object to compare with. 
        const ELEMENT* element;
    };
};

#endif  // DFS_TPTR_SLIST_H
