// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2004 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System - RogueWave->dfsStdLib port
// AUTHOR:    Neil Ferguson - ESO/DMD/DFS
// SUBSYSTEM: dfsStdLib RWTPtrSlistIterator replacement
//
// $Revision: 1.4 $
// $Log: DfsTPtrSlistIterator.h,v $
// Revision 1.4  2004/08/17 15:54:22  nferguso
// Documentation updates.
//
// Revision 1.3  2004/08/10 15:56:38  nferguso
// Removed redundant #include directives.
//
// Revision 1.2  2004/03/26 15:07:40  nferguso
// Copyright years updated.
//
// Revision 1.1  2004/03/11 15:46:45  nferguso
// First Check-In
//
// 
//------------------------------------------------------------------------------

#ifndef DFS_TPTR_SLIST_ITERATOR_H
#define DFS_TPTR_SLIST_ITERATOR_H

// Standard C includes. 
#include <cassert>

// Includes from this library.
#include <DfsTPtrSlist.h>

// Required to resolve adverse include file ordering. 
template<typename ELEMENT> class DfsTPtrSlist;

/// @brief The DfsTPtrSlistIterator class is a partial replacement of the 
/// RWTPtrSlistIterator class. 
/** Only methods required by existing DFS code have been reimplemented, as 
 * new DFS code should be implemented using STL capabilities. 
 * 
 * As DfsTPtrSlistIterator methods should correspond exactly to the original 
 * RogueWave API, this documentation is sparse. Consult the RogueWave 
 * documentation for more details. 
 */ 
template<typename ELEMENT>
class DfsTPtrSlistIterator
{
public:
    /// Creates a new DfsTPtrSlistIterator for the given @a parent list. 
    DfsTPtrSlistIterator<ELEMENT>(DfsTPtrSlist<ELEMENT>& parent) : 
        parentList(&parent), 
        currentPos(parent.stlList.begin()),
        preBegin(true)
    {
        // This constructor does nothing beyond member initialisation. 
    }
    
    // The default destructor suffices for this class. 

    /// Advances this DfsTPtrSlistIterator to the next element. 
    /** Returns the value referenced by the iterator after it has been 
     * advanced, or a null pointer if that value is undefined. 
     */
    ELEMENT* operator()()
    {
        if(preBegin)
        {
            // Conceptual advance to the first element. 
            preBegin = false; 
        }
        else if(currentPos != parentList->stlList.end())
        {
            // At least one position follows the current one - advance to 
            // the next position. 
            ++currentPos;
        }
        
        if(currentPos == parentList->stlList.end())
        {
            // The iterator has advanced beyond the last element - return 
            // a null pointer.
            return(0);
        }
        else
        {
            // The iterator is referencing an element - return its value. 
            return(static_cast<ELEMENT*>(*currentPos));
        }
    }

    /// @brief Returns the DfsTPtrSlist element currently referenced by this 
    /// DfsTPtrSlistIterator. 
    ELEMENT* key()
    {
        return(static_cast<ELEMENT*>(*currentPos));
    }
    
protected:
    
private:
    /// The DfsTPtrSlist that this DfsTPtrSlistIterator refers to. 
    DfsTPtrSlist<ELEMENT>* parentList;

    /// @brief This iterator refers to the parent DfsTPtrSlist's encapsulated 
    /// STL list; it already provides most of the required functionality. 
    std::list<void*>::iterator currentPos; 

    /// @brief True if this DfsTPtrSlistIterator is (conceptually) set to the 
    /// undefined position just before the first element. 
    bool preBegin; 

    /// @brief The copy constructor is not provided by RogueWave and should 
    /// never be used. 
    DfsTPtrSlistIterator<ELEMENT>(const DfsTPtrSlistIterator<ELEMENT>& source) 
    {
        assert(0);
    }

    /// @brief The assignment operator is not provided by RogueWave and should 
    /// never be used. 
    DfsTPtrSlistIterator<ELEMENT>& operator=(
        const DfsTPtrSlistIterator<ELEMENT>& source) 
    {
        assert(0);
    }
};

#endif  // DFS_TPTR_SLIST_ITERATOR_H
