// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2003, 2004 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System - RogueWave->dfsStdLib port
// AUTHOR:    Neil Ferguson - ESO/DMD/DFS
// SUBSYSTEM: dfsStdLib RWTPtrSortedVector replacement
//
// $Revision: 1.8 $
// $Log: DfsTPtrSortedVector.h,v $
// Revision 1.8  2004/08/17 15:54:22  nferguso
// Documentation updates.
//
// Revision 1.7  2004/08/10 15:56:38  nferguso
// Removed redundant #include directives.
//
// Revision 1.6  2004/04/27 11:23:07  nferguso
// Added a note to the append() method documentation.
//
// Revision 1.5  2004/03/26 15:07:40  nferguso
// Copyright years updated.
//
// Revision 1.4  2004/02/23 13:42:39  nferguso
// Extra constness in PtrSort.
//
// Revision 1.3  2004/02/20 17:45:37  nferguso
// Tweaked documentation comments.
//
// Revision 1.2  2004/02/20 10:14:57  nferguso
// Implementation completed.
//
// Revision 1.1.1.1  2003/11/28 14:21:16  nferguso
// First Check-In
//
// 
//------------------------------------------------------------------------------

#ifndef DFS_TPTR_SORTED_VECTOR_H
#define DFS_TPTR_SORTED_VECTOR_H

// Standard C++ includes
#include <algorithm>
#include <functional>
#include <vector>

// Includes from this library.
#include <DfsUtils.h>

/// @brief The DfsTPtrSortedVector class is a partial replacement of the 
/// RWTPtrSortedVector class. 
/** Only methods required by existing DFS code have been reimplemented, 
 * as new DFS code should be implemented using STL capabilities. 
 * 
 * NB: The RogueWave version of this template used by DFS does not accept a 
 * SORT parameter. Instead, it always sorts using less-than ordering, with 
 * the smallest item first. 
 *
 * As DfsTPtrSortedVector methods should correspond exactly to the original 
 * RogueWave API, this documentation is sparse. Consult the RogueWave 
 * documentation for more details. 
 */ 
template<typename ELEMENT, typename SORT = std::less<ELEMENT> >
class DfsTPtrSortedVector
{
public:
    /// Creates an empty DFSTPtrSortedVector. 
    DfsTPtrSortedVector<ELEMENT, SORT>() 
    {
        // This constructor does not need to do anything. 
    }
    
    /// @brief Creates an empty DFSTPtrSortedVector with preallocated storage 
    /// space for the given number of @a elements. 
    DfsTPtrSortedVector<ELEMENT, SORT>(size_t elements)
    {
        // Reserve STL vector memory for elements. 
        stlVector.reserve(elements);
    }

    /// Copy constructor. 
    DfsTPtrSortedVector<ELEMENT, SORT>(const DfsTPtrSortedVector<ELEMENT, SORT>& source) :
        stlVector(source.stlVector)
    {
        // This constructor does nothing beyond member initialisation. 
    }

    // The default destructor suffices for this class. 

    /// @brief Replaces the contents of this DfsTPtrSortedVector with a copy 
    /// of the @a source vector's contents. 
    DfsTPtrSortedVector<ELEMENT, SORT>& operator=(const DfsTPtrSortedVector<ELEMENT, SORT>& source)
    {
        if(&source != this)
        {
            stlVector = source.stlVector;
        }

        return(*this);
    }

    /// Returns the pointer at the given @a index. 
    /** This operator always performs bounds checking on the index, and throws
     * a DfsBoundsErr exception for index values which are out of range. 
     */
    ELEMENT*& operator[](size_t index) throw(DfsBoundsErr)
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(index, stlVector.size());
        
        return(*reinterpret_cast<ELEMENT**>(&stlVector[index]));
    }

    /// Returns the pointer at the given @a index. 
    /** This operator always performs bounds checking on the index, and throws
     * a DfsBoundsErr exception for index values which are out of range. 
     */
    ELEMENT* const& operator[](size_t index) const throw(DfsBoundsErr)
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(index, stlVector.size());

        return(*reinterpret_cast<ELEMENT* const*>(&stlVector[index]));
    }

    /// @brief Inserts the given @a element at the appropriate point determined
    /// by the sorting criterion. 
    /** This method is not documented by RogueWave, but it works in their 
     * implementation, so the DFS version provides it for compatibility. 
     * The official, approved insert() method is slightly more efficient. 
     * 
     * NOTE: The RogueWave append() method inserts the new element at the 
     * end of the vector, regardless of the sort order. However, the DFS 
     * version always inserts the element at the appropriate point dictated 
     * by the sorting criterion. This is necessary to ensure that methods 
     * which depend on the vector being sorted can always be called safely. 
     */ 
    void append(ELEMENT* element)
    {
        // Insert the element. 
        insert(element);
    }

    /// @brief Returns the number of entries this vector will accommodate 
    /// without reallocating. 
    size_t capacity() const
    {
        return(stlVector.capacity());
    }

    /// @brief Clears all stored pointers, without affecting the objects they 
    /// point to. 
    void clear()
    {
        stlVector.clear();
    }

    /// @brief Clears all stored pointers, destroying the objects they point 
    /// to by calling the @c delete operator. 
    /** <B>WARNING:</B> This method will <I>not</I> call @c delete[] for an 
     * array ELEMENT type. Array types must be deleted by the user. 
     */ 
    void clearAndDestroy()
    {
        // Loop through all stored pointers, deleting their objects. 
        int numElements = stlVector.size();

        for(int i = 0; i < numElements; ++i)
        {
            delete static_cast<ELEMENT*>(stlVector[i]);
        }

        stlVector.clear();    // Clear the pointers themselves. 
    }

    /// Returns the number of entries in this DfsTPtrSortedVector. 
    size_t entries() const
    {
        return(stlVector.size());
    }

    /// @brief Inserts the given @a element at the appropriate point determined
    /// by the sorting criterion. 
    /** @return Always returns @c true. */
    bool insert(ELEMENT* element)
    {
        // Find the point where this element should be inserted. 
        std::vector<void*>::iterator 
            insertPoint(lower_bound(stlVector.begin(), 
                                    stlVector.end(), 
                                    static_cast<void*>(element),
                                    PtrSort()));

        // Insert the element. 
        stlVector.insert(insertPoint, static_cast<void*>(element));

        return(true);
    }

    /// Returns the number of entries in this DfsTPtrSortedVector. 
    size_t length() const
    {
        return(stlVector.size());
    }

    /// Removes and returns the element at the given @a index. 
    /** This method performs bounds checking on the index, and throws a 
     * DfsBoundsErr exception for index values which are out of range. 
     */
    ELEMENT* removeAt(size_t index) throw(DfsBoundsErr)
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(index, stlVector.size());
        
        // Store the required element before... 
        ELEMENT* element = static_cast<ELEMENT*>(stlVector[index]);
        
        // ...removing it from the vector. 
        stlVector.erase(stlVector.begin() + index);

        return(element);
    }

    /// Removes and returns the first element. 
    /** This method throws a DfsBoundsErr exception if the DfsTPtrSortedVector 
     * is empty. 
     */
    ELEMENT* removeFirst() throw(DfsBoundsErr)
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(0, stlVector.size());
        
        // Store the first element before... 
        ELEMENT* element = static_cast<ELEMENT*>(stlVector[0]);
        
        // ...removing it from the vector. 
        stlVector.erase(stlVector.begin());

        return(element);
    }

private:
    /// @brief This class adapts the object sorting criterion SORT, so that it 
    /// can also be used to sort pointers to those objects. 
    class PtrSort : std::binary_function<const void*, const void*, bool> 
    {
    public:
        /// @brief This operator returns the result of applying the object 
        /// sorting criterion to the objects pointed to by its parameters. 
        bool operator()(const void* param1, const void* param2)
        {
            return(ObjSort(*static_cast<const ELEMENT*>(param1), 
                           *static_cast<const ELEMENT*>(param2)));
        }

    private:
        /// An instance of the object sorting criterion. 
        SORT ObjSort;
    };

    /// @brief This vector stores the contents of a DfsTPtrSortedVector, and 
    /// provides part of its functionality. 
    std::vector<void*> stlVector;
};

#endif  // DFS_TPTR_SORTED_VECTOR_H 
