// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2004 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System - RogueWave->dfsStdLib port
// AUTHOR:    Neil Ferguson - ESO/DMD/DFS
// SUBSYSTEM: dfsStdLib RWTValSlist replacement
//
// $Revision: 1.7 $
// $Log: DfsTValSlist.h,v $
// Revision 1.7  2004/09/24 15:04:49  nferguso
// Added first() method.
//
// Revision 1.6  2004/09/01 13:50:05  nferguso
// DfsListBase is inherited as a dependent template instantiation - base
// class member accesses must be qualified with 'this'.
//
// Revision 1.5  2004/08/23 08:53:01  nferguso
// Extracted common list functionality from DfsTPtrDlist, DfsTPtrSlist,
// DfsTValDlist, and DfsTValSlist into a new DfsListBase base class,
// which is inherited by the four list classes.
//
// Revision 1.4  2004/08/17 15:54:22  nferguso
// Documentation updates.
//
// Revision 1.3  2004/08/10 15:56:38  nferguso
// Removed redundant #include directives.
//
// Revision 1.2  2004/03/26 15:07:40  nferguso
// Copyright years updated.
//
// Revision 1.1  2004/03/14 18:39:16  nferguso
// First Check-In
//
// 
//------------------------------------------------------------------------------

#ifndef DFS_TVAL_SLIST_H
#define DFS_TVAL_SLIST_H

// Standard C++ includes
#include <algorithm>

// Includes from this library.
#include <DfsListBase.h>
#include <DfsTValSlistIterator.h>

// Required to resolve adverse include file ordering. 
template<typename ELEMENT> class DfsTValSlistIterator;

/// @brief The DfsTValSlist class is a partial replacement of the 
/// RWTValSlist class. 
/** Only methods required by existing DFS code have been reimplemented, 
 * as new DFS code should be implemented using STL capabilities. 
 * 
 * The STL does not provide a singly-linked list type, so DfsTValSlist 
 * is implemented using a doubly-linked list instead. 
 *
 * NB: Indexed element access has linear complexity for lists. Although 
 * DfsTValSlist provides sequential access optimisation, vector indices
 * are still faster. Non-sequential access should be avoided altogether. 
 * 
 * As DfsTValSlist methods should correspond exactly to the original 
 * RogueWave API, this documentation is sparse. Consult the RogueWave 
 * documentation for more details. 
 */ 
template<typename ELEMENT>
class DfsTValSlist : public DfsListBase<ELEMENT>
{
    /// DfsTValSlistIterators need access to the private data members. 
    friend class DfsTValSlistIterator<ELEMENT>;

public:
    /// Creates an empty DfsTValSlist. 
    DfsTValSlist<ELEMENT>() : 
        DfsListBase<ELEMENT>()
    {
        // This constructor does nothing beyond base class initialisation. 
    }

    /// Copy constructor. 
    DfsTValSlist<ELEMENT>(const DfsTValSlist<ELEMENT>& source) : 
        DfsListBase<ELEMENT>(source)
    {
        // This constructor does nothing beyond base class initialisation. 
    }
    
    // The default destructor suffices for this class. 

    /// @brief Replaces the contents of this DfsTValSlist with a copy 
    /// of the @a source list's contents. 
    DfsTValSlist<ELEMENT>& operator=(const DfsTValSlist<ELEMENT>& source)
    {
        // Delegate assignment to the base list class. 
        this->DfsListBase<ELEMENT>::operator=(source); 
        return(*this); 
    }

    /// Returns the element at the given @a index. 
    /** This method always performs bounds checking on the index, and throws 
     * a DfsBoundsErr exception for index values which are out of range. 
     */
    ELEMENT& operator[](size_t index) throw(DfsBoundsErr)
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(index, this->listLength);

        // Move the last indexed element iterator to the new index, and 
        // return a reference to the element stored at that location. 
        return(*(this->moveLastElementPos(index)));
    }
    
    /// Returns the element at the given @a index. 
    /** This method always performs bounds checking on the index, and throws 
     * a DfsBoundsErr exception for index values which are out of range. 
     */
    const ELEMENT& operator[](size_t index) const throw(DfsBoundsErr)
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(index, this->listLength);

        // Move the last indexed element iterator to the new index, and 
        // return a reference to the element stored at that location. 
        return(*(this->moveLastElementPos(index)));
    }

    /// Inserts the given @a element at the end of this DfsTValSlist. 
    void append(const ELEMENT& element)
    {
        (this->stlList).push_back(element);
        ++(this->listLength); 
    }
    
    /// Clears all stored elements. 
    void clear()
    {
        (this->stlList).clear();                // Clear the elements. 

        // Reset the last indexed element position. 
        this->lastElementPos = (this->stlList).begin();
        this->lastElementIndex = 0;
        this->lastElementValid = false;

        this->listLength = 0;                   // List length is zero. 
    }

    /// @brief Returns @c true if this DfsTValSlist contains an element
    /// that compares equal to @a element, @c false otherwise. 
    /** NB: This method performs a linear search through the list. */
    bool contains(const ELEMENT& element) const
    {
        // Find the appropriate element, if one exists. 
        typename std::list<ELEMENT>::const_iterator foundPos(
            find((this->stlList.begin()), (this->stlList.end()), element)); 

        return(foundPos != (this->stlList.end()));
    }

    /// Returns a reference to the first element in the list. 
    const ELEMENT& first() const
    {
        return(*(this->stlList.begin()));
    }

    /// Inserts the given @a element at the end of this DfsTValSlist. 
    /** @return Always returns @c true. */
    bool insert(const ELEMENT& element)
    {
        // Insert the element. 
        (this->stlList).push_back(element);
        ++(this->listLength); 
        return(true);
    }

    /// Inserts @a element into this DfsTValSlist at @a index. 
    /** This method always performs bounds checking on the index, and throws 
     * a DfsBoundsErr exception for index values which are out of range. 
     *
     * NB: this method takes advantage of sequential access optimisation to 
     * locate the element at @a index, but it does not move the last accessed 
     * element position. This ensures that sequentially indexed list accesses 
     * will not be disrupted by randomly intervening insertAt() calls. 
     */
    void insertAt(size_t index, const ELEMENT& element) throw(DfsBoundsErr) 
    {
        // Delegate range check on index. 
        DfsStd::checkIndex(index, this->listLength + 1);

        // Locate the position to insert at, and insert the element there. 
        (this->stlList).insert(this->locateNewElementPos(index), element);
        ++(this->listLength); 

        if(index <= (this->lastElementIndex))
        {
            // Inserted an element before the last indexed element, so the 
            // last index must be incremented. The iterator is unaffected. 
            ++(this->lastElementIndex); 
        }
    }

    /// Returns a reference to the last element in the list. 
    const ELEMENT& last() const
    {
        return(*(--(this->stlList.end())));
    }

    /// @brief Returns the number of elements in the list that compare equal 
    /// to @a element. 
    /** NB: This method performs a linear pass through the entire list. */
    size_t occurrencesOf(const ELEMENT& element) const
    {
        return(count((this->stlList.begin()), 
                     (this->stlList.end()), 
                     element));
    }

protected:
    
private:
    
};

#endif  // DFS_TVAL_SLIST_H
