// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2006 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   Archive Systems Utilities
// AUTHOR:    Neil Ferguson - ESO/DMD/DFS
// SUBSYSTEM: DfsTime
//
// $Revision: 1.4 $
// $Log: DfsTime.h,v $
// Revision 1.4  2007/04/20 14:23:19  nferguso
// Changed formatting of date/time fields from internal alignment to right
// alignment, sidestepping a bug in the Standard C++ Library. More details
// of the bug are available at:
//
//     http://gcc.gnu.org/bugzilla/show_bug.cgi?id=11352
//
// Revision 1.3  2006/08/28 17:26:05  nferguso
// The following changes were made:
// - new variant of DfsTime::asISOTimeString() added.
// - new method DfsTime::asUnixTimeString() added.
//
// Revision 1.2  2006/06/20 15:32:01  nferguso
// Minor documentation tweak.
//
// Revision 1.1  2006/05/30 16:02:48  jknudstr
// Added DfsTime.h DfsTime.C.
//
//------------------------------------------------------------------------------

#ifndef DFSTIME_H
#define DFSTIME_H

// C++ includes. 
#include <string>

// Library includes. 
#include <ace/Time_Value.h>             // For the ACE_Time_Value class. 

/** @file */

/** @def DFS_TIME_STRING_LENGTH 
 * The minimum number of characters needed to store any string result created 
 * by a DfsTime utility function. 
 */
#define DFS_TIME_STRING_LENGTH  30

/// The DfsTime namespace contains utility functions for working with dates 
/// and times. 
/** @b WARNING: Due to a bug in the Standard C++ Library affecting GCC 
 * versions up to and including 3.3.4, these methods use right alignment 
 * for numeric fields when they should be using internal alignment. In 
 * some cases, this can result in misformatted output. More information 
 * about the bug is available at: 
 * 
 *     http://gcc.gnu.org/bugzilla/show_bug.cgi?id=11352 
 * 
 * The documentation for each method gives the specific date/time ranges 
 * that will be misformatted. 
 */ 
namespace DfsTime
{
    /// @brief Converts an ACE_Time_Value into a character string in the 
    /// ISO date and time format. 
    /** The output format is @c YYYY-MM-DDTHH:MM:SS.SSSSSS with seconds 
     * given to microsecond resolution. 
     * 
     * @b WARNING: Dates with years in the range -99 to -1, inclusive, will 
     * be misformatted. See the DfsTime namespace documentation for details. 
     *
     * @param timeValue  the ACE_Time_Value to convert. 
     * @param timeString an array of at least @c ::DFS_TIME_STRING_LENGTH 
     *                   characters where the result will be stored. 
     *
     * @retval timeString the same pointer that was passed as the 
     *                    @a timeString parameter. 
     */
    ACE_TCHAR* asISOTimeString(const ACE_Time_Value& timeValue, 
                               ACE_TCHAR* const timeString); 

    /// @brief Converts an ACE_Time_Value into an STL string in the ISO date 
    /// and time format. 
    /** The output format is @c YYYY-MM-DDTHH:MM:SS.SSSSSS with seconds 
     * given to microsecond resolution. 
     * 
     * @b WARNING: Dates with years in the range -99 to -1, inclusive, will 
     * be misformatted. See the DfsTime namespace documentation for details. 
     *
     * @param timeValue the ACE_Time_Value to convert. 
     */
    std::string asISOTimeString(const ACE_Time_Value& timeValue); 

    /// @brief Converts an ACE_Time_Value into a UNIX time, then converts 
    /// that into an STL string. 
    /** The output format is @c S.ssssss - i.e. seconds since 00:00:00 on 
     * January 1st, 1970, given to microsecond resolution. 
     * 
     * @b NOTE: This method should format all time values correctly. 
     * 
     * @param timeValue the ACE_Time_Value to convert. 
     */
    std::string asUnixTimeString(const ACE_Time_Value& timeValue); 
}

#endif  // DFSTIME_H
