// -*-c++-*-

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Christophe Moins- ESO/SDD/DFS
// SUBSYSTEM: arcsys - Archive Systems
//
// $Revision: 1.4 $
// $Log: FileCopyHandler.h,v $
// Revision 1.4  2007/07/04 11:38:43  cmoins
// DFS04052 : transfert replaced by transfer
//
// Revision 1.3  2007/04/26 12:50:21  cmoins
// doxygen comments correction
//
// Revision 1.2  2007/04/25 14:59:34  cmoins
// performCopy replaced by getFile and uploadFile
//
// Revision 1.1  2007/03/05 10:03:39  cmoins
// Creation of a class interface dealing with file copy
//
//
// Class that handles files copying. Only the GlobalInit operation is not
// thread safe. It should be called once when no thread is running. A 
// FileCopyHandler instance should not be shared among different threads.
// 
// 
//------------------------------------------------------------------------------

#ifndef FILE_COPY_HANDLER_H
#define FILE_COPY_HANDLER_H

#include <string>

using namespace std;

/// Class  handling files copy
/**
 * This class allows to get a file from a remote host or to upload files.
 * It implements the pimpl idiom so that client classes do not rely on 
 * the implementation of the file transfer.
 * Note : globalInit and globalRelease are not thread safe.
*/
class FileCopyHandler
{
public:

  /// \brief Constructor
  FileCopyHandler();
 
  /// \brief Destructor
  ~FileCopyHandler();

  /// \brief Performs a global init of the file transfer system. 
  /**  To be called once when no thread is running as it is not thread safe.
   *  \return 0 if successful, -1 otherwise
   */
  static int globalInit();

  /// \brief Performs a global release of the file transfer system.
  /** To be called once when no thread is running as it is not thread safe.
   *  \return 0 if successful, -1 otherwise
   */
  static int globalRelease();


  /// \brief Gets a file from a remote machine
  /**  \param hostName         name of the machine hosting the file to be copied 
   *   \param originFile       file to be copied
   *   \param targetFile       target file of the copy
   *   \param rtargetFileSize  size of the target file 
   *   \return 0 if successful, -1 otherwise
   */
  int getFile(string hostName,
	      string originFile,
	      string targetFile,
	      int&   rtargetFileSize);

  /// \brief Performs a file upload to another machine
  /** \param hostName        name of the machine the file shall be uploaded to
   *  \param fileHost        name of the machine hosting the file
   *  \param originFile      file to be uploaded
   *  \param targetFile      target file of the uploading
   *  \param rtargetFileSize size of the uploaded file
   *  \return 0 if successful, -1 otherwise
   */
  int uploadFile(string hostName,
		 string fileHost,
		 string originFile,
		 string targetFile,
		 int&   rtargetFileSize);

private :
  /// Forward declaration
  class FileCopyHandlerImpl ;   
  /// file copy handler implementation
  FileCopyHandlerImpl* pimpl_;  
};
#endif // FILE_COPY_HANDLER_H
