// -*-c++-*-

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    C.Moins - ESO/SDD/DFS
// SUBSYSTEM: DFS Archive System
//
// $Name: arcsys-1_2_8 $
// $Revision: 1.2 $
// $Log: FitsUtils.h,v $
// Revision 1.2  2007/10/17 08:13:25  cmoins
// test of NAXIS1 and NAXIS2 keywords existence
//
// Revision 1.1  2007/09/28 13:33:51  cmoins
// creation of a namespace storing methods to handle fits file
//
//
//-----------------------------------------------------------------------------

#ifndef FITS_UTILS_H
#define FITS_UTILS_H

#include <string>
#include "fitsio.h"

/// Creation of a name space that will provide wrappers to cftisio methods
namespace ArcSysFits 
{

  typedef fitsfile FitsFile;

  extern const int RECORD_MAX_LENGTH;

  /// \Brief Opens a fits file
  /**
   * \param rpfitsFile  pointer to the content of the fits file
   * \param fileName   name of the fits file to open
   * \param mode       type of access : READONLY or READWRITE
   * return 0 if the operation if successful, a non null value otherwise
   */
  int openFitsFile(FitsFile*&  rpfitsFile,
		   std::string fileName,
		   int         mode);

  /// \Brief Closes a fits file
  /**
   * \param pfitsFile  pointer to the content of the fits file
   * return 0 if the operation if successful, a non null value otherwise
   */
  int closeFitsFile(FitsFile* pfitsFile);

  /// \Brief reads a value associated to a keyword in a fits file header
  /**
   * \param pfitsFile  pointer to the content of the fits file
   * \param hduNumber  number of the hdu where to fetch the data (first
   *                   hdu index is equal to 1)
   * \param dataType   type of the value
   * \param keyName    keyword to be fetched
   * \param rvalue     value associated to the keyword
   * return 0 if the operation if successful, a non null value otherwise
   */
  template <class ValueType>
  int readFitsKey(FitsFile*   pfitsFile,
		  int         hduNumber,
		  int         dataType,
		  std::string keyName,
		  ValueType&  rvalue)
  {
    int retStatus = 0;
    int hduType;

    // moves to the desired hdu
    fits_movabs_hdu(pfitsFile, hduNumber, &hduType, &retStatus);

    if (!retStatus)
      {
	char* keyNameChar = strdup(keyName.c_str());
	fits_read_key(pfitsFile,
		      dataType,
		      keyNameChar,
		      (void*)&rvalue,
		      NULL,
		      &retStatus);

	free(keyNameChar);
      }

    return retStatus;
  }

  /// \Brief template specialization of readFitsKey for string
  template <> inline int readFitsKey(FitsFile*    pfitsFile,
				     int          hduNumber,
				     int          dataType,
				     std::string  keyName,
				     std::string& rvalue)
  {
    int retStatus = 0;

    int hduType;

    // moves to the desired hdu
    fits_movabs_hdu(pfitsFile, hduNumber, &hduType, &retStatus);

    if (!retStatus)
      {
	char* keyNameChar = strdup(keyName.c_str());
	char valueChar[80];

	fits_read_key(pfitsFile,
		      dataType,
		      keyNameChar,
		      (void*)valueChar,
		      NULL,
		      &retStatus);
	if (!retStatus)
	  {
	    rvalue = valueChar;
	  }
	free(keyNameChar);
      }
    return retStatus;
  }

  /// \Brief Function to convert a hdr file into a fits one.
  /**
   * \param fptr         pointer to the fits file obtained from the hdr file
   * \param hdrFilename  name of the hdr file to convert 
   * \param fitsFilename name of the result fits file 
   */
  int hdr2fits(FitsFile    **fptr, 
	       std::string hdrFilename,
	       std::string fitsFilename);




}

#endif // FITS_UTILS_H
// EOF

