// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2006 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Christophe Moins - ESO/SDD/DFS
// SUBSYSTEM: arcsys - archive Systems
//
// $Revision: 1.2 $
// $Log: GenConfig.h,v $
// Revision 1.2  2008/01/30 17:19:15  mpadovan
// Added an optional parameter to suppress logging while parsing the command
// line. Logging enabled by default.
//
// Revision 1.1  2007/10/01 15:16:43  cmoins
// creation of a general configuration class that parse command lines to get
// the list of arguments. All configuration classes associated to a specific class
// will inherit from it.
//
//
//------------------------------------------------------------------------------

#ifndef GEN_CONFIG_H
#define GEN_CONFIG_H

// C++ includes. 
#include <vector>
#include <map>
#include <string>
#include <sstream>

/**
 * Configuration base class that all configuration classes associated to a 
 * specific clas will inherit from. Mainly, it provides methods for command line
 * arguments processing. It allows to parse a command line to retrieve all 
 * the arguments. Subclasses will implement methods that allow to retrieve
 * a specific argument.
 */
class GenConfig 
{
public:

  /// tag associated to non optional arguments
  static const std::string NON_OPTIONAL_ARG;
  /// tag to indicate that there is no short name associated to an option
  static const char NO_SHORT_NAME;

  /// type defining a vector of value
  typedef std::vector<std::string> ValueVect;  

  /// Type defining the structure of an option
  struct OptionStruct {

    OptionStruct(){}

    OptionStruct(char shortName,
		 std::string argument,
		 bool mandatory,
		 bool argRequired,
		 std::string helpText,
     std::string defaultValue = "") :
      shortName_(shortName),
      argument_(argument),
      mandatory_(mandatory),
      argRequired_(argRequired),
      helpText_(helpText),
      defaultValue_(defaultValue)
    {}
    /// option short name
    char   shortName_;   
    /// argument name used for the help display
    std::string argument_;
    /// mandatory option flag
    bool   mandatory_;
    /// argument required flag
    bool   argRequired_;
    /// help text describing the option
    std::string helpText_;
    /// values associated to this option
    ValueVect availValues_;
    /// default value
    std::string defaultValue_;
  };
  /// Type defining a map of options with their data
  typedef std::map<std::string,OptionStruct> OptionMap;

  /// Constructor : sets the number of arguments along with a list of them
  /** 
   * \param argc number of elements 
   * \param argv List of arguments
   */
  GenConfig(int argc,
	    const char* argv[]);

  /// Destructor
  virtual ~GenConfig();

  /// copy constructor
  GenConfig(const GenConfig& rgenConfig);

  /// assignment operator
  GenConfig& operator = (const GenConfig& rgenConfig);

  /**
   * Parses the command line arguments and matches options with arguments
   *
   * By default it is allowed to log messages; when set to false all
   * errors/info/warning messages are suppressed.
   *
   * returns 0 if successful, -1 otherwise.
   */
  int parseCommandLine(bool loggingAllowed = true) ;


  /// \brief displays a help message relying on the option map
  /// returns 0 if successful, -1 otherwise
  int displayHelp() const;

  /// \brief displays the command line that launched the program
  /**
   * returns 0 if successful, -1 otherwise
   */
  int displayCommandLine();

protected :
  /// map of the available options
  OptionMap optionMap_;
  /// List of non optional arguments
  ValueVect argumentVect_;

private :
  /// number of elements in the configuration line
  int argc_;   
  /// configuration line
  char**  argv_; 
};
#endif  // GEN_CONFIG_H
