// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observator
// LICENSE:   ESO internal use only
//
// PROJECT:   VLT Data Flow Syste
// AUTHOR:    Christophe Moins- ESO/SDD/DF
// SUBSYSTEM: arcsys
//
// $Revision: 1.12 $
// $Log: GenObsProg.h,v $
// Revision 1.12  2009/06/15 12:30:10  vforchi
// Removed daemon mode, added reduced data handling
//
// Revision 1.11  2008/09/11 08:31:07  vforchi
// fixed compilation problems with gcc 4.2.4
//
// Revision 1.10  2007/08/14 07:37:47  cmoins
// DFS04052 : available replaced by retrieved
//
// Revision 1.9  2007/07/12 08:16:25  cmoins
// removal of useless comment
//
// Revision 1.8  2007/07/10 08:55:39  cmoins
// removal of the previously used instruments lists when fetching
// calibration files
//
// Revision 1.7  2007/07/06 14:44:41  cmoins
// DFS04087 : file location checking of files already handled will not
// be performed twice.
//
// Revision 1.6  2007/07/04 11:58:07  cmoins
// a specific message is displayed when all the retrieved files are already available in a package
//
// Revision 1.5  2007/05/24 11:38:24  cmoins
// previously retrieved files are taken into account when looking for package files
//
// Revision 1.4  2007/05/04 12:31:17  cmoins
// removal of storeData, updateProgDB, getMaxTime implementation
//
// Revision 1.3  2007/05/03 08:19:26  cmoins
// use of OMS_MS constant defined in DPPackerDef
//
// Revision 1.2  2007/04/26 12:57:01  cmoins
// adding of a StorageService template parameter
//
// Revision 1.1  2007/04/25 15:57:41  cmoins
// *** empty log message ***
//
// 
//------------------------------------------------------------------------------

#ifndef GENOBSPROG_H
#define GENOBSPROG_H

#include <string>
#include <sstream>
#include <iomanip>
#include <DfsLog.h>
#include <math.h>
#include <boost/tokenizer.hpp>
#include "ObsProg.h"
#include "RawDataProduct.h"

#include "ReducedDataService.h"

using namespace std;

/// General program template class.
/**
 * A general program contains science and calibration files. It will use an
 * ObservationService template parameter to remain independent of the data
 * retrieval layer. 
 */
template <class ObservationService,class StorageService>
class GenObsProg : public ObsProg<StorageService>
{

public :

#if defined UNIT_TEST
  friend class TestGenObsProg;
  friend class TestObsProgPackage;
#endif 

  /// \brief Constructor
  /**
   * \param progId program identifier
   * \param connAlias connAlias data repository connection parameter
   */
  GenObsProg(string progId,
	     string connAlias):
    ObsProg<StorageService>(progId,connAlias)
  {}

  /// \brief Destructor
  ~GenObsProg()
  {}

  /// \brief fetches a list of files associated to the program
  /**
   * \param startTime start time of the files retrieval
   * \param startTimeAdjustment correction to apply on start time for 
   *                            calibration files retrieval
   * \param endTime end time of the files retrieval
   * \param endTimeAdjustment   correction to apply on end time for 
   *                            calibration files retrieval
   * \return 0 if successful, -1 otherwise 
   */
  virtual int fetchData(double  startTime,
			double  startTimeAdjustment,
			double  endTime,
			double  endTimeAdjustment,
			const MediaFile::ResultSet& rpreviousFileSet,
			vector<string> reducedDataLocations);

  /// \brief instantiates a copy of the object
  /// returns the address of the copy
  virtual ObsProg<StorageService>* clone() const;
  
};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Fetches a list of files associated to the program
//------------------------------------------------------------------------------
template <class ObservationService,class StorageService>
int
GenObsProg<ObservationService,StorageService>::
fetchData(double startTime,
	  double  startTimeAdjustment,
	  double endTime,
	  double  endTimeAdjustment,
	  const MediaFile::ResultSet& rpreviousFileSet,
          vector<string> reducedDataLocations)
{
  typedef boost::tokenizer<boost::char_separator<char> > tokenizer;

  int retStatus = 0;
  string connAlias = ObsProg<StorageService>::getAlias();
  ObservationService observationService(connAlias);  

  int nbCalibFiles=0;
  
  DFS_TRACE("GenObsProg<ObservationService>::fetchData");

  RawDataProduct::ResultSet dataFileSet;
  string progId = ObsProg<StorageService>::getProgId();
  int nbScienceFiles = 
    observationService.findScienceFiles(progId,
					startTime,
					endTime,
					dataFileSet);
  if (nbScienceFiles >= 0)
    {
      double minTime=0;
      double maxTime=0;
      string obsProgUsedInstrument;
      for (RawDataProduct::pRawDataProduct  pdataFile = dataFileSet.begin();
	   pdataFile != dataFileSet.end();pdataFile++)
	{
	  // looking for min and max times for time refining afterwards
	  if (pdataFile == dataFileSet.begin())
	    {
	      minTime = pdataFile->mjdObs_;
	      maxTime = pdataFile->mjdObs_;
	    }
	  else if (pdataFile->mjdObs_ < minTime)
	    minTime = pdataFile->mjdObs_;
	  else if (pdataFile->mjdObs_ > maxTime)
	    maxTime = pdataFile->mjdObs_;

	  // Getting the list of used instruments
	  if (obsProgUsedInstrument.find(pdataFile->insId_) == string::npos)
	    {
	      if (!obsProgUsedInstrument.empty())
		obsProgUsedInstrument += ",";
	      obsProgUsedInstrument += pdataFile->insId_;
	    } 
	}
      if (!obsProgUsedInstrument.empty())
	{
	  // some sciences files have been retrieved for this block.
	  // with the list of the instruments they are associated to,
	  // calibration files are retrieved using the list of files
	  // they are associated to.
	  // Proceeds to time refining 
	  // if the time adjustment is below one millisecond, 
	  // no adjustment is performed and the same time as the calibration
	  // files is used. 
	  if (fabs(startTimeAdjustment) > DPPackerDef::ONE_MS)
	    startTime = minTime + startTimeAdjustment;
	  if (fabs(endTimeAdjustment) > DPPackerDef::ONE_MS)
	    endTime = maxTime + endTimeAdjustment;
	  nbCalibFiles = 
	    observationService.findCalibFilesByInst(obsProgUsedInstrument,
						    startTime,
						    endTime,
						    dataFileSet);
	}
      else
	{ 
	  // No science files was found, only calibration files associated
	  // to this obs id are retrieved.
	  startTime = startTime + startTimeAdjustment;
	  endTime = endTime + endTimeAdjustment;
	  nbCalibFiles = 
	    observationService.findCalibFiles(progId,
					      startTime,
					      endTime,
					      dataFileSet);
	}	

    }

    // retrieve reduced files associated to this raw files
    ReducedDataService *rds = new ReducedDataService();
    rds->findReducedFiles(progId,
              startTime,
              endTime,
              dataFileSet,
              reducedDataLocations);
   // exclusion of files already retrieved
  RawDataProduct::pRawDataProduct pnextDataFile;
  for (RawDataProduct::pRawDataProduct pdataFile = dataFileSet.begin();
       pdataFile != dataFileSet.end();)
    {
      pnextDataFile = pdataFile;
      pnextDataFile++;
      for (MediaFile::pconstMediaFile pmediaFile = rpreviousFileSet.begin();
	   pmediaFile != rpreviousFileSet.end();pmediaFile++)
	{
	  
	  if (pmediaFile->fileId_ == pdataFile->dpId_)
	    {
	      DFS_DEBUG("File %s already retrieved",pdataFile->dpId_.c_str());
	      dataFileSet.erase(pdataFile);
	      break;
	    }
	}
      pdataFile = pnextDataFile;
    }
  if (!dataFileSet.empty())
    {
      DFS_INFO("Files retrieval succeeded for %s",progId.c_str());
      ObsProg<StorageService>::setDataFileSet(dataFileSet);
    }
  else
    {
      if ((nbScienceFiles >= 0) || (nbCalibFiles >= 0))
	DFS_INFO("Files already retrieved for %s",progId.c_str());
      else
	{
	  DFS_ERROR("Files retrieval failed for %s",progId.c_str());
	  retStatus = -1;
	}
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// instantiates a copy of the object
//------------------------------------------------------------------------------
template <class ObservationService,class StorageService>
ObsProg<StorageService>*
GenObsProg<ObservationService,StorageService>::
clone()  const
{
  return new GenObsProg<ObservationService,StorageService>(*this);
}
#endif /* GENOBSPROG_H */
