// -*-c++-*-

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Jens Knudstrup - ESO/DMD/DFS
// SUBSYSTEM: DFS Archive System
//
// $Name: arcsys-1_2_8 $
// $Revision: 1.13 $
// $Log: GenTask.h,v $
// Revision 1.13  2007/06/01 07:23:45  jknudstr
// Changes in connection with testing/finalizing of SA->PA Rep Agent.
//
// Revision 1.12  2007/05/03 15:32:07  jknudstr
// Intermediate check-in while developing the SA->PA Archive Replicator.
//
// Revision 1.11  2007/01/11 15:35:10  jknudstr
// -
//
// Revision 1.10  2006/08/08 15:29:22  jknudstr
// -
//
//-----------------------------------------------------------------------------

#ifndef GENTASK_H
#define GENTASK_H

// Allowed when using ACE? Problem is that it is not possible to use ACE_Mutex
// as the value in an STL map.
#include <pthread.h>

// External packages includes
#include "ace/Task.h"

// Local includes
#include <RepAgent.h>
#include <RepReqMsg.h>



/** Queue used to pass messages between the Replication Agent ACE Tasks. 
 *
 * Message queue type definition used within the framework of the
 * Replication Agent instances. 
 */
typedef ACE_Message_Queue<ACE_MT_SYNCH> MsgQueue;


/** Base class for the Replication Agent Task classes.
 *
 * Base class for all task class types within the context of the Replication
 * Agent project.
 */
class GenTask: public ACE_Task<ACE_MT_SYNCH>
{
  public:
    /// Constructor registering common instance of the Replication Agent Class
    /// and the output message queue to use for this task.
    GenTask(const string  taskId,
	    RepAgent&     repAgent,
	    MsgQueue&     msgOutputQueue);

    /// Destructor method.
    ~GenTask(void);

    /// Return ID of task.
    const string& taskId(void);

    /// Activate the task. The number of threads to be spawned should be given.
    /// Note: The subclasses of this class should specify in how many
    ///       instances (number of threads) they can be running.
    void serve(const int noOfThreads);

    /// The server loop. Implements persistent looping.
    /// Persistant looping here refers to a loop that doesn't break even if
    /// an exception is raised within the code executed in the loop.
    void srvLoop(const float loopSleep = 0);

    /// The business logic of the server function.
    virtual void srvBusiness(void) {return;};

    /// The server initialization loop. Implements persistent looping.
    /// Persistant looping here refers to a loop that doesn't break even if
    /// an exception is raised within the code executed in the loop.
    void initLoop(void);

    /// The business logic of the server initialization.
    virtual void initBusiness(void) {return;};

    /// Return reference to the internal Input Message Queue.
    MsgQueue& msgInputQueue(void);

    /// Get reference to associated Replication Agent Class.
    RepAgent& repAgent(void);

    /// Returns true if the queue contains message(s) to be picked up.
    bool msgInQueue(void);

    /// Get message from the associated input message queue if available. If a
    /// message was obtained, true is returned, otherwise false.
    bool getFromQueue(RepReqMsg*& msg);

    /// Put a message object back into the task's input queue.
    void putBackInQueue(RepReqMsg& msg);
    
    /// Put a message in the associated output message queue.
    void putInOutputQueue(RepReqMsg& msg);

    /// Get reference to Replication Request map.
    map<string, RepReqMsg*>& getRepReqMap(void);

    /// Put keys for all Replication Requests in queue in the provided vector.
    GenTask&  getRepReqKeys(vector<string>& repReqKeys);

    /// Factory method that creates new Replication Request Message objects.
    /**
     * Factory method that creates a new Replication Request Message object and
     * registers it in the internal map.
     *
     * \param repReq Reference to the ReplicationRequest instance containing
     * information about the Replication Request.
     *
     * \return Pointer to the new RepReqMsg instance.
     */
    RepReqMsg*  newRepReq(ReplicationRequest& repReq);

    /// Method to insert a Replication Request object into the local map.
    GenTask& insertRepReq(RepReqMsg&  repReq);

    /// Get pointer to RepReqMsg object associated with given key.
    /**
     * Get pointer RepReqMsg object associated with given key. If not found,
     * NULL is returned.
     *
     * \param key Name of key for the Replication Request.
     *
     * \return Pointer to the RepReqMsg associated with the key, if found.
     */
    RepReqMsg* getRepReq(const string&  key);

    /// Remove the RepReqMsg entry from the internal RepAgent repository.
    /**
     * Remove the RepReqMsg entry from the internal RepAgent repository.
     * This should normally only be done by the Report Generation Task
     * (RepGenTask).
     *
     * \param key Name of key for the Replication Request.
     *
     * \return Reference to object itself.
     */
    GenTask&  removeRepReqEntry(const string&  key);

  private:
    string                   taskId_;
    RepAgent*                repAgent_;
    MsgQueue*                outputQueue_;
    map<string, RepReqMsg*>  repReqs_;

    // Used to control the initialization, such that this is only executed
    // once for each group of task.
    // Note, there is one instance of the task classes per type of classes
    // even if several instances of that type of class is running.
    ACE_Mutex         initMutex_;
    bool              initFlag_;
};

#endif // GENTASK_H

// EOF
