/// -*-c++-*-

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Christophe Moins- ESO/SDD/DFS
// SUBSYSTEM: arcsys - Archive Systems
//
// $Revision: 1.8 $
// $Log: HdrHandler.h,v $
// Revision 1.8  2008/05/08 10:04:36  cmoins
// Ticket DFS05348 : only primary header option available when creating hdr
// file + all options become lower case
//
// Revision 1.7  2008/04/15 15:02:34  cmoins
// adding of createImageExtension
//
// Revision 1.6  2008/03/03 15:27:33  cmoins
// Changing of the command line tool interface
//
// Revision 1.5  2008/02/15 11:15:03  cmoins
// adding of options
//
// Revision 1.4  2008/01/09 13:51:46  cmoins
// use of dbrc file to retrieve database connection parameters
//
// Revision 1.3  2008/01/08 12:48:20  cmoins
// database parameters become configurable
//
// Revision 1.2  2008/01/04 15:17:48  cmoins
// no hdr file is created when no data is retrieved in the database for a
// given data product id.
//
// Revision 1.1  2007/12/14 16:40:22  cmoins
// first version
//
// Revision 1.1  2007/12/14 15:58:20  cmoins
// first "operation version" of the header of the fly tool
//
//
// 
//------------------------------------------------------------------------------


#ifndef HDR_HANDLER_H
#define HDR_HANDLER_H

#include <string>
#include "FrameHandler.h"

namespace headerOnTheFly {

  /**
   *  Implementation of a hdr file handler. Its main task will be to 
   *  substitute records inside the header of all the extensions of a hdr file
   *  by database records.
   */ 
  class HdrHandler : public FrameHandler
  {
  public:

    /// Default constructor
    HdrHandler(){}
    /**
     * Constructor
     * Gets the name of the  hdr file to be created
     * \param hdrFileName name of the hdr file to handle
     * \param dbServer database server
     * \param dbUser database user
     * \param dbPasswd user passwd
     * \param onlyPrimExt only primary extension handling flag
     */
    HdrHandler(std::string hdrFileName,
	       std::string dbServer,
	       std::string dbUser,
	       std::string dbPasswd,
	       bool onlyPrimExt);

    /**
     * Destructor 
     * Removes the temporary fits file
     */
    ~HdrHandler();

    /**
     * Substitutes the records of all extensions headers by a set of record
     * coming from the database. It assumes that the database records are 
     * sorted according to their extension and keyword identifiers
     * returns 0 if successful, -1 otherwise
     */
    virtual int substituteExtHeader() ;

    /**
     * returns the successful message of the frame
     */
    virtual std::string getSuccessMessage();

    /**
     * returns the error message of the frame
     */
    virtual std::string getErrorMessage();

  private :

    /// \Brief creates a hdr file from the HDU headers of a fits file
    int createHdrFile();

    /**
     * Creates an image extension inside a fits file
     * \param fptr pointer to the fits file where the extension is created
     * \param pkeyword iterator on the keyword repository values
     * \param endIterator end iterator
     * \param fits operations status 
     */
    void createImageExtension(fitsfile *fptr,
			      KeywordRep::ResultSet::iterator pkeyword,
			      KeywordRep::ResultSet::iterator endIterator,
			      int* pfitsStatus);

    /// id of the hdr file to handle
    std::string hdrFileId_;
    /// name of the temporary fits file that will be used to build the hdr file
    std::string tempFitsFile_;
    /// name of the hdr file to be created   
    std::string hdrFileName_;
    /// flag that indicates that only the primary header should be replaced
    bool onlyPrimary_;

  };
}
#endif // HDR_HANDLER_H
