// -*-c++-*-

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Christophe Moins- ESO/SDD/DFS
// SUBSYSTEM: arcsys - Archive Systems
//
// $Revision: 1.6 $
// $Log: MediaActionDAO.h,v $
// Revision 1.6  2007/07/12 08:14:51  cmoins
// media actions are foun only according to their media request id
// no control is done in insert to check that a media action is unique
//
// Revision 1.5  2007/07/04 11:42:26  cmoins
// insert and delete return the number of handled records
// + a debug message is issued if no record was handled
//
// Revision 1.4  2007/05/24 11:34:08  cmoins
// two media actions becomes different if the media request id and the process id are different
//
// Revision 1.3  2007/05/15 09:57:16  cmoins
// replacement of info logging messages by debug ones
//
// Revision 1.2  2007/04/27 15:41:12  cmoins
// logging message correction
//
// Revision 1.1  2007/04/27 13:24:19  cmoins
// DAO class to insert, retrieve or delete in the DB actions associated to media request
//
// 
//------------------------------------------------------------------------------

#ifndef MEDIA_ACTION_DAO_H
#define MEDIA_ACTION_DAO_H

#include <string>
#include <sstream>
#include <DfsLog.h>
#include <sys/types.h>
#include <unistd.h>
#include "dataAccessDef.h"
#include "MediaAction.h"

using namespace std;

/// Data Access Object associated to MediaAction objects.
/**
 * This class allows to insert, retrieve or delete in a table any information 
 * related on actions being performed associated to a media creation request.
 */

template <class ConnectionHandle> class MediaActionDAO
{
public:
 
  /// ingestion action value in the database
  static const string INGESTION_ACTION;

  /// \brief Constructor that sets the media request action table
  /// \param mediaTable media request table
  MediaActionDAO(string mediaActionTable):
    mediaActionTable_(mediaActionTable)
  {}

  /// \brief Destructor
  MediaActionDAO(){}
  
  /// \brief insert a media action  in the database.
  /**
   * \param rmediaAction media request action to be inserted
   * \param rconnectionHandle data repository connection handle
   * \return the number of inserted rows if successful, -1 otherwise
   */
  int insert(MediaAction& rmediaAction,
	     const ConnectionHandle& rconnectionHandle) const;

  /// \brief looks if a media action is associated to a request
  /**
   * \param mediaId identifier of the media request
   * \param rmediaAction  media action data
   * \param rconnectionHandle data repository connection handle
   * \returns the number of retrieved media action if successful, -1 otherwise
   */
  int findMediaAction(int mediaId,
		      MediaAction::ResultSet& rmediaActionSet,
		      const ConnectionHandle& rconnectionHandle) const;

  /// \brief deletes a media action from the database
  /**
   * \param mediaRequestId identifier of the media request
   * \param rconnectionHandle data repository connection handle
   * \returns the number of deleted rows if successful, -1 otherwise
   */
  int deleteMediaAction(MediaAction& rmediaAction,
			const ConnectionHandle& rconnectionHandle) const;
 
private:

  /// media id column in the table
  static const string MEDIA_ID_FIELD;
  /// media unit name column in the table
  static const string MEDIA_UNIT_NAME_FIELD;
  /// action column in the table
  static const string ACTION_FIELD;
  /// time column in the table
  static const string TIME_FIELD;
  /// process id column in the table
  static const string PROCESS_ID_FIELD;

  /// table storing data about media requests action
  string mediaActionTable_; 
};

template <class ConnectionHandle> const string 
MediaActionDAO<ConnectionHandle>::INGESTION_ACTION = "I";

template <class ConnectionHandle> const string 
MediaActionDAO<ConnectionHandle>::MEDIA_ID_FIELD = "media_request_id";

template <class ConnectionHandle> const string 
MediaActionDAO<ConnectionHandle>::MEDIA_UNIT_NAME_FIELD = "media_unit_name";

template <class ConnectionHandle> const string 
MediaActionDAO<ConnectionHandle>::ACTION_FIELD = "action";

template <class ConnectionHandle> const string 
MediaActionDAO<ConnectionHandle>::TIME_FIELD = "time";

template <class ConnectionHandle> const string 
MediaActionDAO<ConnectionHandle>::PROCESS_ID_FIELD = "process_id";


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// insert informations related to a media request creation inside the database
// Return 0 if the operation was successful, -1 otherwise
//------------------------------------------------------------------------------
template <class ConnectionHandle>
int
MediaActionDAO<ConnectionHandle>::
insert(MediaAction& rmediaAction,
       const ConnectionHandle& rconnectionHandle) const
{
  int retStatus = 0;

  DFS_TRACE("MediaActionDAO<ConnectionHandle>::insert");

  rmediaAction.processId_ = getpid(); 

  //No media action is associate to the media for this process
  string insertInto = "insert into " + mediaActionTable_ + " ";
  string columnList = "(" + MEDIA_ID_FIELD + "," + ACTION_FIELD + "," 
    + TIME_FIELD + "," + PROCESS_ID_FIELD;

  if (rmediaAction.mediaUnitName_ != dataAccessDef::NO_INIT_STRING_FIELD)
    columnList += "," + MEDIA_UNIT_NAME_FIELD;

  columnList += ")";

  // It is considered that the time the action was started corresponds
  // more or less to the insertion time in the database.
  stringstream insertValues;
 
  insertValues << "values(" << rmediaAction.mediaRequestId_ 
	       << ",'" << rmediaAction.action_ << "'"
	       <<  ",getdate()" 
	       << "," << rmediaAction.processId_;

  if (rmediaAction.mediaUnitName_ != dataAccessDef::NO_INIT_STRING_FIELD)
    insertValues << ",'" << rmediaAction.mediaUnitName_ << "'";

  insertValues << ")";

  string sqlStatement = insertInto + columnList + insertValues.str();
  typename ConnectionHandle::ResultSet res =  
    rconnectionHandle.execute(sqlStatement);

  if (res.commandSucceeded())
    {
      retStatus = res.getRowCount();
      //mediaRequest data insertion in the database was successful 
      DFS_DEBUG("MediaAction creation successful for %d",
		rmediaAction.mediaRequestId_);
    }
  else
    {
      DFS_ERROR("Sql request %s failed",sqlStatement.c_str());
      retStatus = -1;
    }
  return retStatus;
} 

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// retrieves the data of a media action from a media request id
//------------------------------------------------------------------------------
template <class ConnectionHandle>
int
MediaActionDAO<ConnectionHandle>::  
findMediaAction(int mediaId,
		MediaAction::ResultSet& rmediaActionSet,
		const ConnectionHandle& rconnectionHandle) const
{
  int retStatus = 0;
  string mediaUnitName;
  string action;
  string time;
  int processId;
  stringstream mediaIdStr;

  typename ConnectionHandle::ResultSet res;
  typename ConnectionHandle::ListOfFieldIndex lofi;

  // SQL statement building
  string select = "select  " + MEDIA_ID_FIELD + "," + MEDIA_UNIT_NAME_FIELD
    + "," + ACTION_FIELD + "," + TIME_FIELD + "," + PROCESS_ID_FIELD + " ";
 
  string from   =  "from " + mediaActionTable_ + " ";
  mediaIdStr << mediaId;
  string where  =  "where " + MEDIA_ID_FIELD + "=" + mediaIdStr.str();
  string sqlStatement  = select + from + where;

  if (rconnectionHandle.executeSelect(sqlStatement,
				      res,
				      lofi) != -1)
    {
      bool hasResultSet = res.hasResultSet();     
      if (hasResultSet)
	{
	  while(res.next())
	    {
	      retStatus++;
	      mediaUnitName = 
		rconnectionHandle.getStringValue(res,
						 lofi[MEDIA_UNIT_NAME_FIELD]);
	      action = 
		rconnectionHandle.getStringValue(res,
						 lofi[ACTION_FIELD]);
	      time = 
		rconnectionHandle.getDateStringValue(res,
						     lofi[TIME_FIELD]);
	      processId = 
		rconnectionHandle.getLongValue(res,
					       lofi[PROCESS_ID_FIELD]);
	      rmediaActionSet.push_back(MediaAction(mediaId,
						    mediaUnitName,
						    action,
						    time,
						    processId));
            }
	}
      DFS_DEBUG("%d Media action data for request %d successfully retrieved",
		retStatus,mediaId);
    }
  else
    {
      DFS_ERROR("Sql request %s failed",sqlStatement.c_str());
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// deletes a media action associated to a media request id
//------------------------------------------------------------------------------
template <class ConnectionHandle>
int
MediaActionDAO<ConnectionHandle>::  
deleteMediaAction(MediaAction& rmediaAction,
		  const ConnectionHandle& rconnectionHandle) const
{
  int retStatus = 0;
  stringstream mediaIdStr;
  stringstream processIdStr;

  rmediaAction.processId_ = getpid();
  mediaIdStr << rmediaAction.mediaRequestId_;
  processIdStr << rmediaAction.processId_;

  // SQL statement building
  string sqlDelete = "delete  ";
  string from   =  "from " + mediaActionTable_ + " ";
  string where  =  "where " + MEDIA_ID_FIELD + "=" + mediaIdStr.str()
    + " AND " + PROCESS_ID_FIELD + "=" + processIdStr.str();

  string sqlStatement  = sqlDelete + from + where;

  typename ConnectionHandle::ResultSet res = 
    rconnectionHandle.execute(sqlStatement);
  retStatus = res.getRowCount();
  if (res.commandSucceeded())
    {
      DFS_DEBUG("%d media request action successfully deleted", 
		rmediaAction.mediaRequestId_);
    }
  else
    {
      DFS_ERROR("Sql request %s failed",sqlStatement.c_str());
      retStatus = -1;
    }

  return retStatus;
}

#endif // MEDIA_ACTION_DAO_H
// -*-++-*
