// -*-c++-*-

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Christophe Moins- ESO/SDD/DFS
// SUBSYSTEM: arcsys - Archive Systems
//
// $Revision: 1.7 $
// $Log: MediaFileDAO.h,v $
// Revision 1.7  2009/06/15 13:34:47  vforchi
// Fixed DFS7081
//
// Revision 1.6  2007/08/01 16:13:05  cmoins
// adding of suspend and restore files functionalities
//
// Revision 1.5  2007/07/25 07:27:51  cmoins
// doxygen comments correction
//
// Revision 1.4  2007/07/05 13:51:53  cmoins
// DFS04061 : adding of a delete files method
//
// Revision 1.3  2007/07/04 12:21:49  cmoins
// returns the number of inserted rows
//
// Revision 1.2  2007/05/15 15:31:00  cmoins
// program id is not inserted in the data repository
//
// Revision 1.1  2007/04/25 15:10:44  cmoins
// creation of DAO class associated to the MediaFile structure
//
// 
//------------------------------------------------------------------------------

#ifndef MEDIA_FILE_DAO_H
#define MEDIA_FILE_DAO_H

#include <string>
#include <iostream>
#include <sstream>
#include <DfsLog.h>
#include "MediaFile.h"

using namespace std;


/// Data Access Object associated to MediaFile objects.
/**
 * This class allows to insert or retrieve media files related informations 
 * into the asto database. It will rely on a table storing informations
 * about files stored on a media.
 */
template <class ConnectionHandle> class MediaFileDAO
{
public:
 
  /// \brief Constructor that sets the media_file table
  /// \param mediaFileTable stores data about files stored on a media
  MediaFileDAO(string mediaFileTable);

  /// \brief Destructor
  ~MediaFileDAO(){}
  
  /// \brief inserts a file inside the data repository
  /**
   * \param rdataFile file to be inserted
   * \param rconnectionHandle data repository connection handle
   * \return the number of inserted rows if successful, -1 otherwise 
   */
  int insert(const MediaFile& rdataFile,
	     const ConnectionHandle& rconnectionHandle) const;

  /// \brief retrieves a set of files associated to a media id.
  /**
   * \param mediaId id of the media whose files shall be retrieved
   * \param rdataFileSet set of files associated to the media
   * \param rconnectionHandle data repository connection handle
   * \return the number of retrieved files if successful, -1 otherwise 
   */
  int findMediaFiles(int mediaId,
		     MediaFile::ResultSet& rdataFileSet,
		     const ConnectionHandle& rconnectionHandle) const;

  /// \brief removes a file belonging to a specific media request
  /// from the data repository 
  /**
   * \param mediaId id of the media request
   * \param mediaFileId id of the file
   * \param rconnectionHandle data repository connection handle
   * \return the number of deleted files, -1 otherwise
   */
  int deleteFile(int mediaId,
		 string mediaFileId,
		 const ConnectionHandle& rconnectionHandle) const;

  /// \brief suspends a file belonging to a specific media request
  /**
   * \param mediaId id of the media request
   * \param mediaFileId id of the file
   * \param rconnectionHandle data repository connection handle
   * \return the number of suspended files, -1 otherwise
   */
  int suspendFile(int mediaId,
		  string mediaFileId,
		  const ConnectionHandle& rconnectionHandle) const;

  /// \brief restores a file belonging to a specific media request
  /**
   * \param mediaId id of the media request
   * \param mediaFileId id of the file
   * \param rconnectionHandle data repository connection handle
   * \return the number of restored files, -1 otherwise
   */
  int restoreFile(int mediaId,
		  string mediaFileId,
		  const ConnectionHandle& rconnectionHandle) const;

private:

  /// Set of fields used in the query
  static const string MEDIA_ID_FIELD;
  static const string MEDIA_REQUEST_NAME_FIELD;
  static const string FILE_ID_FIELD;
  static  const string FILENAME_FIELD; 
  static  const string DATE_NOTIFIED_FIELD;
  static const string SOURCE_ID_FIELD;
  static const string COMPRESSION_FLAG_FIELD;
  static const string FILESIZE_FIELD;
  static const string UNCOMP_FILESIZE_FIELD;
  static const string STATUS_FIELD;
  static const string NO_SOURCE_ID;
  
  /// table storing data on media files
  string mediaFileTable_; 
};

// constant creation
template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::MEDIA_ID_FIELD =  "media_request_id";

template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::MEDIA_REQUEST_NAME_FIELD = "media_request_name";

template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::FILE_ID_FIELD = "file_id";

template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::FILENAME_FIELD = "filename";

template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::DATE_NOTIFIED_FIELD = "date_notified";

template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::SOURCE_ID_FIELD = "source_id";

template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::COMPRESSION_FLAG_FIELD = "compression_flag";

template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::FILESIZE_FIELD = "filesize";

template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::UNCOMP_FILESIZE_FIELD = "uncomp_filesize";

template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::STATUS_FIELD = "status";

template <class ConnectionHandle> const string 
MediaFileDAO<ConnectionHandle>::NO_SOURCE_ID = "";

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Constructor
//------------------------------------------------------------------------------
template <class ConnectionHandle>
MediaFileDAO<ConnectionHandle>::MediaFileDAO(string mediaFileTable):
  mediaFileTable_(mediaFileTable)
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Inserts the file given as a parameter in the data repository
//------------------------------------------------------------------------------
template <class ConnectionHandle>
int
MediaFileDAO<ConnectionHandle>::
insert(const MediaFile& rdataFile,
       const ConnectionHandle& rconnectionHandle) const
{
  int retStatus = 0;
  string sqlStatement;
  typename ConnectionHandle::ResultSet res;
  
  DFS_TRACE("MediaFileDAO<ConnectionHandle>::insert");

  string insertInto = "insert into " + mediaFileTable_ + " ";
  string columnList =
    "("+ FILE_ID_FIELD + "," + FILENAME_FIELD + "," + MEDIA_ID_FIELD 
    + "," + DATE_NOTIFIED_FIELD + "," + SOURCE_ID_FIELD
    + "," + COMPRESSION_FLAG_FIELD + "," + FILESIZE_FIELD + "," 
    + UNCOMP_FILESIZE_FIELD + "," + STATUS_FIELD + ") ";
	 
  stringstream insertValues ;
  insertValues << "values('" << rdataFile.fileId_ << "'"
	       << ",'" << rdataFile.fileName_ << "'"
	       << "," << rdataFile.mediaRequestId_ 
      	       << ",getdate(),'" << NO_SOURCE_ID << "'"
	       << ",'" << rdataFile.compressionFlag_ << "'" 
       	       << "," << rdataFile.fileSize_ 
       	       << "," << rdataFile.uncompFileSize_
       	       << ",'" << rdataFile.status_ << "')";

  sqlStatement = insertInto + columnList + insertValues.str();
  res = rconnectionHandle.execute(sqlStatement);
  if (res.commandSucceeded())
    {
      retStatus = res.getRowCount();
    }
  else
    {
      DFS_ERROR("Error executing SQL command %s",
		sqlStatement.c_str());
      retStatus = -1;
    }
  return retStatus;
}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// retrieves a set of files associated to a media id
//------------------------------------------------------------------------------
template <class ConnectionHandle>
int
MediaFileDAO<ConnectionHandle>::  
findMediaFiles(int mediaId,
	       MediaFile::ResultSet& rdataFileSet,
	       const ConnectionHandle& rconnectionHandle) const
{
  int retStatus = 0;
  int nbFiles = 0;
  string currentFileId;
  string currentFileName;
  string currentStatus;
  typename ConnectionHandle::ResultSet res;
  typename ConnectionHandle::ListOfFieldIndex listOfFieldIndex;
  stringstream mediaIdStr;
  mediaIdStr << mediaId;

  DFS_TRACE("MediaFileDAO<ConnectionHandle>::findMediaFiles");

  // SQL statement building
  string select = "select  " + FILE_ID_FIELD + "," + FILENAME_FIELD + ","
    + STATUS_FIELD + " ";
  string from   =  "from " + mediaFileTable_ + " ";
  string where  =  "where " + MEDIA_ID_FIELD  + "=" + mediaIdStr.str();
  string sqlStatement = select + from + where;
  if (rconnectionHandle.executeSelect(sqlStatement,
				      res,
				      listOfFieldIndex) != -1)
    {
      bool hasResultSet = res.hasResultSet();
      if (hasResultSet)
	{
	  while( res.next())
	    {
	      currentFileId =  
                 rconnectionHandle.getStringValue(res,
						  listOfFieldIndex[FILE_ID_FIELD]);
	      currentFileName =  
                 rconnectionHandle.getStringValue(res,
						  listOfFieldIndex[FILENAME_FIELD]);
	      currentStatus =  
                 rconnectionHandle.getStringValue(res,
						  listOfFieldIndex[STATUS_FIELD]);
	      MediaFile fileToInsert(currentFileId,currentFileName,0,0,'U',
				     currentStatus[0]);
	      rdataFileSet.push_back(fileToInsert);
	      nbFiles++;
	    }
	}
      retStatus = nbFiles;
    }
  else 
    {
      DFS_ERROR("Could not retrieve any files");
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// removes a file from the data repository 
//------------------------------------------------------------------------------
template <class ConnectionHandle>
int
MediaFileDAO<ConnectionHandle>::  
deleteFile(int mediaId,
	   string mediaFileId,
	   const ConnectionHandle& rconnectionHandle) const
{
  int retStatus = 0;
  stringstream mediaIdStr;

  DFS_TRACE("MediaFileDAO<ConnectionHandle>::deleteFile");

  // SQL statement building
  string sqlDelete = "delete ";
  string from   =  "from " + mediaFileTable_ + " ";
  mediaIdStr << mediaId;
  string where  =  "where " + MEDIA_ID_FIELD + "=" + mediaIdStr.str()
    + " AND " + FILE_ID_FIELD  + "='" + mediaFileId + "'";
  string sqlStatement = sqlDelete + from + where;

  typename ConnectionHandle::ResultSet res = 
    rconnectionHandle.execute(sqlStatement);
  retStatus = res.getRowCount();
  if (res.commandSucceeded() && (retStatus > 0))
    {
      DFS_DEBUG("file with identifier %s successfully deleted from database",
		mediaFileId.c_str());
    }
  else if (retStatus == 0)
    {
      DFS_DEBUG("file with identifier %s not deleted from database",
		mediaFileId.c_str());
    }
  else 
    {
      DFS_ERROR("Sql request %s failed",sqlStatement.c_str());
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  suspends a file belonging to a specific media request
//------------------------------------------------------------------------------
template <class ConnectionHandle>
int
MediaFileDAO<ConnectionHandle>::  
suspendFile(int mediaId,
	   string mediaFileId,
	   const ConnectionHandle& rconnectionHandle) const
{
  const string  SUSPENDED_STATE = "S";
  int retStatus = 0;
  stringstream mediaIdStr;

  DFS_TRACE("MediaFileDAO<ConnectionHandle>::suspendFile");

  // SQL statement building
  string sqlUpdate = "update " + mediaFileTable_ + " ";
  string sqlSet = "set " + STATUS_FIELD + "='" + SUSPENDED_STATE + "'";
  mediaIdStr << mediaId;
  string where  =  "where " + MEDIA_ID_FIELD + "=" + mediaIdStr.str()
    + " AND " + FILE_ID_FIELD  + "='" + mediaFileId + "'";
  string sqlStatement = sqlUpdate + sqlSet + where;

  typename ConnectionHandle::ResultSet res = 
    rconnectionHandle.execute(sqlStatement);
  retStatus = res.getRowCount();
  if (res.commandSucceeded() && (retStatus > 0))
    {
      DFS_DEBUG("file with identifier %s successfully suspended",
		mediaFileId.c_str());
    }
  else if (retStatus == 0)
    {
      DFS_DEBUG("file with identifier %s not suspended",
		mediaFileId.c_str());
    }
  else 
    {
      DFS_ERROR("Sql request %s failed",sqlStatement.c_str());
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//  restores a file belonging to a specific media request
//------------------------------------------------------------------------------
template <class ConnectionHandle>
int
MediaFileDAO<ConnectionHandle>::  
restoreFile(int mediaId,
	   string mediaFileId,
	   const ConnectionHandle& rconnectionHandle) const
{
  const string  RESTORED_STATE = "R";
  int retStatus = 0;
  stringstream mediaIdStr;

  DFS_TRACE("MediaFileDAO<ConnectionHandle>::restoreFile");

  // SQL statement building
  string sqlUpdate = "update " + mediaFileTable_ + " ";
  string sqlSet = "set " + STATUS_FIELD + "='" + RESTORED_STATE + "'";
  mediaIdStr << mediaId;
  string where  =  "where " + MEDIA_ID_FIELD + "=" + mediaIdStr.str()
    + " AND " + FILE_ID_FIELD  + "='" + mediaFileId + "'";
  string sqlStatement = sqlUpdate + sqlSet + where;

  typename ConnectionHandle::ResultSet res = 
    rconnectionHandle.execute(sqlStatement);
  retStatus = res.getRowCount();
  if (res.commandSucceeded() && (retStatus > 0))
    {
      DFS_DEBUG("file with identifier %s successfully restored",
		mediaFileId.c_str());
    }
  else if (retStatus == 0)
    {
      DFS_DEBUG("file with identifier %s not restored",
		mediaFileId.c_str());
    }
  else 
    {
      DFS_ERROR("Sql request %s failed",sqlStatement.c_str());
      retStatus = -1;
    }
  return retStatus;
}

#endif // MEDIA_FILE_DAO
// -*-++-*
