// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Stefano Zampieri - ESO/DMD/DFS
// SUBSYSTEM: arcsys - Archive Systems
//
// $Revision: 1.7 $
// $Log: NgamsClient.h,v $
// Revision 1.7  2007/06/01 07:23:45  jknudstr
// Changes in connection with testing/finalizing of SA->PA Rep Agent.
//
// Revision 1.6  2007/05/10 15:34:23  jknudstr
// Testing ArcRepSA2PA ...
//
// Revision 1.5  2007/04/11 09:56:31  jknudstr
// Implemented uncompressing only for Ngas2LocalCopy Task.
//
// Revision 1.4  2006/08/10 07:46:22  jknudstr
// -
//
// Revision 1.3  2006/06/01 14:09:21  szampier
// added fileVersion, diskId, timeoutSecs to retrieve method
//
// Revision 1.2  2006/05/30 14:47:02  jknudstr
// Added NgamsStatus::outFile(), init of listerner in defaul constructor
// for NgamsClient.
//
// Revision 1.1  2006/05/17 16:28:56  szampier
// first check-in
//
// 
// 
// C++ wrapper classes for the NGAMS C Client Library
// 
//------------------------------------------------------------------------------

// TODO: Consider to remove all handling of server multiplexing to the NG/AMS
//       C-API. For instance remove the types server_t, server_list_t and
//       server_list_iterator.


#ifndef NGAMS_CLIENT_H
#define NGAMS_CLIENT_H

#include <string>
#include <sstream>
#include <vector>

#include "ngams.h"
#include "DfsLog.h"

// forward declaration
class NgamsClientListener;
class NgamsStatus;

/// C++ wrapper for Ngams C Client
/** The NgamsClient class can be used to send commands to an NGAMS server. */

class NgamsClient
{
public:
  /// Equivalent of std::pair<std::string, int>
  typedef std::pair<std::string, int> server_t;
  /// Equivalent of std::vector<std::pair<std::string, int>>
  typedef std::vector<server_t> server_list_t;
  /// Equivalent of std::vector<std::pair<std::string, int>>::iterator
  typedef std::vector<server_t>::iterator server_list_iterator;
  
  /// Default constructor
  NgamsClient() :
    listener_(0)
  {
    DFS_TRACE("NgamsClient default constructor");
  }
  
//   /// \brief Constructor with host/port
//   /// \param host the host name
//   /// \param port the port number
//   NgamsClient(const std::string& host,
// 	      const int port) :
//     listener_(0)
//   {
//     DFS_TRACE("NgamsClient host/port constructor");
//     server_t srv(host, port);
//     servers_.push_back(srv);
//   }
  
  /// \brief Constructor with server list (host:port,host:port,...)
  /// \param clientId  ID for the client, used to refer to this server list.
  /// \param srvList   List of servers to contact.
  NgamsClient(const std::string&  clientId,
	      const std::string&  srvList) :
    listener_(0)
  {
    DFS_TRACE("NgamsClient clientId/srvList constructor");
    setServerList(clientId, srvList);
  }
  
  /// \brief Set the internal list of NGAS servers.
  /// \param clientId  ID for the client, used to refer to this server list.
  /// \param srvList   List of servers to contact.
  void setServerList(const std::string&  clientId,
		     const std::string&  srvList)
  {
    DFS_TRACE("NgamsClient::setServerList");
    //servers_.clear();
    //server_t srv(srvList, -1);
    //servers_.push_back(srv);
    ngamsParseSrvListId(clientId.c_str(), srvList.c_str());
  }
  

//   /// \brief Add a server definition (host, port) to the list of servers
//   /// \param host the host name
//   /// \param port the port number
//   void addServer(const std::string& host,
// 		 const int port)
//   {
//     DFS_TRACE("NgamsClient::addServer");
//     server_t srv(host, port);
//     servers_.push_back(srv);
//   }
  
  /// \brief Set the listener
  /// \param listener the NGAMS client listener
  void addListener(NgamsClientListener *listener)
  {
    DFS_TRACE("NgamsClient::addListener");
    listener_ = listener;
  }
  
  /// \brief Return status of NGAMS server
  /// \return NgamsStatus object
  NgamsStatus getStatus(const std::string& clientId);
  
  /// \brief Retrieve a file from the NGAMS server
  /// \param fileId the file ID
  /// \param outdir the output directory
  /// \return NgamsStatus object
  NgamsStatus retrieve(const std::string& clientId,
		       const std::string& fileId,
		       const std::string& outdir);
  
  NgamsStatus retrieve(const std::string& clientId,
		       const std::string& fileId,
		       int fileVersion,
		       const std::string& diskId,
		       int timeoutSecs,
		       const std::string& outdir);

    /// \brief Send Restore Request to associated NGAS Servers.
    /// \param fileId The ID of the reference file.
    /// \param fileVersion Version of the reference file.
    /// \param sourceFile Name of data file to submit with the request.
    /// \return NgamsStatus object
    NgamsStatus restore(const std::string&  clientId,
			const std::string&  sourceFile,          
			const std::string&  fileId,
			const int           fileVersion,
			const int           timeoutSecs);
  
//   /// \brief Return a string representation of this object
//   std::string asString()
//   {
//     std::stringstream buf;
//     server_list_iterator iter;
//     for (iter=servers_.begin(); iter<servers_.end(); iter++)
//       buf << (*iter).first << ":" << (*iter).second << " ";
//     return buf.str();
//   }
  
private:
//   /// Get next server from the list of servers
//   server_t getNextServer()
//   {
//     DFS_TRACE("NgamsClient::getNextServer");
//     return servers_[0];
//   }
  
private:
    //server_list_t servers_;
  NgamsClientListener *listener_;
};


/// Represents a status message returned by the NGAMS server

class NgamsStatus
{
 public:
  /// Default constructor
  NgamsStatus();
  
  /// \brief Constructor with ngamsSTATUS
  /// \param s the ngamsSTATUS record
  NgamsStatus(ngamsSTATUS& s);
  
  /// \brief Get status
  /// \return the status
  int status() { return status_; }
  
  /// \brief Get status string
  /// \return the status string
  const std::string& statusString() { return statusString_; }
  
  /// \brief Get date
  /// \return the date as string
  const std::string& date() { return date_; }
  
  /// \brief Get host
  /// \return the host name
  const std::string& host() { return host_; }
  
  /// \brief Get message
  /// \return the message  
  const std::string& message() { return message_; }
  
  /// \brief Get state
  /// \return the state  
  const std::string& state() { return state_; }
  
  /// \brief Get sub-state
  /// \return the status
  const std::string& subState() { return subState_; }

  /// \brief Get version
  /// \return the version
  const std::string& version() { return version_; }

  /// \brief Get a string representation of this object
  /// \return a string representation of this
  std::string asString();  

  /// \brief  Set name of output file resulting from a retrieve request.
  /// \return void.
  void outFile(const std::string& outputFile) { outFile_ = outputFile; }

  /// \brief  Return name of output file resulting from a retrieve request.
  /// \return Output filename.
  const std::string& outFile(void) {return outFile_;}

 private:
  int status_;
  std::string statusString_;
  std::string date_;
  std::string host_;
  std::string message_;
  std::string state_;
  std::string subState_;
  std::string version_;

  std::string outFile_;
};


/// \brief Provides a call back model for executing
/// actions before or after a given NgamsClient command
/// is executed.

class NgamsClientListener
{
 public:
  /// \brief Call back method invoked after file retrieval
  /// \param filename the file name
  /// \return 0 if the operation was successful, -1 otherwise
  virtual int afterRetrieve(const std::string& filename,
			    std::string&       finalFilename) = 0;
};

/// \brief This listener is used to uncompress the files upon
/// retrieval from the archive

class UncompressListener : public NgamsClientListener
{
 public:
  /// \brief Uncompress file after retrieval
  /// \copydoc NgamsClientListener::afterRetrieve
  int afterRetrieve(const std::string& filename,
		    std::string&       finalFilename);
};

#endif
