// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Christophe Moins- ESO/DMD/DFS
// SUBSYSTEM: arcsys - Archive Systems
//
// $Revision: 1.4 $
// $Log: NgasFileDAOSyb.h,v $
// Revision 1.4  2006/12/19 09:41:38  cmoins
// removal of the database nane as a parameter
//
// Revision 1.3  2006/12/15 15:22:34  cmoins
// DAO & Service Design modification
//
// Revision 1.2  2006/12/14 13:55:15  cmoins
// cvs header adding
//
//
// Sybase Data Access Object for NgasFile objects
// 
//------------------------------------------------------------------------------

#ifndef NGAS_FILE_DAO_SYB_H
#define NGAS_FILE_DAO_SYB_H

#include <string>
#include <SybUtils.h>
#include "NgasFile.h" 
#include <vector>

class ServiceConnectionSyb;

/// Sybase Data Access Object for NgasFile objects
/**
 * This class can be used to access NgasFile objects stored
 * in a Sybase table.
 */
class NgasFileDAOSyb
{
public:
 
  /// \brief Constructor
  /// \param tableName the name of the ngas file table
  NgasFileDAOSyb( const std::string& tableName );
  
  /// \brief Insert the given NgasFile object
  /// \param rngasFile the replication request
  /// \param rconnectionHandle database connection handle
  /// \return the number of inserted objects (0 or 1), or -1 in case of error
  int insert(const NgasFile&             rngasFile,
             const ServiceConnectionSyb& rconnectionHandle);

  /// \Insert data in the ngas files table by fetching rows associated to 
  /// \rsrcdiskId, rfileId and fileVersion, setting for each of them the disk_id
  /// \field to rdestdiskId and the creation date to the current one and 
  /// \inserting the result into ngas_files.
  /// \param rsrcdiskId  : the source disk
  /// \param rfileId     : file id of the files to be retrieved
  /// \param fileVersion : file version of the file to be retrieved
  /// \param rdestdiskId : the destination disk
  /// \param rconnection database connection handle
  /// \return the number of inserted objects (0 or 1), or -1 in case of error
  int insert(const std::string&          rsrcdiskId,
             const std::string&          rfileId,
             int                         fileVersion,
             const std::string&          rdestdiskId,
             const ServiceConnectionSyb& rconnectionHandle);

  
  /// \brief Update the NgasFile object with given fileId
  /// \param fileId the file ID
  /// \param status the new status
  /// \param rconnectionHandle database connection handle
  /// \return the number of updated objects (0 or 1), or -1 in case of error
  int update(const std::string&          rfileId,
	     NgasFile::NgasFileStatus    status,
             const ServiceConnectionSyb& rconnectionHandle);
  
  /// \brief Update the NgasFile object with given fileId,fileVersion,diskId
  /// \param fileId the file ID
  /// \param fileVersion the file version
  /// \param diskId the disk ID
  /// \param status the new status
  /// \param rconnectionHandle database connection handle
  /// \return the number of updated objects (0 or 1), or -1 in case of error
  int update(const std::string&          rfileId,
	     int                         fileVersion,
	     const std::string&          rdiskId,
	     NgasFile::NgasFileStatus    status,
             const ServiceConnectionSyb& rconnectionHandle);
  
  /// \brief Remove the NgasFile object with given fileId
  /// \param fileId the file ID
  /// \param fileVersion the file version
  /// \param diskId the disk ID
  /// \param rconnectionHandle database connection handle
  /// \return the number of removed objects (0 or 1), or -1 in case of error
  int remove(const std::string&          rfileId,
	     int                         fileVersion ,
	     const std::string&          rdiskId ,
             const ServiceConnectionSyb& rconnectionHandle);

  /// \brief Retrieve from the DB the NgasFile object with given fileId
  /// \param[in] fileId the file ID
  /// \param[out] rngasFileSet the selected NgasFile objects
  /// \param rconnectionHandle database connection handle
  /// \return the number of retrieved objects, or -1 in case of error
  int find(const std::string&          rfileId, 
           NgasFile::ResultSet&                rngasFileSet,
           const ServiceConnectionSyb& rconnectionHandle);
  
  /// \brief Retrieve from the DB the NgasFile objects with given status
  /// \param[in] status the request status
  /// \param[out] rs the selected replication request objects
  /// \param rconnectionHandle database connection handle
  /// \return the number of retrieved objects, or -1 in case of error
  int find(NgasFile::NgasFileStatus    status, 
           NgasFile::ResultSet&        rngasFileSet,
           const ServiceConnectionSyb& rconnectionHandle);
 
  /// \brief Retrieve from the DB the NgasFile objects with a given 
  /// file id, disk id and version number.
  /// \param rfileId the file ID
  /// \param fileVersion the file version
  /// \param rdiskId the disk ID
  /// \param rngasFileSet set of retrieved files
  /// \param rconnectionHandle database connection handle
  /// \return the number of retrieved objects, or -1 in case of error
  int find(const std::string&          rfileId,
	   int                         fileVersion,
	   const std::string&          rdiskId,
           NgasFile::ResultSet&        rngasFileSet,
           const ServiceConnectionSyb& rconnectionHandle);
 
  /// \brief Retrieve all ReplicationRequest objects from the DB
  /// \param[out] rngasFileSet the selected ngas file objects
  /// \param rconnectionHandle database connection handle
  /// \return the number of retrieved objects, or -1 in case of error
  int loadAll(NgasFile::ResultSet&        rngasFileSet,
              const ServiceConnectionSyb& rconnectionHandle);
  

private:
  /// \brief Select NgasFile objects from the DB
  /// \param[out] rngasFileSet the selected ngas file objects
  /// \param[in] whereClause the optional where clause
  /// \param rconnectionHandle database connection handle
  /// \return the number of selected objects, or -1 in case of error
  int select(const ServiceConnectionSyb& rconnectionHandle,
             NgasFile::ResultSet& rngasFileSet, 
             const std::string& whereClause = "");
  
  /// \brief Return a comma separated list of all fields in the 
  /// ngas file table
  std::string allFields();
  
  /// \brief Convert a SybaseResultSet object to a NgasFile object
  /// \param[in] res the ResultSet
  /// \param[out] rngasFile the NgasFile
  /// \return 0 if the conversion was succesful, -1 otherwise
  int convertResultSetToNgasFile(SybaseResultSet&  res,
				 NgasFile&         rngasFile);

  /// \Executes the query given in parameter
  /// \Return the number of objects issued by the query execution
  /// \or -1 in case of error.
  /// \param query query to be executed
  /// \param rconnectionHandle database connection handle
  int executeQuery(std::string query,
                   const ServiceConnectionSyb& rconnectionHandle);

  /// Field Names
  static const std::string fileIdField; 
  static const std::string fileNameField; 
  static const std::string fileVersionField; 
  static const std::string diskIdField; 
  static const std::string formatField; 
  static const std::string fileSizeField; 
  static const std::string uncompressedFileSizeField; 
  static const std::string compressionField; 
  static const std::string ingestionDateField; 
  static const std::string ignore; 
  static const std::string checksum; 
  static const std::string checksumPluginField; 
  static const std::string fileStatusField; 
  static const std::string creationDateField; 

  std::string tableName_;

};

#endif //NGAS_FILE_DAO_SYB_H
