// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    M. Padovan - ESO/DMD/DFS
// SUBSYSTEM: arcsys - Archive Systems
//
//
// Sybase Data Access Object for the NgasFilesSafe table
// 
//------------------------------------------------------------------------------

#ifndef NGAS_FILES_SAFE_DAO_SYB_H
#define NGAS_FILES_SAFE_DAO_SYB_H

#include <string>
#include <SybUtils.h>
#include <vector>

class ServiceConnectionSyb;

/// Sybase Data Access Object for NgasFile objects
/**
 * This class can be used to access NgasFile objects stored
 * in a Sybase table.
 */
class NgasFilesSafeDAOSyb
{
public:
 
  /// \brief Constructor
  /// \param tableName the name of the ngas file table
  NgasFilesSafeDAOSyb( const std::string& tableName );

  /// \brief Retrieve from the DB the NgasFile object with given fileId
  /// \brief and version.
  /// \param[in] version the file version
  /// \param[in] fileId the file ID
  /// \param[out] rngasFileSet the selected NgasFile objects
  /// \param rconnectionHandle database connection handle
  /// \return the number of retrieved objects, or -1 in case of error
  int find(const std::string&          rfileId, 
           int version,
           /*NgasFile::ResultSet&        rngasFileSet,*/
           const ServiceConnectionSyb& rconnectionHandle);
  
private:

  /// \brief Select NgasFile objects from the DB
  /// \param[out] rngasFileSet the selected ngas file objects
  /// \param[in] whereClause the optional where clause
  /// \param rconnectionHandle database connection handle
  /// \return the number of selected objects, or -1 in case of error
  int select(const ServiceConnectionSyb& rconnectionHandle,
             /*NgasFile::ResultSet& rngasFileSet, */
             const std::string& whereClause = "");

  /// \brief Return a comma separated list of all fields in the 
  /// ngas file table
  std::string allFields();
/*
  /// \brief Convert a SybaseResultSet object to a NgasFile object
  /// \param[in] res the ResultSet
  /// \param[out] rngasFile the NgasFile
  /// \return 0 if the conversion was succesful, -1 otherwise
  int convertResultSetToNgasFile(SybaseResultSet&  res,
         NgasFile&         rngasFile);
*/

  /// \brief Executes the given query.
  /// \param query Query to be executed.
  /// \param rconnectionHandle Database connection handle.
  /// \return The number of affected rows or -1 in case of error.
  int executeQuery(std::string query,
                   const ServiceConnectionSyb& rconnectionHandle);

  /// Field Names
  static const std::string fileIdField; 
//   static const std::string fileNameField; 
  static const std::string fileVersionField; 
//   static const std::string diskIdField; 
//   static const std::string formatField; 
//   static const std::string fileSizeField; 
//   static const std::string uncompressedFileSizeField; 
//   static const std::string compressionField; 
//   static const std::string ingestionDateField; 
//   static const std::string ignore; 
//   static const std::string checksum; 
//   static const std::string checksumPluginField; 
//   static const std::string fileStatusField; 
//   static const std::string creationDateField;

  const std::string tableName_;

};

#endif //NGAS_FILES_SAFE_DAO_SYB_H
