// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observator
// LICENSE:   ESO internal use only
//
// PROJECT:   VLT Data Flow Syste
// AUTHOR:    Christophe Moins- ESO/SDD/DF
// SUBSYSTEM: arcsys
//
// $Revision: 1.4 $
// $Log: ObsBlock.h,v $
// Revision 1.4  2007/05/15 15:35:11  cmoins
// removal of block id to build media files
//
// Revision 1.3  2007/05/15 10:13:36  cmoins
// modification of logging messages
//
// Revision 1.2  2007/05/04 11:57:34  cmoins
// adding ofd storeData, updateProgDB and getMaxTime implementation
//
// Revision 1.1  2007/04/25 15:57:41  cmoins
//
// 
//------------------------------------------------------------------------------

#ifndef OBSBLOCK_H
#define OBSBLOCK_H

#include <string>
#include <set>
#include "RawDataProduct.h"
#include "FileCopy.h"
#include "DPPackerDef.h"
#include "Media.h"
#include "MediaFile.h"


using namespace std;

/// Observation block base class
/**
 * All the block classes in charge of retrieving files associated to a certain
 * type of block, to store them in a staging area and to update a data 
 * repository will inherit it. It can not be instantiate.
 * It uses a Storage Service template parameter to remain independent of the 
 * storage service layer.
 */
template <class StorageService>
class ObsBlock
{
public :
 
  /// Set of observation blocks
  typedef set<ObsBlock<StorageService>*> ResultSet; 
  typedef typename ResultSet::iterator pObsBlock;
 
  /// \brief Constructor
  /// \param blockId block identifier
  /// \param connAlias database connection alias
  ObsBlock(int blockId,
	   string connAlias):
    blockId_(blockId),
    connAlias_(connAlias)
  {}
 
  /// \brief Destructor
  virtual ~ObsBlock()
  {}

  /// \brief fetches a list of files associated to the block
  /**
   * Note : when giving a list of instruments whose calibration files
   * should be excluded, the format shall be the following :
   * Inst1 start1 end1,Inst2 start2 end2,..
   * \param startTime start time of the files retrieval
   * \param startTimeAdjustment correction to apply on start time for 
   *                            calibration files retrieval
   * \param endTime end time of the files retrieval
   * \param endTimeAdjustment   correction to apply on end time for 
   *                            calibration files retrieval
   * \param rlistOfUsedInstruments list of instruments that should be 
   *                               excluded to retrieve calibration files
   * \return 0 if successful, -1 otherwise 
   */
  virtual int fetchData(double  startTime,
			double  startTimeAdjustment,
			double  endTime,
			double  endTimeAdjustment,
			string& rlistOfUsedInstruments) = 0;

  /// \brief stores the content of the block inside a staging area.
  /**
   * \param destHost host where the files shall be stored
   * \param stagingArea directory where the files will be stored.
   * \return 0 if successful, -1 otherwise
   */
  virtual int storeData(string destHost,
			string stagingArea);

  ///\brief updates the data repository with the files in the staging area
  /**
   * rmedia media request files will be associated to
   * \return 0 if successful, -1 otherwise
   */
  virtual int updateObsBlockDB(Media& rmedia) const;

  /// \brief gives the time of the most recent file 
  /**
   * \param rmaxTime time of the most recent file
   * \return 0 if successful, -1 otherwise
   */
  virtual int getMaxTime(double& rmaxTime) const;

  /// \brief instantiates a copy of the object
  /// returns the address of the copy
  virtual ObsBlock* clone() const = 0;
protected :

  /// \brief set the connection alias data member
  /// \param alias connection alias to assign to the data member
  void setAlias(string alias)
  {
    connAlias_ = alias;
  }

  /// \brief get the connection alias data member
  string getAlias()
  {
    return connAlias_;
  }

  /// \brief set the block identifier data member
  /// \param blockId block identifier to assign to the data member
  void setBlockId(int blockId)
  {
    blockId_ = blockId;
  }
  ///\brief get the block identifier data member
  int getBlockId()
  {
    return blockId_;
  }

  /// \brief set the data file set data member
  /// \param dataFileSet set of data files to assign to the data member
  void setDataFileSet(const RawDataProduct::ResultSet& dataFileSet)
  {
    dataFileSet_ = dataFileSet;
  }
  /// \brief get the data file set data member
  /// \return the data file set data member
  RawDataProduct::ResultSet getDataFileSet()
  {
    return dataFileSet_;
  }

private :
  /// block identifier
  int blockId_;   
  /// database connection parameters alias
  string connAlias_;
  /// set of files of the block
  RawDataProduct::ResultSet dataFileSet_; 

};

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Stores the content of a block inside a staging area.
//------------------------------------------------------------------------------
template <class StorageService>
int
ObsBlock<StorageService>::storeData(string destHost,
				    string stagingArea)
{
  int retStatus = 0;
  int nbCopiedFiles=0;
  stringstream blockIdStr;
  string storageDirectory;

  DFS_TRACE("ObsBlock<ObservationService>::storeData"); 

  blockIdStr << blockId_;
  storageDirectory = stagingArea + "/" + blockIdStr.str();

  for (RawDataProduct::pRawDataProduct pdataFile = dataFileSet_.begin(); 
       pdataFile != dataFileSet_.end();pdataFile++)
    {
      if (!(DPPackerCopy::copyFile(destHost,storageDirectory,pdataFile)))
	{
	  // complete file name is updated to the copy one
	  const_cast<string&>(pdataFile->magDiskFile_.fname_) = 
	    blockIdStr.str() + "/" + pdataFile->dpCat_ + "/" +
	    pdataFile->magDiskFile_.fname_;
	  nbCopiedFiles++;
	}
    }

  if ((unsigned int)nbCopiedFiles != dataFileSet_.size())
    {
      DFS_ERROR("Some files of %d block have not been copied ",blockId_);
      retStatus = -1;
    }
  else if (dataFileSet_.empty())
    {
      DFS_ERROR("No file available");
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// updates the data repository with the files in the staging area
//-----------------------------------------------------------------------------
template <class StorageService>
int
ObsBlock<StorageService>::updateObsBlockDB(Media& rmedia)  const
{
  int retStatus = 0;
  string fileName;
  StorageService storageService(connAlias_);
  MediaFile::ResultSet mediaFileSet;

  DFS_TRACE("ObsBlock<ObservationService>::updateObsBlockDB");

  if (!dataFileSet_.empty())
    {
      
      for (RawDataProduct::pRawDataProduct pdataFile = dataFileSet_.begin();
	   pdataFile != dataFileSet_.end();pdataFile++)
	{
	  if (pdataFile->magDiskFile_.status_ == DPPackerDef::FILE_RETRIEVED)
	    {
	      fileName =  pdataFile->magDiskFile_.fname_;
	      mediaFileSet.push_back(MediaFile(pdataFile->dpId_,
					       fileName,
					       pdataFile->magDiskFile_.fsize_,
					       pdataFile->magDiskFile_.uncompFSize_,
					       DPPackerDef::NO_COMPRESSION_FLAG,
					       pdataFile->magDiskFile_.status_));
	    }

	}
      if (!mediaFileSet.empty() &&
	  !storageService.insertMediaFiles(rmedia,
					   mediaFileSet))
	{
	  DFS_INFO("Media Files insertion succeeded for %d",blockId_);
	}
      else
	{
	  DFS_ERROR("Media Files insertion failed for %d",blockId_);
	  retStatus = -1;
	}
    }
  else
    {
      DFS_ERROR("No file Available");
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// gives the time of the most recent file
//------------------------------------------------------------------------------
template <class StorageService>
int
ObsBlock<StorageService>::getMaxTime(double& rmaxTime) const
{
  int retStatus = 0;
  double maxExpTime = -1;
        
  DFS_TRACE("ObsBlock<ObservationService>::updateObsBlockDB");

  for (RawDataProduct::pRawDataProduct pdataFile = dataFileSet_.begin();
       pdataFile != dataFileSet_.end();pdataFile++)
    {
      if (pdataFile->mjdObs_ > maxExpTime)
	maxExpTime = pdataFile->mjdObs_;
    }
  if (maxExpTime > 0)
    rmaxTime = maxExpTime;
  else
    retStatus = -1;
  return retStatus;
}

#endif /* OBSBLOCK_H */
