// -*-c++-*-

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Christophe Moins- ESO/SDD/DFS
// SUBSYSTEM: arcsys - Archive Systems
//
// $Revision: 1.13 $
// $Log: ObservationService.h,v $
// Revision 1.13  2008/02/01 14:59:45  mpadovan
// The new remove() method now takes also the host name parameter.
//
// Revision 1.12  2008/01/29 16:06:12  mpadovan
// The new remove() method must work with magDiskFileDAO_, not dataFileDAO_.
//
// Revision 1.11  2008/01/28 15:19:49  cmoins
// asto database becomes configurable
//
// Revision 1.10  2008/01/08 14:23:25  mpadovan
// Added a new method (remove) to remove files with the specified file id.
//
// Revision 1.9  2007/11/05 15:38:27  mpadovan
// Added new constructor overload that accepts ServerName, DBname, userName,
// and password.
//
// Revision 1.8  2007/07/25 07:28:10  cmoins
// doxygen comments correction
//
// Revision 1.7  2007/07/10 09:03:22  cmoins
// the list of previously used intruments is not used any more
// when retrieving the list of calibration files
//
// Revision 1.6  2007/05/24 11:34:44  cmoins
// more info in error logging messages
//
// Revision 1.5  2007/05/21 12:51:32  cmoins
// Modification of logging message
//
// Revision 1.4  2007/05/15 15:32:58  cmoins
// replacement of info logging messages by debug ones
//
// Revision 1.3  2007/05/02 09:38:08  cmoins
// adding of a dao class to objects representing files on magnetic disks
//
// Revision 1.2  2007/04/26 15:18:53  cmoins
// modification of doxygen comments
//
// Revision 1.1  2007/04/25 15:11:30  cmoins
// creation of a service layer associated to the observation data repository
//
// 
//------------------------------------------------------------------------------

#ifndef OBSERVATION_SERVICE_H
#define OBSERVATION_SERVICE_H

#include <string>
#include <DfsLog.h>
#include "RawDataProduct.h"

using namespace std;

/// Data Access Service layer that handles observation programs or blocks files
/**
 * This class allows to retrieve files scientific or calibration files
 * associated to an observation program or block.
 * It will rely on three tables :
 * - an observation table storing program related data on files
 * - a file on magnestisk disk table to know files location
 * - a magnetic disk table to know disk location
 *
 * There are three template parameters :
 * - ConnectionHandle : type of connection handle to the data repository so that 
 *                      the layer remains independent of any storage system
 * - DataFileDAO : type of Data Access Object class to the observation  files 
 * - DataFileOnDiskDAO :  type of Data Access Object class to the files on disk
 *
 * Note : when giving a list of instruments and time range already handled,
 * the format shall be the following :
 * Inst1 start1 end1,Inst2 start2 end2,..
 */
template 
<class ConnectionHandle, 
 template <typename> class DataFileDAO,
 template <typename> class DataFileOnDiskDAO>
class ObservationService
 {
 public:
 
   /// \brief Constructor
   /// \param connAlias data repository connection keyword
   ObservationService(string connAlias);

   /// \brief Constructor
   /**
    * \param dbServer DB server name
    * \param dbUser   DB user name
    * \param dbPasswd DB password
    * \param dbName   DB name
    */
   ObservationService(	std::string dbServer,
		   				std::string dbUser,
		   				std::string dbPasswd,
		   				std::string dbName	);   
   
   /// \brief Destructor
   ~ObservationService(){};

   /// \brief Finds the file associated to a given file id
   /**
    * \param fileId file identifier
    * \param rdataFileSet set of files
    * \return the number of files retrieved if successful, -1 otherwise
    */
   int findFiles(string fileId,
		 RawDataProduct::ResultSet& rdataFileSet) const;


   /// \brief Finds a list of scientific files inside a program
   /**
    * \param progId identifier of the observation program
    * \param startTime start time of the retrieval
    * \param endTime end time of the retrieval
    * \param rdataFileSet set of science files
    * \return the number of files retrieved if successful, -1 otherwise
    */
   int findScienceFiles(string progId,
			double startTime,
			double endTime,
			RawDataProduct::ResultSet& rdataFileSet) const;

   /// \brief Finds a list of calibration files inside a program
   /**
    * \param progId identifier of the observation program
    * \param startTime start time of the retrieval
    * \param endTime end time of the retrieval
    * \param rdataFileSet set of calibration files
    * \return the number of files retrieved if successful, -1 otherwise
    */
   int findCalibFiles(string progId,
		      double startTime,
		      double endTime,
		      RawDataProduct::ResultSet& rdataFileSet) const;

   /// \brief Finds a list of calibration files associated to a set of instruments
   /**
    * \param instList list of instruments whose calibration files
    *                 shall be retrieved
    * \param startTime start time of the retrieval
    * \param endTime end time of the retrieval
    * \param rdataFileSet set of calibration files
    * \return the number of files retrieved if successful, -1 otherwise
    */
   int findCalibFilesByInst(string instList,
			    double startTime,
			    double endTime,
			    RawDataProduct::ResultSet& rdataFileSet) const;
  
   /// \brief  Finds a list of scientific files inside an observation block
   /**
    * \param obsId identifier of the observation block
    * \param startTime start time of the retrieval
    * \param endTime end time of the retrieval
    * \param rdataFileSet set of science files
    * \return the number of files retrieved if successful, -1 otherwise
    */
   int findObsBlockScienceFiles(int obsId,
				double startTime,
				double endTime,
				RawDataProduct::ResultSet& rdataFileSet) const;
 
   /// \brief Finds a list of calibration files inside an observation block 
   /**
    * \param obsId identifier of the observation block
    * \param startTime start time of the retrieval
    * \param endTime end time of the retrieval
    * \param rdataFileSet set of calibration files
    * \return the number of files retrieved if successful, -1 otherwise
    */
   int findObsBlockCalibFiles(int obsId,
			      double startTime,
			      double endTime,
			      RawDataProduct::ResultSet& rdataFileSet) const;

   /// \brief Removes the record with the given file id and host name.
   /**
    * \param rfileId id of the file to be removed
    * \param hostName name of the host where the file is located.
    * \return the number of files removed if successful, -1 otherwise
    */
   int remove(const std::string& rfileId,
              const std::string& hostName) const;

 private :

   //static constant declaration
   /// observation product table
   static const string OBSERVATION_TABLE; 
   /// files on magnetic disk table
   static const string MDFILES_TABLE;     
   /// magnetic disk table 
   static const string MDTABLE;           

   typedef DataFileDAO<ConnectionHandle> DataFileDAOCH;

   /// Finds the location of a set of file
   /**
    * \param rdataFileSet set of files whose location shall be retrieved
    * \param rconnectionHandle connection handle to the database
    * \return the number of files whose location was retrieved
    */
   int findFileLocation(RawDataProduct::ResultSet& rdataFileSet,
			ConnectionHandle& rconnectionHandle) const;
 
   /// Data repository connection keyword
   //string connAlias_; 

   /// Connection handle to the data repository
   mutable ConnectionHandle dsConnection_;
   /// Data Access Object to retrieve observation files
   DataFileDAOCH dataFileDAO_; 
   /// Data access object to retrieve files on magnetic disks
   DataFileOnDiskDAO<ConnectionHandle> magDiskFileDAO_;
 };

// static constant creation
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
	  template <typename> class DataFileOnDiskDAO >
const string  ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>
::OBSERVATION_TABLE = "observations..data_products";

template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
	  template <typename> class DataFileOnDiskDAO> 
const string  ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>
::MDFILES_TABLE = "mdfiles";

template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
const string  ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>
::MDTABLE = "mds";

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Constructor
// The database where the disk and disk files table are stored is configurable 
//------------------------------------------------------------------------------
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>::
ObservationService(string connAlias):
  dsConnection_(connAlias),
  dataFileDAO_(OBSERVATION_TABLE),
  magDiskFileDAO_(dsConnection_.getDbName()+".." + MDFILES_TABLE,
		  dsConnection_.getDbName()+".." + MDTABLE)
{
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Constructor
// The database where the disk and disk files table are stored is configurable 
//------------------------------------------------------------------------------
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>::
ObservationService(	std::string dbServer,
					std::string dbUser,
					std::string dbPasswd,
					std::string dbName	) : 
  dsConnection_(dbServer, dbUser, dbPasswd, dbName),
  dataFileDAO_(OBSERVATION_TABLE),
  magDiskFileDAO_(dbUser + ".." + MDFILES_TABLE,
		  dbUser + ".." + MDTABLE)
{
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// \brief Finds the file associated to a given file id
//-----------------------------------------------------------------------------
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
int 
ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>::
findFiles(string fileId,
	  RawDataProduct::ResultSet& rdataFileSet) const
{
  int retStatus = 0;
  int nbFiles;

  DFS_TRACE("ObservationService<ConnectionHandle,DataFileDAO,"\
	    "DataFileOnDiskDAO>::findFiles");

  //ConnectionHandle dsConnection(connAlias_);
  int connectionStatus = dsConnection_.acquire();
  if (!connectionStatus)
    {
      nbFiles = dataFileDAO_.findByFileId(fileId,
					  rdataFileSet,
					  dsConnection_);
      if ((nbFiles > 0) && (findFileLocation(rdataFileSet, dsConnection_) > 0))

	{
	  DFS_DEBUG("Files retrieval succeeded for %s",fileId.c_str());
	  retStatus = nbFiles;
	}
      else if (!nbFiles)
	{
	  DFS_DEBUG("No files retrieved for %s",fileId.c_str());
	  retStatus = 0;
	}
      else
	{
	  DFS_ERROR("Files retrieval failed for %s",fileId.c_str());
	  retStatus = -1;
	}
      dsConnection_.release();
    }
  else
    {
      DFS_ERROR("ObservationService::findFiles : Could not acquire connection");
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Finds a list of scientific files inside a program
//------------------------------------------------------------------------------
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
int
ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>::
findScienceFiles(string progId,
		 double startTime,
		 double endTime,
		 RawDataProduct::ResultSet& rdataFileSet) const
{
  int retStatus = 0;
  int nbFiles;

  DFS_TRACE("ObservationService<ConnectionHandle,DataFileDAO,"\
            "DataFileOnDiskDAO>::findScienceFiles");

  //ConnectionHandle dsConnection(connAlias_);
  int connectionStatus = dsConnection_.acquire();
  if (!connectionStatus)
    {
      nbFiles = dataFileDAO_.findByCatgAndProgId(progId,
						 DataFileDAOCH::SCIENCE,
						 startTime,
						 endTime,
						 rdataFileSet,
						 dsConnection_);
      if ((nbFiles > 0) && (findFileLocation(rdataFileSet, dsConnection_) > 0))
	{
	  DFS_DEBUG("Science files retrieval succeeded for %s",progId.c_str());
	  retStatus = nbFiles;
	}
      else if (!nbFiles)
	{
	  DFS_DEBUG("No Science files retrieved for %s",progId.c_str());
	  retStatus = 0;
	}
      else
	{
	  DFS_ERROR("Science files retrieval failed for %s",progId.c_str());
	  retStatus = -1;
	}
      dsConnection_.release();
    }
  else
    {
      DFS_ERROR("ObservationService::findScienceFiles."\
		"Could not acquire connection");
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Finds a list of calibration files inside a program
//------------------------------------------------------------------------------
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
int
ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>::
findCalibFiles(string progId,
	       double startTime,
	       double endTime,
	       RawDataProduct::ResultSet& rdataFileSet) const
{
  int retStatus = 0;
  int nbFiles;

  DFS_TRACE("ObservationService<ConnectionHandle,DataFileDAO,"\
	    "DataFileOnDiskDAO>::findCalibFiles");

  //ConnectionHandle dsConnection(connAlias_);
  int connectionStatus = dsConnection_.acquire();
  if (!connectionStatus)
    {
      nbFiles = dataFileDAO_.findByCatgAndProgIdAndInst(progId,
							DataFileDAOCH::NO_INST,
							DataFileDAOCH::CALIB,
							startTime,
							endTime,
							rdataFileSet,
							dsConnection_);
      if ((nbFiles > 0) && (findFileLocation(rdataFileSet, dsConnection_) > 0))
	{
	  DFS_DEBUG("Calibration files retrieval succeeded for %s",
		   progId.c_str());
	  retStatus = nbFiles;
	}
      else if (!nbFiles)
	{
	  DFS_DEBUG("No calibration files retrieved for %s",progId.c_str());
	  retStatus = 0;
	}
      else
	{
	  DFS_ERROR("Calibration files retrieval failed for %s",
		    progId.c_str());
	  retStatus = -1;
	}
      dsConnection_.release();
    }
  else
    {
      DFS_ERROR("ObservationService::findCalibFiles."\
		"Could not acquire connection");
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Finds a list of calibration files associated to a set of instruments 
//------------------------------------------------------------------------------
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
int
ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>::
findCalibFilesByInst(string instList,
		     double startTime,
		     double endTime,
		     RawDataProduct::ResultSet& rdataFileSet) const
{
  int retStatus = 0;
  int nbFiles;

  DFS_TRACE("ObservationService<ConnectionHandle,DataFileDAO,"\
            "DataFileOnDiskDAO>::findCalibFilesByInst");

  //ConnectionHandle dsConnection(connAlias_);
  int connectionStatus = dsConnection_.acquire();
  if (!connectionStatus)
    {
      nbFiles = dataFileDAO_.findByCatgAndInst(instList,
					       DataFileDAOCH::CALIB,
					       startTime,
					       endTime,
					       rdataFileSet,
					       dsConnection_);
      if ((nbFiles > 0) && (findFileLocation(rdataFileSet, dsConnection_) > 0))
	{
	  DFS_DEBUG("Calibration files retrieval succeeded for %s",
		   instList.c_str());
	  retStatus = nbFiles;
	}
      else if (!nbFiles)
	{
	  DFS_DEBUG("No calibration files retrieved for %s",instList.c_str());
	  retStatus = 0;
	}
      else
	{
	  DFS_ERROR("Calibration files retrieval failed for %s",
		    instList.c_str());
	  retStatus = -1;
	}
      dsConnection_.release();
    }
  else
    {
      DFS_ERROR("ObservationService::findCalibFilesByInst."\
		"Could not acquire connection");
      retStatus = -1;
    }

  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Finds a list of scientific files inside an observation block.
//------------------------------------------------------------------------------
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
int
ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>::
findObsBlockScienceFiles(int obsId,
			 double startTime,
			 double endTime,
			 RawDataProduct::ResultSet& rdataFileSet) const
{
  int retStatus = 0;
  int nbFiles;

  DFS_TRACE("ObservationService<ConnectionHandle,DataFileDAO,"\
            "DataFileOnDiskDAO>::findObsBlockScienceFiles");

  //ConnectionHandle dsConnection(connAlias_);
  int connectionStatus = dsConnection_.acquire();
  if (!connectionStatus)
    {
      nbFiles = dataFileDAO_.findByCatgAndObsId(obsId,
						DataFileDAOCH::SCIENCE,
						startTime,
						endTime,
						rdataFileSet,
						dsConnection_);
      if ((nbFiles > 0) && (findFileLocation(rdataFileSet,dsConnection_) > 0))
	{
	  DFS_DEBUG("Scientific files retrieval succeeded for %d",obsId);
	  retStatus = nbFiles;
	}
      else if (!nbFiles)
	{
	  DFS_DEBUG("No Science files retrieved for %d",obsId);
	  retStatus = 0;
	}
      else
	{
	  DFS_ERROR("Scientific files retrieval failed for %d",obsId);
	  retStatus = -1;
	}
      dsConnection_.release();
    }
  else
    {
      DFS_ERROR("ObservationService::findObsBlockScienceFiles."\
		"Could not acquire connection");
      retStatus = -1;
    }
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Finds a list of calibration files inside an observation block 
//------------------------------------------------------------------------------
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
int
ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>::
findObsBlockCalibFiles(int obsId,
		       double startTime,
		       double endTime,
		       RawDataProduct::ResultSet& rdataFileSet) const
{
  int retStatus = 0;
  int nbFiles;

  DFS_TRACE("ObservationService<ConnectionHandle,DataFileDAO,"\
            "DataFileOnDiskDAO>::findObsBlockCalibFiles");

  //ConnectionHandle dsConnection(connAlias_);
  int connectionStatus = dsConnection_.acquire();
  if (!connectionStatus)
    {
      nbFiles = dataFileDAO_.findByCatgAndObsIdAndInst(obsId,
						       DataFileDAOCH::NO_INST,
						       DataFileDAOCH::CALIB,
						       startTime,
						       endTime,
						       rdataFileSet,
						       dsConnection_);
      if ((nbFiles > 0) && (findFileLocation(rdataFileSet, dsConnection_) > 0))
	{
	  DFS_DEBUG("Calibration files retrieval succeeded for %d",obsId);
	  retStatus = nbFiles;
	}
      else if (!nbFiles)
	{
	  DFS_DEBUG("No Calibration files retrieved for %d",obsId);
	  retStatus = 0;
	}
      else
	{
	  DFS_ERROR("Calibration files retrieval failed for %d",obsId);
	  retStatus =-1;
	}
      dsConnection_.release();
    }

  else
    {
      DFS_ERROR("ObservationService::findObsBlockCalibFiles."\
		"Could not acquire connection");
      retStatus = -1;
    }

  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Finds the location of a set of file
//------------------------------------------------------------------------------
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
int
ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>::
findFileLocation(RawDataProduct::ResultSet& rdataFileSet,
		 ConnectionHandle& rconnectionHandle) const
{
  int retStatus = 0;
  int nbFiles = 0;

  DFS_TRACE("ObservationService<ConnectionHandle,DataFileDAO,"\
            "DataFileOnDiskDAO>::findFileLocation");
  
  for (RawDataProduct::pRawDataProduct pconstfile = rdataFileSet.begin(); 
       pconstfile != rdataFileSet.end(); pconstfile++)
    {
      RawDataProduct* pfile = (const_cast<RawDataProduct*>(&(*pconstfile)));
      if (!magDiskFileDAO_.findFileLocation(pfile->magDiskFile_,
					    rconnectionHandle))
	{
	  DFS_DEBUG("Find file location successful for %s", 
		   pfile->dpId_.c_str());
	  nbFiles++;
	}
      else
	{
	  DFS_ERROR("Find file location failed for %s name %s size %d ", 
		    pfile->dpId_.c_str(),pfile->magDiskFile_.fname_.c_str(),
		    pfile->magDiskFile_.fsize_);
	}
    }
  retStatus = nbFiles;
  return retStatus;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Remove the object with given fileId
// Return the number of removed objects or -1 in case of error
//------------------------------------------------------------------------------
template <class ConnectionHandle, 
          template <typename> class DataFileDAO,
          template <typename> class DataFileOnDiskDAO>
int ObservationService<ConnectionHandle,DataFileDAO,DataFileOnDiskDAO>::
remove(const std::string& rfileId, const std::string& rdiskId) const
{
  int rowCount = -1; 
  
  DFS_TRACE("ObservationService<ConnectionHandle,DataFileDAO,"\
            "DataFileOnDiskDAO>::remove");

  int connectionStatus = dsConnection_.acquire();
  if (!connectionStatus)
  {
    rowCount = magDiskFileDAO_.remove(rfileId, rdiskId, dsConnection_);

    dsConnection_.release();
  }
  else
  {
    DFS_ERROR("Could not acquire connection");
  }

  return rowCount;
}

#endif // OBSERVATION_SERVICE_H
