// -*-c++-*-

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) 2006 European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Christophe Moins - ESO/SDD/DFS
// SUBSYSTEM: arcsys - archive Systems
//
// $Revision: 1.5 $
// $Log: PackageConfig.h,v $
// Revision 1.5  2007/08/14 07:43:06  cmoins
// DFS04052 : adding of a function to keep a track of the command line used to launch
// the program.
//
// Revision 1.4  2007/07/03 12:53:23  cmoins
// constructor takes const char* as an argument and initConfig returns the number
// of valid options if successful
//
// Revision 1.3  2007/06/14 13:02:18  cmoins
// non optional argument handling
//
// Revision 1.2  2007/05/21 12:54:59  cmoins
// adding of OptionStruct type
// VS: ----------------------------------------------------------------------
//
// Revision 1.1  2007/04/25 15:57:41  cmoins
// *** empty log message ***
//
//
//------------------------------------------------------------------------------

#ifndef PACKAGECONFIG_H
#define PACKAGECONFIG_H

// C++ includes. 
#include <vector>
#include <map>
#include <string>

using namespace std;

/// Package configuration class
/**
 * This class contains all the configuration options of an observation program  
 * package along with the methods used to load and access their values. 
 */
class PackageConfig 
{
public:

  /// tag associated to non optional arguments
  static const string NON_OPTIONAL_ARG;

  typedef vector<string> StringValList;
  typedef StringValList::const_iterator pconstStringValue;
  typedef vector<double> DoubleValList;
  typedef DoubleValList::const_iterator pconstDoubleValue;
  typedef vector<int> IntValList;
  typedef IntValList::const_iterator pconstIntValue;

  /// Type defining the structure of an option
  struct OptionStruct {

    OptionStruct(){}

    OptionStruct(char shortName,
		 string argument,
		 bool mandatory,
		 bool argRequired,
		 string helpText):
      shortName_(shortName),
      argument_(argument),
      mandatory_(mandatory),
      argRequired_(argRequired),
      helpText_(helpText)
    {}
    char   shortName_;
    string argument_;
    bool   mandatory_;
    bool   argRequired_;
    string helpText_;
  };

  /// type that allows to associate a definition to an option
  typedef map<string,OptionStruct> OptionMap;

  /// Constructor 
  /** 
   * Performs a global init of FileCopyHandler. As this operation is 
   * not thread safe, PackageConfig shall be instantiated 
   * when only the main thread is running.
   * The configuration parameters need to be initialised using the 
   * initConfig() method. 
   * \param argc number of elements in the configuration line
   * \param argv configuration line
   */
  PackageConfig(int argc,
		const char* argv[]);

  /// Destructor
  /**
   * Performs a global clean up of FileCopyHandler. As this operation is
   * not thread safe, a FileCopyHandler instance shall be deleted 
   * when only the main thread is running.
   */
  ~PackageConfig();

  /// copy constructor
  PackageConfig(const PackageConfig& rpackageConfig);

  /// assignment operator
  PackageConfig& operator = (const PackageConfig& rpackageConfig);

  /**
   * Parses the command line arguments and matches options with arguments
   *
   * \param roptionMap : options map
   * returns the number of valid options and arguments if successful, 
   * -1 otherwise.
   */
  int initConfig(OptionMap& roptionMap);

  /// \brief displays a help message relying on the option map
  /**
   * \param roptionMap : options map
   * returns 0 if successful, -1 otherwise
   */
  int displayHelp(const OptionMap& roptionMap);

  /// \brief displays the command line that launched the program
  /**
   * returns 0 if successful, -1 otherwise
   */
  int displayCommandLine();

  /// gives the string value list of non optional arguments
  /**
   * rvalueList list of non optional arguments
   * returns 0 if successful, -1 otherwise
   */
  int getArgString(StringValList& rvalueList) const;

  /// gives the string value  associated to a parameter 
  /**
   * parameter parameter whose value shall be retrieved
   * rvalue string value associated to the parameter
   * returns 0 if successful, -1 otherwise
   */
  int getString(string parameter,
		string& rvalue) const;

  /// gives the time value associated to a parameter
  /**
   * parameter format is : YYYY-MM-DD or YYYY-MM-DDThh:mm:ss
   * \param parameter parameter whose time value shall be retrieved
   * \param rvalue time value associated to the parameter 
   * \param utcTime set to true if the time is indicated in the date
   * returns 0 if successful, -1 otherwise
   */
  int getTime(string parameter,
	      double& rvalue,
	      bool& utcTime) const;

  /// gives the integer value associated to a parameter
  /**
   * parameter parameter whose value shall be retrieved
   * rvalue int value associated to the parameter
   * returns 0 if successful, -1 otherwise
   */
  int getInt(string parameter,
	     int& rvalue) const;

  /// gives the string value list associated to a parameter 
  /**
   * parameter parameter whose value shall be retrieved
   * rvalueList list of value associated to the parameter
   * returns 0 if successful, -1 otherwise
   */
  int getString(string parameter,
		StringValList& rvalueList) const;

  /// gives the time value list associated to a parameter
  /**
   * parameter parameter whose value shall be retrieved
   * rvalueList list of double values associated to the parameter
   * returns 0 if successful, -1 otherwise
   */
  int getTime(string parameter,
	      DoubleValList& rvalueList) const;

  /// gives the integer value list associated to a parameter
  /**
   * parameter parameter whose value shall be retrieved
   * rvalueList list of Integer values associated to the parameter
   * returns 0 if successful, -1 otherwise
   */
  int getInt(string parameter,
	     IntValList& rvalueList) const;
    
private :

  /// argument values list type
  typedef vector<string> ValueList;  

  /// package arguments list type
  typedef map<string,ValueList> ArgList; 
  typedef ArgList::const_iterator pconstArgument;

  /// list of arguments of a package
  ValueList packageArgList_;    

  /// list of optional arguments of a package
  ArgList packageOptArgList_; 

  int argc_;   // number of elements in the configuration line
  char**  argv_; // configuration line

};

#endif  // PACKAGECONFIG_H
