// -*-c++-*-

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Jens Knudstrup - ESO/DMD/DFS
// SUBSYSTEM: DFS Archive System
//
// $Name: arcsys-1_2_8 $
// $Revision: 1.24 $
// $Log: ProTrieve.h,v $
// Revision 1.24  2007/09/25 15:38:23  jknudstr
// Implemented HTML encoding of Global Keys with _ in them.
//
// Revision 1.23  2007/06/01 07:23:45  jknudstr
// Changes in connection with testing/finalizing of SA->PA Rep Agent.
//
// Revision 1.22  2007/05/03 15:32:07  jknudstr
// Intermediate check-in while developing the SA->PA Archive Replicator.
//
// Revision 1.21  2007/04/11 09:40:16  jknudstr
// Implemented support for uncompressing Z compressed files upon retrieval.
//
// Revision 1.20  2007/01/11 15:35:34  jknudstr
// -
//
// Revision 1.19  2006/11/09 07:34:24  jknudstr
// -
//
// Revision 1.18  2006/11/06 11:06:36  jknudstr
// -
//
// Revision 1.17  2006/08/08 15:29:22  jknudstr
// -
//
//-----------------------------------------------------------------------------

#ifndef PROTRIEVE_H
#define PROTRIEVE_H


// STL includes
#include <string>
#include <vector>
#include <map>
#include <sstream>

// External packages includes
#include "ace/Mutex.h"
#include "boost/filesystem/path.hpp"
#include "boost/filesystem/operations.hpp"

// Local includes
#include <ArcSysEx.h>

// Namespaces
namespace fs = boost::filesystem;

// Forward declarations.
class ProTrieveArchiveRequest;
class ProTrieveArchReqsMap;
class ProTrieveFileRequest;


/** Class to interface to pro_trieve.
 *
 * The ProTrieve Class implements an interface to archive and retrieve data
 * objects to/from a tape library via the controller SW pro_trieve
 * (http://www.pro-trieve.de/).
 * 
 * <p>
 * When archiving data into the TL, an object ProTrieveArchReqsMap is used to
 * keep track of which responses are pending. Note: The response for an archive
 * request is generated asynchronously and will not returned as an immediate
 * response to the archive request.
 * 
 * <p>
 * Also for retrieve requests, the files requested may be delivered some time
 * after it was requested. An object of type ProTrieveFileRequest is used to
 * keep track of a pending request.
 */


/// Class providing an interface to the pro_trieve daemon.
class ProTrieve
{
  public:
    /// Constructor method.
    ProTrieve(void);

    /// Constructor method initializing basic members.
    /**
     * Constructor method setting up reference to client ID, archive
     * system ID and pro_trieve base directory.
     *
     * \param clientId  ID of this client.
     * \param sysId     System wide ID. Used to tag data when archiving into
     *                  pro_trieve.
     * \param baseDir   Base (staging) directory used by pro_trieve.
     */
    ProTrieve(const string&  clientId, 
	      const string&  sysId, 
	      const string&  baseDir);

    /// Destructor method.
    ~ProTrieve(void);

    /// Set the internal basic parameters.
    /**
     * Set the internal basic parameters.
     *
     * \param clientId  ID of this client.
     * \param sysId     System wide ID. Used to tag data when archiving into
     *                  pro_trieve.
     * \param baseDir   Base (staging) directory used by pro_trieve.
     *
     * \return Reference to object itself.
     */
    ProTrieve& setPars(const string&  clientId,
		       const string&  sysId, 
		       const string&  baseDir);
    
    /// Get value of Client ID member.
    /**
     * Get value of Client ID member.
     *
     * \return          Client ID.
     */
    const string& getClientId(void);

    /// Get value of System ID member. 
    /**
     * Get value of System ID member. 
     *
     * \return          System ID.
     */
    const string& getSysId(void);

    /// Get value of pro_trieve base directory.
    /**
     * Get value of pro_trieve base directory.
     *
     * \return          Base/staging Directory for pro_trieve.
     */
    const string& getBaseDir(void);

    /// Get input staging directory.
    /**
     * Get input staging directory.
     *
     * \return          Input Staging Area for pro_trieve.
     */
    const string& getInputStagingDir(void);

    /// Get request directory.
    /**
     * Get request directory.
     *
     * \return        Request Directory for pro_trieve.
     */
    const string& getRequestDir(void);

    /// Encode a Global Key to a pro_trieve compatible format.
    /** 
     * Encode a Global Key,
     *
     * \return     Encoded Global Key.
     */
    static string encodeGlobalKey(const string& globalKey);

    /// Decode a Global Key from pro_trieve compatible format to normal format.
    /** 
     * Decode a Global Key,
     *
     * \return     Decoded Global Key.
     */
    static string decodeGlobalKey(const string& globalKey);

    /// Archive a file into the TL/pro_trieve.
    /**
     * Archive a file into the associated tape library via pro_trieve.
     *
     * \param context Context of the data file being archived, e.g. ESO-SA. 
     * \param duration Associated time reference for the data object.
     * \param globalKey Global Key of data object in the pro_trieve name 
     *                  space.
     * \param fileName Name of file to archive in the pro_trieve Input Staging
     *                 Area.
     * \param virtualVolId Reference to external volume ID.
     * \param archReqObj Instance of ProTrieveArchiveRequest, which will be
     *                   filled with the appropriate information.
     *
     * \return Reference to object itself.
     */
    ProTrieve& archive(const string&             context,
		       const string&             duration, 
		       const string&             globalKey, 
		       const string&             fileName,
		       const string&             virtualVolId,
		       ProTrieveArchiveRequest&  archReqObj);

    /// Archive a file into the TL/pro_trieve.
    /**
     * Archive a file into the associated pro_trieve controlled tape library.
     * The information necessary to handle the request further is stored in 
     * the map.
     *
     * \param context Context of the data file being archived, e.g. ESO-SA. 
     * \param duration Associated time reference for the data object.
     * \param globalKey Global Key of data object in the pro_trieve name 
     *                  space.
     * \param fileName Name of file to archive in the pro_trieve Input Staging
     *                 Area.
     * \param virtualVolId Reference to external volume ID.
     * \param archReqMap Instance of type ProTrieveArchReqsMap used to keep
     *                   ProTrieveArchiveRequest objects currently being
     *                   archived.
     *
     * \return Reference to object itself.
     */
    ProTrieve& archive(const string&                context,
		       const string&                duration,
		       const string&                globalKey,
		       const string&                fileName,
		       const string&                virtualVolId,
		       ProTrieveArchReqsMap&        archReqMap);

    /// Check the status of the pending archive requests.
    /**
     * Check the status of the pending archive requests within the context
     * of this instance of ProTrieve.
     *
     * \param cleanUp If true, the pro_trieve status files will be deleted
     *                after reading in the contents. If false the host 
     *                application should delete these.
     * \param archReqMap Instance of type ProTrieveArchReqsMap used to keep
     *                   ProTrieveArchiveRequest objects currently being
     *                   archived.
     * \param okStatusObjs  Contains objects with information for successful
     *                      archive requests.
     * \param errStatusObjs Contains objects with information for failed
     *                      archive requests.
     *
     * \return Reference to object itself.
     */
    ProTrieve& checkArchStatus(const bool                       cleanUp,
			       ProTrieveArchReqsMap&            archReqMap,
			       vector<ProTrieveArchiveRequest>& okStatusObjs,
			       vector<ProTrieveArchiveRequest>& errStatusObjs);
    
    /// Clean up data in connection with handled archive requests.
    /**
     * Remove handled status files in connection with an archive request
     * or a set of archive requests.
     *
     * \param  statusObjs Vector of objects with information for handled
     *                    archive requests.
     *
     * \return Reference to object itself.
     */
    ProTrieve& removeArchStatusFiles(vector<ProTrieveArchiveRequest>& 
				     statusObjs);

    /// Request a file from the tape library.
    /**
     * Request a file from pro_trieve. 
     *
     * \param reqId  ID for this specific request. The files will be delivered
     *               into a sub-directory within the pro_trieve Output Staging
     *               Area with that name.
     * \param reqList Vector with ProTrieveFileRequest objects specifying
     *                which files to request.
     * \param reqFileName Name of the pro_trieve Request File produced.
     * \param reqObjMap Map that contains ProTrieveFileRequest instances for
     *                  requested files.
     *
     * \return Reference to object itself.
     */
    ProTrieve& request(const string&                      reqId,
		       vector<ProTrieveFileRequest>&      reqList,
		       string&                            reqFileName,
		       map<string, ProTrieveFileRequest>& reqObjMap);

    /// Check the status of pending file requests.
    /**
     * Check the status of pending Retrieval Requests. If files have been 
     * delivered, the ProTrieveFileRequest objects for these will be put in 
     * the delivFiles vector. The objects for failed requests will be put in 
     * the failFiles vector.
     *
     * \param reqId ID for this request.
     * \param reqFileName Name of the Request File issued to pro_trieve.
     * \param reqObjMap Map that contains ProTrieveFileRequest instances for
     *                  request files.
     * \param delivFiles Vector with information for delivered files.
     * \param failFiles  Vector with information failed requests.
     * \param complete Will be set to true if there are no further pending 
     *                 retrieve requests.
     *
     * \return Reference to object itself.
     */
    ProTrieve& checkReqStatus(const string&                       reqId,
			      const string&                       reqFileName,
			      map<string, ProTrieveFileRequest>&  reqObjMap,
			      vector<ProTrieveFileRequest>&       delivFiles,
			      vector<ProTrieveFileRequest>&       failFiles,
			      bool&                               complete);

    /// Check status of delivered files.
    /**
     * See checkReqStatus().
     */
    ProTrieve& 
    checkReqDelivStatus(const string&                       reqId,
			map<string, ProTrieveFileRequest>&  reqObjMap,
			vector<ProTrieveFileRequest>&       delivFiles);

    /// Check status of request error status files.
    /**
     * See checkReqStatus().
     */
    ProTrieve&
    checkReqErrorStatus(const string&                       reqId,
			map<string, ProTrieveFileRequest>&  reqObjMap,
			vector<ProTrieveFileRequest>&       failFiles);

    /// Remove a file request object from the internal map.
    /**
     * \param fileReqObj File request object.
     */
    ProTrieve& removeFileRequest(const string&                      reqId,
				 ProTrieveFileRequest&              fileReqObj,
				 map<string, ProTrieveFileRequest>& reqObjMap);

    /// Generate a new, unique index.
    unsigned long getUniqueIndex(void);
 
    // Constants.
    /// Version of pro_trieve interface.
    static const string LAB_FILE_VER;
    /// Name of pro_trieve Input Staging Directory.
    static const string IN_DIR;
    /// Extension for pro_trieve Label Files.
    static const string LAB_EXT;
    /// Delimiter for Global Key (<File ID><Delimiter><File Version>).
    static const string GLOB_KEY_DELIM;
    /// Used to encode '_' in Global Keys.
    static const string GLOB_KEY_UNDERSCORE;
    /// Name of pro_trieve user directory, containing error messages from 
    /// archiving requests.
    static const string USR_DIR;
    /// Name of pro_trieve Output Staging Directory.
    static const string OUT_DIR;
    /// Name of pro_trieve Request Directory.
    static const string REQ_DIR;
    ///  Extension for pro_trieve Request Files.
    static const string REQ_EXT;
    /// Extension for temporary files.
    static const string TMP_EXT;
    /// Extension for handled files.
    static const string DONE_EXT;
    /// Extension for successfully handled message files.
    static const string ARCH_STAT_OK_EXT;
    /// Extension for unsuccessfully handled message files.
    static const string ARCH_STAT_ERR_EXT;

    /// Template for generating pro_trieve Label Files.
    static const string LABEL_FILE_TPL;
    
    /// Template for generating pro_trieve file request entries.
    static const string REQ_FILE_TPL;

  private:

    /// Create a Label File in connection with an archive request.
    /**
     * See ProTrieve::archive() for documentation about the input parameters.
     */
    ProTrieve& createLabelFile_(const string&  context, 
				const string&  duration, 
				const string&  globalKey,
				const string&  fileName,
				const string&  virtualVolId,
				string&        labelFileName);
			    
    /// Create request file in connection with a request for one or more files.
    /**
     * \param reqId ID for this request.
     * \param reqList Vector with information about files to be requested.
     * \param reqFileName Name of the pro_trieve Request File created.
     */
    ProTrieve& createReqFile_(const string&                 reqId,
			      vector<ProTrieveFileRequest>& reqList,
			      string&                       reqFileName);
    
    string        clientId_;
    string        sysId_;
    string        baseDir_;
    string        inputStagingDir_;  
    string        requestDir_ ;
    unsigned long uniqueCount_; // Used to generate a unique integer value.
    ACE_Mutex     uniqueCountMutex_; 
};


/// Class to handle information in connection with a file ingestion.
class ProTrieveArchiveRequest
{
  public:
    /// Default constructor.
    ProTrieveArchiveRequest() {}

    /// Constructor setting the internal members.
    /**  
     * Set the internal data members in connection with a pro_trieve 
     * archive request.
     *
     * \param virtualVolId Name of volume.
     * \param globalKey Global Key requested for data object.
     * \param dataFile Staging file to be archived.
     */
    ProTrieveArchiveRequest(const string&    virtualVolId, 
			    const string&    globalKey,
			    const string&    dataFile)
    { 
	DFS_TRACE("ProTrieveArchiveRequest::ProTrieveArchiveRequest()");

	setAll(virtualVolId, globalKey, dataFile);
    }

    /// Set all data members.
    /**
     * Set all data members of the object. This includes deriving the 
     * names for the associated pro_trieve Label File, OK Status File and
     * Error Status file.
     *
     * \param virtualVolId  Name pertaining volume.
     * \param globalKey Global Key requested for data object.
     * \param dataFile File to be archived from staging area.
     *
     * \return Reference to object itself.
     */
    ProTrieveArchiveRequest& setAll(const string&    virtualVolId, 
				    const string&    globalKey,
				    const string&    dataFile)
    {
	DFS_TRACE("ProTrieveArchiveRequest::setAll(1)");

	string labelFile = dataFile + "." + ProTrieve::LAB_EXT;
	fs::path tmpFile(labelFile, fs::no_check);
	string okStatFile = tmpFile.branch_path().native_file_string() + "/" +
	    ProTrieve::USR_DIR + "/" + tmpFile.leaf() + "." +
	    ProTrieve::ARCH_STAT_OK_EXT;

	string errStatFile = tmpFile.branch_path().native_file_string() + "/" +
	    ProTrieve::USR_DIR + "/" + tmpFile.leaf() + "." +
	    ProTrieve::ARCH_STAT_ERR_EXT;

	setAll(globalKey, virtualVolId, dataFile, labelFile, okStatFile, 
	       errStatFile);

	return *this;
    }


    /// Set all data members of the class.
    /**
     * Set all internal data members.
     *
     * \param virtualVolId  Name pertaining volume.
     * \param globalKey Global Key requested for data object.
     * \param dataFile Staging file to be archived.
     * \param labelFile Name of pro_trieve Label File.
     * \param okStatFile Name of pro_trieve OK Status File.
     * \param errStatFile Name of pro_trieve Error Status File.
     *
     * \return Reference to object itself.
     */
    ProTrieveArchiveRequest& setAll(const string&  globalKey,
				    const string&  virtualVolId,
				    const string&  dataFile,
				    const string&  labelFile,
				    const string&  okStatFile,
				    const string&  errStatFile)
    {
	DFS_TRACE("ProTrieveArchiveRequest::setAll(2)");

	DFS_DEBUG("ProTrieveArchiveRequest::setAll: Setting members of "
		  "ProTrieveArchiveRequest instance. Global Key: |%s|, "
		  "Virtual Volume ID: |%s|, Data File: |%s|, "
		  "Label File: |%s|, OK Status File: |%s|, "
		  "Error Status File: |%s|", globalKey.c_str(), 
		  virtualVolId.c_str(), dataFile.c_str(), labelFile.c_str(),
		  okStatFile.c_str(), errStatFile.c_str());

	globalKey_    = globalKey;
	virtualVolId_ = virtualVolId;
	dataFile_     = dataFile;
	labelFile_    = labelFile;
	okStatFile_   = okStatFile;
	errStatFile_  = errStatFile;	

	return *this;
    }

    
    /// Copy over data from another instance.
    /**
     * Copy over the data from another instance of this class.
     *
     * \param archReq 
     *
     * \return Reference to object itself.
     */
    ProTrieveArchiveRequest& copy(ProTrieveArchiveRequest& archReq)
    {
	DFS_TRACE("ProTrieveArchiveRequest::copy()");	

	setAll(archReq.getGlobalKey(), archReq.getVirtualVolId(), 
	       archReq.getDataFile(), archReq.getLabelFile(), 
	       archReq.getOkStatFile(), archReq.getErrStatFile());
	
	return *this;
    }


    /// Return Global Key of this object.
    const string& getGlobalKey(void) 
    {
	DFS_TRACE("ProTrieveArchiveRequest::getGlobalKey()");
	
	return globalKey_;
    }


    /// Return reference to the name of the file to be archived.    
    const string& getDataFile(void) 
    {
	DFS_TRACE("ProTrieveArchiveRequest::getDataFile()");

 	return dataFile_;
    }


    /// Return reference to  the Global Key set internally in the object.
    const string& getVirtualVolId(void) 
    {
	DFS_TRACE("ProTrieveArchiveRequest::getVirtualVolId()");
	
	return virtualVolId_;
    }


    /// Return reference to the associated Label Filename.
    const string& getLabelFile(void) 
    {
	DFS_TRACE("ProTrieveArchiveRequest::getLabelFile()");	

	return labelFile_;
    }


    /// Return reference to the associated OK Status Filename.
    const string& getOkStatFile(void) 
    {
	DFS_TRACE("ProTrieveArchiveRequest::getOkStatFile()");
	
	return okStatFile_;
    }


    /// Return reference to the associated Error Status Filename.
    const string& getErrStatFile(void) 
    {
	DFS_TRACE("ProTrieveArchiveRequest::getErrStatFile()");
	
	return errStatFile_;
    }


    /// Set the internal status of the request.
    ProTrieveArchiveRequest& setStatus(const string& status)
    {
	DFS_TRACE("ProTrieveArchiveRequest::setStatus()");
	
	status_ = status;

	return *this;
    }


    /// Return reference to the internal status of the request.
    const string& getStatus(void)
    {
	DFS_TRACE("ProTrieveArchiveRequest::getStatus()");

	return status_;
    }


    /// Dump the information contained in the object into a string buffer.
    ProTrieveArchiveRequest& dump(stringstream&  buf)
    {
	DFS_TRACE("ProTrieveArchiveRequest::dump()");

	buf.str("");
	buf << "Global Key:        " << getGlobalKey() << endl;
	buf << "Virtual Volume ID: " << getVirtualVolId() << endl;	
	buf << "Data File:         " << getDataFile() << endl;
	buf << "Label File:        " << getLabelFile() << endl;
	buf << "OK Status File:    " << getOkStatFile() << endl;
	buf << "Error Status File: " << getErrStatFile() << endl;
	buf << "Status:            " << endl << getStatus() << endl;	

	return *this;
    }

  private:
    string  globalKey_;         // Global key for data object.
    string  virtualVolId_;      // Associated volume ID.
    string  dataFile_;          // Source data file.
    string  labelFile_;         // Associated Label File.
    string  okStatFile_;        // Name of OK Status File.
    string  errStatFile_;       // Name of Error Status File.
    string  status_;            // Status of request.
};


/// Class used to hold information about one pro_trieve Archive Request.
class ProTrieveArchReqsMap
{
  public:
    /// Constructor method.
    ProTrieveArchReqsMap(void) 
    {
	DFS_TRACE("ProTrieveArchReqsMap::ProTrieveArchReqsMap()");	
    }


    /// Destructor method.
    ~ProTrieveArchReqsMap(void);


    /// Fill vector with all keys defined.
    ProTrieveArchReqsMap& getKeys(vector<string>&  keys);


    /// Add an entry in the Archive Requests Map.
    ProTrieveArchReqsMap& addEntry(ProTrieveArchiveRequest&  archReqObj);


    /// Remove entry.
    ProTrieveArchReqsMap& removeEntry(const string& key);


    /// Check if a file is concerned by the pending Archive Requests for this
    /// instance.
    /**
     * \param filename Name of file to check.
     */
    ProTrieveArchiveRequest* fileConcerned(const string& filename);


    /// Return the number of pending Archive Requests registered in the object.
    unsigned int size(void);


    /// Get reference to internal map.
    map<string, ProTrieveArchiveRequest*>& getMap(void);

    // Constants.
    /// The number of book keeping mappings created per file ingested into
    /// pro_trieve.
    static const unsigned int MAPPINGS_PER_REQ;

  private:
    // Map with info about pending requests.
    map<string, ProTrieveArchiveRequest*>  archReqMap_; 
    // Mutal exclusion protection for archive request map.
    ACE_Mutex                              mutex_;    
};


/// Class to handle information in connection with a request for a file.
class ProTrieveFileRequest
{	
  public:
    /// Constructor method setting internal data members.
    /**
     * Constructor method setting the Global Key + Linkname reference in the
     * object.
     *
     * \param globalKey Global Key for data object to retrieve.
     * \param linkName Output filename requested.
     */
    ProTrieveFileRequest(const string&  globalKey,
			 const string&  linkName) 
    { 
	DFS_TRACE("ProTrieveFileRequest::ProTrieveFileRequest()");
	
	globalKey_ = globalKey;
	linkName_  = linkName;
	delivered_ = false;
	status_    = "";
    }


    /// Set the value of the internal Global Key.
    /**
     * Method setting the Global Key reference in the object.
     *
     * \param globalKey Global Key for data object to retrieve.
     *
     * \return Reference to object itself.
     */
    ProTrieveFileRequest& setGlobalKey(const string& globalKey)
    {
	DFS_TRACE("ProTrieveFileRequest::setGlobalKey()");

	globalKey_ = globalKey;
	return *this;
    }


    /// Return reference to the internal Global Key.
    const string& getGlobalKey(void)
    {
	DFS_TRACE("ProTrieveFileRequest::getGlobalKey()");

	return globalKey_;
    }


    /// Set the internal Linkname.
    /**
     * Method setting the link name reference in the object.
     *
     * \param linkName Global Key for data object to retrieve.
     *
     * \return Reference to object itself.
     */
    ProTrieveFileRequest& setLinkName(const string& linkName)
    {
	DFS_TRACE("ProTrieveFileRequest::setLinkName()");

	linkName_ = linkName;
	return *this;
    }


    /// Return reference to the internal linkname.
    const string& getLinkName(void)
    {
	DFS_TRACE("ProTrieveFileRequest::getLinkName()");
	
	return linkName_;
    }

    
    /// Set the status of the request to delivered/!delivered.
    /**
     * Method setting the delivered flag.
     *
     * \param delivered If the data object was delivered this will be true.
     *
     * \return Reference to object itself.
     */
    ProTrieveFileRequest& setDelivered(const bool delivered)
    {
	DFS_TRACE("ProTrieveFileRequest::setDelivered()");

	delivered_ = delivered;

	return *this;
    }


    /// Get the status of the request.
    const bool getDelivered()
    {
	DFS_TRACE("ProTrieveFileRequest::getDelivered()");

	return delivered_;
    }


    /// Set the internal status of the request (text message).
    /**
     * Set internal status message for this request.
     *
     * \param status Status message.
     *
     * \return Reference to object itself.
     */
    ProTrieveFileRequest& setStatus(const string& status)
    {
	DFS_TRACE("ProTrieveFileRequest::setStatus()");
	
	status_ = status;

	return *this;
    }


    /// Return reference to the internal status message.
    const string& getStatus()
    {
	DFS_TRACE("ProTrieveFileRequest::getStatus()");

	return status_;
    }

  private:
    string  globalKey_;      
    string  linkName_;
    bool    delivered_;
    string  status_;
};


#endif // PROTRIEVE_H

// EOF

