/// -*-c++-*-

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// COPYRIGHT (c) European Southern Observatory
// LICENSE:   ESO internal use only
// 
// PROJECT:   VLT Data Flow System
// AUTHOR:    Christophe Moins- ESO/SDD/DFS
// SUBSYSTEM: arcsys - Archive Systems
//
// $Revision: 1.3 $
// $Log: Qc0FrameAnalyserConfig.h,v $
// Revision 1.3  2008/10/20 09:00:55  cmoins
// DFS05918 : adding of nolog options + display of NO_OBS_CONSTR_SET message
// when no constraint is defined for an observation block
//
// Revision 1.2  2007/10/18 08:03:22  cmoins
// use of a conversion factor for parameters
//
// Revision 1.1  2007/10/17 08:11:15  cmoins
// Replacement of Qcm by Qc0
//
//
// 
//------------------------------------------------------------------------------

/// Configuration class of Qc0FrameAnalyser
#ifndef QC0_FRAME_ANALYSER_CONFIG_H
#define QC0_FRAME_ANALYSER_CONFIG_H

#include <string>
#include <map>
#include "GenConfig.h"

class Qc0FrameAnalyserConfig : public GenConfig
{
public:
  

  struct ParamKeyFactor {
    ParamKeyFactor(std::string keyword,double convFactor):
      keyword_(keyword),convFactor_(convFactor)
    {}
    std::string keyword_;
    double convFactor_;
  };
  /// type that allows to associate a keyword and a factor to a parameter id 
  typedef std::map<std::string,ParamKeyFactor> ParamKeyList;

  /// \brief Default Constructor
  /// \param argc number of arguments of the command line
  /// \param argv command line
  Qc0FrameAnalyserConfig(int argc,
			 const char* argv[]);

  /// \brief Destructor
  virtual ~Qc0FrameAnalyserConfig()
  {}

  /// indicates if the version is requested
  /// returns 1 if the version is requested, 0 otherwise
  bool isVersion() const;

  /// indicates if only measured constraints are requested
  /// returns 1 if only measured constraints are requested, 0 otherwise
  bool isOnlyMeasured() const;

  /// gives the output filename
  /// returns the filename if set, an empty string otherwise
  std::string getResultFile() const;

  /**
   * indicates if logging messages should be produced
   * return true if no logging message should be produced, false otherwise
   */
  //bool Qc0FrameAnalyserConfig::noLogging() const;
  bool noLogging() const;

  /// gives the polled directory name ended by a "/"
  /// returns the polled directory name if set, the current one otherwise
  std::string getPolledDir() const;

  /// gives the log file name. The default one is : QC0./YYYY-MM-DD
  /// returns the log file name if successful, the default one otherwise
  std::string getLogFile() const;

  /// gives the verbose level.
  /// returns the verbose level if set, -1 otherwise
  int getVerboseLevel() const;

  /// gives the ambient database connection alias
  /// returns the alias if set, the default value (AMBIREP) otherwise
  std::string getAmbientAlias() const;

  /// gives the obrep2 database connection alias
  /// returns the alias if set, the default value (OBREP) otherwise
  std::string getObRep2Alias() const;

  /// gives the ARSystem database connection alias
  /// returns the alias if set, the default value (ARSYSTEM) otherwise
  std::string getArSystemAlias() const;
 
  /// gives the list of required parameters with their keywords
  /// return 0 if successful, -1 otherwise
  int getParamKeyList(ParamKeyList& rparamList) const;
  
  /// indicates if the header shall be put in the report
  bool isHeader() const;

  /// indicates if the help shall be displayed
  bool isHelp() const;

private :
  /// static constants defining the available options
  static const std::string VERSION_OPTION;
  static const std::string ONLY_MEASURED_OPTION;
  static const std::string RESULT_FILE_OPTION;
  static const std::string DIRECTORY_OPTION;
  static const std::string LOG_FILE_OPTION;
  static const std::string NO_LOG_OPTION;
  static const std::string VERBOSE_LEVEL_OPTION;
  static const std::string HEADER_OPTION;
  static const std::string AMBIENT_ALIAS_OPTION;
  static const std::string OB_REP2_ALIAS_OPTION;
  static const std::string ARSYSTEM_ALIAS_OPTION;
  static const std::string HELP_OPTION;
};
#endif // QC0_FRAME_ANALYSER_CONFIG_H
