#ifndef _DFSLOG_h_
#define _DFSLOG_h_

#ifdef __cplusplus
extern "C" {
#endif
/*
 * E.S.O. - VLT project/ESO Dfs
 * $Id: dfslog.h,v 1.1 2004/09/06 14:33:18 cguirao Exp $
 *
 * Standard routines for generating messages and forwarding the messages
 * to DataFlow Logging environment.
 *
 * Intended Purpose and Guidelines for Design:
 *
 * This set of routines provides the following capabilities:
 *     - standard means to classify, address and format messages from
 *       applications running in the Data Flow System environment.
 *     - encapsulation of message forwarding functionality, allowing
 *       applications to forward messages to the Data Flow System logger.
 *
 * In order to achieve the above, the set of routines must obey the
 * following guidelines:
 *     - be simple to incorporate into DFS applications, using languages
 *       and features which are available to the majority of applications.
 *     - be lightweight in execution, so as to minimize its impact on the
 *       application.
 *     - be flexible and configurable in operation.
 *
 * The DFS Message:
 *
 * As per the DFS Logging and Monitoring Document, and similar to the
 * Unix syslog messages and VLT messages, DFS Messages have
 * the following parameters:
 *
 *     - date and time of message generation
 *     - priority of message (see below).
 *     - address of sender (ASCII strings containing no white space characters):
 *         - host name where application is executing,
 *         - component name (typically the application name),
 *         - label (name to differentiate between applications of the same name
 *           and/or different components/threads within an application).
 *     - message (a single line of ASCII text).
 * A DFS Message has the following format:
 *
 *   DD/MM/YYYY HH:MM:SS [PRIORITY] HOST COMPONENT LABEL MESSAGE
 *
 * For example:
 *
 *   12/12/1998 10:12:34 [Warning] wu0pl DhsSubscribe PIPE Cleaning up...
 *   12/01/1999 23:03:43 [Error] wu0dhs dhs dhs Keywords not located
 *   14/12/1999 14:18:06 [Info] wu0dhs testapp debug files open: 4, status ok
 *
 * The current set of routines provides the following:
 *
 *     - accessors for getting/setting default information: priority,
 *       host, component and label names.
 *     - switch on/off of message transmission to DFS Logging system (syslogd).
 *     - access to formatted messages.
 *     - optional forwarding of messages to STDOUT or STDERR.
 *     - request of version information
 *
 * Disabling DFSLog:
 * 
 *    The dfslogInitialize() and dfslogMessage() calls will return immediately,
 *    and not initialize/send a message if:
 *    the environment variable DFSLOG is set to "No", "NO", "no" or "disabled". 
 *    Any other setting for DFSLOG environment variable indicates that it is enabled.
 *
 * who                  when       what
 * -------------------  ------------------------------------------------
 * N. Kornweibel        14/01/99   Original Version
 */

/* include files */

#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>




/*
   Definitions of message field lengths.
   The definitions do not include the null character.
*/
 
#define DfsLog_MaxWorkstationNameLength 15
#define DfsLog_MaxComponentNameLength 25
#define DfsLog_MaxLabelNameLength 25
#define DfsLog_MaxMessageLength 160
#define DfsLog_MaxDateTimeLength 19
#define DfsLog_MaxPriorityStringLength 9




/* version number and string accessors */

const char *dfslogGetVersionString();
int dfslogGetVersionNumber();




/* enum of user priority states */

typedef enum DfsLogPriorityStates { DfsLogEmergency = 0, DfsLogAlert,
    DfsLogCritical, DfsLogError, DfsLogWarning, DfsLogNotice,
    DfsLogInfo, DfsLogDebug } DfsLogPriorityState;



/*****************************************************************************
    Conversion of priority states to strings, as follows:

    DfsLogEmergency    ->    "Emergency"
    DfsLogAlert        ->    "Alert"
    DfsLogCritical     ->    "Critical"
    DfsLogError        ->    "Error"
    DfsLogWarning      ->    "Warning"
    DfsLogNotice       ->    "Notice"
    DfsLogInfo         ->    "Info"
    DfsLogDebug        ->    "Debug"
*****************************************************************************/

const char *dfslogPriorityToString(DfsLogPriorityState);

/* returns 0 on success and -ve on failure to convert. */
int dfslogStringToPriority(const char *, DfsLogPriorityState*);




/**************************************************************************
  Setting default information:
    default host: if not set, will use $DFSLOG_HOST
                  if not set, will use $HOST.
                  else "".
    default component: if not set, will use $DFSLOG_COMPONENT
                  else "".
    default label: if not set, will be: PID=pid, where pid is the process
                   id from getpid().
    default priority: if not set, will be DfsLogInfo.

    Returns the string/value written.

**************************************************************************/

const char *dfslogSetDefaultHost(const char *);
const char *dfslogSetDefaultComponent(const char *);
const char *dfslogSetDefaultLabel(const char *);
DfsLogPriorityState dfslogSetDefaultPriority(DfsLogPriorityState);




/* Getting default information: */

const char *dfslogGetDefaultHost();
const char *dfslogGetDefaultComponent();
const char *dfslogGetDefaultLabel();
DfsLogPriorityState dfslogGetDefaultPriority();




/***************************************************************************
    This function returns the minimum priority a message may have in
    order to be sent to the DFSLog. While all messages are written to
    file/stdout/stderr as per initialisation, only messages of this minimum
    priority and above are communicated to the DFSLog (and only when
    DFSLOG_LOGMESSAGE has been specified.
 
    The value may be set with dfslogSetLowestLoggedPriority(), and is by
    default set at DfsLogError.
***************************************************************************/

DfsLogPriorityState dfslogGetLowestLoggedPriority();
DfsLogPriorityState dfslogSetLowestLoggedPriority(DfsLogPriorityState);





/******************************************************************************
    Initializing the logging system:
    If not called explicitly, it will be called on first call to
    dfslogMessage(), using default parameters.

    The int parameter allows definition of the following:
        - forward (or not) log messages to DFS Logger.
        - forward (or not) log messages to stdout(F) and/or stderr(F).
    The int value is built by or-ing the following definitions:

    DFSLOG_LOGMESSAGES   to send messages to dfslog.
    DFSLOG_STDOUT        to send messages to stdout(F).
    DFSLOG_STDERR        to send messages to stderr(F).

    The default value is: DFSLOG_LOGMESSAGES | DFSLOG_STDERR, logging to
    the dfsLog system and stderr..

    The char * parameter is a filepath. The file will be created if
    necessary, then opened for appending.

    The default value is (char *)0, indicating that no file is to be
    used.

    It may be called at any time during execution to modify behavior.
******************************************************************************/

extern const int DFSLOG_LOGMESSAGES;   /* to send messages to dfslog.   */
extern const int DFSLOG_STDOUT;        /* to send messages to stdout(F).*/
extern const int DFSLOG_STDERR;        /* to send messages to stderr(F).*/

int dfslogEnabled();
int dfslogInitialize(int, const char *);




/*****************************************************************************
    sending messages:
    - uses printf style variable argument lists for formatting the
    message.
    - the message will be terminated at the first new line character and
      has a maximum length (not set just yet).
    - the message will be pre-pended with date, time, priority and address
      information, using current system time and default information.
    - resultant message is returned.
    - there is no means to determine if the message has been
      successfully logged/transmitted.
******************************************************************************/

char *dfslogMessage(char *, ...);




/*****************************************************************************
    Useful functions to send messages as priority:
    Notice, Warning, Error or Critical, without modifying the default
    priority.
*****************************************************************************/

char *dfslogNoticeMessage(char *, ...);
char *dfslogWarningMessage(char *, ...);
char *dfslogErrorMessage(char *, ...);
char *dfslogCriticalMessage(char *, ...);

#ifdef __cplusplus
}
#endif

#endif /* _DFSLOG_h_ */
/* end of file dfslog.h */

