#ifndef SYBCONECTION_H
#define SYBCONECTION_H
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// $Id: sybconnection.h,v 1.12 2008/09/05 14:47:26 cguirao Exp $
// 
// COPYRIGHT (c) European Southern Observatory
//
// PROJECT:   VLT Data Flow System
// AUTHOR :   Nick Kornweibel - ESO/DMD
// SUBSYSTEM: Pipeline
//
// $Revision: 1.12 $
// $Log: sybconnection.h,v $
// Revision 1.12  2008/09/05 14:47:26  cguirao
// Fixing error:  error: extra qualification 'SybUtils::' on member 'EmptyConnection'
//
// Revision 1.11  2006/12/18 13:05:49  cmoins
// replacement of localtime by localtime_r
// adding of convertFromUnixTimeval
//
// Revision 1.10  2006/05/05 16:49:49  szampier
// added methods to support high resolution time - unix timeval struct -
// added getRowCount method
//
// Revision 1.9  2005/08/18 17:03:27  nferguso
// Bugfixes connected to the DFS port, including:
// - Indices off by 1 in SybConn getXXXValue() calls.
// - Duplicate quotemarks in SQL queries.
// - Error handling now includes DB server errors (courtesy of Stefano Zampieri).
//
// Revision 1.8  2005/08/11 17:37:53  szampier
// added support for error handler compatible with the RW DBtools one
//
// Revision 1.7  2004/04/16 15:50:37  nkornwei
// Added getDataLength() to SybaseResultSet and updated unit tests
//
// Revision 1.6  2004/04/07 09:09:27  nkornwei
// Updated header comment
//
// Revision 1.5  2004/03/31 11:51:56  nkornwei
// Modified error handling: only after calling dbresult()
// for each SQL instruction can it be determined if the
// SQL command succeeded or not. Thus the full result set
// must be scanned before any possible errors will be reported
// to the library.
//
// Revision 1.4  2004/03/30 16:07:47  nkornwei
// Slightly modified executeAndDisplay() parameter interface
//
// Revision 1.3  2004/03/25 14:40:56  nkornwei
// Added note to header about string_rtrucation setting.
//
// Revision 1.2  2004/03/23 12:51:01  nkornwei
// Added interface to specify MM/DD/YY or DD/MM/YY format for date output
//
// Revision 1.1.1.1  2004/03/09 09:13:18  nkornwei
// first version
//
//
// C++ Wrapper Classes for the SYBASE DB-Library.
//
//
// Supported platforms:
//
// Linux 6.4, Sybase 11.9
// Linux 7.3, Sybase 11.9
// Solaris 8, Sybase 12.0
//
//
//
//-----------------------------------------------------------------------------

#include <string>
#include <ctime>
#include "SybaseStatus.h"

using namespace std;

extern "C" {
#include <sybfront.h>
#include <stdlib.h>
#include <sybdb.h>
}


// forward declarations

class SybaseMessageListener;
class DefaultMessageListener;
class SybaseResultSet;

///
/// SybaseConnection represents a connection to a sybase server.
/// The connection can be used to execute() commands and return
/// SybaseResultSet objects.
/// <p>
/// Updates/inserts to TEXT values can be performed in one of two 
/// ways:<br>
/// Regular SQL Statement, using SybaseConnection::execute()
/// <p>
/// This library creates a connection where string_rtrucation
/// is on. This means that string types that are too long for
/// the field will simply be trucated. To set it off, run the
/// command: "set string_rtrucation off" on the connection
/// prior to using the connection. Setting the property off will
/// result in a rejected SQL operation if the field is too long.
/// <p>
/// Update of a specific text column through the SybaseResultSet.
/// This means of writing to columns of type TEXT is somewhat odd,
/// and requires first performing a select on the text column 
/// - please see the header of SybaseResultSet for details.
/// <p>
/// Ensure that commandSucceeded() calls are made after
/// releasing the results set (as this is the only way to be sure there
/// were no problems). The correct sequence for checking the success
/// of a SQL command, when the result set contents are not of interest
/// is:
/// <p>
///     SybaseResultSet rs = conn.execute(sqlcommmand);
///     rs.releaseResults();
///     if (!rs.commandSucceeded()) {
///         // handle the error...
///         throw runtime_error("Error executing SQL Command");
///     }
/// <p>
/// @warning This library is not thread safe.
/// @warning By default this library does not check string lengths
///

class SybaseConnection {

   private:

        DBPROCESS            *dbproc;  // connection with SQL Server

        int commandIndex;

        // singleton, static message listener used by all SybaseConections
        // to report errors.
	
        static DefaultMessageListener defaultListener;
        static SybaseMessageListener* listener;
	
	// call back method to be invoked when a Sybase error occours
	static SybaseStatus::ErrorHandler errorHandler;
	
        static int errHandler(DBPROCESS *dbproc, int severity, int dberr,
            int oserr, char *dberrstr, char *oserrstr);

        static int msgHandler( DBPROCESS *dbproc, DBINT msgno,
            int msgstate, int severity, char *msgtext, char *srvname,
            char *procname, int line);


            // values of commandIndex used to determine if this
            // is initialized or not.

        static const int NotInitialized  = -2;
        static const int Initialized     = -1;
        static const int Connected       =  0;

        void clearLastResults();

        string username;
        string password;
        string db;

    public:

            /// Create an unconfigured, unconnected SybaseConnection.
  
        SybaseConnection();

           /// Close the connection, if open, and free resources.

        virtual ~SybaseConnection();

            /// Connect to the specified database
            /// Details of connection errors are typically passed through the 
            /// SybaseMessageListener object.
            /// @param server the name of the server, as appears in the interfaces file
            /// @param user the name of the user account on Sybase
            /// @param pw the password for the Sybase user account
            /// @return 1 on success, else 0.

        int connect(string, string, string);

            /// Connect to the specified database
            /// @param server the name of the server, as appears in the interfaces file
            /// @param user the name of the user account on Sybase
            /// @param pw the password for the Sybase user account
            /// @return 1 on success, else 0.

        int connect(const char *, const char *, const char *);

            /// Disconnect from the database

        void disconnect();

            /// Execute the given SQL query or update.
            /// The result set is valid until a new command is executed or
            /// the connection is dropped.
            /// @param cmd SYBASE SQL command to execute
            /// @return a SybaseResultSet containing the results.

        SybaseResultSet execute(const string&);

            /// Execute the given SQL query or update.
            /// The result set is valid until a new command is executed or
            /// the connection is dropped.
            /// @param cmd SYBASE SQL command to execute
            /// @return a SybaseResultSet containing the results.

        SybaseResultSet execute(const char *);
	
	static void setErrorHandler(SybaseStatus::ErrorHandler handler);

            /// Set the message listener (extending SybaseMessageListener).
            /// There can be only one of these objects set, the next call
            /// replaces the previously set object.
            /// @param l the message listener object

        static void setMessageListener(SybaseMessageListener*);

            /// Get the currently set message listener
            /// @return a SybaseMessageListener object, or NULL if there is not one

        static SybaseMessageListener* getMessageListener();

            /// Get the current connection command number
            /// @return a unique integer value for the last command, or -1 if
            /// there was no last command (used by the SybaseResultSet).

        int getCommandIndex();

            /// Determine if the SybaseConnection is connected.
            /// @return 1 if connected, else 0

        int isConnected();

            /// Get the DBPROCESS object of the DB-Libaray
            /// This call provides access to the SYBASE connection
            /// handle and may be used to call the SYBASE library
            /// directly, if required to perform a task not provided by the
            /// SybaseConnection class. It is not intended for typical use of the
            /// SybaseConnection class.
            /// @return the DBPROCESS object of the DB-Libaray

        DBPROCESS *getDbProcess();

            /// Get the current user name
            /// @return username, or "" if not connected.

        string getUsername();

            /// Get the current password.
            /// @return password, or "" if not connected.
        string getPassword();

            /// Get the current DB name.
            /// @return db server, or "" if not connected.

        string getDb();
};


///
/// The result and return code of SybaseConnection::execute().
/// <p>
/// Represents the results from a query/update, including the
/// result set, if appropriate. Not all SQL commands return
/// a result set, in which case this class simply contains 
/// an indication of the success/failure of the command.
/// <p>
/// The caller must free the result set when finished.
/// A result set becomes invalid after the next call to
/// the execute() on the same SybaseConnection and should not be used,
/// after the call.
/// <p>
/// This class is more than the result set, it is the return
/// code (success/failure) of the query as well as all result sets
/// returned. It should therefore be renamed to SybaseResults, or
/// something more indicative of its purpose.
/// <p>
/// The correct way to use the SybaseResultSet returned from
/// SybaseConnection::execute() is to check if the query/update
/// succeeded, then process all result sets returned, as follows:
/// <p>
/// Execute the statement and display all results.
/// This function demonstrates the correct way to process result sets:
/// <PRE>
///    SybaseResultSet rc = conn.execute("SELECT * FROM BOOKS");<br>
///    if (rc.commandSucceeded()) {<br>
///        while (rc.hasResultSet()) {<br>
///            while (rc.next()) {<br>
///                // process result set...<br>
///            }<br>
///        }<br>
///    }<br>
/// </PRE>
/// <p>
/// As odd as it sounds, the SybaseResultSet is also an interface to writing
/// data to TEXT type fields in Sybase. The DB-Library requires a select be
/// performed on the table and column containing the TEXT type to be written to, 
/// followed by the write call.
/// <p>
/// Of course you could always use the SybaseConnection::execute(), as well.
/// <p>
/// Here are some more details: http://manuals.sybase.com/onlinebooks/group-cn/cng1250e/dblib/@Generic__BookView
/// <p>
/// @todo Find a better name for this class
/// @todo Replace/remove the writeTextValue() method, if possible.
///

class SybaseResultSet {

            // used to determine if this result set is still valid.

        int myCommandIdx;
   
            // the return code of the SQL call.

        int retcode;

            // the connection associated with this result set.

        SybaseConnection *myConnection;

           // value of retcode when command has failed.

        static const int CommandFailed = -1;
 
          // value of retcode when the command succeded

        static const int CommandSucceeded = 1;

    public:

            /// default constructor
            /// creates an invalid SybaseResultSet
            /// There is no need to create result sets
            /// directly, they are returned from
            /// SybaseConnection::execute().

        SybaseResultSet();

            /// Create a result set from the connection give.
            /// Usual method to get a result set is from
            /// There is no need to create result sets
            /// directly they are returned from
            /// SybaseConnection::execute();

        SybaseResultSet(SybaseConnection *, int);

            /// Determine if this result set is still valid. A
            /// result set becomes invalid if the connection closes
            /// or another query/update is executed on the connection.
            /// @return 1 if this result set is still valid, else 0

        int isValid();

            /// Determine if a command succeeded or not.
            /// @return 1 if the query/update creating this result set succeeded, else 0

        int commandSucceeded();

            /// move the the next row in the current result set
            /// @return 1 if there is a row, 0 if there is no next row.

        int next();

            /// Determine if there are any (more) result sets.
            /// If there is a result set to process, it is processed with next().
            /// When next() returns 0, the result set is processed.
            /// There maybe multiple result sets returned in a single
            /// SybaseResultSet.
            /// @return 1 if there is an an unprocessed result set, else 0.

        int hasResultSet();

            /// Skip to the current result set.
            /// @return 1 if it was skipped, else 0.

        int skipResultSet();

            /// skip the specified number of result sets.
            /// @return the actual number of result sets skipped.

        int skipResultSet(int);

            /// cancel all remaining result rows and result sets.

        void releaseResults();

            /// Get the number of columns in the current result set
            /// @return the number of column in the result set, or -1 on error

        int getColumnCount();
	
	/// Get the number of rows affected by the last command
	
	int getRowCount() { return DBCOUNT(myConnection->getDbProcess()); };

            /// Get the name of the specified result set column
            /// @param idx the column index, starting at 1.
            /// @return the name of the column specified, or NULL on error

        char *getColumnName(int);

            /// Get the type of the column
            /// @see SybaseType for list of types.
            /// @param idx the column number, from 1.
            /// @return the column type

        int getColumnType(int);

            /// Get the length of the column.
            /// @param idx the column number, from 1.
            /// @return the column length

        int getColumnLength(int);

            /// Get the data length.
            /// This is  the length of data for the current row,
            /// for the specified column.
            /// @param col the column number, from 1.
            /// @return the length of the data, or -1 on error.

        int getDataLength(int);

            /// Determine if a cell is null.
            /// @param idx the column number, from 1.
            /// @return 1 if the column has a null value, else 0

        int isNull(int);

            /// Get the pointer to the raw cell data 
            /// for "behind the scenes" access to data.
            /// @warning: do not free the returned data.
            /// @param idx the column number, from 1.
            /// @return the raw pointer to the data 

        void *getData(int);

            /// Get the cell value as a string
            /// @param idx the column number, from 1.
            /// @return the cell value as a string, or "" if column is not appropriate type

        string getStringValue(int);

            /// Get the cell value as a double or 0.0 if inappropriate type.
            /// @param idx the column number, from 1.
            /// @return the cell value as a double or 0.0 if inappropriate type.

        double getDoubleValue(int);

            /// Get a date/time cell as a readable date string - or 
            /// "" if inappropriate type
            /// @param idx the column number, from 1.
            /// @return a date/time cell as a readable date string - or 

        string getDateStringValue(int);
        
            /// Get the value of a text column
            /// @param val string to contain the text value
            /// @param idx the column number, from 1.
            /// @return 1 if succeeded, else 0.

        int getTextValue(int, string&);

            /// Get a date/time column as a unit time_t value
            /// or (time_t)-1 if inappropriate type.
            /// @param idx the column number, from 1.
            /// @return a date/time column as a unit time_t value, or (time_t)-1 if inappropriate type

        time_t getDateAsTimeTValue(int);

            /// Get a date/time column as a unix timeval struct
            /// @param idx the column number, from 1.
            /// @param tval the output unix timeval struct
            /// @return 1 if succeeded, else 0.

	int getDateAsTimevalValue(int, struct timeval&);


            /// Get the value as a long, or 0 if inappropriate type.
            /// @param idx the column number, from 1.
            /// @return the value as a long, or 0 if inappropriate type.

        long getLongValue(int);

            /// Get boolean value (as an int), or 0 if inappropriate type.
            /// 1 is true, 0 is false.
            /// @param idx the column number, from 1.
            /// return boolean value as an int, or 0 if inappropriate type.

        int getBooleanValue(int);
        
            /// Write Text to the specified column TEXT type column. 
            /// This is an alternate means to write a TEXT type column.
            /// The simplest means is to use SybaseConnection::execute() with
            /// an update or insert statement.
            /// This method ONLY works
            /// if this SybaseResultSet is the result of a SELECT on the appropriate
            /// table and column.
            /// If not, the DB-Library can not handle the request
            /// The given string is writen to the specified text column
            /// at the specified row.
            /// For example to write to the text col "txt" in table "txttbl"
            /// this result set must be the result set from:
            /// <PRE>
            /// SELECT txt from txttbl...<br>
            /// </PRE>
            /// In the above case, the parameters to the writeTextValue() 
            /// call would be as follows:
            /// the columnSpec value would be "txttbl.txt", and colIdx = 1
            /// The columnSpec and colIdx must match the SELECT call executed
            /// to generate this result set.
            /// Currently text column timestamps are NOT checked.
            /// @param colSpec the name of the table and column (with tzpe TEXT) to write text to.
            /// @param colIdx to index of column in the result set, starting with 1.
            /// @param value the text value to be written to the database.
            /// @return 1 on success, else 0.
        int writeTextValue(const char *, int, string&); 
};

///
/// The SybaseMessageListner can be set to receive all messages
/// and errors from the DB_Library (default is to send to stdout/stderr).
/// <p>
/// Message Listener that is to be extended by the user and set to
/// the SybaseConnection object using setMessageListener().
/// <p>
/// This call back model is used to receive error and info messages
/// from Sybase.
/// 

class SybaseMessageListener {

   public:

           /// Called when an error message has been received
           /// @param int severity
           /// @param int dberr Database error number
           /// @param int oserro OS error number
           /// @param char *dberrstr Database error message
           /// @param char *oserrstr OS error message

       virtual void errorMessage(int, int, int, char *, char *) = 0;

           /// Called when an information message has been received.
           /// @param long msgno the message number
           /// @param int msgstate the message state
           /// @param int severity the message severity
           /// @param char *msgtext the message text
           /// @param char *srvname the server name
           /// @param char *procname the procedure name
           /// @param int line the associated line in the SQL code

       virtual void infoMessage(int, int, int, char *, char *, 
           char *, int) = 0;

           /// Destructor

       virtual ~SybaseMessageListener() {};

};

///
/// Default message listener (print to stdout/stderr),
/// set as default message listener in a SybaseConnection,
/// and may be replaced by calling SybaseConnection::setMessageListener().
///

class DefaultMessageListener : public SybaseMessageListener {
   public:
       void errorMessage(int, int, int, char *, char *);
       void infoMessage(int, int, int, char *, char *,
           char *, int);

};

///
/// Definition of all supported Sybase types.
/// Class used for static constant values only.
/// Contains a (not-exhaustive) set of data types
///

class SybaseType {

    public:

            // data type definitions - not exhaustive.

        static const int Double  = 1;
        static const int Date    = 2;
        static const int String  = 3;
        static const int Boolean = 4;
        static const int Text    = 5;
        static const int Image   = 6;
        static const int Tiny    = 7;
        static const int Short   = 8;
        static const int Long    = 9;

            // Money, Images, Boundaries, etc...
        static const int NotSupported   = 0;
};

///
/// Utilities class containing static methods
/// for general purpose operations.
///

class SybUtils {

    private:

            // show debug output or not - default to no.
            // applies to all classes in the Sybase Connection suite.

        static bool debug;

            // has the environment been checked to see if SYBDEBUG is set > 0.

        static bool envChecked;

        static string padNumber(int n, int padSize = 2);

    public:

            /// definition of an empty connection

        static SybaseConnection EmptyConnection;

            /// Escape all "'" characters in a string and return it.
            /// This should be used to render data "SQL safe".
            /// @param val the string to be SQL escaped.
            /// @return string an 'SQL escaped' copy of the given string. 

        static string escapeForSQL(string); 
        
            /// Enclose the given string in SQL single quotes, escaping any 
            /// which are already present, and return it. This should be used 
            /// to obtain a quoted string suitable for use in an SQL query. 
            /// @param val the string to be quoted for SQL. 
            /// @return string an SQL quoted version of the given string. 

        static string quoteForSQL(string); 

            /// Migrate the data in the Sybase DBDATEREC into the
            /// given unix tm struct.
            /// @param sybtm the input Sybase DB-Library date/time structure
            /// @param tm the output unix tm struct

        static void convertToUnixTime(const DBDATEREC&, tm&);

            /// Migrate the data in the Sybase DBDATEREC into the
            /// given unix timeval struct.
            /// @param sybtm the input Sybase DB-Library date/time structure
            /// @param timeval the output unix timeval struct

        static void convertToUnixTimeval(const DBDATEREC&, struct timeval&);
	
            /// Convert tm to one of the accepted Sybase SQL datetime formats
            /// @param time the unix tm stuct containing the date time data
            /// @return a standard SQL format time/date for use in SQL statements

        static string convertFromUnixTime(const tm& utm, bool dmy=false);

            /// Convert time_t to one of the accepted Sybase SQL datetime formats
            /// @param time the unix time_t value containing the time
            /// @return a standard SQL format time/date for use in SQL statements

        static string convertFromUnixTime(const time_t ut, bool dmy=false);

            /// Convert timeval to one of the accepted Sybase SQL datetime formats
            /// @param time the unix timeval value containing the time
            /// @return a standard SQL format time/date for use in SQL statements
        static string convertFromUnixTimeval(const timeval& rtimeval);
	
            /// Convert time_t to human readable format
            /// @param time the unix time_t value containing the time
            /// @return a human readable time/date format
        
        static string dateToString(const time_t);

            /// Is the library is verbose mode or not.
            /// @return true if debug output is turned on, off.

        static bool isVerbose();

            /// turn debug output on/off.
            /// @param bool true if verbose mode is to be turned on

        static void setVerbose(bool);

            /// utility routine execute a given SQL string on given connection and print results.
            /// @param conn the connection to use
            /// @param sql the SQL code to be executed

        static void executeAndDisplay(SybaseConnection&, const string&);

            /// utility routine execute a given SQL string on given connection and print results.
            /// @param conn the connection to use
            /// @param sql the SQL code to be executed

        static void executeAndDisplay(SybaseConnection&, const char*);

            /// Display and out of memory error and abort() the program.
            /// This method is called by routines when memory allocation fails.

        static void outOfMemory();
};

// end of file.
#endif
