package org.eso.ohs.core.dbb.client;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.util.Collections;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Vector;

import javax.swing.Box;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.Border;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;

import org.apache.log4j.Logger;
import org.eso.ohs.core.dbb.sql.DbbInvalidSqlOperatorType;
import org.eso.ohs.core.dbb.sql.DbbSqlChunk;
import org.eso.ohs.core.utilities.GTreeNode;

/** 
 * DbbPanel is the class defining the general behavior of the browser
 * panels containing the sql widgets, the column selection widgets,
 * the sorting criteria widgets etc.<br> 
 * This class implements a DbbWidgetListener in order to listen all
 * the widget events and to append the generating widgets to a ordered
 * dynamical list. This list keeps trace of which widget generated an
 * event and in which order the events have been generated.
 * 
 * @see DbbWidget
 * @see DbbWidgetEvent
 * @see DbbWidgetListener
 *
 * @version $Revision: 1.16 $
 * @author Carlo Boarotto, ESO 19-Jul-1999
 * 
 */
public class DbbPanel extends JPanel implements DbbWidgetListener {

    private static final String rcsid = "$Id: DbbPanel.java,v 1.16 2006/04/28 08:48:21 gdonino Exp $";

    /** Constraints to bgbc used by the GridBagLayout layout manager */
    protected GridBagConstraints gbc_;

    /**
     * The keyword storing the name of a widget 
     */
    public static final String NAME_KEY =  "name"; 
    /**
     * The  keyword storing the value of a widget 
     */   
    public static final String VALUE_KEY =  "value"; 
    /**
     * The  keyword storing the order of a widget 
     */   
    public static final String ORDER_KEY =  "order";    
    /**
     * The  keyword storing the position of a widget in the panel.
     * the -1 value means that the widget is not displayed
     */ 
    public static final String POSITION_KEY = "position";

    /** Number of displaying columns */
    private int colNumber_ = -1;


    /** Number of displaying columns */
    private int adjLabelLength_ = 8;

    /** Current column: used when laying out the widgets in the panel */
    private int currentCol_ = 0;

    /** Current row: used when laying out the widgets in the panel */
    private int currentRow_ = 0;

    private static Logger stdlog_ = Logger.getLogger( DbbPanel.class );
    
    private String dbbPanelType_;

    /**
     * List of widgets contained in the DbbColumnPanel object.
     * @see DbbWidget
     */
    private Vector<DbbWidget> widgetList_ = new Vector<DbbWidget>();

    /** Border that surrounds the panel when displaying a title. */
    private TitledBorder titled_ = null;

    /**
     * Dynamical ordered list of DbbWidgets objects contained in the
     * DbbPanel. DbbWidegts, not empty that fired a DbbWidgetEvent
     * event, are appended to the list; if empty they are removed from
     * the list.
     * @see DbbWidget
     */
    private Vector<DbbWidget> activeWidgetList_ = new Vector<DbbWidget>();


    //----------------------------------------
    // Constructors
    //----------------------------------------

    /**
     * Create a panel with a title and a grid of DBB widgets.<BR>
     * Widgets should be added to this panel with   
     * <code>addWidget( DbbWidget, int, int )</code>.
     *
     * @param title		Title label of the panel.
     *
     * @see #addWidget( DbbWidget, int, int )
     */
    public DbbPanel( String title ) {
		this( title, 0 );
	}

    /**
     * Create a panel with a specific title and a grid of 
     * <code>col_number</code> columns.<BR>
     * Each column contains zero or more DbbWidgets.<BR>
     * Widgets should be added to this panel with   
     * <code>addWidget( DbbWidget )</code>.
     *
     * @param title		Title label of the panel.
     * @param col_number	Number of widget columns.  
     *
     * @see #addWidget( DbbWidget )
     */
    public DbbPanel(String title, int col_number) {

	super(true);
	colNumber_ = col_number;

	setLayout(new GridBagLayout());
	setOpaque(true);

	gbc_ = new GridBagConstraints();
	gbc_.insets = new Insets(0,0,0,0);

	gbc_.gridx = 0;
	gbc_.gridy = 0;

	// Display components within the cells
	// gbc_.fill = GridBagConstraints.HORIZONTAL;
	gbc_.fill = GridBagConstraints.NONE;

	displayLabel(title);

	gbc_.weightx = 1;
	gbc_.weighty = 1;
    }

    /**
     * Create a panel without title and just one column of widgets.
    public DbbPanel() {
	this(null, 1);
    }
     */

    //----------------------------------------
    // Protected Methods
    //----------------------------------------

    /**
     * Create a label for the panel.
     * @param label	Label string text.
     */
    protected void displayLabel(String label) {

	Border etched = new EtchedBorder();

	if(label == null || label.equals("")) {
	    this.setBorder(etched);
	}
	else {        
	    titled_ = new TitledBorder(etched, 
				       label,
				       TitledBorder.LEFT,
				       TitledBorder.TOP);
            //addCompForBorder(titled_,label);
 //
            // LABEL2 - jump through hoops to prevent title cropping
            //
           
	    this.setBorder(titled_);

            
	}
    }

    public void addCompForBorder(Border border,
                          String description
                          ) {
        JPanel comp = new JPanel(false);
        JLabel label = new JLabel(description, JLabel.CENTER);
        comp.setLayout(new GridLayout(1, 1));
        comp.add(label);
        comp.setBorder(border);
        this.add(Box.createRigidArea(new Dimension(0, 10)));
        this.add(comp);
    }


    //----------------------------------------
    // Public methods
    //----------------------------------------

    /*
    public void addWidget() {
	// Add label
	gbc_.fill = GridBagConstraints.HORIZONTAL;
	// gbc_.weightx = 0;
	this.add(new JLabel(""), gbc_);  

	// Add the widget
	gbc_.gridx++;
	// gbc_.weightx = 5;
	gbc_.fill = GridBagConstraints.NONE;
	this.add(new JLabel(""), gbc_);  

	// Increment row and column counters
	// colNumber_ is the number of columns of widgets
	if(gbc_.gridx < colNumber_ * 2)
	    gbc_.gridx++;
	else {
	    gbc_.gridx = 0;
	    gbc_.gridy++;
	}
    }
    */

    /** 
     * Add a DbbWidget to the column selection panel.  Widgets are
     * added left-to-right; once the last column of the panel is
     * filled, a new row of widgets is created.
     *
     * @param w DbbWidget object to be added.  
     */
    public void addWidget(DbbWidget w) {

	addWidget( w, currentRow_, currentCol_ );

	currentCol_++;		// we added one column

	// If necessary, start a new row
	//----------------------------------------
	if( currentCol_ >= colNumber_ ) {
	    currentCol_ = 0;
	    currentRow_++;
	}
    }
    /** 
     * Add a DbbWidget to the column selection panel.  Widgets are
     * added left-to-right; once the last column of the panel is
     * filled, a new row of widgets is created.
     *
     * @param w DbbWidget object to be added.  
     * @param labelWeight
     * @param widgetWeight 
     */
    public void addWidgetWeight(DbbWidget w, int labelWeight,int widgetWeight) {

	addWidget( w, currentRow_, currentCol_ ,labelWeight,widgetWeight);

	currentCol_++;		// we added one column

	// If necessary, start a new row
	//----------------------------------------
	if( currentCol_ >= colNumber_ ) {
	    currentCol_ = 0;
	    currentRow_++;
	}
    }

	/** 
	 * Add a DbbWidget to the  panel.  The widget doesn't 
	 * need to be added in the panel, (e.g. is already in another panel)
	 *
	 * @param w   DbbWidget object to be added.  
	 * @param row Row number
	 * @param col Column number
	 *
	 */
	public void addWidgetNoGui( DbbWidget w)  {
	// Add to the internal DbbWidget list
	widgetList_.addElement(w);
	w.addListener(this);
	}
    

	/** 
     * Add a DbbWidget to the column selection panel.  Widget is
     * placed on the grid in position (row,col).
     *
     * @param w   DbbWidget object to be added.  
     * @param row Row number
     * @param col Column number
     *
     */
    public void addWidget( DbbWidget w, int row, int col )  {
    	addWidget(w,row,col,1,1);
    }

	/** 
     * Add a DbbWidget to the column selection panel.  Widget is
     * placed on the grid in position (row,col).
     *
     * @param w   DbbWidget object to be added.  
     * @param row Row number
     * @param col Column number
     * @param labelWeight 
     * @param widgetWeight
     */
    public void addWidget( DbbWidget w, int row, int col ,int labelWeight, int widgetWeight)  {

	// Add to the internal DbbWidget list
	widgetList_.addElement(w);
    stdlog_.debug("Panel : "+titled_.getTitle()+" the widget: "+w.label_.getText());
	// Register the current object (this) as listener
	w.addListener(this);

	// Add the label
	gbc_.gridx  = col*2;
	gbc_.gridy  = row;
	gbc_.weightx = labelWeight;
	gbc_.fill   = GridBagConstraints.RELATIVE;
	gbc_.anchor = GridBagConstraints.EAST;  // anchor to the right
	gbc_.insets = new Insets( 2, 0, 2, 2 ); // top,bot,right padding 
	this.add( w.getLabel(), gbc_ );


        int width = w.getLabel().getPreferredSize().width + 
            w.getWidget().getPreferredSize().width;
        int minSize = titled_.getMinimumSize(new JLabel(titled_.getTitle())).width;
        
        // What this?? the title is not the length of the panel
        if(minSize > width)
            w.getLabel().setPreferredSize(new Dimension(
            (minSize-w.getWidget().getPreferredSize().width)- getAdjLabelLength(),
            w.getLabel().getPreferredSize().height));

	// Add the widget
    gbc_.weightx = widgetWeight;
	gbc_.gridx  = GridBagConstraints.RELATIVE;
	gbc_.fill   = GridBagConstraints.HORIZONTAL;
	gbc_.anchor = GridBagConstraints.WEST;  // anchor to the left
	gbc_.insets = new Insets( 2, 2, 2, 0 ); // top,bot,left padding 
	this.add( w.getWidget(), gbc_ );
        

    }


    /** 
     *
     * Add a DbbWidget to the column selection panel.
     *
     * @param w       DbbWidget object to be added.  
     * @param new_row If true, the DbbWidget is placed on the first
     *                column of a ne row; otherwise on the next column 
     *                of the current row.
     */
    public void addWidget(DbbWidget w, boolean new_row) {

	if(new_row) {
	    currentCol_ = 0;
	    currentRow_++;
	}

	addWidget(w);
    }

    /**
     * For each DbbWidget we manage, reset that widget to its
     * default state; also, clear the list of active widgets.
     */
    public void reset() {
	activeWidgetList_.removeAllElements();
	int len = widgetList_.size();
	for(int i=0; i<len; i++) { 
	    DbbWidget w = (widgetList_.elementAt(i));
	    w.reset();
	}

    
    }

    /** 
     * @return Array of DbbWidget objects, corresponding to the all the
     *         widgets.  
     */
    public DbbWidget[] getAllWidgets() {

	int len = widgetList_.size();
	DbbWidget[] return_array = new DbbWidget[len];
	widgetList_.copyInto(return_array);

	return return_array;
    }




  
    /**
     * Sets as visible only the widgets found in the list.
     * The value foreach widget is set to the corrispondent value in the
     * newDefValue list.<br> Should a newVisibleWidget[i] value not to be changed,
     * the newDefValue[i] must point to null; if the newDefValue points to null
     * all the values won't be modified.<br>
     * A widget is identified by a label widget.getLabel().getText()
     * The value is accessible from the widget.[get|set]().<br>
     * Remmember to use   view.repaint(); view.validate(); after the calling of the methods
     * @param newVisibleWidget widget list to be shown
     * @param  newDefValue widget list values to be set
     */
    public void setVisibleWidgets(String[] newVisibleWidget,
				  String[] newDefValue) {
	
	DbbWidget[] panelWidget = null;
	panelWidget = getAllWidgets();
	int numVisible = 0;
	if (newVisibleWidget != null) {
	    numVisible = newVisibleWidget.length;
	}
	
 	// search the new widget and save to a new panel 
	for ( int i = 0; i < panelWidget.length; i++) {
		
	    panelWidget[i].setVisible(false);
	    panelWidget[i].reset();
		
		
	    // search the widget and try to set the value
	    for ( int j = 0; j< numVisible; j++) {
		if (panelWidget[i].getLabel().getText().equals( newVisibleWidget[j])) {
		    if (newDefValue != null && 
			newDefValue.length > j && 
			newDefValue[j] != null) {
			panelWidget[i].set(newDefValue[j]);
		   }
		    panelWidget[i].setVisible(true);  
		}
	    }
	panelWidget[i].fireDbbWidgetAction();
		 
	}
	repaint();
	validate();
	
    }


  
    /**
     * Sets values for only the widgets found in the list.
     * No validity check is performed to the proper value associated with a widget 
     * (e.g. a CheckBox should have as value "y" or "n").
     * The GTreeNode structure must have as children the widgets to be modified.
     * The name of the children is not important. For each child must be two leaves,
     * ("name","value") having as values the widget name and the new value to be assigned.
     * The leaf "order" is optional and assinges an order to the selected widgets.
     * This places the columns in the proper order automatically.
     *<p>
     * E.g. cols.w1.name "ID"
     *      cols.w1.value "y"
     *      cols.w1.order "1"
     *  "cols" node is passed to the method, the widget named "ID" will be set to "y"
     * 
     *<p>
     * If the name or the value node is not present, the widget won't be selected.<br>
     *
     * A widget is identified by a label widget.getLabel().getText()
     * The value is  set by DbbWidget.set(String) method.
     * @param  listWidgetNode starting node
     */
    @SuppressWarnings("unchecked")
	public void setValueWidgets( GTreeNode listWidgetNode) {
	
		 java.util.List<WidgetConfig> listWidget = new LinkedList<WidgetConfig>();
		 
		 if (listWidgetNode!= null) {
		 
			 for (Enumeration<GTreeNode> e = listWidgetNode.getChildren() ; e.hasMoreElements() ;){
				 GTreeNode nextWidget = e.nextElement();
				 GTreeNode nameNode = nextWidget.getChild(NAME_KEY);
				 GTreeNode valueNode = nextWidget.getChild(VALUE_KEY);
				 GTreeNode orderNode = nextWidget.getChild(ORDER_KEY);
				 GTreeNode positionNode = nextWidget.getChild(POSITION_KEY);
				 if (nameNode != null && valueNode  != null ){
					 int orderIndex =  0;
					 int positionIndex = 0;
					 
					 if (orderNode != null){
						 try {
							 orderIndex = Integer.decode(orderNode.getValue()).intValue();
						 } catch (Exception en){
							 stdlog_.warn ("error order value "+orderNode.getValue()+"assign default:", en);
						 }
					 }
					 
					 if (positionNode != null){
						 try {
							 positionIndex = Integer.decode(positionNode.getValue()).intValue();
						 } catch (Exception en){
							 stdlog_.warn ("error order value "+orderNode.getValue()+"assign default:", en);
						 }
					 }
					 
					 WidgetConfig widgetConf = new WidgetConfig(nameNode.getValue(),valueNode.getValue(), orderIndex, positionIndex);
					 listWidget.add(widgetConf);		     
				 }
		     }
		 }
		 
		 Collections.sort(listWidget);
		 String [] listWidgetNameArray = new String[listWidget.size()];
		 String [] listWidgetValueArray = new String[listWidget.size()];
		 int [] listWidgetOrderArray = new int[listWidget.size()];
		 int [] listWidgetPositionArray = new int[listWidget.size()];
		 for (int i = 0; i < listWidget.size(); i++) {
		     listWidgetNameArray[i] = (listWidget.get(i)).getName();
		     listWidgetValueArray[i] = (listWidget.get(i)).getValue();
		     listWidgetOrderArray[i] = (listWidget.get(i)).getOrder();
		     listWidgetPositionArray[i] = (listWidget.get(i)).getPosition();
		     stdlog_.debug("i) " + listWidgetNameArray[i] + " = " + listWidgetValueArray[i] + " (Order " + listWidgetOrderArray[i] + ")(Position " + listWidgetPositionArray[i] + ")");
		 }
		
		 
		 
		 setValueWidgets(listWidgetNameArray,listWidgetValueArray);
    }

    /**
     * Sets values for only the widgets found in the list.
     * A widget is identified by a label widget.getLabel().getText()
     * The value is accessible from the widget.[get|set]().
     * @param newWidget widget list to be set
     * @param  newDefValue widget list values to be set
     */
    public void setValueWidgets(String[] newWidgetNameArray,
				  String[] newWidgetValueArray) {
	
	DbbWidget[] panelWidget = null;
	panelWidget = getAllWidgets();
	int numW= 0;
	if (newWidgetNameArray != null) {
	    numW  = newWidgetNameArray.length;
	}
	
	// search the widget and try to set the value
	for ( int j = 0; j< numW; j++) {
	    // search the new widget and save to a new panel 
	    for ( int i = 0; i < panelWidget.length; i++) {	

		if (panelWidget[i].getLabel().getText().equals( newWidgetNameArray[j])) {
		    if (newWidgetValueArray != null && 
			newWidgetValueArray.length > j && 
			newWidgetValueArray[j] != null) {

			// RESET and then set for assigning the order in proper way 
			//  if was set the order didn't change
			panelWidget[i].reset();
			panelWidget[i].fireDbbWidgetAction();
			panelWidget[i].set(newWidgetValueArray[j]);
			panelWidget[i].fireDbbWidgetAction();
			
		    }
		
		}
	    }		 
	}	
    }



 
    /**
     * Sets to default all the widgets added to the panel.
     * The widgets listed won't be modified.
     * @param newWidget list of widgets to leave as they are
     */
    public void unsetValueWidgets(String[] newWidget) {
	DbbWidget[] panelWidget = null;
	panelWidget = getAllWidgets();
	int numKeepUnchanged = 0;
	boolean resetWidget = true;
	if (newWidget != null) {
	    numKeepUnchanged = newWidget.length;
	}
	for ( int i = 0; i < panelWidget.length; i++) {
	    resetWidget = true;
	    for ( int j = 0; j< numKeepUnchanged; j++) {
		if (panelWidget[i].getLabel().getText().equals( newWidget[j])) {
		    resetWidget = false;
		    break;
		}
	    }
	    if (resetWidget) {
		panelWidget[i].reset();
		panelWidget[i].fireDbbWidgetAction();
	    } 
	}
	
    }




 
    /**
     * Sets visible all the widgets added to the panel.
     * The value foreach widget is set to the corrispondent value in the
     * newDefValue list.  if the newDefValue or newVisibleWidget points to null
     * all the values won't be modified. 
     *<p>
     * A widget is identified by a label widget.getLabel().getText()
     * The value is accessible from the widget.[get|set]()<br>
     * Remmember to use   view.repaint(); view.validate(); after the calling of the methods
     * @param newVisibleWidget widget list which values we want to set
     * @param  newDefValue widget list values to be set

     */
    public void setAllVisibleWidgets(String[] newVisibleWidget,
				     String[] newDefValue) {
	DbbWidget[] panelWidget = null;
	panelWidget = getAllWidgets();
	
	for ( int i = 0; i < panelWidget.length; i++) {

	    panelWidget[i].setVisible(true);	    
	    panelWidget[i].reset();

	    // search the widget and try to set the value
	    if (newVisibleWidget != null) {
		for ( int j = 0; j< newVisibleWidget.length; j++) {
		    if (panelWidget[i].getLabel().getText().equals( newVisibleWidget[j])) {
			if (newDefValue != null && 
			    newDefValue.length > j && 
			    newDefValue[j] != null) {
			    panelWidget[i].set(newDefValue[j]); 
			}
		    }
		}
	    }

	    panelWidget[i].fireDbbWidgetAction(); 
	}
	repaint();
	validate();
	
    }

    /** 
     * @return Array of DbbWidget objects, corresponding to the active
     *         widgets.  
     */
    public DbbWidget[] getActiveWidgets() {

	int len = activeWidgetList_.size();
	DbbWidget[] return_array = new DbbWidget[len];
	activeWidgetList_.copyInto(return_array);

	return return_array;
    }

    /** 
     * @return Array of SQL chunks, corresponding to the active widgets.
     * @see DbbSqlChunk 
     */
    public DbbSqlChunk[] getActiveChunks() {

	DbbWidget widget;
	int len = activeWidgetList_.size();
	DbbSqlChunk[] ret = new DbbSqlChunk[len];

	for(int i=0; i<len; i++) {

	    try {
		widget = activeWidgetList_.elementAt(i);
		ret[i] = widget.getSqlChunk();
	    }
	    catch(DbbInvalidSqlOperatorType e) { 
		stdlog_.warn("",e);
	    }
	}

	return ret;
    }


    /** 
     * Get an array of additional chunks (if any) defined for the
     * active widgets.
     *
     * @return an array of DbbSqlChunk  
     */
    public DbbSqlChunk[] getAdditionalChunks() {

	DbbWidget widget;
	Vector<DbbSqlChunk> chunk_list = new Vector<DbbSqlChunk>();

	for(int i=0; i<activeWidgetList_.size(); i++) {

	    try {
		widget = activeWidgetList_.elementAt(i);

		/* Retrieve the widget additional chunks (may be more than
		   one) */
		DbbSqlChunk[] widget_chunks = widget.getAdditionalSqlChunks();

		for(int k=0; k<widget_chunks.length; k++)
		    chunk_list.addElement(widget_chunks[k]);
	    }
	    catch(DbbInvalidSqlOperatorType e) { 
		stdlog_.warn("",e);
	    }
	}

	/* Copy into the return chunk array */
	int len = chunk_list.size();
	DbbSqlChunk[] ret = new DbbSqlChunk[len];
	chunk_list.copyInto(ret);

	return ret;
    }


    //----------------------------------------
    // Event methods
    //----------------------------------------

    /**
     * Implement the interface method.
     *
     * When a DbbWidgetChangedEvent has been delivered, the DbbWidget
     * object generating the event has to be checked. If it has been
     * just activated, add it to activeWidgetList_; if it has been just
     * deactivated, remove it from activeWidgetList_.
     *
     * @param ev Sql widget changed event.  
     */
    public void dbbWidgetAction(DbbWidgetEvent ev) {

	DbbWidget source = (DbbWidget)ev.getSource();

	if( source.isEmpty() ) {
      
	    // Inactive: remove from the list
	    if(activeWidgetList_.contains(source))
			activeWidgetList_.removeElement(source);
      
	} else {
      
	    // Active: add to the list
	    if(!activeWidgetList_.contains(source))
		activeWidgetList_.addElement(source);
	}
    }

	/**
	 * Method setAdjLabelLength. the length to which adjust the label widget
	 * length. I don't think is useful to have it, but kept for backward
	 * compatibility. By default is 8, I would suggest to set to 0
	 * @param adjLabelLength_
	 */
	public void setAdjLabelLength(int adjLabelLength) {
		this.adjLabelLength_ = adjLabelLength;
	}

	public int getAdjLabelLength() {
		return adjLabelLength_;
	}

	/**
	 * It returns the vector containing the list of widget.
	 * It's similar to getAllWidgets();
	 * @return the list of widget.
	 */
	public Vector<DbbWidget> getWidgetList() {
		return widgetList_;
	}
	
	/**
	 * Set the widget List.
	 * Added to manage user widgetList .
	 * @param widgetList
	 */
	public void setWidgetList(Vector<DbbWidget> widgetList) {
		this.widgetList_ = widgetList;
	}

	public TitledBorder getTitled() {
		return titled_;
	}

	public int getColNumber() {
		return colNumber_;
	}

	public Vector<DbbWidget> getActiveWidgetList() {
		return activeWidgetList_;
	}

	public void setActiveWidgetList(Vector<DbbWidget> activeWidgetList) {
		this.activeWidgetList_ = activeWidgetList;
	}

	/**
	 * Set the DbbPanel Type.
	 * The allowed type are: ColumnsPanel, RowsPanel, SortingPanel.
	 * The type is taken by the XML element name and the XSD has to check for it
	 * @param type
	 */
	public void setType(String type) {
		dbbPanelType_ = type;
		
	}
	public String getType() {
		return dbbPanelType_;
		
	}
	
	
	
}
