package org.eso.ohs.core.dbb.client;

// $Id: DbbTableModel.java,v 1.11 2006/04/27 13:24:18 tcanavan Exp $

import java.util.Vector;

import javax.swing.event.TableModelEvent;

import org.apache.log4j.Logger;
import org.eso.ohs.core.dbb.sql.DbbSqlQueryResult;
/** 
 * DbbTableModel implements the table model storing the data that
 * shall be displayed in the DbbTable.
 *
 * @author Carlo Boarotto, ESO 24-Sep-1999
 * 
 **/

public class DbbTableModel extends AbstractDbbTableModel {

    /** Debug boolean flag */
    public boolean debug = false;

    private static Logger stdlog_ = Logger.getLogger(DbbTableModel.class);

    /** Initial number of rows. */
    protected static int START_NUM_ROWS = 0;

    /** Current number of columns. */
    protected int numColumns_ = 0;

    /** Current number of rows. */
    protected int numRows_ = 0;

    /** First empty row, used when adding a new row. */
    protected int nextEmptyRow_ = 0;

    /**
     * The Vector of Array of Object values. The column number is
     * almost fixed for each query, since the user cannot add columns
     * but running a new query.  
     */
    protected Vector<Object> dataVector = null;

    /**
     * The array of column identifiers. The column number is almost
     * fixed for each query, since the user cannot add columns but
     * running a new query.  
     */
    protected String[] columnIdentifiers = null;

    /** Array containing the classes associated to the columns */
    protected Class[] columnClasses = null;

    /** Store the column max sizes */
    protected int[] columnSizes = null;

    /** Flag to be set to true when the table has to be redrawn. */
    protected boolean tableCleared_ = true;

    //----------------------------------------
    // Constructors
    //----------------------------------------

    /** 
     * Create a DbbTableModel setting its internal variables to their
     * initial values.
     */
    public DbbTableModel() {
        numRows_ = 0;
        numColumns_ = 0;
        dataVector = new Vector<Object>();
    }

    public Vector<Object> getDataVector() {
        return dataVector;
    }

    public void setDataVector(Vector<Object> dataVec) {
        dataVector = dataVec;
        fireTableDataChanged();
    }

    //----------------------------------------
    // Public methods
    //----------------------------------------

    /**
     * Get the column name for a given column index.
     * @param column Column index.
     * @return Column name.
     */
    public String getColumnName(int column) {
    	String result = "";
    	
        if (column < numColumns_ && column >= 0)
            result = columnIdentifiers[column];
        
        return result;
    }

    /**
     * Get the column width for a given column index.
     * @param column Column index.
     * @return Column width.
     */
    public int getColumnSize(int column) {
    	int result = 0;
    	
		if (column < numColumns_)
            result = columnSizes[column];
        
		return result;
    }

    /** @return Model's number of columns. */
    public synchronized int getColumnCount() {
        return numColumns_;
    }

    /** @return Model's number of rows. */
    public synchronized int getRowCount() {
    	int result = numRows_;
    	
        if (numRows_ < START_NUM_ROWS)
            result = START_NUM_ROWS;
        
        return result;
    }

    /**
     * It defines the way to access the table model data and it's used
     * by JTable to get the data it's going to display.
     *
     * @param row		Table row.
     * @param column	Table column.
     *
     * @return An Object object.
     */
    public synchronized Object getValueAt(int row, int column) {
    	Object result = "";
    	
        try {
            if (row < numRows_ && column < numColumns_) {
                Object[] row_array = (Object[]) dataVector.elementAt(row);
                result = row_array[column];
            }
        } catch (Exception e) {

            if (debug)
                stdlog_.debug(
                    "ERROR when getting element (" + row + "," + column + ")");
        }

        return result;
    }
	/**
	 * Reorder the vector containing the data according to the new indexes
	 * @param newIndexes
	 */
	@SuppressWarnings("unchecked")
	public void reorderContents(int[] newIndexes){
		Vector<Object> oldDataVector = (Vector<Object>) dataVector.clone();
		
		for (int i = 0; i < newIndexes.length; i++)	
			dataVector.set(i,  oldDataVector.get(newIndexes[i]));	
	}

    public void setValueAt(Object obj, int row, int column) {

        stdlog_.debug(
            "DbbTableModel: Set element (" + row + "," + column + ")");
   //     boolean datachanged = false;

        synchronized (this) {
            try {
                if (row < numRows_ && column < numColumns_) {
                    Object[] row_array = (Object[]) dataVector.elementAt(row);
                    row_array[column] = obj;
                    dataVector.setElementAt(row_array, row);
         //           datachanged = true;
                }

                stdlog_.debug(
                    "CANNOT set element (" + row + "," + column + ")");

            } catch (Exception e) {
                stdlog_.warn(
                    "ERROR when setting element (" + row + "," + column + ")",
                    e);
            }
        }
        fireTableChanged(new TableModelEvent(this, row, row, column));
        //fireTableDataChanged();	
        //fireTableStructureChanged();	
    }

    /**
     * Get the java class associated with a specific column.
     * @param columnIndex Column index
     * @return A java class associated with the column.
     */
    @SuppressWarnings("unchecked")
	public Class getColumnClass(int columnIndex) {

        if (columnIndex < numColumns_) {
            //stdlog_.debug("DbbTableModel::getColumnClass " +
            //			 columnIdentifiers[columnIndex] +
            //			 " " + columnClasses[columnIndex]);
            return columnClasses[columnIndex];
        }
        return Object.class;
    }

    /**
     * Method called by DbbTable to force the model's update.
     *
     * @param result DbbSqlQueryResult object containing the query
     *               result.
     */
    public void updateModel(DbbSqlQueryResult result) {
        synchronized (this) {
            numRows_ = result.getRows();
            numColumns_ = result.getColumns();

            if (debug)
                stdlog_.debug("Rows: " + numRows_ + " Cols: " + numColumns_);

            dataVector = result.getDataVector();
            columnIdentifiers = result.getColumnNames();
            columnSizes = result.getColumnSizes();
            columnClasses = result.getColumnTypes();

            nextEmptyRow_ = numRows_;
        }
        fireTableRowsInserted(0, numRows_ - 1);
    }

    // It works !!!
    public void addColumn(String col_name, Object[] items, Class col_class) {

        synchronized (this) {
            int col_width = 0;

            for (int i = 0; i < numRows_; i++) {

                Object[] new_row = new Object[numColumns_ + 1];
                Object[] c_row = (Object[]) dataVector.elementAt(i);

                for (int k = 0; k < numColumns_; k++)
                    new_row[k] = c_row[k];

                if (i < items.length && items[i] != null) {
                    new_row[numColumns_] = items[i];
                    int w = items[i].toString().length();
                    if (w > col_width)
                        col_width = w;
                } else
                    new_row[numColumns_] = "";

                dataVector.setElementAt(new_row, i);
            }

            Class[] new_classes = new Class[numColumns_ + 1];
            int[] new_sizes = new int[numColumns_ + 1];
            String[] new_names = new String[numColumns_ + 1];

            for (int k = 0; k < numColumns_; k++) {

                new_classes[k] = columnClasses[k];
                new_sizes[k] = columnSizes[k];
                new_names[k] = columnIdentifiers[k];

                // 	    stdlog_.debug("new_classes[k]: "+
                // 			       new_classes[k] +
                // 			       " columnClasses[k]: "+
                // 			       columnClasses[k]);
            }

            new_classes[numColumns_] = col_class;
            new_sizes[numColumns_] = col_width;
            new_names[numColumns_] = col_name;

            columnClasses = new_classes;
            columnSizes = new_sizes;
            columnIdentifiers = new_names;

            numColumns_++;
        }
        fireTableDataChanged();
        fireTableStructureChanged();
    }

    protected synchronized Object[] insertElement(
        Object[] in_array,
        Object new_item,
        int pos) {
        int len = in_array.length;
        Object[] out_array = new Object[len + 1];

        for (int i = 0; i < pos; i++) {
            out_array[i] = in_array[i];
        }

        for (int i = len - 1; i >= pos; i--) {
            out_array[i + 1] = in_array[i];
        }

        out_array[pos] = new_item;

        if (debug) {
            stdlog_.debug("Inserting element in the array " + in_array);
            for (int k = 0; k < len + 1; k++)
                stdlog_.debug(k + ": " + out_array[k]);
        }

        return out_array;
    }

    protected synchronized int[] insertIntElement(
        int[] in_array,
        int new_item,
        int pos) {
        int len = in_array.length;
        int[] out_array = new int[len + 1];

        for (int i = 0; i < pos; i++) {
            out_array[i] = in_array[i];
        }

        for (int i = len - 1; i >= pos; i--) {
            out_array[i + 1] = in_array[i];
        }

        out_array[pos] = new_item;

        if (debug) {
            stdlog_.debug("Inserting element in the INT array " + in_array);
            for (int k = 0; k < len + 1; k++)
                stdlog_.debug(k + ": " + out_array[k]);
        }

        return out_array;
    }

    // It works now
    public void addColumn(
        String col_name,
        Object[] items,
        Class col_class,
        int pos) {

        synchronized (this) {
            int col_width = 0;
            numColumns_++;
            Vector<Object> new_data = new Vector<Object>();

            if (debug)
                stdlog_.debug("Inserting element in the DATA array");

            for (int i = 0; i < numRows_; i++) {

                Object[] c_row = (Object[]) dataVector.elementAt(i);
                Object new_item;

                if (i < items.length && items[i] != null) {
                    new_item = items[i];
                    int w = items[i].toString().length();
                    if (w > col_width)
                        col_width = w;
                } else {
                    new_item = new String("");
                }

                Object[] new_row = insertElement(c_row, new_item, pos);

                new_data.addElement(new_row);
            }

            // Update column names
            if (debug)
                stdlog_.debug(
                    "Inserting element in the columnIdentifiers array");
            Object[] new_names =
                insertElement(columnIdentifiers, col_name, pos);

            columnIdentifiers = new String[numColumns_];
            for (int i = 0; i < numColumns_; i++)
                columnIdentifiers[i] = (String) new_names[i];

            // Update column sizes
            if (debug)
                stdlog_.debug("Inserting element in the columnSizes array");
            int[] new_sizes = insertIntElement(columnSizes, col_width, pos);

            columnSizes = new int[numColumns_];

            for (int i = 0; i < numColumns_; i++)
                columnSizes[i] = new_sizes[i];

            // Update column classes 
            if (debug)
                stdlog_.debug("Inserting element in the columnClasses array");
            Object[] new_classes = insertElement(columnClasses, col_class, pos);

            columnClasses = new Class[numColumns_];

            for (int i = 0; i < numColumns_; i++)
                columnClasses[i] = (Class) new_classes[i];

            dataVector = new_data;
        }
        fireTableDataChanged();
        fireTableStructureChanged();
    }

    public void setColumn(
        String col_name,
        Object[] items,
        Class col_class,
        int pos) {
        synchronized (this) {
            int col_width = 0;
            Vector<Object> new_data = new Vector<Object>();

            if (debug)
                stdlog_.debug("Setting element in the DATA array");

            for (int i = 0; i < numRows_; i++) {

                Object[] c_row = (Object[]) dataVector.elementAt(i);
                Object new_item;

                if (i < items.length && items[i] != null) {
                    new_item = items[i];
                    int w = items[i].toString().length();
                    if (w > col_width)
                        col_width = w;
                } else {
                    new_item = new String("");
                }

                c_row[pos] = new_item;
                new_data.addElement(c_row);
            }

            // Update column names
            if (debug)
                stdlog_.debug("Setting element in the columnIdentifiers array");
            columnIdentifiers[pos] = col_name;

            // Update column sizes
            if (debug)
                stdlog_.debug("Setting element in the columnSizes array");
            columnSizes[pos] = col_width;

            // Update column classes 
            if (debug)
                stdlog_.debug("Setting element in the columnClasses array");
            columnClasses[pos] = col_class;

            dataVector = new_data;
        }
        fireTableDataChanged();
        fireTableStructureChanged();
    }

    /**
     * Add rows from a given sql result to the table model and fire
     * the proper events.
     *
     * @param result Sql result.
     * @see org.eso.ohs.core.dbb.sql.DbbSqlQueryResult 
     */
    public void addRows(DbbSqlQueryResult result) {
        int old_num = numRows_;
        synchronized (this) {
            if (tableCleared_) {
                numColumns_ = result.getColumns();
                columnIdentifiers = result.getColumnNames();
                columnSizes = result.getColumnSizes();
                columnClasses = result.getColumnTypes();

                tableCleared_ = false;
            }

            Vector<Object> new_rows = result.getDataVector();
            int len = new_rows.size();

            for (int i = 0; i < len; i++)
                dataVector.addElement(new_rows.elementAt(i));

            numRows_ = dataVector.size();
            nextEmptyRow_ = numRows_;
        }
        fireTableRowsInserted(old_num, numRows_ - 1);
    }

    public void insertRow(Object[] new_row, int pos) {
        if (new_row == null
            || new_row.length != numColumns_
            || pos >= numRows_) {
            stdlog_.debug("DbbTableModel insertRow failed");
            return;
        }

        dataVector.insertElementAt(new_row, pos);
        numRows_++;
        fireTableRowsInserted(pos, pos - 1);
    }

    /**
     * Removes a set of rows from the model
     * @param rows specifies a set of row's indeces to be removed
     */
    public void removeRows(int[] rows) {
        synchronized (this) {
            for (int i = rows.length - 1; i >= 0; i--) {
                dataVector.removeElementAt(rows[i]);
                numRows_ = numRows_ - 1;
                // Generate notification   
            }
        }
        for (int i = rows.length - 1; i >= 0; i--) {
            fireTableRowsDeleted(rows[i], rows[i]);
        }
    }

    /**
      * Move up a set of rows
      * @param rows specifies a set of row's indexes to be moved up
      */
    
    public void moveUpRows(int[] rows) {
        int insertPosition = rows[0] - 1;
        moveUpRows(rows, insertPosition);
    }

    public void moveUpRows(int[] rows, int insertPosition) {
        if (insertPosition < 0)
            return; // Nothing to move

        // Do the move by first removing the row, then reinserting it
        Object[] removed = new Object[rows.length];

        for (int i = rows.length - 1; i >= 0; i--) {
            removed[i] = dataVector.elementAt(rows[i]);
            dataVector.removeElementAt(rows[i]);
        }

        for (int i = rows.length - 1; i >= 0; i--) {
            dataVector.insertElementAt(removed[i], insertPosition);
        }

        // Generate notification
        fireTableDataChanged();

    }

    public void moveRows(int[] rows, int insertPosition) {

        // org.eso.ohs.utilities.DebugUtils.stackTrace();
        if (insertPosition < 0)
            return; // Nothing to move

        // Do the move by first removing the row, then reinserting it
        Object[] removed = new Object[rows.length];

        for (int i = rows.length - 1; i >= 0; i--) {
            removed[i] = dataVector.elementAt(rows[i]);
            dataVector.removeElementAt(rows[i]);
        }

        for (int i = rows.length - 1; i >= 0; i--) {
            dataVector.insertElementAt(removed[i], insertPosition);
        }

        // Generate notification
        fireTableDataChanged();

    }

    /**
       * Move up a set of rows
       * @param rows specifies a set of row's indeces to be moved up
       */
    public void moveToTop(int[] rows) {
        moveUpRows(rows, 0);
    }

    public void moveDownRows(int[] rows) {
        int insertPosition = rows[rows.length - 1] + 1;
        moveDownRows(rows, insertPosition);
    }

    /**
     * Moves downs a set of rows
     * @param rows specifies a set of row's indeces to be moved up
     */
    public void moveDownRows(int rows[], int insertPosition) {
        //int insertPosition = rows[rows.length-1]+1;

        if (insertPosition >= getRowCount())
            return; //nothing to move

        Object[] removed = new Object[rows.length];

        for (int i = 0; i < rows.length; i++) {
            removed[i] = dataVector.elementAt(rows[i] - i);
            dataVector.removeElementAt(rows[i] - i);
        }

        for (int i = 0; i < rows.length; i++) {
            dataVector.insertElementAt(
                removed[i],
                insertPosition - removed.length + i + 1);
        }

        fireTableDataChanged();
    }

    /**
     * Remove all the data from the model and fire the proper
     * events. 
     */
    public void clear() {
        int nr = numRows_;
        synchronized (this) {
            nr = numRows_;
            dataVector.removeAllElements();
            numRows_ = 0;
            nextEmptyRow_ = numRows_;

            tableCleared_ = true;
        }
        fireTableRowsDeleted(0, numRows_);
    }

    /** Allow the user to edit the cells. */
    public boolean isCellEditable(int row, int col) {
        // In principal no cell editing should be allowed (at least
        // for the early versions)
        //return true;
        return false;
    }
}
