package org.eso.ohs.core.dbb.server;

import java.util.ArrayList;
import java.util.Iterator;

import org.eso.ohs.core.utilities.MsgManager;

public class DbbOrderingClauseInfo {

	private ArrayList<String> columns = null;
	private ArrayList<String> directions = null;
	
	public DbbOrderingClauseInfo(){
		columns = new ArrayList<String>();
		directions = new ArrayList<String>();
	}
	/**
	 * add a couple (columnName, direction)
	 * @param column a db column name in the dbname.tablename.columnname form
	 * @param direction ASC or DESC
	 */
	public void add(String column, String direction){
		if (column == null) throw new IllegalArgumentException("column name is null");
		if (column.length() <= 0) throw new IllegalArgumentException("column name is empty");
		if (direction == null) throw new IllegalArgumentException("direction name is null");
		if (!checkDirection(direction)) throw new IllegalArgumentException("direction value must be ASC or DESC");
		
		columns.add(column);
		directions.add(direction.trim().toUpperCase());
	}
	
	/**
	 * 
	 * @return an columns iterator
	 */
	public Iterator<String> getColumns(){
		return columns.iterator();
	}
	
	/**
	 * 
	 * @return directions iterator
	 */
	public Iterator<String> getDirections(){
		assert (checkDataStructure()): MsgManager.errInvariant("Wrong size! 'columns' has " + columns.size() + " elements while 'directions' has " + columns.size() + " elements");
		return directions.iterator();
	}
	
	/**
	 * 
	 * @param index
	 * @return the column name at the given index
	 */
	public String getColumnAt(int index){
		if (( index < 0) || (index > columns.size())) throw new IllegalArgumentException ("Index out of bound. Index requested: " + index + " columns size: " + columns.size());
		assert (checkDataStructure()): MsgManager.errInvariant("Wrong size! 'columns' has " + columns.size() + " elements while 'directions' has " + columns.size() + " elements");
		String result = columns.get(index);
		return result;
	}
	
	/**
	 * 
	 * @param index
	 * @return the direction at the given index
	 */
	public String getDirectionAt(int index){
		if (( index < 0) || (index > directions.size())) throw new IllegalArgumentException ("Index out of bound. Index requested: " + index + " directions size: " + directions.size());
		assert (checkDataStructure()): MsgManager.errInvariant("Wrong size! 'columns' has " + columns.size() + " elements while 'directions' has " + columns.size() + " elements");
		String result = directions.get(index);
		return result;
	}
	
	/**
	 * 
	 * @param column
	 * @return the proper ordering direction for the given column name.
	 */
	public String getDirectionFor(String columnName){
		if (columnName == null) throw new IllegalArgumentException("column name is null");
		if (columnName.length() <= 0) throw new IllegalArgumentException("column name is empty");
		assert (checkDataStructure()): MsgManager.errInvariant("Wrong size! 'columns' has " + columns.size() + " elements while 'directions' has " + columns.size() + " elements");
		String result = null;
		for (int i = 0; i < columns.size(); i++){
			String item = columns.get(i);
			
			if (item.equals(columnName)){
				result = directions.get(i);
			}
		}
		return result;
	}
	
	/**
	 * 
	 * @param column column name
	 * @return true if the column name belongs to the columns ArrayList
	 */
	public boolean containsColumn(String columnName){
		if (columnName == null) throw new IllegalArgumentException("column name is null");
		if (columnName.length() <= 0) throw new IllegalArgumentException("column name is empty");
		assert (checkDataStructure()): MsgManager.errInvariant("Wrong size! 'columns' has " + columns.size() + " elements while 'directions' has " + columns.size() + " elements");
		
		boolean result = columns.contains(columnName);
		
		return result;
	}
	
	/**
	 * Clear the data structure.
	 *
	 */
	public void clear(){
		columns.clear();
		directions.clear();
	}
	
	/**
	 * the method return the index of a given column name in the internal structure
	 * The method is used while setting the parameters in the prepared sql statement.
	 * @param columnName a String
	 * @return an index
	 */
	public int indexOf(String columnName){
		if (columnName == null) throw new IllegalArgumentException("column name is null");
		if (columnName.length() <= 0) throw new IllegalArgumentException("column name is empty");
		assert (checkDataStructure()): MsgManager.errInvariant("Wrong size! 'columns' has " + columns.size() + " elements while 'directions' has " + columns.size() + " elements");
		
		int result = columns.indexOf(columnName);
		
		return result;
	}
	
	/**
	 * The instance is "ready" when it contains at least 1 column name.
	 * This method is used to check the invariant condition while building the WHERE 
	 * clause.
	 * @return true if the instance can be used to 
	 * build both the ORDER BY and WHERE clause for pagination
	 * purpose.
	 */
	public boolean isReady(){
		boolean result = false;
		 
		if ((columns.size() > 0) && (directions.size() > 0))
			if (checkDataStructure())
				result = true;
		
		return result;
	}
	
	/*
	 * Check over data structures size.
	 */
	private boolean checkDataStructure() {
		boolean result = false;
		
		if (columns.size() == directions.size())
			result = true;
		
		return result;
	}
	
	/*
	 * The method contains the logic to validate a direction
	 * admitted values: ASC, DESC.
	 */
	private boolean checkDirection(String direction){
		assert (direction != null): MsgManager.errPrecondition("direction is null");
		boolean result = false;
		
		direction = direction.trim().toUpperCase();
		if (direction.equals("ASC") || (direction.equals("DESC")))
			result = true;
		
		return result;
	}
	
	
}
