package org.eso.ohs.core.dbb.server;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;

import org.apache.log4j.Logger;
import org.eso.ohs.core.dbb.sql.DbbSqlQueryResult;

/**
 * The DbbQueryImpl class allow to read the data coming from a result
 * set and to get the result within a DbbSqlQueryResult object
 * <p>
 * $Id: DbbQueryImpl.java,v 1.4 2006/03/20 08:34:48 gdonino Exp $
 * @version $Revision: 1.4 $
 * @author Carlo Boarotto, ESO 19-Jul-1999
 * @author A M Chavan, ESO 26-Jul-1999
 * <p>
 * @see DbbSqlQueryResult
 */
public class DbbQueryImpl implements DbbQuery {

	private static final int NO_PAGINATION = -1;
    private static Logger stdlog_ = Logger.getLogger(DbbQueryImpl.class);

    public static final String rcsid =
        "$Id: DbbQueryImpl.java,v 1.4 2006/03/20 08:34:48 gdonino Exp $";

    /** Array of column names */
    private String[] columnNames;

    /** Array of column types */
    private Class[] columnTypes;

    /** Number of columns */
    private int cols;
    
    /** first row of the resultset that will be displayed */
    private int startRow_ = NO_PAGINATION;

    /**
     * Default number of rows to display on the table per time when
     * retrieving data from the database
     * <br>
     * CAVEAT: Modifying this number can be alter the performance of
     * the database browser: a huge number speeds up the performance
     * but allow the user to break after a alrge amount of row; on the
     * other hand a small number allow the user to interrupt the query
     * as soon as he presses the break button but it slows down the
     * performances.
     */
    private int numOfRowsToGet = 500;

    /** Database result set reference */
    private ResultSet rs_;

    /** True if there are more rows to be read from the result set */
    private boolean moreToRead_ = true;

    /**
     * The DbbSession reference has to be used when closing the sql query,
     * because it has been delegated to release the database connection
     */
    private DbbSession session_;

    /**
     * DbbSqlQueryResult data structure to store temporarly the sql result
     */
    protected DbbSqlQueryResult queryResult_ = null;


    //----------------------------------------
    // Constructors
    //----------------------------------------

    /**
     * Create a DbbQuery object passing it a reference to the database
     * result set and to the database session
     *
     * @param rs        Database result set to be read
     * @param session   Database session
     * @exception SQLException if a jdbc call throws an exception
     */
    public DbbQueryImpl(ResultSet rs, DbbSession session) throws SQLException {

        rs_ = rs;
        session_ = session;
        if (rs != null) {
            ResultSetMetaData rsmd = rs.getMetaData();
            cols = rsmd.getColumnCount();

            columnNames = new String[cols];
            columnTypes = new Class[cols];

            for (int i = 1; i <= cols; i++) {
                columnNames[i - 1] = rsmd.getColumnName(i);
                columnTypes[i - 1] =
                    DbbSession.getDataTypeClass(rsmd.getColumnType(i));
            }
        }
    }
    
    /**
     * Create a DbbQuery object passing it a reference to the database
     * result set and to the database session
     *
     * @param rs        Database result set to be read
     * @param session   Database session
     * @param startRow	position of the first row of the resultSet that will be displayed
     * @exception SQLException if a jdbc call throws an exception
     */
    public DbbQueryImpl(ResultSet rs, DbbSession session, int startRow) throws SQLException {
   	 
    	 rs_ = rs;
         session_ = session;
         startRow_ = startRow;
         if (rs != null) {
             ResultSetMetaData rsmd = rs.getMetaData();
             cols = rsmd.getColumnCount();

             columnNames = new String[cols];
             columnTypes = new Class[cols];

             for (int i = 1; i <= cols; i++) {
                 columnNames[i - 1] = rsmd.getColumnName(i);
                 columnTypes[i - 1] =
                     DbbSession.getDataTypeClass(rsmd.getColumnType(i));
             }
         }
	}

    public DbbQueryImpl(ResultSet rs) throws SQLException {

        rs_ = rs;
     //   session_ = session;
        if (rs != null) {
            ResultSetMetaData rsmd = rs.getMetaData();
            cols = rsmd.getColumnCount();

            columnNames = new String[cols];
            columnTypes = new Class[cols];

            for (int i = 1; i <= cols; i++) {
                columnNames[i - 1] = rsmd.getColumnName(i);
                columnTypes[i - 1] =
                    DbbSession.getDataTypeClass(rsmd.getColumnType(i));
            }
        }
    }
    //----------------------------------------
    // Public methods
    //----------------------------------------

    


	/**
     * Set the number of rows to be displayed per time
     * <br>
     * CAVEAT: Modifying this number can be alter the performance of
     * the database browser: a huge number speeds up the performance
     * but allow the user to break after a alrge amount of row; on the
     * other hand a small number allow the user to interrupt the query
     * as soon as he presses the break button but it slows down the
     * performances.
     */
    public void setNumOfRowsToGet(int n) {
        numOfRowsToGet = n;
    }

    /**
     * @return True if there are more rows to read from the result set
     */
    public boolean hasMoreRows() {
        return moreToRead_;
    }

    /**
     * Used internally by the view in order to get the result data structure
     *
     * @return DbbSqlQueryResult data structure containing the query result
     */
    public DbbSqlQueryResult getResult() {
        return queryResult_;
    }

    /**
     * Read rows from the result set and put the data into the
     * queryResult internal variable
     *
     * @exception SQLException if a jdbc call throws an exception
     */
    public void readRows() throws SQLException {

        int counter = 0;

        queryResult_ = new DbbSqlQueryResult();

        // if (stdlog_.isDebugEnabled())
        // stdlog_.debug("Created query result set");

        queryResult_.setColumns(cols);

        // if (stdlog_.isDebugEnabled())
        // stdlog_.debug("Column s " + cols);

        queryResult_.setColumnNames(columnNames);

        // if (stdlog_.isDebugEnabled())
        // stdlog_.debug("Created Column names");

        queryResult_.setColumnTypes(columnTypes);

        if (!moreToRead_) {
            close();
            return;
        }
        if (startRow_ == NO_PAGINATION){
	        while (moreToRead_ = rs_.next()) {
	            counter++;
	            queryResult_.addRow(rs_);
	
	            if (counter >= numOfRowsToGet) {
	                break;
	            }
	        }
        }else{
        	while (moreToRead_ = rs_.next()) {
	            counter++;
	            if ( counter >= startRow_)
	            	queryResult_.addRow(rs_);
        	}
        }

        // if (stdlog_.isDebugEnabled())
        // stdlog_.debug("After a partial result set or a break");

        if (!moreToRead_)
            close();
    }

    /**
     * Read rows from the result set and append the data into the
     * queryResult internal variable
     *
     * @exception SQLException if a jdbc call throws an exception
     */
    public DbbSqlQueryResult appendRows(DbbSqlQueryResult queryResult)
        throws SQLException {

        int counter = 0;

        if (!moreToRead_) {
            close();
            return queryResult;
        }

        while (moreToRead_ = rs_.next()) {
            counter++;
            queryResult.addRow(rs_);

            if (counter >= numOfRowsToGet) {
                break;
            }
        }

        // if (stdlog_.isDebugEnabled())
        // stdlog_.debug("After a partial result set or a break");

        if (!moreToRead_)
            close();

        return queryResult;
    }

    public DbbSqlQueryResult getEmptyResult() {

        DbbSqlQueryResult queryResult = new DbbSqlQueryResult();

        // if (stdlog_.isDebugEnabled())
        // stdlog_.debug("Created query result set");

        queryResult.setColumns(cols);

        // if (stdlog_.isDebugEnabled())
        // stdlog_.debug("Column s " + cols);

        queryResult.setColumnNames(columnNames);

        // if (stdlog_.isDebugEnabled())
        // stdlog_.debug("Created Column names");

        queryResult.setColumnTypes(columnTypes);

        return queryResult;
    }

    /**
     * Release the current database connection.
     * To be called when all the data have been read from the result set.
     *
     * @exception SQLException if a jdbc call throws an exception
     */
    public void close() throws SQLException {
        if (rs_ != null) {
            rs_.close();
            session_.releaseCurrentConnection();
            rs_ = null;
        }
    }
}

