package org.eso.ohs.core.dbb.xml;

import java.io.IOException;
import java.io.Reader;
import java.sql.SQLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;

import org.eso.ohs.core.dbb.htmlclient.DbbHtmlCheckBox;
import org.eso.ohs.core.dbb.htmlclient.DbbHtmlComboBox;
import org.eso.ohs.core.dbb.htmlclient.DbbHtmlMultiBox;
import org.eso.ohs.core.dbb.htmlclient.DbbHtmlPanel;
import org.eso.ohs.core.dbb.htmlclient.DbbHtmlTextField;
import org.eso.ohs.core.dbb.htmlclient.DbbHtmlThreeState;
import org.eso.ohs.core.dbb.htmlclient.DbbHtmlView;
import org.eso.ohs.core.dbb.htmlclient.DbbHtmlWidget;
import org.eso.ohs.core.dbb.server.DbbSession;
import org.eso.ohs.core.dbb.server.DbbSqlEngineImpl;
import org.eso.ohs.core.dbb.sql.DbbInvalidSqlOperatorType;
import org.eso.ohs.core.dbb.sql.DbbSqlChunk;
import org.eso.ohs.core.dbb.sql.DbbSqlOperator;
import org.eso.ohs.core.dbb.sql.DbbSqlTable;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import org.xml.sax.Locator;
import org.xml.sax.helpers.DefaultHandler;

/**
 * An XML-based factory for DBB objects: builds DBB
 * instances by parsing an XML file.
 * <BR>
 * Implements the Singleton pattern.
 *
 * @version $Revision: 1.10 $
 * @author amchavan 
 */

// $Id: DbbHtmlFactory.java,v 1.10 2006/04/27 13:24:18 tcanavan Exp $

public class DbbHtmlFactory extends DefaultHandler {
	/** DbbPanel Co-ordinates */
	String type_;
	String position_;
	boolean visible_;
	int x_;
	int y_; 
    
	/** Singleton instance */
	private static DbbHtmlFactory factory = null;

	/** The fabricated DbbView */
	private DbbHtmlView dbbView = null;
    
	/** The list of dafault column chunks. */
	private List<DbbSqlChunk> defcols = null;
    
	/** A registry of objects we create from the XML file.
	 * This objects will be referenced by other objects
	 * in different areas of the DBB.
	 */
	private HashMap<String, Object> elementRegistry = null;

	/** To keep track of where we are in the file */
	private Locator locator = null;

	/** The fabricated DbbSession */
	private DbbSession session = null;
    
	/** A registry of DBB widgets we create from the XML file.
	 */
	private HashMap<Object, DbbHtmlWidget> widgetRegistry = null;
    
	/** A temporary list of DbbWidgets we create from the XML file. */
	private List widgetsTMP = null;
    
	/** Private constructor (this is a Singleton) */
	private DbbHtmlFactory() {
		//;
	}
    
	/** Clean up all internal data structures */
	private void clear() {
		defcols         = new LinkedList<DbbSqlChunk>();
		elementRegistry = new HashMap<String, Object>();
		widgetRegistry  = new HashMap<Object, DbbHtmlWidget>();
		dbbView   = null;
		session   = null;
	}
	
	/** @return The current input line. */
	private int currentLine() {
		return -99;
	}

	/**
	 * @param text Text of the error message.
	 * @return A nicely formatted error message, including
	 *         input line number.
	 */
	private String errmsg( String text ) {
		StringBuffer sb = new StringBuffer();
		sb.append( "XML Input Line Number" );
		sb.append( currentLine() );
		sb.append( ": " );
		sb.append( text );
		return sb.toString();
	}

	/**
	 * Utility method: filter out all non-Element children
	 * of an Element.
	 * 
	 * @param element A JDOM Element.
	 * @return An iterator over all child Elements of the
	 *         input Element.
	 */
	private Iterator<Element> getChildElements( Element element ) {
        
		Vector<Element>   children = new Vector<Element>();
		List      content = element.getContent();
		Iterator iterator = content.iterator();
        
		while( iterator.hasNext() ) {
			Object o = iterator.next();
			if( o instanceof Element ) {
				Element child = (Element) o;
				children.add( child );
			}
		}
		return children.iterator();
	}


	/** @return The fabricated DbbSession */
	public DbbSession getDbbSession() {
		return session;
	}
    
	/** @return The fabricated DbbView */
	public DbbHtmlView getHtmlDbbView() {
		return dbbView;
	}
    
	/**
	 * Retrieve a DBB element by id.
	 * 
	 * @param id   Key to retrieve the object. If the input XML document
	 *             contains for instance an element like:
	 *             <pre>
	 *               &lt;DbbSqlChunk id="chunk1" column="surname_dl" ... /&gt;
	 *             </pre>
	 *             this method allows to retrieve the corresponding
	 *             <code>DbbSqlChunk</code> using key <em>chunk1</em>.
	 * 
	 * @return The object registered with the input id, if any;
	 *         <tt>null</tt> otherwise.
	 */
	public Object getElement( String id ) {
		return elementRegistry.get( id );
	}

	/** @return The current element registry. It is public for debugging
	 *          purposes. */
	public HashMap<String, Object> getElementRegistry() {
		return elementRegistry;
	}
    
	/**
	 * @return All the elements defined in the input XML document.
	 */
	public Iterator<Object> getElements() {
		Collection<Object> values = elementRegistry.values();
		Iterator<Object> iterator = values.iterator();
		return iterator;
	}
    
	/**
	 * Retrieve a widget by ID.
	 * 
	 * @param id   Key to retrieve the object. If the input XML document
	 *             contains for instance an element like:
	 *             <pre>
	 *               &lt;DbbTextField chunk="chunk1" /&gt;
	 *             </pre>
	 *             this method allows to retrieve the that <code>DbbWidget</code>
	 *             using key <em>chunk1</em>.
	 * 
	 * @return The object registered with the input key, if any;
	 *         <tt>null</tt> otherwise.
	 */
	public DbbHtmlWidget getWidget( Object id ) {
		return widgetRegistry.get( id );
	}
    
	/** @return The registry of <code>DbbWidget</code>s created
	 *          by this factory when parsing the input XML document.
	 *          This method is public for debugging purposes.
	 */
	public HashMap<Object, DbbHtmlWidget> getWidgetRegistry() {
		return widgetRegistry;
	}    
    
	/**
	 * @return All the widgets created by this factory when parsing
	 *         the input XML document.
	 */
	public Iterator<DbbHtmlWidget> getWidgets() {
		Collection<DbbHtmlWidget> values = widgetRegistry.values();
		Iterator<DbbHtmlWidget> iterator = values.iterator();
		return iterator;
	}
    
	/** makeDbbSortButton */ 
	private DbbHtmlThreeState makeDbbSortButton( HashMap<String, String> attrs ) throws DbbXMLException {
	  if( attrs.size() == 0 )
		  throw new DbbXMLException( errmsg( "no attributes" ));
    
		String aLabeltext = attrs.get( "labeltext" );
		String aChunk     = attrs.get( "chunk" );
    
		if( aChunk == null || aChunk.length() == 0 )
			throw new DbbXMLException( errmsg( "Empty chunk id" ) );
		DbbSqlChunk c = (DbbSqlChunk) getElement( aChunk );
		if( c == null )   
			throw new 
				DbbXMLException( errmsg( "Chunk id '" + aChunk + "' not found" ));
                
		if( aLabeltext == null || aLabeltext.length() == 0 )
			aLabeltext = c.getTable().getTable() + "." + c.getColumn();

		DbbHtmlThreeState ab = new DbbHtmlThreeState(aLabeltext, c);
		putWidget( aChunk, ab );
		return ab;
	}
	
	
	/** 
	 * DbbComboBox
	 * The xml tag should look something like this:
	 * <DbbComboBox chunk="chunk0" labeltext="Telescope" items="UT1, UT2" delim=","/>
	 * */
	private DbbHtmlComboBox makeDbbComboBox (HashMap<String, String> attrs) throws DbbXMLException {
		if(attrs == null)
			throw new DbbXMLException(errmsg("no attributes"));
		
		DbbHtmlComboBox result;
			
		String aLabeltext = 	attrs.get("labeltext");
		String aChunk = 		attrs.get("chunk");
		String aItems = 		attrs.get("items");
		String aDelim =	   		attrs.get("delim");
		
		//chunk
		if(aChunk == null || aChunk.length() == 0)
			throw new DbbXMLException(errmsg("Empty chunk id"));
		DbbSqlChunk c = (DbbSqlChunk)getElement(aChunk);
		if(c == null)
			throw new DbbXMLException(errmsg("Chunk id '" + aChunk + "' not found"));

		//labeltext
		if(aLabeltext == null || aLabeltext.length() == 0)
			aLabeltext = c.getTable().getTable() + "." + c.getColumn();

		//~~~~~~~~~~~~~~~~~~~~~~~~~TODO~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		if(aItems == null || aItems.length() == 0 || aDelim == null || aDelim.length() == 0)
			throw new DbbXMLException("Empty Items list");
		
		StringTokenizer st = new StringTokenizer(aItems, aDelim);
		Vector<String> v = new Vector<String>();								
		while (st.hasMoreElements()){
			v.addElement(st.nextToken());
		}
		Object[] obj = v.toArray(); 
		result = new DbbHtmlComboBox( aLabeltext, c, obj);
		putWidget( aChunk, result );
		return result;
		//~~~~~~~~~~~~~~~~~~~~~~~~~TODO~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~			 											
	}
	
	
	/** 
	 * DbbHtmlComboBox 
	 * Populate it with data resulting from an DbbItemList.
	 * xml tag should look something like this:
	 * <DbbItemList id="telescopes"
	 * 		select distinct tel from sched_rep 
	 * </DbbItemList>
	 * <DbbSqlComboBox chunk="chunk0" labeltext="Telescope" items="telescopes"/>
	 * */	
	private DbbHtmlComboBox makeDbbSqlComboBox (HashMap<String, String> attrs) throws DbbXMLException {
		if(attrs == null)
			throw new DbbXMLException(errmsg("no attributes"));

		String aChunk = 		attrs.get("chunk");					
		String aLabeltext = 	attrs.get("labeltext"); 
		String aItems = 		attrs.get("items");
		
		if(aChunk == null || aChunk.length() == 0)
			throw new DbbXMLException(errmsg("Empty chunk id"));
		DbbSqlChunk c = (DbbSqlChunk)getElement(aChunk);
		if(c == null)
			throw new DbbXMLException(errmsg("Chunk id '" + aChunk + "' not found"));
		
		if(aLabeltext == null || aLabeltext.length() == 0)
			aLabeltext = c.getTable().getTable() + "." + c.getColumn();
			
		if(aItems == null || aItems.length() == 0)
			throw new DbbXMLException(errmsg("Chunk id '" + aChunk + "' not found"));
		String[] items = (String[])getElement(aItems);
		
		DbbHtmlComboBox cb = new DbbHtmlComboBox( aLabeltext, c, items );
		putWidget( aChunk, cb );
		return cb;						
	}

	
	/** 
	 * List of items for DbbHtmlComboBox 
	 * xml tag should look something like this
	 * <DbbItemList id="telescopes" >
	 * 		select distinct tel from sched_rep
	 * </DbbItemList>
	 * */
	private void makeDbbItemList (HashMap<String, String> attrs, Element element) throws DbbXMLException {
		if( attrs.size() == 0 )
			throw new DbbXMLException( errmsg( "no attributes" ));
        
		String[] items = null;
        
		String aId         	= attrs.get( "id" );	
		String aItems 		= element.getText();
				
		if( aId == null || aId.length() == 0)
			throw new DbbXMLException( errmsg( "No id specified" ) );
			
		if( aItems == null || aItems.length() == 0 )
			throw new DbbXMLException( errmsg( "No items specified" ) );

		try {
			items = DbbXmlDataHelper.getInstance().readStringArray(aItems);			
		} catch (ObjectIOException e) {
			e.printStackTrace();
		}
		putElement(aId, items);							
	}
	
	
	/** 
	 * DbbMultiBox
	 * */ 
	private DbbHtmlMultiBox makeDbbMultiBox (HashMap<String, String> attrs) throws DbbXMLException {
		if(attrs == null)
			throw new DbbXMLException(errmsg("no attributes"));
			
			String aLabeltext = 	attrs.get("labeltext");
			String aChunk = 		attrs.get("chunk");
			String aItems = 		attrs.get("items");
			String aDelim =	   		attrs.get("delim");
		
			//chunk
			if(aChunk == null || aChunk.length() == 0)
				throw new DbbXMLException(errmsg("Empty chunk id"));
			DbbSqlChunk c = (DbbSqlChunk)getElement(aChunk);
			if(c == null)
				throw new DbbXMLException(errmsg("Chunk id '" + aChunk + "' not found"));

			//labeltext
			if(aLabeltext == null || aLabeltext.length() == 0)
				aLabeltext = c.getTable().getTable() + "." + c.getColumn();

			//TODO - may use a list of items like the DbbHtmlComboBox TBD
			if(aItems == null || aItems.length() == 0 || aDelim == null || aDelim.length() == 0)
				throw new DbbXMLException("Empty Items list");
			else 
			{
				StringTokenizer st = new StringTokenizer(aItems, aDelim);
				Vector<String> v = new Vector<String>();								
				while (st.hasMoreElements()){
					v.addElement(st.nextToken());
				}
				Object[] obj = v.toArray(); 
				DbbHtmlMultiBox cb = new DbbHtmlMultiBox( aLabeltext, c, obj);
				putWidget( aChunk, cb );
				return cb;
			}			 											
		}
			

	private DbbHtmlCheckBox makeDbbCheckBox( HashMap<String, String> attrs ) throws DbbXMLException {
    
		if( attrs.size() == 0 )
			throw new DbbXMLException( errmsg( "no attributes" ));
    
		String aLabeltext = attrs.get( "labeltext" );
		String aChunk     = attrs.get( "chunk" );
    
		if( aChunk == null || aChunk.length() == 0 )
			throw new DbbXMLException( errmsg( "Empty chunk id" ) );
			DbbSqlChunk c = (DbbSqlChunk) getElement( aChunk );
			if( c == null )   
				throw new 
					DbbXMLException( errmsg( "Chunk id '" + aChunk + "' not found" ));
                
		if( aLabeltext == null || aLabeltext.length() == 0 )
			aLabeltext = c.getTable().getTable() + "." + c.getColumn();                    
		
		DbbHtmlCheckBox cb = new DbbHtmlCheckBox( aLabeltext, c, true );
		putWidget( aChunk, cb );
		return cb;
	}
	
		
	private DbbHtmlPanel makeDbbPanel( HashMap<String, String> attrs, Element element )
		throws DbbXMLException {
                
		if( attrs.size() == 0 )
			throw new DbbXMLException( errmsg( "no attributes" ));
        

		String aTitle   = attrs.get( "title" );
		String aColNum  = attrs.get( "colnumber" );
		String id       = attrs.get( "id" );
		String type     = attrs.get( "type" );
		String pos      = attrs.get( "position" );
		String aVisible = attrs.get( "visible" );
		String aX       = attrs.get( "x" );
		String aY       = attrs.get( "y" );                
        
		// Now some checks
		//-------------------------------------
                    
		if( type == null )
			throw new DbbXMLException( errmsg( "Missing DbbPanel type." ));
		else type_ = type;			

		if( aTitle == null )
			aTitle = "";
        
		int colNum = 0;
		if( aColNum != null && aColNum.length() != 0 )
			colNum = Integer.parseInt(aColNum);

		if( pos == null )
		   pos = "N";
		   position_ = pos;
        
		boolean visible = true;
		if( aVisible != null )
			visible = Boolean.valueOf( aVisible ).booleanValue();
			visible_ = visible;
        
		int x = -1;
		if( aX != null )
			x = Integer.valueOf( aX ).intValue();
			x_ = x;
        
		int y = -1;
		if( aY != null )
			y = Integer.valueOf( aY ).intValue();
			y_ = y;
            
		DbbHtmlPanel panel = new DbbHtmlPanel( aTitle, colNum );
		
		// Make this panel accessible from outside the factory
		putElement( aTitle, panel );
        
		// Now recurse down to the panel's children
		Iterator<Element> iterator = getChildElements( element );
		while( iterator.hasNext() ) {
			Element child = iterator.next();
			Object object = visit( child ); 
			if( isDbbWidget(object))
				panel.addWidget( (DbbHtmlWidget)object );
		}
		return panel;
	}
	
	private void makeDbbSession( HashMap<String, String> attrs ) throws DbbXMLException {
    
		if( attrs.size() == 0 )
			throw new DbbXMLException( errmsg( "no attributes" ));

		String url   = attrs.get( "url" );
		String defdb = attrs.get( "defdb" );
		String usr  = attrs.get( "usr" );
		String pswd  = attrs.get("pswd");
    
		if( url == null || url.length() == 0 )
			throw new DbbXMLException( errmsg( "JDBC URL missing or empty" ));
                
		if( defdb == null || defdb.length() == 0 )
			throw new DbbXMLException( errmsg( "Default database missing or empty" ));
			
		if( usr == null || usr.length() == 0 )
			throw new DbbXMLException( errmsg( "Default database user missing or empty" ));
					
		if( pswd == null || pswd.length() == 0 )
			throw new DbbXMLException( errmsg( "Default database password missing or empty" ));								                    
		
		session = new DbbSession( url, defdb, usr, pswd );
	}	


	private void makeDbbSqlChunk( HashMap<String, String> attrs ) throws DbbXMLException {
                
		if( attrs.size() == 0 )
			throw new DbbXMLException( errmsg( "no attributes" ));
        
		String id         = attrs.get( "id" );		
		String aTable     = attrs.get( "table" );
		String column     = attrs.get( "column" );
		String label      = attrs.get( "label" );
		String aOperator  = attrs.get( "operator" );
		String aData_type = attrs.get( "data_type" );
		String function   = attrs.get( "function" );
		String aDefault   = attrs.get( "default" );
                
		DbbSqlOperator op = null;
		DbbSqlTable table = null;
		Class   data_type = null;
        
		// Now some checks
		//-------------------------------------
        
		if( aTable == null )
			throw new DbbXMLException( errmsg( "No table specified" ));
        
		table = (DbbSqlTable) getElement( aTable );
		if( table == null )
			throw new DbbXMLException( errmsg( "table '" + aTable + "' not found"));
            
		if( column == null || column.length() == 0 )
			throw new DbbXMLException( "Column name not found" );
            
		if( aData_type == null || aData_type.length() == 0 )
			data_type = null;
		else
			try {
				data_type = Class.forName( aData_type );
			}
			catch (ClassNotFoundException e) {
				throw new DbbXMLException(
					errmsg( "class '" + aData_type + "' not found" ));
			}
            
		boolean def = false;
		if( aDefault != null && aDefault.length() > 0 )
			def = Boolean.valueOf(aDefault).booleanValue();
        
        
		// Some chunks have no operator.
		// If one is specified, look for it: if we don't find it
		// we try to create one.
		// If *that* fails, well, we're in trouble.
		//--------------------------------------------
		if( aOperator != null ) {
			op = (DbbSqlOperator) getElement( aOperator );
			if( op == null ) {
				try {
					op = new DbbSqlOperator( aOperator );
				}
				catch (DbbInvalidSqlOperatorType e1) {      
					throw new DbbXMLException( errmsg( e1.getMessage() ));
				}
			}
		}
            
		DbbSqlChunk chunk =
			new DbbSqlChunk( table, column, label, op, data_type, function );
        
		if( id != null )
			putElement( id, chunk );
		if( def ) {
			defcols.add( chunk );
		}
	}
	

	private void makeDbbSqlOperator( HashMap<String, String> attrs ) throws DbbXMLException {
    
		String value = null;
        
		if( attrs.size() == 0 )
			throw new DbbXMLException( errmsg( "no attributes" ));
        
		value = attrs.get( "value" );    
		if( value == null || value.length() == 0 )
			throw new DbbXMLException( errmsg( "empty operator value" ));
    
		DbbSqlOperator operator;
		try {
			operator = new DbbSqlOperator(value);
		} catch (DbbInvalidSqlOperatorType e) {
			throw new DbbXMLException( 
				errmsg( "invalid operator: '" + value + "'" ));
		}
		putElement( value, operator );
	}

	/**
	 * Create a DbbSqlTable and puts it in the elementRegistry.
	 * @param attrs
	 * @throws DbbXMLException
	 */
	private void makeDbbSqlTable( HashMap<String, String> attrs ) throws DbbXMLException {
    
		if (attrs == null)
			throw new DbbXMLException(errmsg("no attributes"));
     
		String db_name    = attrs.get("dbname");
		String owner_name = attrs.get("ownername");
		String table_name = attrs.get("tablename");
    
		if (table_name == null || table_name.length() == 0)
			throw new DbbXMLException(errmsg("Empty table name"));
    
		DbbSqlTable table = new DbbSqlTable(db_name, owner_name, table_name);
		putElement( table.toString(), table );
	}
	
	
	private DbbHtmlTextField makeDbbTextField( HashMap<String, String> attrs ) throws DbbXMLException {
    
		if( attrs.size() == 0 )
			throw new DbbXMLException( errmsg( "no attributes" ));
    
		String aLabeltext = attrs.get( "labeltext" );
		String aChunk     = attrs.get( "chunk" );
		String aWidth     = attrs.get( "width" );
    
		if( aChunk == null || aChunk.length() == 0 )
			throw new DbbXMLException( errmsg( "Empty chunk id" ) );
		DbbSqlChunk c = (DbbSqlChunk) getElement( aChunk );
		if( c == null )   
			throw new 
				DbbXMLException( errmsg( "Chunk id '" + aChunk + "' not found" ));
                
		if( aLabeltext == null || aLabeltext.length() == 0 )
			aLabeltext = c.getTable().getTable() + "." + c.getColumn();
        
		int width;
		if( aWidth == null || aWidth.length() == 0 )
			width = DbbHtmlTextField.COLUMNSIZE;
		else
			width = Integer.parseInt( aWidth );
            
		DbbHtmlTextField tf = new DbbHtmlTextField( aLabeltext, c);
		putWidget( aChunk, tf );
		return tf;
	}
	
   
	private void makeDbbView( Element element ) throws DbbXMLException {
   
		if( session == null )
			throw new DbbXMLException( errmsg( "no DbbSession found" ));
        
		// The DbbView constructor requires the following parameters:
		//      sqlEngine SQL engine of the browser [MANDATORY]
        
		DbbSqlEngineImpl e = null;
		try {
			e = new DbbSqlEngineImpl(session, false);
		}
		catch (SQLException e1) {
			// this in fact should never happen!
			e1.printStackTrace();
		}
        
		// List defcols is populated with DbbSqlChunks
		while( defcols.size() > 0 )           
			e.addDefaultColumnsChunk( defcols.remove( 0 ));

		dbbView = new DbbHtmlView( e );
		
		//Now recurse down to add DbbPanel(s) to the view
		Iterator<Element> iterator = getChildElements( element );
		while( iterator.hasNext() ) {
			Element child = iterator.next();
			Object object = visit( child );
			if( isDbbPanel(object))
			{ 			
				if(type_.equals("COLS")) {
					dbbView.setColumnsPanel((DbbHtmlPanel)object, position_, visible_, x_, y_);
				}
				else if(type_.equals("ROWS")){
					dbbView.setWherePanel((DbbHtmlPanel)object, position_, visible_, x_, y_);
				}					
				else if(type_.equals("SORT")) {
					dbbView.setSortPanel((DbbHtmlPanel)object, position_, visible_, x_, y_);
				}
				else{	
					throw new DbbXMLException(errmsg("Unknown DbbPanel type: " +
													 "should be one of COLS, ROWS, SORT."));										
				}
			}			
		}
	}	

    	
	/**
	 * Parse an XML file and build all the DBB elements
	 * and widgets described therein.
	 * <BR>
	 * DBB elements like its view and session can be retrieved
	 * using accessor methods <code>getDbbSession()</code>,
	 * <code>getDbbView()</code>, <code>getElements()</code>
	 * and <code>getWidgets()</code>.
	 * 
	 * @param url  A Reader for the XML file/text to parse.
	 * 
	 * @throws JDOMException   If input file is not well-formed.
	 * @throws IOException     If input file cannot be read.
	 * @throws DbbXMLException If input file is not a valid DBB descriptor.
	 * 
	 * @see #getDbbSession()
	 * @see #getDbbView()
	 * @see #getElements()
	 * @see #getWidgets()
	 */    
	public void parse( Reader reader ) 
		throws JDOMException, IOException, DbbXMLException {
		//	org.jdom.input.SAXBuilder parser = new org.jdom.input.SAXBuilder();
			SAXBuilder parser = new SAXBuilder();
		Document document = parser.build( reader );
		visit( document );
	}

	/**
	 * Parse an XML file and build all the DBB elements
	 * and widgets described therein.
	 * <BR>
	 * DBB elements like its view and session can be retrieved
	 * using accessor methods <code>getDbbSession()</code>,
	 * <code>getDbbView()</code>, <code>getElements()</code>
	 * and <code>getWidgets()</code>.
	 * 
	 * @param url  URL of the XML file to parse.
	 * 
	 * @throws JDOMException   If input file is not well-formed.
	 * @throws IOException     If input file cannot be read.
	 * @throws DbbXMLException If input file is not a valid DBB descriptor.
	 * 
	 * @see #getDbbSession()
	 * @see #getDbbView()
	 * @see #getElements()
	 * @see #getWidgets()
	 */    
	public void parse( String url )
		throws JDOMException, IOException, DbbXMLException {
		SAXBuilder parser = new SAXBuilder();
		Document document = parser.build( url );
		visit( document );
	}

	/**
	 * Register an element for future reference.
	 * @param key  Key to retrieve the object
	 * @param o    Object to register
	 * @throws DbbXMLException   If some object is already registered
	 *                           with the same key.
	 */
	private void putElement( String key, Object o ) throws DbbXMLException {
		if( getElement( key ) != null )
			throw new DbbXMLException( errmsg( "duplicate key: " + key ));
            
		elementRegistry.put( key, o );
	}
    
	/**
	 * Register a widget for future reference.
	 * @param key  Key to retrieve the object
	 * @param o    Object to register
	 * @throws DbbXMLException   If some object is already registered
	 *                           with the same key.
	 */
	private void putWidget( Object key, DbbHtmlWidget w ) throws DbbXMLException {
		if( getWidget( key ) != null )
			throw new DbbXMLException( errmsg( "duplicate key: " + key ));
            
		widgetRegistry.put( key, w );
	}

	/**
	 * Converts a list of Attribute objects to a HashMap
	 * of name/value pairs.
	 */
	private HashMap<String, String> toHashMap( List attributes ) {
		HashMap<String, String> map = new HashMap<String, String>(attributes.size());

		Iterator iterator = attributes.iterator();
		while (iterator.hasNext()) {
			Attribute attr = (Attribute) iterator.next();
			String   value = attr.getValue();
			String     key = attr.getName();
			map.put( key, value );
		}
		return map;
	}
	

	/**
	 * Process a JDOM document resulting from parsing a DBB descriptor
	 * and build all corresponding DBB elements.
	 * 
	 * @param document The input JDOM document.
	 * @return The input JDOM document, unchanged.
	 * 
	 * @throws DbbXMLException If input file is not a valid DBB descriptor.
	 */
	private void visit( Document document ) throws DbbXMLException {
		Element root = document.getRootElement();
		List content = root.getContent();
		visit( root );
	}


	/**
	 * Process a JDOM tree node and build the corresponding DBB elements.
	 * 
	 * @param element  Input JDOM tree node.
	 * @return A DbbWidget if one was created; 
	 *         <code>null</code> otherwise.
	 */
	private Object visit( Element element ) throws DbbXMLException {

		String   name = element.getQualifiedName();
		List   lattrs = element.getAttributes();
		HashMap<String, String> attrs = toHashMap( lattrs );
		
		Object object = null;
        
		if( element.isRootElement() ) {
			Iterator<Element> iterator = getChildElements( element );
			while( iterator.hasNext() ) {
				Element child = iterator.next();
				visit( child );
			}
		}
		else if( name.equals( "DbbSqlChunk" ))    makeDbbSqlChunk( attrs );
		else if( name.equals( "DbbItemList" ))	  makeDbbItemList( attrs, element );
		else if( name.equals( "DbbView" ))        makeDbbView( element );
		else if( name.equals( "DbbPanel" ))       object = makeDbbPanel( attrs, element );
		else if( name.equals( "DbbSqlTable" ))    makeDbbSqlTable( attrs );
		else if( name.equals( "DbbSqlOperator" )) makeDbbSqlOperator( attrs );
		else if( name.equals( "DbbTextField" ))   object = makeDbbTextField( attrs );
		else if( name.equals( "DbbCheckBox" ))    object = makeDbbCheckBox( attrs );
		else if( name.equals( "DbbSortButton" ))  object = makeDbbSortButton( attrs );
		else if( name.equals( "DbbComboBox" ))    object = makeDbbComboBox( attrs );
		else if( name.equals( "DbbMultiBox" ))	  object = makeDbbMultiBox( attrs );
		else if( name.equals( "DbbSqlComboBox" )) object = makeDbbSqlComboBox( attrs );	 
		else if( name.equals( "DbbSession" ))     makeDbbSession( attrs );

		return object;
	}
	
	/** @return true if input object is a DbbPanel, false otherwise */
	@SuppressWarnings("unchecked")
	private boolean isDbbPanel( Object o ) {
		  boolean b;
		  if(o == null) b = false;
		  else {
			  Class   c = o.getClass();
			  Class   p = DbbHtmlPanel.class;
			  b = p.isAssignableFrom( c );
		  }
		  return b;
	}
	  
	     
	/** @return true if input object is a DbbWidget, false otherwise */
	@SuppressWarnings("unchecked")
	private boolean isDbbWidget( Object o ) {
		boolean b;
		if(o == null) b = false;
		else {
		  Class   c = o.getClass();
		  Class   w = DbbHtmlWidget.class;
		  b = w.isAssignableFrom( c );
		}
		return b;
	}
	
	/** @return A DbbFactory object (Singleton pattern). */
	public static DbbHtmlFactory getInstance() {
		if( factory == null )
			factory = new DbbHtmlFactory();
			factory.clear();
		return factory;
	}
}
