package org.eso.ohs.core.docview.datatrans;

/**
 * A transducer to link to a String property but including a mapping
 * between displayed and stored values.
 *
 * <p>
 * For example, users will call it's
 * setValue/getValue methods.  Those calls will be translated into
 * calls to setX/getX on the object given to the constructor.
 * Values in the displayed_ array will first be mapped to the
 * corresponding value in displayed_, or vice versa.
 *
 * <p>
 * If the displayed/stored value is not present in the array,
 * it is passed through unchanged.
 * 
 * <p>
 * $Id: MappingAdaptor.java,v 1.1 2004/02/17 12:37:59 tcanavan Exp $
 * <br>
 * @author Bob Kemp
 */
public class MappingAdaptor extends ObjAdaptor {

    /** internal (stored) value corresponding to the value in displayed_ */
    private String[] internal_;
 
    /** displayed value corresponding to the value in internal_ */
    private String[] displayed_;


    /**
     * Constructs an adaptor that maps between values in internal[]
     * in the object and displayed[] in the GUI component and vice versa.
     * Values not found in the corresponding array are passed through
     * unchanged.
     * 
     * @param internal	array of values, corresponding to entries in
     * 			displayed, that are stored in the object.
     * @param displayed	array of values, corresponding to entries in
     * 			internal, that are to be displayed to the user
     */
    public MappingAdaptor(String[] internal, String[] displayed) {
	super(String.class);

	if (internal.length != displayed.length) {
	    throw new IllegalArgumentException(
		"MappingAdaptor: arrays of unequal length"
	    );
	}

	internal_  = internal;
	displayed_ = displayed;
    }

    /**
     * Maps elements of displayed_ to corresponding element of internal_.
     * Values not found in the corresponding array are passed through
     * unchanged.
     * 
     * @param propVal	value of property as might be displayed to the user.
     */
    public Object toMetaLevelObject(String propVal) {
	for (int j=0; j < displayed_.length; j++) {
	    if (displayed_[j].equals(propVal)) {
		return internal_[j];
	    }
	}

        return propVal;
    }

    /**
     * Maps elements of internal_ to corresponding element of displayed_.
     * Values not found in the corresponding array are passed through
     * unchanged.
     * 
     * @param propVal	value of property as might be stored in the object.
     */
    public String fromMetaLevelObject(Object obj) {
	for (int j=0; j < internal_.length; j++) {
	    if (internal_[j].equals(obj)) {
		return displayed_[j];
	    }
	}

       return obj.toString();
    }

  

}

