package org.eso.ohs.core.docview.examples;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import javax.swing.AbstractAction;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.eso.ohs.core.docview.datatrans.DocViewObject;
import org.eso.ohs.core.docview.gui.DocView;

/**
 * Constructs several windows, each displaying fields from
 * one of several objects.  Each window contains a JList
 * of object names on the left and sever DocView objects on
 * the right.  When the user selects an object from the list,
 * it's fields are displayed on the right.  If several windows
 * are displaying the same object, updates in one will be
 * visible in the others.
 * 
 * <p>
 * To execute this example, type
 * <blockquote>
 * java org.eso.ohs.docview.datatrans.examples.MultiWin
 * </blockquote>
 * 
 * <p>
 * $Id: MultiWin.java,v 1.1 2004/02/17 12:38:00 tcanavan Exp $
 * <br>
 * @author Bob Kemp
 */
public class MultiWin extends JFrame {

    /** List of objects that can be displayed in this window */
    private JList list_;

    /** Collection of DocView objects to display the object's fields */
    private ObjView objview_;

    /** object currently displayed in this window */
    private TestObject currObj_;

    /** List of objects that can be displayed in this window: see list_ */
    private TestObject[] objs_;

    /** 
     * Construct a window with a list of object names on the left
     * and a few DocView objects on the left.
     */
    public MultiWin(int winNo, TestObject[] objs) {
	objs_= objs;

	list_ = new JList(objs_);
	currObj_ = (objs_ == null || objs_.length == 0) ? null : objs_[0];

	objview_ = new ObjView();

	setTitle("window " + winNo);

	Container content = getContentPane();

	content.setLayout(new BorderLayout());

	JLabel lab = new JLabel(
			"Updates are made when the component loses focus"
		    );

	content.add(lab, BorderLayout.NORTH);
	content.add(list_, BorderLayout.WEST);
	content.add(objview_, BorderLayout.EAST);

	list_.addListSelectionListener(new ObjSelector(objs_));

	ListSelectionModel lsm = list_.getSelectionModel();
	lsm.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
	lsm.addSelectionInterval(0, 0);

	JMenuBar mb = new JMenuBar();
	JMenu menu = new JMenu("File");

	menu.add(new QuitAction());
	mb.add(menu);
	setJMenuBar(mb);
    }

    /**
     * Listens to selections in the list and sets the object
     * displayed accordingly.
     */
    public class ObjSelector implements ListSelectionListener {
	private TestObject[] objs_;
	private int winNo_;

	public ObjSelector(TestObject[] objs) {
	    objs_= objs;
	}

	public void valueChanged(ListSelectionEvent ev) {
	    if ( ! ev.getValueIsAdjusting() ) {
		JList list = (JList)ev.getSource();
		int index = list.getSelectedIndex();

		objview_.setObject(objs_[index]);
	    }
	}
    }

    /** Quit action handler */
    public class QuitAction extends AbstractAction {
	public QuitAction() {
	    super("Quit");
	}

	public void actionPerformed(ActionEvent evt) {
	    System.exit(0);
	}

    }


    /**
     * Entry point to the program.
     * Creates objects and starts windows.
     * 
     * @param args	Command-line arguments -- ignored
     */
    public static void main(String[] args) {
	TestObject[] objs = new TestObject[5];

	for (int j=0; j < objs.length; j++) {
	    objs[j] = new TestObject(j);
	}

	MultiWin mw;

	for (int j=1; j <= 3; j++) {
	    mw = new MultiWin(j, objs);
	    mw.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
	    mw.pack();
	    mw.setLocation(760, 200+220*(j-1));
	    mw.setVisible(true);
	}

    }


    /**
     * Collection of DocView objects to display a fixed set of fields
     * of an object.
     */
    public static class ObjView extends JPanel {

	/**
	 * Array of DocView objects displaying various fields
	 * of the selected object.
	 */
	private DocView[] props_;


	/**
	 * Constructor: builds and displays DocViews
	 */
	public ObjView() {
	    setLayout(new GridLayout(0, 1));

	    props_ = new DocView[5];

	    props_[0] = new DocView(new JTextField());
	    props_[1] = new DocView(new JTextField());
	    props_[2] = new DocView(new JTextArea());
	    props_[3] = new DocView(new JSlider());
	    props_[4] = new DocView(new JTextField());

	    for (int j=0; j < props_.length; j++) {
		add(props_[j]);
	    }
	}

	/**
	 * Constructor: builds and displays DocViews
	 */
	public void setObject(TestObject obj) {
	    props_[0].setObjectProperty(obj, "IntProperty");
	    props_[1].setObjectProperty(obj, "FloatProperty");
	    props_[2].setObjectProperty(obj, "IntProperty");
	    props_[3].setObjectProperty(obj, "IntProperty");
	    props_[4].setObjectProperty(obj, "StringProperty");
	}

    }

    /**
     * Test object implementing DocViewObject interface.
     * Note that addPropertyChangeListener() and
     * removePropertyChangeListener() are required by the
     * DocViewObject.  All changes to the object are made
     * through the set/get methods.
     */
    public static class TestObject implements DocViewObject {

	private int objectNo_;

	private int intProperty_;

	private float floatProperty_;

	private String stringProperty_;

	private PropertyChangeSupport support_
			= new PropertyChangeSupport(this);


	public TestObject(int objNo) {
	    objectNo_ = objNo;
	}

	public void setIntProperty(int value) {
	    intProperty_ = value;
	    firePropertyChange();
	}

	public int getIntProperty() {
	    return intProperty_;
	}

	public void setFloatProperty(float value) {
	    floatProperty_ = value;
	    firePropertyChange();
	}

	public float getFloatProperty() {
	    return floatProperty_;
	}

	public void setStringProperty(String value) {
	    stringProperty_ = value;
	    firePropertyChange();
	}

	public String getStringProperty() {
	    return stringProperty_;
	}


	/**
	 * Adds a new listener to be notified of changes.
	 * 
	 * @param lnr the listener
	 */
	public void addPropertyChangeListener(PropertyChangeListener lnr) {
	    support_.addPropertyChangeListener(lnr);
	}

	/**
	 * Removes a new listener, which will no longer be notified of changes.
	 * 
	 * @param lnr the listener
	 */
	public void removePropertyChangeListener(PropertyChangeListener lnr) {
	    support_.removePropertyChangeListener(lnr);
	}

	/**
	 * Informs listeners of a property change.
	 * No details of which property of the old/new value are given.
	 */
	protected void firePropertyChange() {
	    support_.firePropertyChange(null, null, null);
	}


	/** Print meaningful name for JList */
	public String toString() {
	    return "TestObject " + objectNo_;
	}

    }

}

