package org.eso.ohs.core.gui.baseaction;

import javax.swing.JFrame;
import javax.swing.SwingUtilities;

import org.apache.log4j.Logger;
import org.eso.ohs.core.gui.utilities.Utilities;
import org.eso.ohs.core.gui.widgets.OutOfMemoryReport;
import org.eso.ohs.core.gui.widgets.StoppableProgressDialog;


/**
 * Utility base class to perform actions on multiple objects,
 * eg duplication of multiple BusinessObject instances.
 * 
 * <p>
 * $Id: IncrementalActionAutomaton.java,v 1.2 2006/03/08 12:35:02 pnunes Exp $
 * <p>
 * @author Bob Kemp
 */
public class IncrementalActionAutomaton implements Runnable {

    private static Logger stdlog_ =
        Logger.getLogger(IncrementalActionAutomaton.class);

    /**
     * If more than this many window are to be actioned,
     * hide the detail view windows beforehand and restore
     * them after the actions are complete.
     */
    public final static int SHORT_ACTION_THRESHOLD = 5;

    /** Counter specifying the index of the affected object. */
    protected int counter_ = -2;

    /** Maximum number of affected objects. */
    protected int maximum_ = 100;

    /** Dialog box to allow the user to stop more items from being affected. */
    protected StoppableProgressDialog dialog_;

    /** Memento used to remember which windows were initially visible. */
    //private Object memento_;

    /** Base frame to use for dialog boxes, eg dialog_ */
    protected JFrame baseFrame_;

    /** Object that will actually perform the steps of the large action */
    protected IncrementalAction action_;

    protected int num_steps_before_pausing_;

    /**
     * Protected default constructor - useful for extending classes. Not part of the 
     * public interface.
     */
    protected IncrementalActionAutomaton() {
    }

    /**
     * Constructor.
     * 
     * @param action	Object that will perform the steps of a larger action
     * @param frm	Base frame to use for dialog boxes such as dialog_
     * @param msg	Message to display in the progress bar dialog (dialog_)
     */
    public IncrementalActionAutomaton(
        IncrementalAction action,
        JFrame frm,
        String msg) {
        init(action, frm, msg);
    }

    protected void init(IncrementalAction action, JFrame frm, String msg) {
        baseFrame_ = frm;
        action_ = action;
        maximum_ = action.getNumStepsRemaining(0);

        if (maximum_ > 1) {
            dialog_ = new StoppableProgressDialog(baseFrame_, msg, 0, maximum_);

            int num_pauses = (int) (Math.sqrt(maximum_) + 0.5) * 2;

            if (num_pauses > 20) {
                num_pauses = 20;
            }
            num_steps_before_pausing_ =
                (maximum_ + num_pauses - 1) / num_pauses;
        } else {
            dialog_ = null;
            num_steps_before_pausing_ = 1000;
        }

        runLater();
    }

    public StoppableProgressDialog getProgressDialog() {
        return dialog_;
    }

    /**
     * Utility method to queue this Runnable object for execution by
     * the Swing Event Queue Thread.  It is the run() method that is
     * actually executed.
     */
    protected void runLater() {
        if (counter_ >= 0 && counter_ % num_steps_before_pausing_ != 0) {
            SwingUtilities.invokeLater(this);
        } else {
            new SlightPause(this);
        }
    }

    protected void showProgressBar() {
        if (dialog_ != null) {
            dialog_.show();
        }
    }

    protected void setProgressBarValue(int value) {
        if (dialog_ != null) {
            dialog_.setValue(value);
            
        }
    }

    protected boolean shouldStop() {
        if (dialog_ != null) {
            return dialog_.shouldStop();
        } else {
            return false;
        }
    }

    protected void closeViews() {

    }

    /**
     * Method to manage each incremental step of a larger action.
     * It calls action_.performStep() to perform each step.
     */
    public void run() {
        counter_++;

        if (counter_ == -1) {
            action_.startOfAction();
            Utilities.waitCursor(baseFrame_);
            closeViews();
            showProgressBar();
            runLater();
        } else {
            int stepsRemaining = action_.getNumStepsRemaining(counter_);

            setProgressBarValue(maximum_ - stepsRemaining);

            if (!shouldStop() && stepsRemaining > 0) {
                try {
                    action_.performStep(counter_);
                            			
                    
                } catch (OutOfMemoryError e) {
                    e.printStackTrace();
                    OutOfMemoryReport.showMessage();
                    cleanUpAtEnd();
                    throw e;
                }
                runLater();
            } else {
                // All done: just cleanup and return
                new SlightPause(new Runnable() {
                    public void run() {
                        cleanUpAtEnd();
                    }
                }, 250);
                return;
            }
        }
    }

    public void cleanUpAtEnd() {
        //if (memento_ != null) {
        //MainWindow.getView().restoreViews(memento_);
        //	}
        //MainWindow.getView().setDisplaySuppressed(false);
        if (dialog_ != null) {
            dialog_.dispose();
        }
        Utilities.normalCursor(baseFrame_);
        action_.endOfAction();
    }

    /** Utility class to execute the given Runnable object afer a delay */
    public static class SlightPause extends Thread {
        private Runnable runner_;
        private int millisecDelay_;

        public SlightPause(Runnable run, int delay) {
            millisecDelay_ = delay;
            runner_ = run;

            start();
        }

        public SlightPause(Runnable run) {
            this(run, 10);
        }

        public void run() {
            try {
                sleep(millisecDelay_); // Pause for Display to catch up.
            } catch (InterruptedException e) {
                // do nothing -- ignore
            }
            SwingUtilities.invokeLater(runner_);
        }
    }

}
