package org.eso.ohs.core.gui.widgets;

import java.awt.Component;
import java.awt.Container;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.sql.SQLException;
import java.sql.SQLWarning;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JTextArea;

import org.eso.ohs.core.utilities.TextUtils;

/**
 * Utility methods for reporting error messages.
 * 
 * <p>
 * $Id: ErrorMessages.java,v 1.4 2006/03/20 14:09:03 ddorigo Exp $
 */

public class ErrorMessages {

    /**
     * Inform the user that an unrecoverable I/O error has occurred.
     * 
     * @param window
     *            window upon which the dialog box should be based.
     * 
     * @param e
     *            exception containing details of the I/O error.
     */
    public static void announceIOError(Component window, Exception e) {
        announceError(window,
                new String[]
                { "An error occurred while reading/writing data", " ",
                        e.getMessage() }, e);
    }

    /**
     * Inform the user that a storable object could not be retrieved because
     * because it could not be found.
     * 
     * @param window
     *            window upon which the dialog box should be based.
     * 
     * @param e
     *            exception containing details of the I/O error.
     */
    public static void announceNoObject(Component window, Exception e) {
        announceError(window,
                new String[]
                { "An internal consistency check has failed",
                        "The selected object failed to be loaded.", " ",
                        e.getMessage() }, e);
    }

    /**
     * Inform the user that an error occurred.
     * 
     * @param window
     *            window upon which the dialog box should be based.
     * @param text
     *            extra explanatory text.
     * @param e
     *            exception containing details of the I/O error.
     */
public static void announceError(Component window, String[] text,
							Exception e) {
	final JTextArea detailsArea = new JTextArea();
	JButton okButton = new JButton("OK");
	Object[] buttonsList = new Object[] { okButton };

	String sb = e.getMessage();


	    if (sb.length() > 0) {
		JButton detailsButton = new JButton("Details");

		detailsArea.setText(sb.toString());
		detailsArea.setEditable(false);
		detailsArea.setOpaque(false);
		buttonsList = new Object[] { okButton, detailsButton };

		detailsButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ev) {
			    JComponent src = (JComponent)ev.getSource();
			    Container cont = src.getTopLevelAncestor();

			    JOptionPane.showMessageDialog(
				cont,
				detailsArea,
				"Error details",
				JOptionPane.ERROR_MESSAGE
			    );
			}
		    }
		);
	    }

	okButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent ev) {
		    JComponent comp = (JComponent)ev.getSource();
		    Window win = (Window)comp.getTopLevelAncestor();

		    win.setVisible(false);
		    win.dispose();
		}
	    }
	);

	JOptionPane.showOptionDialog(
	    window,
	    text,
	    "Error",
	    JOptionPane.DEFAULT_OPTION,
	    JOptionPane.ERROR_MESSAGE,
	    null,
	    buttonsList,
	    okButton
	);

    }
    /**
     * Display an SQL error message <br>
     * <strong>ChangeLog</strong>: amchavan, 08-Jan-2002: added
     * 
     * @param e
     *            The SQLException object to display: the message text is broken
     *            up to fit in a reasonably small panel.
     * 
     * @param w
     *            The component in front of which the error panel is displayed;
     *            mey be <tt>null</tt>, in which case the panel is displayed
     *            in the center of the screen.
     * 
     * @see org.eso.ohs.core.utilities.TextUtils#breakString(String,int)
     */
    public static void sqlError(SQLException e, Component w) {
        JOptionPane
                .showMessageDialog(w,
                        TextUtils.breakString(e.getMessage(), 40), "SQL error "
                                + e.getErrorCode(), JOptionPane.ERROR_MESSAGE);
    }

    /**
     * Display an SQL warning message
     * 
     * @param The
     *            SQLWarning object to display: the message text is broken up to
     *            fit in a reasonably small panel.
     * 
     * @param The
     *            component in front of which the error panel is displayed; mey
     *            be <tt>null</tt>, in which case the panel is displayed in
     *            the center of the screen.
     * 
     * @see org.eso.ohs.core.utilities.TextUtils#breakString(String,int)
     * 
     * <br>
     * ChangeLog: amchavan, 08-Jan-2002: added
     */
    public static void sqlWarning(SQLWarning e, Component w) {
        JOptionPane.showMessageDialog(w, TextUtils.breakString(e.getMessage(),
                40), "SQL warning " + e.getErrorCode(),
                JOptionPane.WARNING_MESSAGE);
    }

}
