/* @(#) ImageCanvas2D.java 1.1 99/01/11
 * Copyright (c) 1999 Lawrence Rodrigues
 */
package org.eso.ohs.core.gui.widgets.imageviewer;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.print.PrinterJob;

import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.MediaSizeName;
import javax.swing.JFrame;

/** Displays a manipulated image.
 * This class doesn't perform image manipulation, but has
 * several properties to hold the manipulation attributes.
 * To manipulate an image, check out the AffineTransform attribute, concatenate it,
 * and then check it in. Call the applyTransform() method to apply the resultinng transformation
 * to the current image.
 * @version 1.0  18 Nov 1999
 * @author Lawrence Rodrigues
 **/
public class ImageCanvas2D extends ImageCanvas implements ImageManipulator {

    protected AffineTransform atx = new AffineTransform();
    protected AffineTransform dispModeAtx = new AffineTransform();
    protected AffineTransform flipAtx = new AffineTransform();

    //Pan variables
    protected Point panOffset = new Point(0, 0);
    //zoom variables
    protected boolean magOn = true;
    protected double magFactor = 1.0;
    protected int magCenterX = 0;
    protected int magCenterY = 0;
    protected Point zoomOffset = new Point(0, 0);
    //Rotation variables
    protected double rotationAngle = 0.0;
    protected boolean rotateOn = true;
    protected int rotationCenterX = 0;
    protected int rotationCenterY = 0;
    //Shear variables
    protected boolean shearOn = true;
    protected double shearFactor = 0.0;
    protected double shearX = 0.0, shearY = 0.0;
    //Off screen image width and height
    private int bufImageWidth = -1;
    private int bufImageHeight = -1;
    protected int interpolationType = AffineTransformOp.TYPE_NEAREST_NEIGHBOR;

    public ImageCanvas2D() {
        init();
    }

    protected void init() {
        setDoubleBuffered(false);
        setBackground(Color.black);
        setForeground(Color.white);
        this.setSize(200, 150);
    }

    public void setMagFactor(double magFactor) {
        firePropertyChange(
            "MagFactor",
            new Double(this.magFactor),
            new Double(magFactor));
        this.magFactor = magFactor;
    }
    public double getMagFactor() {
        return magFactor;
    }
    public void setShearFactor(double shearFactor) {
        firePropertyChange(
            "ShearFactor",
            new Double(this.shearFactor),
            new Double(shearFactor));
        this.shearFactor = shearFactor;
    }

    public double getShearFactor() {
        return shearFactor;
    }
    public double getShearFactorX() {
        return shearX;
    }
    public double getShearFactorY() {
        return shearY;
    }
    
    


    public void setRotationAngle(double rotationAngle) {
        firePropertyChange(
            "RotationAngle",
            new Double(this.rotationAngle),
            new Double(rotationAngle));
        this.rotationAngle = rotationAngle;
    }
    public double getRotationAngle() {
        return rotationAngle;
    }

    public void setPanOffset(Point panOffset) {
        firePropertyChange("PanOffset", this.panOffset, panOffset);
        this.panOffset = panOffset;
    }

    public Point getPanOffset() {
        return panOffset;
    }

    public void setInterpolationType(int interType) {
        interpolationType = interType;
    }
    public int getInterpolationType() {
        return interpolationType;
    }

    public AffineTransform getTransform() {
        return atx;
    }
    public void setTransform(AffineTransform at) {
        atx = at;
    }

    public void setMagOn(boolean onOff) {
        magOn = onOff;
    }
    public boolean getMagOn() {
        return magOn;
    }

    /** Applies the current display mode and flip mode settings to the displayed image.
       * The display mode and flip mode settings are applied by
       * using the affine transformation.
       * First computes an affine transformation for current display mode setting.
       * It then creates a BufferedImage object size of which is compted using this transformation,
       * and draws the current image on this BufferedImage.
       *
       * This method then calls the createFlipTransform() method to create an affine transformtion
       * for the current flip mode setting. It then applies the flips transforamtion to the BufferedImage
       * object, and then displays it.
       * @param imWid the image width.
       * @param imHt the image height.
       * @param if successful returns true.
       */
    public synchronized boolean doDisplayModeAndFlip(
        int imageWidth,
        int imageHeight) {
        width = this.getBounds().width;
        height = this.getBounds().height;
        double magX = (double) width / (double) imageWidth;
        double magY = (double) height / (double) imageHeight;
        int bufferWid = width, bufferHt = height;
        dispModeAtx = new AffineTransform();
        switch (displayMode) {
            case DisplayMode.ORIG_SIZE :
                bufferWid = imageWidth;
                bufferHt = imageHeight;
                break;
            case DisplayMode.SCALED :
                double mag = (magY > magX) ? magX : magY;
                dispModeAtx.setToScale(mag, mag);
                bufferWid = (int) (imageWidth * mag);
                bufferHt = (int) (imageHeight * mag);
                break;
            case DisplayMode.TO_FIT :
                dispModeAtx.setToScale(magX, magY);
                bufferWid = width;
                bufferHt = height;
            default :
                break;
        }
        //BufferedImage bi;
        if(bufferWid < 0 || bufferHt < 0) {
             //bigc.drawImage(awtImage, dispModeAtx, this);
             return false;         
        }
        BufferedImage bi = new BufferedImage(bufferWid, bufferHt, imageType);
        Graphics2D bigc = bi.createGraphics();
        if (originalImageType == TYPE_AWT_IMAGE)
            bigc.drawImage(awtImage, dispModeAtx, this);
        else
            bigc.drawImage(bufferedImage, dispModeAtx, this);

        flipAtx = createFlipTransform(flipMode, bufferWid, bufferHt);
        AffineTransformOp atop =
            new AffineTransformOp(flipAtx, interpolationType);
        offScrImage = atop.filter(bi, null);
        offScrGc = offScrImage.createGraphics();
        applyTransform(offScrImage, atx);
        repaint();
        return true;
    }

    /** Creates a flip transform. It first creates a reflection and then translates to
      * the current quadrant.
      * @param mode the specified flip mode.
      * @param imageWid the width of the BufferedImage.
      * @param imageHt  the height of the BufferedImage.
      * @return the flip transformation encapsulated in the AffineTransform object.
     */
    static public AffineTransform createFlipTransform(
        int mode,
        int imageWid,
        int imageHt) {
        AffineTransform at = new AffineTransform();
        switch (mode) {
            case FlipMode.NORMAL :
                break;
            case FlipMode.TOP_BOTTOM :
                at = new AffineTransform(new double[] { 1.0, 0.0, 0.0, -1.0 });
                at.translate(0.0, -imageHt);
                break;
            case FlipMode.LEFT_RIGHT :
                at = new AffineTransform(new double[] { -1.0, 0.0, 0.0, 1.0 });
                at.translate(-imageWid, 0.0);
                break;            
            case FlipMode.TOP_BOTTOM_LEFT_RIGHT :
                at = new AffineTransform(new double[] { -1.0, 0.0, 0.0, -1.0 });
                at.translate(-imageWid, -imageHt);
                break;
            default :
        }
        return at;
    }

    /** Applies a specified transformation to an input image, and then displays it on the canvas
      * @param bi the buffered image to which the transformation is applied.
      * @param atx the transformation to be applied.
      */
    protected void applyTransform(BufferedImage bi, AffineTransform atx) {
        if (offScrImage == null)
            return;
        if (displayImage == null)
            createDisplayImage();
        AffineTransformOp atop = new AffineTransformOp(atx, interpolationType);
        dispGc.setColor(Color.black);
        dispGc.fillRect(
            0,
            0,
            displayImage.getWidth(),
            displayImage.getHeight());
        dispGc.setClip(clipShape);
        if (clipShape == null) {
            atop.filter(bi, displayImage);
        } else {
            dispGc.drawImage(bi, atx, this);
        }            
    }

    /** Applies a specified transformation to the current off screen image,
      * and then displays it on the canvas.
      * @param atx the transformation to be applied.
      */
    public void applyTransform(AffineTransform atx) {
        applyTransform(offScrImage, atx);
        this.atx = atx;
        repaint();
    }
    
    
    public AffineTransform getAffindTransform() {
        return atx;
    }

    public void reset() {
        panOffset = new Point(0, 0);
        magCenterX = 0;
        magCenterY = 0;
        magFactor = 1.0;
        rotationAngle = 0.0;
        shearX = 0.0;
        shearY = 0.0;
        atx = new AffineTransform();
        paintImage();
    }

    /** Resets the manipulation attributes.
      * Applies the default attributes to display the current image.
      */
    public void resetManipulation() {
        panOffset = new Point(0, 0);
        magCenterX = 0;
        magCenterY = 0;
        magFactor = 1.0;
        shearX = 0.0;
        shearY = 0.0;
        rotationAngle = 0.0;
        atx = new AffineTransform();
        paintImage();
        repaint();
    }
    
   
    
    public void printContents() {
        PrintRequestAttributeSet attributes = new HashPrintRequestAttributeSet();
        attributes.add(MediaSizeName.ISO_A4);
       
        PrinterJob pj = PrinterJob.getPrinterJob();
        pj.setPrintable(this);        
        if (!pj.printDialog(attributes))
            return;
        try{
           pj.print();
        }catch(Exception e){System.out.println(e);}
    }

    
    
    public static void main(String[] args) {
        ImageCanvas2D iv = new ImageCanvas2D();
        JFrame fr = new JFrame();
        fr.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.exit(0);
            }
        });
        fr.setTitle("ImageCanvas2D Bean");
        fr.getContentPane().add(iv);
        fr.pack();
        fr.setSize(400, 300);
        fr.show();
        iv.paintComponent(iv.getGraphics());
        fr.repaint();
    }
    
}
