package org.eso.ohs.core.utilities;

import java.util.ArrayList;
/**
 * This class handle censorship over strings.
 * It holds a Forbidden Index of dangerous words and check a string (SqlStatement) for every words in the Forbidden Index.
 * The class is not case sensitive.
 * @author gdonino
 *
 */
// Congregatio Indicis Sql Prohibitorum
 
public class CISP {

	/**
	 * List of forbidden words
	 */
	private ArrayList<String> forbiddenIndex = null;
	
	/**
	 * CISP constructor
	 * 
	 * @param addDefault True if you want to add dafault forbidden words to the forbidden index (drop, delete, update)
	 */
	public CISP(boolean addDefault){
		initialise(addDefault);
		
	}
	/**
	 * Cisp Constructor
	 * 
	 * @param addDefault True if you want to add dafault forbidden words to the Forbidden Index (drop, delete, update)
	 * @param fIndex An ArrayList of words to be added to the Forbidden Index
	 */
	public CISP(boolean addDefault, ArrayList<String> fIndex){
		/* Design by Contract */
		if (fIndex == null) 
			throw new IllegalArgumentException(MsgManager.errPrecondition("Forbidden Index is null"));
		if (fIndex.size() == 0)
			throw new IllegalArgumentException(MsgManager.errPrecondition("Forbidden Index is empty"));
		/* DbC */
		initialise(addDefault);
		forbiddenIndex.addAll(fIndex);
	}
	
	private void initialise(boolean addDefault){
		if (forbiddenIndex == null)
			forbiddenIndex = new ArrayList<String>();
		if (addDefault)
			addDefaultItems();
	}
	/**
	 * Add deafult items to the Forbidden Index (drop, delete, update)
	 *
	 */
	private void addDefaultItems(){
		forbiddenIndex.add("drop");
		forbiddenIndex.add("delete");
		forbiddenIndex.add("update");
	}
	/**
	 * Add one item to the Forbidden Index
	 * @param item A String
	 */
	public void addItem(String item){
		/* Design by Contract */
		if (item == null) 
			throw new IllegalArgumentException(MsgManager.errPrecondition("item is null"));
		if (item.length() == 0)
			throw new IllegalArgumentException(MsgManager.errPrecondition("item is empty"));
		/* DbC */
		forbiddenIndex.add(item);		
	}
	/**
	 * Add a collection of "dangerous words" to the Forbidden Index
	 * @param items An ArrayList
	 */
	public void addItems(ArrayList<String> items){
		/* Design by Contract */
		if (items == null) 
			throw new IllegalArgumentException(MsgManager.errPrecondition("items collection is null"));
		if (items.size() == 0)
			throw new IllegalArgumentException(MsgManager.errPrecondition("items collection is empty"));
		for (int i = 0; i < items.size(); i++){
			if (items.get(i).length() == 0)
				throw new IllegalArgumentException(MsgManager.errPrecondition("one ore more items in the collection are empty"));	
		}
		/* DbC */
		forbiddenIndex.addAll(items);		
	}
	/**
	 * The method check the sql statement for every forbidden words in the forbidden index
	 * This method can only return "false" if the sql statement doesn't contain any forbidden word else it will raise an IllegalArgumentException
	 * @param originalSql The sql statement we want to check for
	 * @return false is the sql statement does not contain forbidden words
	 * @throws IllegalArgumentException if the sql is "dangerous"
	 */
	public boolean censor(String originalSql) throws IllegalArgumentException{
		/* Design by Contract */
		if (originalSql==null) 
           throw new IllegalArgumentException(MsgManager.errPrecondition("sql statement is null"));
		if (originalSql.length() ==0)
			throw new IllegalArgumentException(MsgManager.errPrecondition("sql statement is empty"));
		if (forbiddenIndex==null) 
			throw new IllegalArgumentException(MsgManager.errPrecondition("Indicis Prohibitorum is null"));
		if (forbiddenIndex.size() == 0)
			throw new IllegalArgumentException(MsgManager.errPrecondition("Indicis Prohibitorum is empty"));
		/* DbC */
		boolean result = true;
		originalSql = originalSql.toLowerCase();
		String sql = new String(originalSql);
		for (int i = 0; i < forbiddenIndex.size(); i++)
			sql = sql.replaceAll(forbiddenIndex.get(i).toString().toLowerCase(), "->" + forbiddenIndex.get(i).toString().toLowerCase() + "<-");
		
		if (sql.compareTo(originalSql)== 0)
			result = false;
		else
			throw new IllegalArgumentException(MsgManager.errCISP(sql));
		
		
		return result;
	}
	
	public static void main(String[] args) {
		CISP c = new CISP(true);
		ArrayList<String> sqlFileToCheck = new ArrayList<String>();
		sqlFileToCheck.add("dROp the whole update word and smile");
		sqlFileToCheck.add("select whatever from tableTwo where Name = 'drop'");
		sqlFileToCheck.add("Select Mail for update where Name like 'Helmut'");
		sqlFileToCheck.add("deLEte aDwarf from Garden and then drop the dwarf and fly");
		sqlFileToCheck.add("select mySelf from table3");
		for (int i = 0; i < sqlFileToCheck.size(); i++){
			try{
				boolean r = c.censor(sqlFileToCheck.get(i).toString());
			}catch(IllegalArgumentException e){
				e.printStackTrace();
				System.exit(0);
			}
		}
	}
	/**
	 * Retrieve the Forbidden Index
	 * @return the Forbidden Index
	 */
	public ArrayList<String> getForbiddenIndex() {
		return forbiddenIndex;
	}
	/**
	 * Set a new Forbidden Index
	 * @param fIndex a Forbidden Index
	 */
	public void setForbiddenIndex(ArrayList<String> fIndex) {
		/* Design by Contract */
		if (fIndex==null) 
			throw new IllegalArgumentException(MsgManager.errPrecondition("Indicis Prohibitorum is null"));
		if (fIndex.size() == 0)
			throw new IllegalArgumentException(MsgManager.errPrecondition("Indicis Prohibitorum is empty"));
		/* DbC */
		this.forbiddenIndex = fIndex;
	}
	

}
