package org.eso.ohs.core.utilities;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.StringTokenizer;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;



/**
 * Utility methods to convert data between formats, etc.
 * <p>
 * <strong>Note</strong>: throughout this class right ascension values are
 *                        expressed in hours, 0:24, while declination
 *                        values are expressed in degrees, -90:+90.
 *                        <br>
 *                        Right ascension values are converted to and 
 *                        from milliseconds, and declination values 
 *                        to and from milliarcseconds.
 *
 * @author A M Chavan, 06-Jun-2001
 * @author A M Chavan, revised, 12-Jan-2006
 * @version $Revision: 1.34.4.1 $
 */

// $Id: Convert.java,v 1.34.4.1 2007/01/11 14:00:35 pnunes Exp $

public class Convert {

    private static final int FIRST_DAY_MONTH = 1;

    // list of months 
    private static final String[] MONTHS_LIST =
    { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct",
            "Nov", "Dec" };

    // the month integer list
    private static int[] MONTHS_LIST_INT =
    { Calendar.JANUARY, Calendar.FEBRUARY, Calendar.MARCH, Calendar.APRIL,
            Calendar.MAY, Calendar.JUNE, Calendar.JULY, Calendar.AUGUST,
            Calendar.SEPTEMBER, Calendar.OCTOBER, Calendar.NOVEMBER,
            Calendar.DECEMBER };

  
    
    
//    private static Logger stdlog_ = Logger.getLogger(Convert.class);

    public final static String ST_TIME_FORMAT = "HH:mm";

    public final static int SEC_PER_MINUTE = 60;

    public final static int SEC_PER_HOUR = 3600;

    public final static int SEC_PER_DAY = 3600 * 24;

    /** Milliseconds per second */
    public final static int MILLISECS_PER_SEC = 1000;

    /** Milliseconds per minute */
    public final static int MILLISECS_PER_MINUTE = 60 * MILLISECS_PER_SEC;

    /** Milliseconds per unit (hour or degree) */
    public final static int MILLISECS_PER_UNIT = 60 * MILLISECS_PER_MINUTE;

    /** Number of decimal places of accuracy */
    public final static int DECIMAL_PLACES_ACCURACY = 3;

    /** Max milliseconds for RA -- corresponds to 23:59:59.999 */
    public final static int MAX_MILLISECS_RA = 86400000;
  
    /** 
     * Difference in seconds between the Java (Unix) epoch,
     * January 1, 1970, 00:00:00 GMT and the STARCAT epoch,
     * exactly ten years after that: 1980-01-01T00:00:00.000 UT
     */
    static final long STARCAT_TO_UNIX_OFFSET = 315532800;
    
    /** Max milliarcseconds for DEC -- corresponds to 90:00:00.000 */
    // amchavan, 10-Jan-2006: renamed from MAX_MILLISECS_DEC
    public final static int MAX_MILLIARCSECS_DEC = 324000000;

    /** Max execution time 23:59:59.999 */
    public final static int EXECTIME_MAX = 86399;
    public final static int EXECTIME_ERR = -86399;
    public final static String EXECTIME_ERR_STRING = "-23:59:59.000";

    /** Greatest date value "2038-01-19T02:14:07" allowed */
    public final static long TIME_MAX = 2147483648L;

    /**
     * Multiplicative conversion factor, from (milli)sec to 
     * (milli)arcsec. Divide by this number to convert (milli)arcsec
     * back to (milli)sec.
     */
    // amchavan, 30-Jan-2006: created
    public final static int SECTOARCSEC = 15;

    public final static long ENDTIME_MAX = 2147483647L;

    /** Standard format for ISO date/time strings */
    // amchavan, 03-Jun-2002: added
    public final static String ISO_DATE_FORMAT = "yyyy-MM-dd'T'HH:mm:ss";
    
    /** Short format for  date strings 
     */ 
    private static final String D_MMM_YY_DATE_FORMAT = "d MMM yy";

    /** 
     * SimpleDateFormat for ISO date/time strings
     * @see #ISO_DATE_FORMAT
     */
    public final static SimpleDateFormat ISO_SIMPLE_DATE_FORMAT = 
        new SimpleDateFormat( ISO_DATE_FORMAT );
    
    /** 
     * SimpleDateFormat for date/time strings
     * @see #DateFormat.MEDIUM
     */
    private final static  DateFormat SIMPLE_DATE_FORMAT = 
        DateFormat.getDateInstance(DateFormat.MEDIUM);
    
    /** 
     * ShortDateFormat for date/time strings
     * @see #D_MMM_YY_DATE_FORMAT
     */
    public final static SimpleDateFormat SHORT_DATE_FORMAT = 
        new SimpleDateFormat(D_MMM_YY_DATE_FORMAT);
    
  
    static {
        ISO_SIMPLE_DATE_FORMAT.setTimeZone(TimeZone.getTimeZone("GMT"));
        SIMPLE_DATE_FORMAT.setTimeZone(TimeZone.getTimeZone("GMT"));
        SHORT_DATE_FORMAT.setTimeZone(TimeZone.getTimeZone("GMT"));
    }
    
 
    /** Standard format for db date/time strings */
    public final static String DB_DATE_FORMAT = "MM-dd-yyyy";

    /** Standard format for duration strings */
    public static final String DURATION_FORMAT = "d'T'HH:mm:ss";

    /** Standard format for duration strings, HH:mm:ss */
    public static final String DURATION_HOURS = "HH:mm:ss";

    /** Standard format for duration strings, days */
    public static final String DURATION_DAYS = "d";

    // amchavan, 09-Jan-2006: used by round()
    private static final double[] bases = { 1e0, 1e1, 1e2, 1e3, 1e4, 1e5, 1e6,
                                            1e7, 1e8, 1e9, 1e10 };

    /**
     * Coordinate regular expression: an optional plus or minus sign, 
     * followed by single real number, like 
     * "-12.345", expressing degrees or hours.
     * @see #initializeCoordinateRegexps()
     */
    // amchavan, 19-Jan-2006: added
    private static Pattern CRE_ONE = null;
    
    /**
     * Coordinate regular expression: an optional plus or minus sign, 
     * followed by an integer, followed by
     * a single real number, like 
     * "+12 34.567", expressing degrees/hours and minutes.
     * Components may be separated by a blank or a colon char.
     * @see #initializeCoordinateRegexps()
     */
    // amchavan, 19-Jan-2006: added
    private static Pattern CRE_TWO = null;
    
    /**
     * Coordinate regular expression: an optional plus or minus sign, 
     * followed by an integer, followed by
     * a second integer followed by a single real number, like 
     * "12 34 56.789", expressing degrees/hours and minutes.
     * Components may be separated by a blank or a colon char.
     * @see #initializeCoordinateRegexps()
     */
    // amchavan, 19-Jan-2006: added
    private static Pattern CRE_THREE = null;
    
    /**
     * Coordinate regular expression: an optional plus or minus sign, 
     * followed by exactly six digits, followed by a fractional part of
     * 1-3 digits, like "123456.789", expressing degrees/hours and minutes.
     * @see #initializeCoordinateRegexps()
     */
    // amchavan, 19-Jan-2006: added
    private static Pattern CRE_NOSEPARATORS = null;
   
    

    /** Return months of the year */
    public static String[] getMonths() {
        String[] monthList = (String[]) MONTHS_LIST.clone();
        return monthList;
    }
    
    
    /** Return days in a month of the year
     * 
     * @param month the calendar int value representing a month: e.g. Calendar.APRIL
     * @param year the year in long syntax: e.g. 2004
     * @return int number of days in that month
     */
    public static int getDays(int month, int year) {
        // Create a calendar object of the desired month
        Calendar cal = new GregorianCalendar(year, month, FIRST_DAY_MONTH);
       // Get the number of days in that month
        int days = cal.getActualMaximum(Calendar.DAY_OF_MONTH); 
        return days;
    }
   
    /**
     * 
     * Return the calender month from its sequence number from Calendar.
     * Eg. 4 --> May
     * if value not found, is returned an empty string 
     * 
     * @param calendarMonth Calendar static variable representing the number: e.g Calendar.DECEMBER
     * @return String month the month in three char format. e.g. "Jan"
     */
    public static String getMonth(int calendarMonth) {
        String month = "";
        if (calendarMonth < MONTHS_LIST_INT.length) {
            for (int i = 0; i < MONTHS_LIST_INT.length; i++) {
                if (calendarMonth == MONTHS_LIST_INT[i]) {
                    month = MONTHS_LIST[i];
                    break;
                }
            }
        }
        return month;
    }


    /**
     * Validates a date format for format "day month year" As a separator,
     * "/:,-". 
     *   Day is the number of days from the first of the month (e.g. 100 is valid).
     *   Month 3 letters case insensitive. 
     *   Year full
     * 
     * @param date
     * @return boolean
     */
    public static boolean validateDateFormatDMY(String date) {
        boolean result = true;
        try {
            Convert.getDateFormatDMY(date);
        } catch (IllegalArgumentException e) {
            result = false;
        }
        return result;
    }

    
    /**
     * The integer that correspond to a month string.
     * if value not found, the int value correspond to Jenuary
     * @param month Month in this syntax "Jan"
     * @return int index of month
     */
    public static int monthToCalendarMonth(String month) {
        int calendarMonth = Calendar.JANUARY;
        for (int i = 0; i < MONTHS_LIST.length; i++) {
            if (MONTHS_LIST[i].equals(month)) {
                calendarMonth = MONTHS_LIST_INT[i];
                break;
            }
        }
        return calendarMonth;
    }
    
    
    /**
     * 
     * Return the list of months in lower case.
     * 
     * @return String lowercase list of months (three letters)
     */
    public static String[] getLowerCaseMonths() {
        String[] months = getMonths();
        String[] lowerCaseMonths = new String[months.length];
        for (int i = 0; i < lowerCaseMonths.length; i++) {
            lowerCaseMonths[i] = months[i].toLowerCase();
        }
        return lowerCaseMonths;
    }
    
    
    /**  
     * current date in number of seconds from the epoch January 1, 1970 UTC.
     * @return Number of seconds since the Unix epoch.
     */
    // amchavan, 10-Jan-2006: OK
    public static long getCurrentTime() {
        return System.currentTimeMillis() / Convert.MILLISECS_PER_SEC;
    }

    
    
    /** 
    * Convert hours and minutes string to seconds.
    * @param hh_mm A string like <code>HH:MM</code>.
    * @throws NumberFormatException
    */
    // amchavan, 09-Jan-2005: some refactoring, unit tested, added minutes check
    public static int convertHHMMTOSeconds( String hh_mm )
            throws NumberFormatException {
        int hours, minutes, seconds;
        String[] array = hh_mm.split( ":" );
        final String errorMessage = 
            "Invalid format: should be HH:MM, not '" + hh_mm + "'";
        if( array == null || array.length != 2 )
            throw new NumberFormatException( errorMessage );
        if( array[0].length() > 2 || array[1].length() > 2 )
            throw new NumberFormatException( errorMessage );
        
        try {
            hours = Integer.valueOf( array[0] ).intValue();
            minutes = Integer.valueOf( array[1] ).intValue();
        }
        catch( NumberFormatException e ) {
            throw new NumberFormatException( errorMessage );
        }
        catch( IndexOutOfBoundsException e ) {
            throw new NumberFormatException( errorMessage );
        }

        if( minutes > 59 ) {
            String msg = "Invalid format: MM should be <= 60";
            throw new NumberFormatException( msg );
        }
        
        seconds = hours * SEC_PER_HOUR + minutes * SEC_PER_MINUTE;
        return seconds;
    }

    /**
     * Converts a string representing a time expression to
     * milliseconds.
     *
     * @param time  String to be converted, represented 
     *              as <code>HH:MM:SS.sss</code>,
     *                 <code>HH&nbsp;MM&nbsp;SS.sss</code> or 
     *                 <code>HHMMSS.sss</code>, 
     *              where: HH is hours (0-23) and can be abbreviated to
     *                     a single H; 
     *                     MM is minutes (0-59); 
     *                     SS is seconds (0-59); 
     *                     sss is milliseconds and may have 0-3 digits
     *                     (that is, it's optional). <P>
     *        
     *        An optional + sign can be prepended. <P>
     *        
     *        Partial strings are accepted as well: <code>HH.hhh</code>
     *        (with optional decimal hours), <code>HH:MM.mmm</code>
     *        and <code>HH&nbsp;MM.mmm</code> (with optional decimal minutes)
     *        are considered valid. Note however that elements within partial
     *        strings must be properly separated: <code>HHMM.mmm</code> 
     *        is <em>not</em> valid syntax.
     *
     *        Valid examples include: <p> 
     *        <code>
     *        02:33:52.1<br>
     *         2:33:52.1<br>
     *        +23 59 59.999<br>
     *        0.00<br>
     *        00 0.000<br>
     *        0 0 0.000<br>
     *        0:0:0.000<br>
     *        +1.000<br>
     *        022734.070<br>
     *        22734.07
     *        </code>
     *        
     * @return		input converted to milliseconds
     *
     * @exception  IllegalArgumentException If the input string is not
     *             properly formatted, or some component is out of range.    
     *
     */
    // @author A M Chavan, 06-Jun-2001; was originally
    //                     convertToMilliArcSeconds() in DegreeAdaptor.java
    // amchavan, 10-Jan-2006: renamed, updated Javadoc
    // amchavan, 19-Jan-2006: updated Javadoc
    public static int HHMMSSToMillisec( String time )
            throws IllegalArgumentException {
        return stringToMilli___sec( time, true, false );
    }

    /**
     * Converts a string representing a right ascension value to
     * milliarcseconds. See <code>HHMMSSToMillisec()</code> for a description
     * of the input syntax.
     * 
     * @return Input converted to milliarcseconds
     *
     * @exception  IllegalArgumentException If the input string is not
     *             properly formatted, or some component is out of range.
     *       
     * @see #HHMMSSToMillisec(String)
     *
     */
    // amchavan, 30-Jan-2006: created
    public static int HHMMSSToMilliarcsec( String time )
            throws IllegalArgumentException {
        return HHMMSSToMillisec( time ) * SECTOARCSEC;
    }

    /**
     * Converts a string representing a time expression to
     * seconds. 
     *
     * @param time  String to be converted; see HHMMSSToMillisec()
     *              for a description of valid inputs.
     *
     * @return		input converted to seconds
     *
     * @exception	IllegalArgumentException on out of bounds and
     *                                              out of format
     * 
     * @see #HHMMSSToMillisec(String)
     */
    // @author cmavros, 10-10-01
    // amchavan, 10-Jan-2006: OK
    // amchavan, 19-Jan-2006: updated Javadoc
    public static int HHMMSSToSec(String time)
        throws IllegalArgumentException {
        int newValue = stringToMilli___sec(time, false, true);
        return (newValue / MILLISECS_PER_SEC);
    }

    /**
     * Converts a string representing an angle expression to
     * milliarcseconds.
     *
     * @param time  String to be converted, represented 
     *              as <code>DD:MM:SS.sss</code>,
     *                 <code>DD&nbsp;MM&nbsp;SS.sss</code> or 
     *                 <code>DDMMSS.sss</code>, 
     *               where DD is degrees (0-90) and can be abbreviated to
     *                     a single D; 
     *                     MM is arcminutes (0-59); 
     *                     SS is arcseconds (0-59); 
     *                     sss is milliarcseconds and may have 0-3 digits
     *                     (that is, it's optional). <P>
     *        
     *        An optional + or - sign can be prepended. <P>
     *        
     *        Partial strings are accepted as well: <code>DD.ddd</code>
     *        (with optional decimal degrees), <code>DD:MM.mmm</code>
     *        and <code>DD&nbsp;MM.mmm</code> (with optional decimal arcminutes)
     *        are considered valid. Note however that elements within partial
     *        strings must be properly separated: <code>DDMM.mmm</code> 
     *        is <em>not</em> valid syntax.
     *
     *        Valid examples include:<p> 
     *        <code>
     *        -02:33:52.1<br>
     *        -2:33:52.1<br>
     *        +23 59 59.999<br>
     *        0.00<br>
     *        00 0.000<br>
     *        0 0 0.000<br>
     *        0:0:0.000<br>
     *        +1.000<br>
     *        022734.070<br>
     *        -041449.900<br>
     *        -41449.9
     *        </code>
     *
     * @return		input converted to milliarcseconds
     *
     * @exception       IllegalArgumentException on out of bounds and
     *                                              out of format
     *
     */
    // @author A M Chavan, 06-Jun-2001; was originally 
    //         convertToMilliArcSeconds() in DegreeAdaptor.java
    // amchavan, 10-Jan-2006: OK
    // amchavan, 19-Jan-2006: updated Javadoc
    public static int DDMMSSToMilliarcsec( String time )
        throws IllegalArgumentException {

        return stringToMilli___sec( time, false, false );
    }

    /**
     * Converts a String in the format +UU:MM:SS.sss or +UUMMSS.sss to
     * milli(arc)seconds, where + is an optional sign, UU is hours or
     * degrees, MM is minutes, SS is seconds and sss is millseconds.
     *
     * @param value	Value to be converted, represented as a String
     * 
     * @param hours	<code>true</code> means convert to milliseconds of hours, 
     *              <code>false</code> to milliarcseconds of degrees
     *
     * @param noCheck <code>true</code> means skip RA/Dec validation check
     *
     * @return		Input converted to a milli(arc)seconds
     *
     * @exception       IllegalArgumentException on out of bounds and
     *                                              out of format
     *
     * @author A M Chavan, 06-Jun-2001; was originally
     *         <code>convertToMilliArcSeconds()</code> in DegreeAdaptor.java
     */
    // amchavan, 10-Jan-2005: renamed; unit tested via its callers.
    private static int stringToMilli___sec( String value,
                                            boolean hours,
                                            boolean noCheck )
        throws IllegalArgumentException {
        
        double[] coordComponents = parseCoordString( value );

        int sign = (int) coordComponents[0];
        double   units = coordComponents[1];
        double minutes = coordComponents[2];
        double seconds = trunc(coordComponents[3],3);
        
        
        
        if (!noCheck) {
            if (hours)
                checkBoundsHHMMSS( sign, units, minutes, seconds );
            else
                checkBoundsDDMMSS( sign, units, minutes, seconds );
        }

        double result = sign * ( units * MILLISECS_PER_UNIT + 
                                 minutes * MILLISECS_PER_MINUTE +                                                                  
                                 seconds* MILLISECS_PER_SEC );

        return (int)Math.floor(result);
    }

    /**
     * Break a coord string down into its components.
     * 
     * @param coords  A coordinate string, see DDMMSSToMilliarcsec()
     *                and HHMMSSToMillisec() for syntax specs
     *                
     * @return An array of four doubles. <br>
     *         Index 0: -1 if the coord string starts with a "-"
     *         char; 1 if it starts with "+" or it has no sign. <br>
     *         Index 1: hours or degrees <br>
     *         Index 2: (arc)minutes <br>
     *         Index 3: (arc)seconds
     * 
     * @throws IllegalArgumentException If the input string has invalid syntax
     * 
     * @see #DDMMSSToMilliarcsec(String)
     * @see #HHMMSSToMillisec(String)
     */
    static double[] parseCoordString( String coords )
        throws IllegalArgumentException {
        
        if( CRE_ONE == null )
            initializeCoordinateRegexps();

        coords = coords.trim();     // ignore surrounding whitespace
        
        double[] ret = new double[4];
        Matcher matcher;
        int start;
        int end;
        String subgroup;

        // The followig constants are used to extract fragments
        // of the input string after it matched one of the patterns
        // The regular expressions we defined (CRE_ONE, etc.)
        // define the following (meaningful) subgroups:
        // 0: '+1 34 56.789' ==> '+1 34 56.789'
        // 1: '+1 34 56.789' ==> '+'
        // 2: '+1 34 56.789' ==> '1'
        // 3: '-41449.900' ==> '41449'
        // 4: '+1 34 56.789' ==> '34'
        // 6: '+1 34 56.789' ==> '56.789'
        //---------------------------------------------------------
        final int subgrSIGN   = 1;  // subgroup index: optional +/- sign
        final int subgrFIRST  = 2;  // subgroup index: first component
        final int subgrNOSEP  = 3;  // subgroup index: non-separated components
        final int subgrSECOND = 4;  // subgroup index: second component
        final int subgrTHIRD  = 6;  // subgroup index: third component
        
        //-------------------------------------------------------------------
        // Case 1: non-separated strings like "022734.070" and "-041449.900"
        //-------------------------------------------------------------------
        matcher = CRE_NOSEPARATORS.matcher( coords );
        if( matcher.matches() ) {
//            int i = 0;
//            while( true ) {
//                try {
//                    start = matcher.start( i );
//                    end = matcher.end( i );
//                    String t = coords.substring(start, end);
//                    System.out.println( i + ": '" + coords + "' ==> '" + t + "'" );
//                    i++;
//                }
//                catch( IndexOutOfBoundsException e ) {
//                    break;
//                }
//            }
            
            // Extract the sign
            start = matcher.start( subgrSIGN );
            end = matcher.end( subgrSIGN );
            subgroup = coords.substring( start, end );
            ret[0] = subgroup.equals( "-" ) ? -1 : +1;
            
            // We need to measure the length of the non-separated part: 
            // "41449" or "041449"
            start = matcher.start( subgrNOSEP );
            end = matcher.end( subgrNOSEP );
            subgroup = coords.substring( start, end );
            boolean isLength5 = subgroup.length() == 5; // can be 5 or 6
            
            // now consider the entire "41449.900" subgroup
            start = matcher.start( subgrFIRST );
            end = matcher.end( subgrFIRST );
            subgroup = coords.substring( start, end );
            if( isLength5 ) {
                StringBuffer sb = new StringBuffer();
                sb.append( "0" ).append( subgroup );
                subgroup = sb.toString();
            }
            
            int l = subgroup.length();
            String uu = subgroup.substring( 0, 2 );
            String mm = subgroup.substring( 2, 4 );
            String ss = subgroup.substring( 4, l );
            ret[1] = Integer.valueOf( uu ).doubleValue();
            ret[2] = Integer.valueOf( mm ).doubleValue();
            ret[3] = Double.valueOf(  ss ).doubleValue();
            
            return ret;
        }
        
        //-------------------------------------------------------------------
        // Case 2: just one number, like "12.345" or "-12.345"
        //-------------------------------------------------------------------
        matcher = CRE_ONE.matcher( coords );
        if( matcher.matches() ) {
//            int i = 0;
//            while( true ) {
//                try {
//                    start = matcher.start( i );
//                    end = matcher.end( i );
//                    String t = coords.substring(start, end);
//                    System.out.println( i + ": '" + coords + "' ==> '" + t + "'" );
//                    i++;
//                }
//                catch( IndexOutOfBoundsException e ) {
//                    break;
//                }
//            }
            
            // Extract the sign
            start = matcher.start( subgrSIGN );
            end = matcher.end( subgrSIGN );
            subgroup = coords.substring( start, end );
            ret[0] = subgroup.equals( "-" ) ? -1 : +1;
            
            // First subgroup is the real number representing hours/degrees
            start = matcher.start( subgrFIRST );
            end = matcher.end( subgrFIRST );
            subgroup = coords.substring( start, end );
            ret[1] = Double.valueOf( subgroup ).doubleValue();
            
            return ret;
        }
        
        //-------------------------------------------------------------------
        // Case 3: Two components, like "12 34.5" or "-12:34.567", separated
        //         by spaces or a colon char
        //-------------------------------------------------------------------
        matcher = CRE_TWO.matcher( coords );
        if( matcher.matches() ) {
//            int i = 0;
//            while( true ) {
//                try {
//                    start = matcher.start( i );
//                    end = matcher.end( i );
//                    t = coords.substring(start, end);
//                    System.out.println( i + ": '" + coords + "' ==> '" + t + "'" );
//                    i++;
//                }
//                catch( IndexOutOfBoundsException e ) {
//                    break;
//                }
//            }
            
            // Extract the sign
            start = matcher.start( subgrSIGN );
            end = matcher.end( subgrSIGN );
            subgroup = coords.substring( start, end );
            ret[0] = subgroup.equals( "-" ) ? -1 : +1;
            
            // First subgroup is the integer representing hours/degrees
            start = matcher.start( subgrFIRST );
            end = matcher.end( subgrFIRST );
            subgroup = coords.substring( start, end );
            ret[1] = Integer.valueOf( subgroup ).doubleValue();
            
            // Second subgroup is the real number representing minutes
            start = matcher.start( subgrSECOND );
            end = matcher.end( subgrSECOND );
            subgroup = coords.substring( start, end );
            ret[2] = Double.valueOf( subgroup ).doubleValue();
            
            return ret;
        }

        
        //-------------------------------------------------------------------
        // Case 4: Three components, like "12 34 5.67" or "-65:32:5.190",
        //         separated by spaces or a colon char
        //-------------------------------------------------------------------
        matcher = CRE_THREE.matcher( coords );
        if( matcher.matches() ) {
//            int i = 0;
//            while( true ) {
//                try {
//                    start = matcher.start( i );
//                    if( start == -1 )
//                        break;
//                    end = matcher.end( i );
//                    if( end == -1 )
//                        break;
//                    t = coords.substring(start, end);
//                    System.out.println( i + ": '" + coords + "' ==> '" + t + "'" );
//                    i++;
//                }
//                catch( IndexOutOfBoundsException e ) {
//                    break;
//                }
//            }
            
            // Extract the sign
            start = matcher.start( subgrSIGN );
            end = matcher.end( subgrSIGN );
            subgroup = coords.substring( start, end );
            ret[0] = subgroup.equals( "-" ) ? -1 : +1;
            
            // First subgroup is the integer representing hours/degrees
            start = matcher.start( subgrFIRST );
            end = matcher.end( subgrFIRST );
            subgroup = coords.substring( start, end );
            ret[1] = Integer.valueOf( subgroup ).doubleValue();
            
            // Second subgroup is the integer representing minutes
            start = matcher.start( subgrSECOND );
            end = matcher.end( subgrSECOND );
            subgroup = coords.substring( start, end );
            ret[2] = Integer.valueOf( subgroup ).doubleValue();
            
            // Third subgroup is the real number representing seconds
            start = matcher.start( subgrTHIRD );
            end = matcher.end( subgrTHIRD );
            subgroup = coords.substring( start, end );
            ret[3] = Double.valueOf( subgroup ).doubleValue();
            
            return ret;
        }

        //---------------------------------------------------------
        // If all else fails, throw your arms in the air and scream
        //---------------------------------------------------------
        throw new IllegalArgumentException( 
                "Invalid coordinate syntax \"" + coords + "\"" );
    }

    /**
     * Converts milliseconds to a String in the format
     * "<code>&nbsp;HH:MM:SS.sss</code>" (note the leading space),
     * where HH is hours, MM is minutes, SS is seconds and sss is
     * milliseconds.
     *
     * @param value	value to be converted, in milliseconds
     * @throws       IllegalArgumentException if input value is out of range
     *
     * @return		Input converted to a String
     *
     */
    // @author A M Chavan, 06-Jun-2001
    // amchavan, 10-Jan-2006: renamed, added unit tests
    public static String millisecToHHMMSS( long value )
            throws IllegalArgumentException {
        return milli___secToString( value, true, false ,true);
    }

    /**
     * Converts milliarcseconds to a String in the format
     * "<code>&nbsp;HH:MM:SS.sss</code>" (note the leading space),
     * where HH is hours, MM is minutes, SS is seconds and sss is
     * milliseconds.
     *
     * @param value	Value to be converted, in milliarcseconds
     * @throws       IllegalArgumentException if input value is out of range
     *
     * @return		Input converted to a String
     *
     */
    // amchavan, 30-Jan-2006: created
    public static String milliarcsecToHHMMSS( long value )
            throws IllegalArgumentException {
        String ret;
        
        try {
            ret = millisecToHHMMSS( value / SECTOARCSEC );
        }
        catch( IllegalArgumentException e ) {
            throw new IllegalArgumentException(
                    "Invalid right ascension value: " + value );
        }
        
        return ret;
    }
    
    
    /**
     * Converts milliseconds to a String in the format
     * "<code>&nbsp;HHMMSS.sss</code>" (note the leading space),
     * where HH is hours, MM is minutes, SS is seconds and sss is
     * milliseconds.
     *
     * @param value	Value to be converted, in milliarcseconds
     * @throws       IllegalArgumentException if input value is out of range
     *
     * @return		Input converted to a String
     *
     */
    public static String millisecToHHMMSSNoColon( long value )
            throws IllegalArgumentException {    	    	
    	 return milli___secToString( value, true, false ,false);        
    }

    /**
     * Converts seconds to a String in the format
     * "<code>&nbsp;HH:MM:SS.sss</code>" (note the leading space),
     * where HH is hours, MM is minutes, SS is seconds and sss is
     * milliseconds.
     *
     * @param value	value to be converted, in seconds
     * @exception	IllegalArgumentException if value is negative
     *
     * @return		Input value converted to a String
     *
     * @see #millisecToHHMMSS(long)
     */
    // @author cmavros, 10-10-01
    // amchavan, 10-Jan-2006: OK
    public static String secToHHMMSS( int value )
        throws IllegalArgumentException {
        long millisecs = (long) value * (long) MILLISECS_PER_SEC;
        return milli___secToString(millisecs, false, true,true);
    }

    /**
     * Converts milliarcseconds to a String in the format
     * "<code>&nbsp;DD:MM:SS.sss</code>" (note the leading space) ,
     * where DD is hours, MM is minutes, SS is seconds and sss is
     * millseconds.
     *
     * @param value	value to be converted, in milliarcseconds
     *
     * @return		input converted to a String
     * @throws       IllegalArgumentException
     *
     */
    // @author A M Chavan, 06-Jun-2001
    // amchavan, 10-Jan-2005: added unit tests
    public static String milliarcsecToDDMMSS( long value )
            throws IllegalArgumentException {
        return milli___secToString( value, false, false,true );
    }
    
    
    /**
     * Converts milliarcseconds to a String in the format
     * "<code>&nbsp;DDMMSS.sss</code>"  (note the leading space),
     * where DD is hours, MM is minutes, SS is seconds and sss is
     * millseconds.
     *
     * @param value	value to be converted, in milliarcseconds
     *
     * @return		input converted to a String
     * @throws       IllegalArgumentException
     *
     */
    // @author A M Chavan, 06-Jun-2001
    // amchavan, 10-Jan-2005: added unit tests
    public static String milliarcsecToDDMMSSNoColon( long value )
            throws IllegalArgumentException {
        return milli___secToString( value, false, false,false);
    }
    

    /**
     * Converts milli(arc)seconds to a String in the format
     * +UU:MM:SS.sss, where + is an optional sign or space, UU is
     * hours or degrees, MM is minutes, SS is seconds and sss is
     * millseconds.
     *
     * @param value	  Value to be converted, in milliarcseconds
     * 
     * @param hours	  <code>true</code> means convert to HH:MM:SS, 
     *                <code>false</code> to DD:MM:SS
     *
     * @param noCheck <code>true</code> means skip RA/Dec validation check
     *
     * @return		Input value converted to a String
     *
     * @author A M Chavan, 06-Jun-2001; was originally
     *         convertToHMS() in DegreeAdaptor.java
     */
    // amchavan, 10-Jan-2006: renamed; fixed Javadoc;
    //                        no reason to unit test this; it is tested via
    //                        milliarcsecToDDMMSS() and millisecToHHMMSS()
    private static String milli___secToString( long value, 
                                               boolean hours, 
                                               boolean noCheck, 
                                               boolean useColons)
        throws IllegalArgumentException {

        // Check input value
        if (!noCheck) {
// amchavan, 31-Jan-2006: commented out, it's useless            
//            if (value > Integer.MAX_VALUE || value < Integer.MIN_VALUE) {
//                throw new IllegalArgumentException(
//                    "Integer value out of bounds: " + value);
//            }
            if (hours) {
                if (!isValidRA((int) value)) {
                    throw new IllegalArgumentException(
                        "Invalid right ascension value: " + value );
                }
            } else if (!isValidDec((int) value)) {
                throw new IllegalArgumentException(
                    "Invalid declination value: " + value );
            }
        }

        char sign = ' ';
        if (value < 0) {
            sign = '-';
            value = -value;
        }

        long units = value / MILLISECS_PER_UNIT;
        value -= units * MILLISECS_PER_UNIT;

        long minutes = value / MILLISECS_PER_MINUTE;
        value -= minutes * MILLISECS_PER_MINUTE;

        String result;
        String colon = ":";
        String dot = ".";
        
        if (useColons) {
		  result = sign
            + TextUtils.zeroPadded(2, units)
            + colon
            + TextUtils.zeroPadded(2, minutes)
            + colon;
        } else {
        	result = sign
            + TextUtils.zeroPadded(2, units)
            + TextUtils.zeroPadded(2, minutes) ;           
        }
        
        result = result +  TextUtils.zeroPadded(2, (value / MILLISECS_PER_SEC))
        	+ dot
        	+ TextUtils.zeroPadded(3, (value % MILLISECS_PER_SEC));
        
        return(result);
        
    }

    /**
     * Check if an RA value in milliseconds is within its valid range.
     * 
     * @param ra   Right ascension in milliseconds.
     * 
     * @return true if the value is within its range, false otherwise.
     * 
     */
    // @author A M Chavan, 04-Jul-2001
    // amchavan, 10-Jan-2006: OK
    public static boolean isValidRA( long ra ) {
        if( ra < 0 || ra >= MAX_MILLISECS_RA )
            return false;
        return true;
    }

    /**
     * Check if a Dec value in milliarcseconds is within its valid range.
     * 
     * @param dec Declination in milliarcseconds.
     * 
     * @return true if the value is within its range, false otherwise.
     */
    // @author A M Chavan, 04-Jul-2001
    // amchavan, 10-Jan-2006: OK
    public static boolean isValidDec( long dec ) {
        if( dec <= -MAX_MILLIARCSECS_DEC || dec > MAX_MILLIARCSECS_DEC )
            return false;
        return true;
    }


    /**
     * Utility method: checks that the fields in a right ascension
     * definition have
     * legal values.
     * If any value is illegal, it throws an exception.
     *
     * @param sign	    must be +1
     * @param hours	    value in the hours field
     * @param minutes	value in the minutes field
     * @param seconds	value in the seconds field
     * 
     * @throws IllegalArgumentException If any values are illegal
     */
    // amchavan, 09-Jan-2006: added check for sign, updated Javadoc
    // amchavan, 19-Jan-2006: rewritten
    static void checkBoundsHHMMSS( int sign,
                                   double hours,
                                   double minutes,
                                   double seconds )
        throws IllegalArgumentException {

        final String illegal = "Illegal HHMMSS value: ";
        
        if( sign != 1 )
            throw new IllegalArgumentException( 
                    illegal + "negative values are not allowed" );
        
        if( hours >= 24 ) {
            throw new IllegalArgumentException( 
                    illegal + "hours should be < 24" );
        }
        
        if( minutes >= 60 ) {
            throw new IllegalArgumentException( 
                    illegal + "minutes should be < 60" );
        }
        
        if( seconds >= 60 ) {
            throw new IllegalArgumentException( 
                    illegal + "seconds should be < 60" );
        }
    }

    /**
     * Utility method: checks that the fields in a angle definition have
     * legal values.
     *
     * @param sign	        -1 if the entry had a minus sign, +1 otherwise
     * @param degrees	    value of the degrees field
     * @param arcminutes	value of the arcminutes field
     * @param arcseconds	value of the arcseconds field
     * 
     * @throws IllegalArgumentException If any values are illegal
     */
    // amchavan, 09-Jan-2006: added check for sign, updated Javadoc
    // amchavan, 19-Jan-2006: rewritten
    static void checkBoundsDDMMSS( int sign,
                                   double degrees,
                                   double arcminutes,
                                   double arcseconds )
        throws IllegalArgumentException {
        
        final String illegal = "Illegal HHMMSS value: ";

        if( sign != -1 && sign != 1 )
            // should never happen, really
            throw new IllegalArgumentException(
                    illegal + "'sign' should be +1 or -1" );
               
        if( degrees > 90 ) {
            throw new IllegalArgumentException( 
                    illegal + "degrees should be < 90" );
        }
        
        if( arcminutes >= 60 ) {
            throw new IllegalArgumentException( 
                    illegal + "minutes should be < 60" );
        }
        
        if( arcseconds >= 60 ) {
            throw new IllegalArgumentException( 
                    illegal + "arcseconds should be < 60" );
        }
        
        // special case...
        if (degrees == 90 && (arcminutes > 0 || arcseconds > 0)) {
            throw new IllegalArgumentException( 
                    illegal + "arcminutes and arcseconds should be zero" );
        }
    }

    /** 
     * Rounds argument db to a number of decimal places.
     * <BR>
     * Example: <code>round(2.32839, 1)</code> returns <code>2.3</code>
     *
     * @param db        Any double number
     * @param places    The number of decimal places to round to, should
     *                  be non-negative
     *
     * @return Input double value rounded to specific decimal places. 
     *         For very large input values and/ot number of decimal places
     *         it may return "infinity" instead. For instance:
     *         <code>Convert.round(1.7e300,10)</code> returns
     *         <code>Double.POSITIVE_INFINITY</code>
     * 
     */
    // @author cmavros, 19-Sept-2001
    // amchavan, 09-Jan-2006: refactored; optimized; fixed bug, was failing
    //                        for large numbers of decimal places
    public static double round( double db, int places ) {
        
        if( places < 0 ) 
            throw new IllegalArgumentException( 
                    "'places' is negative (" + places + ")" );
        
        double base = getRoundBase(places);
        
        double tmp = db * base;
        if( tmp >= 0. ) {
            tmp += 0.5;
            tmp = Math.floor( tmp );
        }
        else {
            tmp -= 0.5;
            tmp = Math.ceil( tmp );
        }

        tmp /= base;
        return tmp;
    }
    
    
    /** 
     * Truncates argument db to a number of decimal places.
     * <BR>
     * Example: <code>round(2.32839, 2)</code> returns <code>2.32</code>
     *
     * @param db        Any double number
     * @param places    The number of decimal places to round to, should
     *                  be non-negative
     *
     * @return Input double value rounded to specific decimal places. 
     *         For very large input values and/ot number of decimal places
     *         it may return "infinity" instead. For instance:
     *         <code>Convert.round(1.7e300,10)</code> returns
     *         <code>Double.POSITIVE_INFINITY</code>
     * 
     */
    public static double trunc ( double db , int places ) {
    	if( places < 0 ) 
            throw new IllegalArgumentException( 
                    "'places' is negative (" + places + ")" );
        
        double base = getRoundBase(places);
        
        double tmp = db * base;
        if( tmp >= 0. ) {
            tmp = Math.floor( tmp );
        }
        else {
            tmp = Math.ceil( tmp );
        }

        tmp /= base;
        return tmp;
    }


	private static double getRoundBase(int places) {
		// use a known value if possible
        double base;
        if( places <= 10 ) {
            base = bases[ places ];
        }
        else {
            base = 1.0;
            for( int i = 1; i <= places; i++ ) {
                base *= 10.;
            }
        }
		return base;
	}

    /** 
     * Convert an ISO date+time string to seconds.
     *
     * The input string is parsed using a subset of the ISO
     * representation, and it is allowed to include less then the
     * entire ISO string; missing information is supplied by the
     * system based on the current time and date.
     * <P>
     *
     * <strong>Note</strong>: Input string is interpreted as UT.
     *
     * @param isodate  A possibly incomplete ISO date+time string
     *
     * @return The number of seconds since January 1, 1970,
     *         01:00:00 GMT represented by the input string.
     *
     * @throws ParseException
     *
     * @see #longToISODate( long )
     * @see #completeISODateString( String )
     *
     */
    // @author amchavan, 09-Oct-2001
    // amchavan, 10-Jan-2005: OK
    public static long ISOdateToLong( String isodate ) throws ParseException {

        long milliSeconds = ISODateToMilliSec( isodate );
        long seconds = milliSeconds / 1000;
        return seconds;
    }
 
    /**
     * Convert a Starcat date to a Java date
     * @param starcatDate  Number of seconds since the Starcat epoch 
     *                     (1980-01-01T00:00:00.000 UT)
     * @return An equivalent Java date (UTC)
     */
    public static Date starcatDateToDate( long starcatDate ) {
        long time = (starcatDate + STARCAT_TO_UNIX_OFFSET) * 1000;
        Date d = new Date(time);
        return d;
    }

    /** 
     * 
     *  Convert a Java date to a Starcat date
     * @param date An equivalent Java date (UTC)
     * @return long starcatDate  Number of seconds since the Starcat epoch 
     *                     (1980-01-01T00:00:00.000 UT)
     */
    public static long dateToStarcatDate( Date date ) {
        long dateLong = date.getTime();
        long unixDate = dateLong / 1000;
        long time = unixDate - STARCAT_TO_UNIX_OFFSET;
        return time;
    };
    
    
    /**
     * convert the date in string format in GMT time
     * @param date starcatDate  Number of seconds since the Starcat epoch 
     *                     (1980-01-01T00:00:00.000 UT)
     * @return String represent a date in simple date format
     * @see #SIMPLE_DATE_FORMAT
     */
    public static String starcatDateToString(long date) {
        String ret = Convert.SIMPLE_DATE_FORMAT.format(Convert.starcatDateToDate(date));
        return ret;
    }

    
    
    /**
     * Convert a date in GMT time to a starcat format.
     * 
     * if invalid string the methods re
     * @param value String represent a date in simple date format
     * @return long  starcatDate  Number of seconds since the Starcat epoch 
     *                     (1980-01-01T00:00:00.000 UT)
     * @see #SIMPLE_DATE_FORMAT
     * @throws IllegalArgumentException  if the string is invalid
     */
    public static long stringToStarcatDate(String value)throws IllegalArgumentException {

        if (value == null) {
            throw new IllegalArgumentException("DATE CONVERSION ERR. null value ");
        }
        long val = 0;
            value = value.trim();
            try {
                Date d = Convert.SIMPLE_DATE_FORMAT.parse(value);              
                val = Convert.dateToStarcatDate(d);
            } catch (ParseException e) {
               throw new IllegalArgumentException("DATE CONVERSION ERR. while converting " + value);
            }
        return val;
    }
      

    /**
     * convert the date 
     * @param date date obj
     * @return String represent a date in simple date format
     * @see #SIMPLE_DATE_FORMAT
     */
    public static String dateToSimpleFormatString(Date dateObj) {
        return Convert.SIMPLE_DATE_FORMAT.format(dateObj);
    }

    
    /**
     * Return a date format for format "day month year" As a separator,
     * "/:,-". Month 3 letters case insensitive. year full
     * 
     * @param date
     * @return boolean
     */
    public static Date getDateFormatDMY(String date) {
        Date d = null;
     
        StringBuffer value = new StringBuffer();
        if (date != null){    
        StringTokenizer tokens = new StringTokenizer(date,"[/:,- ]");
        if (tokens.countTokens() == 3) {
           String day = tokens.nextToken();
           String month = tokens.nextToken();
           String year = tokens.nextToken();
           String start = month.substring(0,1).toUpperCase();
           String end = month.substring(1).toLowerCase();
           value.append(day).append(" ").append(start).append(end).append(" ").append(year); 
        }
        }
        try {
            d = Convert.SHORT_DATE_FORMAT.parse(value.toString());
        } catch (ParseException e) {
          throw new IllegalArgumentException("DATE CONVERSION ERR. while converting " + value);
        }
        return d;
    }
    /**
     * Convert an ISO date+time string to seconds.
     * @see #ISOdateToLong( String )
     * @throws ParseException
     */
    public static long ISODateToMilliSec( String isodate ) 
        throws ParseException {
        
        Date date = ISODateToDate( isodate );
        long milliSeconds = date.getTime();
        return milliSeconds;
    }

    /**
     * Convert an ISO date+time string to a Date object.
     * @see #ISOdateToLong( String )
     * @throws ParseException
     */
    // amchavan, 10-Jan-2005: OK
    public static Date ISODateToDate(String isodate) throws ParseException {
        String fulldate = completeISODateString(isodate);
        SimpleDateFormat iso = ISO_SIMPLE_DATE_FORMAT;
        Date date = iso.parse(fulldate);
        return date;
    }

    /**
     * Convert an integer number of seconds to an ISO date+time
     * string. <BR>
     * All (absolute) times and dates are displayed in OHS applications
     * using the full ISO representation of date/time values,
     * which follows the following syntax:<BR>
     * &nbsp;&nbsp;&nbsp;&nbsp;<code>yyyy-mm-ddThh:mm:ss</code><bR>
     * where:
     * <BR>     
     *
     * <code>yyyy</code> is a four-digit year number, like
     * <code>1997</code>;<BR>
     *
     * <code>mm</code> is the two-digit month number, like
     * <code>08</code> for August;<BR>
     *
     * <code>dd</code> is the two digit day number, like
     * <code>14</code>;<BR>
     *
     * <code>hh</code> is the two digit hour number (24-hour clock),
     * like <code>21</code>;<BR>
     *
     * <code>mm</code> is the two digit minute number, like
     * <code>05</code>;<BR>
     *
     * <code>ss.sss</code> is the fractional number of seconds, like
     * <code>30.5</code>;<BR>
     * <P>
     *
     * <strong>Note</strong>: output string is expressed as UT.
     *
     * @param secs   Seconds since January 1, 1970, 00:00:00 GMT 
     *               to be represented by the output string.
     *
     * @return An ISO date+time string. 
     *
     * @see #ISOdateToLong(String)
     */
    // @author amchavan, 09-Oct-2001: creation.
    // amchavan, 10-Jan-2006: OK
    public static String longToISODate(long secs) {
        long milliSec = secs * 1000;      
        String isoDate = longToISODateMilliSec(milliSec);
        return isoDate;
    }

  
    
    /**
     * Convert an integer number of milliseconds to an ISO date+time string.
     * 
     * @param milliSec Milliseconds since January 1, 1970, 00:00:00 GMT 
     *                 to be represented by the output string.
     * @return An ISO date+time string.
     * @see #longToISODate(long)
     */
    // amchavan, 10-Jan-2006: updated Javadoc
    public static String longToISODateMilliSec( long milliSec ) {
        Date date = new Date( milliSec );
        String isoDate = dateToIsoDate( date );
        return isoDate;
    }
    
    /**
     * @param milliSec Milliseconds since January 1, 1970, 00:00:00 GMT 
     *                 to be represented by the output string.
     * @return A date+time string in database format
     * @see #DB_DATE_FORMAT
     */
    // amchavan, 10-Jan-2006: updated Javadoc
    public static String longToDBDateMilliSec( long milliSec ) {
        Date date = new Date( milliSec );
        String isoDate = dateToDBDate( date );
        return isoDate;
    }

    /**
     * @return An date+time string in ISO format representing the input date.
     * @see #ISO_DATE_FORMAT
     */
    // amchavan, 09-Jan-2005: updated Javadoc
    public static String dateToIsoDate(Date date) {
        SimpleDateFormat iso = ISO_SIMPLE_DATE_FORMAT;                 
        String isoDate = iso.format(date);
        return isoDate;
    }
    
    /**
     * @return An date+time string in DB format representing the input date.
     * @see #DB_DATE_FORMAT 
     */
    // amchavan, 09-Jan-2005: updated Javadoc
    public static String dateToDBDate(Date date) {
        SimpleDateFormat iso = new SimpleDateFormat(DB_DATE_FORMAT);                 
        String isoDate = iso.format(date);
        return isoDate;
    }
    
    /**
     *
     * @param secs The number of seconds since January 1, 1970,
     *             01:00:00 GMT to be represented by the output 
     *             string.
     *
     * @return An ISO date+time string. 
     *
     * @see #ISOdateToLong(String)
     */
    // amchavan, 10-Jan-2006: OK
    public static String longToHHSS(long secs) {

        SimpleDateFormat iso = new SimpleDateFormat(ST_TIME_FORMAT);
        return (iso.format(new Date(secs * 1000)));
    }


    /**
     * Parses a partial date-time string. <P>
     *
     * The input string is parsed using a subset of the ISO
     * representation, and it is allowed to include less then the
     * entire ISO string; missing information is supplied by the
     * system: year and month are assumed to be the same as "today",
     * with hours, minutes and seconds being zero.
     * <br>
     * For more info, see <code>completeISODateString( String, String )</code>
     *
     * @param dt_in    Incomplete string to parse.
     *
     * @throws ParseException
     *
     * @see #completeISODateString( String, String )
     */
    // amchavan, 09-Jan-2005: no need to unit test this method, we'll
    //                        test completeISODateString(String,String)
    //                        instead
    public static String completeISODateString(String dt_in)
        throws ParseException {
        String now = TextUtils.currentISODateTime();
        int idx = now.indexOf('T');
        now = now.substring(0, idx + 1) + "00:00:00";
        return completeISODateString(dt_in, now);
    }

    /**
     * Parses a partial date-time string. <P>
     *
     * The input string is parsed using a subset of the ISO
     * representation, and it is allowed to include less then the
     * entire ISO string; missing information is supplied by parameter
     * <code>defDT</code>. The minimum piece of information to be
     * provided is the day number. The following table gives a few
     * examples of how partial ISO date/time strings are completed,
     * assuming <code>defDT</code> is <code>1997-08-01T00:00:00</code>.
     * <BR>
     * Notice that:
     * <UL>
     * <LI> blanks surrounding the partial strings are ignored, and
     *      are shown here only for clarity;     
     * <LI> one-digit numbers can, but need not, be prefixed by a zero
     *      ("2" is the same as "02"); 
     * <LI> the "T" char is only needed when a time different
     *      from midnight (00:00:00) needs to be specified; 
     * <LI> partial year specification (like "98") is not allowed;
     * <LI> textual month specification (like "Dec") is not allowed;
     * <LI> only years between 1970 and 2037 are supported.
     * </UL>
    <PRE>
            07              1997-08-07T00:00:00
         10-07              1997-10-07T00:00:00
    1998-2-1                1998-02-01T00:00:00
    1998-01-01              1998-01-01T00:00:00
            15T05           1997-08-15T05:00:00
            15T05:30        1997-08-15T05:30:00
            20T5:5:30       1997-08-20T05:05:30
         10-16T21:30        1997-10-16T21:30:00
    1999-12-31T23:59:59.99  1999-12-31T23:59:59.99
    98-2-1                  ERROR: partial year specification
    1969-12-31              ERROR: only years 1970-2037 are supported
    1999-12-31T23:59:61     ERROR: invalid number of seconds
    0001-55-55T66:66:66     ERROR: invalid time specification
    2001-Oct-09             ERROR: textual month specification is not allowed
    </PRE>
     *
     * @param inDT     Incomplete string to parse.
     * @param defDT    A full ISO date+time string; this is where
     *                 default information is taken from.<BR>
     *                 For instance: <code>2001-10-09T10:52:30</code><BR>
     *                 This parameter is not checked for integrity.
     *
     * @throws ParseException
     *
     * @see #completeISODateString( String )
     *
     */
    // @author amchavan, 09-Oct-2001: creation
    // 09-Jan-2005, amchavan: OK
    public static String completeISODateString(String inDT, String defDT)
        throws ParseException {

        String[] inElements = parseISODate(inDT);
        String[] defElements = parseISODate(defDT);
        StringBuffer complete = new StringBuffer();
        String ret = null;

        // Provide default elements
        //----------------------------------------
        for (int i = 0; i < inElements.length; i++)
            if (inElements[i].length() == 0)
                inElements[i] = defElements[i];

        // Build final string
        //----------------------------------------
        complete.append(TextUtils.zeroPadded(4, inElements[0]));
        complete.append('-');
        complete.append(TextUtils.zeroPadded(2, inElements[1]));
        complete.append('-');
        complete.append(TextUtils.zeroPadded(2, inElements[2]));
        complete.append('T');
        complete.append(TextUtils.zeroPadded(2, inElements[3]));
        complete.append(':');
        complete.append(TextUtils.zeroPadded(2, inElements[4]));
        complete.append(':');
        complete.append(TextUtils.zeroPadded(2, inElements[5]));

        ret = complete.toString();

        // Now verify if it makes any sense at all
        //----------------------------------------
        SimpleDateFormat iso = ISO_SIMPLE_DATE_FORMAT;
        iso.setLenient(false);
        Date d = iso.parse(ret);
        if (d == null)
            throw new ParseException(ret, 0);

        try {
            int y = Integer.parseInt(inElements[0]);
            if (y < 1970 || y > 2037)
                throw new ParseException(ret, 0);
        } catch (NumberFormatException e) {
            // should never happen here
        }

        return (ret); // All is OK, return
    }

    /**
     * Breaks up an ISO date+time string into an array of six
     * strings: year, month, day, hour, minutes and seconds.
     * 
     * @throws ParseException
     * @author amchavan, 09-Oct-2001: creation
     */
    // amchavan, 10-Jan-2006: no reason to unit test this; it is tested via
    //                        completeISODateString()
    private static String[] parseISODate(String isodatetime)
        throws ParseException {

        String[] components = new String[6];
        String t = null;
        String isodate = "";
        String isotime = "";

        isodatetime = isodatetime.trim(); // just in case

        // Check if we have a "T"
        int idx = isodatetime.indexOf('T');
        if (idx == -1)
            isodate = isodatetime; // No T, just date 
        else {
            isodate = isodatetime.substring(0, idx);
            isotime = isodatetime.substring(idx + 1);
        }

        // Parse the date part
        //----------------------------------------

        t = isodate;
        for (int i = 2; i >= 0; i--) {

            idx = t.lastIndexOf('-');
            if (idx == -1) {

                components[i] = t;
                t = "";
                continue;
            }

            components[i] = t.substring(idx + 1);
            t = t.substring(0, idx);
        }

        // Parse the time part
        //----------------------------------------

        t = isotime;
        for (int i = 3; i <= 5; i++) {

            idx = t.indexOf(':');
            if (idx == -1) {

                components[i] = t;
                t = "";
                continue;
            }

            components[i] = t.substring(0, idx);
            t = t.substring(idx + 1);
        }

        for (int i = 0; i < components.length; i++) {
            for (int j = 0; j < components[i].length(); j++) {

                char c = components[i].charAt(j);
                if (!Character.isDigit(c) && c != '.')
                    throw new ParseException(components[i], j);
            }
        }

        return (components);
    }
    	
	/**
	 * Check if an RA value in the formats
	 * 
	 * HH.sss
	 * HH MM.sss
	 * HH MM SS.sss
	 * 
	 * is valid and within its valid range.
	 * 
	 * @param ra
	 *            Right ascension in a string format
	 * 
	 * @return true if the value is valid and within its range, false otherwise.
	 * 
	 */
     // @author fsogni, 03-Jun-2005
    // amchavan, 10-Jan-2006: OK
	public static boolean isValidRA(String ra) {
	
		try {
			HHMMSSToMillisec(ra);
		} catch (IllegalArgumentException e) {
			return false;
		}
		return true;
	}
	
	/**
	 * Check if an DEC value in the formats
	 * 
	 * DD.sss
	 * DD MM.sss
	 * DD MM SS.sss
	 * 
	 * is valid and within its valid range.
	 * 
	 * @param dec Declination in a string format
	 * 
	 * @return true if the value is valid and within its range, false otherwise.
	 * 
	 */
    // @author fsogni, 03-Jun-2005
    // amchavan, 10-Jan-2006: OK
	public static boolean isValidDec( String dec ) {

        try {
            DDMMSSToMilliarcsec( dec );
        }
        catch( IllegalArgumentException e ) {
            return false;
        }
        return true;
    }

    
    /**
     * TEST ONLY: check the correctness of the regular expressions
     *            used to parse coordinate strings. 
     *            DO NOT REMOVE, this is used by unit tests.
     * @param string
     */
    static boolean checkCoordSyntax( String string ) {
        if( CRE_ONE == null )
            initializeCoordinateRegexps();
        
        if( CRE_NOSEPARATORS.matcher( string ).matches() )
            return true;
        if( CRE_ONE.matcher( string ).matches() )
            return true;
        if( CRE_TWO.matcher( string ).matches() )
            return true;
        if( CRE_THREE.matcher( string ).matches() )
            return true;
        return false;
    }
    
    /**
     * Initialize the regular expression patterns used to parse
     * coordinate strings.
     */
    // Cannot get quantifiers like \d(1,2) to work, so regular 
    // expressions are complicated :-(
    // amchavan, 19-Jan-2006
    private static void initializeCoordinateRegexps() {
        StringBuffer regexp;
        
        // A plus or minus char, appearing zero or one times
        String reOptionalSign = "([\\+-]??)";
        
        // One or two digits, followed by an optional fractional
        // part of 1-3 digits
        String reRealNumber = "((\\d\\d??)(\\.(\\d(\\d\\d??)??)??)??)";
        
        // One or two digits
        String reInteger = "(\\d\\d??)";
        
        // Any number of spaces, or one colon char
        String reSeparator = "( +|:)";
        
        // Beginning of line
        String reBegin = "^";
        
        // End of line
        String reEnd = "$";
        
        // As used in PAF files: no separators, followed by an optional
        // fractional part of 1-3 digits
        String reNoSeparators = 
            //"((\\d\\d\\d\\d\\d\\d??)(\\.(\\d(\\d\\d??)??)??)??)";
        	"((\\d\\d\\d\\d\\d\\d??)((\\.(\\d+))??))";
        
        regexp = new StringBuffer();
        regexp.append( reBegin )
              .append( reOptionalSign )
              .append( reRealNumber )
              .append( reEnd );
        CRE_ONE = Pattern.compile( regexp.toString() );
        
        regexp = new StringBuffer();
        regexp.append( reBegin )
              .append( reOptionalSign )
              .append( reInteger )
              .append( reSeparator )
              .append( reRealNumber )
              .append( reEnd );
        CRE_TWO = Pattern.compile( regexp.toString() );
        
        regexp = new StringBuffer();
        regexp.append( reBegin )
              .append( reOptionalSign )
              .append( reInteger )
              .append( reSeparator )
              .append( reInteger )
              .append( reSeparator )
              .append( reRealNumber )
              .append( reEnd );
        CRE_THREE = Pattern.compile( regexp.toString() );
        
        regexp = new StringBuffer();
        regexp.append( reBegin )
              .append( reOptionalSign )
              .append( reNoSeparators )
              .append( reEnd );
        CRE_NOSEPARATORS = Pattern.compile( regexp.toString() );
    }
    
    
    /**
     * Convert degrees in a double to milliarcseconds in an integer
     * 
     * @param degrees
     * @return
     */
    public static int degreesToMilliarcsecs( double degrees ) {
    	int res;    	    	
    	res = (int)(0.5 + (degrees * MILLISECS_PER_UNIT));
		return res;
    	
    }
}
