package org.eso.ohs.core.utilities;

/**
 * A generic interface defining a tree model.
 * This interface is intended to be generic enough to act as a general
 * tree model for a range of applications as well as being
 * able to incorporate a range of implementations (JDOM or 
 * some other XML parser, for example).
 *
 * While there is a javax.swing.tree.TreeModel class,
 * it was not used as it requires the inclusion of Swing and is therefore
 * not appropriate for non-graphical applications.
 * 
 * The "G" in "GTreeModel" is supposed to be for "Generic",
 * but its really just to create a name which will not cause conflicts
 * with other "TreeModel" classes.
 *
 * Tree Nodes:
 *
 * The GTreeNode is the basic interface for a node of a GTree.
 * A GTreeNode has children and a parent, which are GTreeNodes.
 * In addition a GTreeNode has a String name and an Object value.
 * GTreeNode a compared based on name alone, so two tree nodes are the
 * same if they have the same name.
 *
 * Tree Paths:
 *
 * The GTreePath interface is a means to address a node in a tree.
 * The interface is intended to be extended to handle any form of
 * tree-path syntax, indepedent of the format of the tree source data.
 * That is, a PAFTreePath could be used to navigate in a GTreeModel 
 * built from a Properties file...for example.
 *
 * The getPath() method of GTreePath returns an array of GTreeNode objects
 * which are used to traverse a GTreeModel.
 *
 * Note that the nodes of a tree path NOT expected to be connected in a
 * parent<->child relationship as in the case of a tree.
 *
 * @author Nick Kornweibel for ESO.
 */

import java.util.Enumeration;

public interface GTreeModel {

    /**
     * @return the root of the tree
     */
    public GTreeNode getRoot();

    /**
     * @return true if node is a leaf, else false.
     */
    public boolean isLeaf(GTreeNode node);

    /**
     * @return the child of node with the name childName, or null if not found.
     */
    public GTreeNode getChild(GTreeNode node, String childName);

    /**
     * @return the child of node with the name child.getName(), or null if not found.
     */
    public GTreeNode getChild(GTreeNode node, GTreeNode child);

    /**
     * @return the number of children of node.
     */
    public int getChildCount(GTreeNode node);

    /**
     * @return an Enumeration of node's children.
     */
    public Enumeration<GTreeNode> getChildren(GTreeNode node);

    /**
     * @return an array of GTreeNode objects, with the 0th element being root and
     * the last element being aNode.
     */
    public GTreeNode[] getPathToRoot(GTreeNode aNode);
 
    /**
     * Add child as a child or aNode (no checking for duplicates).
     */
    public void addChild(GTreeNode aNode, GTreeNode child);

    /**
     * @return (and remove) child from aNode, or null if not found.
     */
    public GTreeNode removeChild(GTreeNode aNode, GTreeNode child);
   
    /**
     * @return (and remove) the first child of aNode with the name childName.
     */
    public GTreeNode removeChild(GTreeNode aNode, String childName);
}
