package org.eso.ohs.core.utilities;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.log4j.Logger;

/**
 * THIS IS THE (BASE) CLASS TO BE USED FOR ALL CONFIGURATION HANDLING FOR OHS
 * APPLICATIONS.
 * 
 * The apps/AppConfig class is an extention of OHSConfig that is suitable ONLY
 * for the P2PP, P2PPAppServer and OT applications.
 * 
 * A tree based data model for storing and retriving string data. Specifically
 * the model allows for the reading in and querying of Param file and Paf file
 * data from multiple Param or Paf files.<BR>
 * 
 * The tree model has an "invisible" root node, which is bypassed on data
 * queries.<BR>
 * 
 * The class has the option of setting an instance, allowing it to be used as a
 * singleton. However other instances of the class may be created that are
 * distinct from the singleton instance.<BR>
 * 
 * The singleton accessor is getInstance(), which creates an instance of
 * OHSConfig and sets if not already set.<BR>
 * 
 * The following is a example of how the OHSConfig could be used to manage
 * configuration data for an application....for real running examples please
 * refer to the unit test file: <BR>
 * </tt>org.eso.ohs.utilities.test.OHSConfig</tt>
 * <P>
 * At start-up of the application
 * 
 * <PRE>
 * 
 * OHSConfig config = OHSConfig.getInstance();
 *  // Read the general application settings. PafFile appParams = new
 * Paramfile("site.cf"); config.addPafData(appParams);
 *  // Read database settings, and place them under "DB". Paramfile dbParams =
 * new Paramfile("database.cf"); config.addParamData(dbParams, "DB");
 *  // Add more config data not available in param files.
 * config.setValue("DB.CONNECTION.TIMEOUT", 30);
 *  // Read in app server settings and place under "APPSERVER" Paramfile
 * dbParams = new Paramfile("appserver.cf"); config.addParamData(dbParams,
 * "APPSERVER");
 * 
 * </PRE>
 * 
 * During running the application...data is queried
 * 
 * <PRE>
 * 
 * OHSConfig dbConfig = OHSConfig.getInstance().getSubConfig("DB");
 *  // database code expects an OHSConfig object.
 * initializeDatabaseConnection(dbConfig);
 * ConnectionPool.getPool().acquireConnection(dbConfig.getValue("URL")); // the
 * previous line could also be acheived with:
 * ConnectionPool.getPool().acquireConnection(
 * OHSConfig.getInstance().getValue("DB.URL"));
 * 
 * </PRE>
 * 
 * During running the application, a connection to a redundant appserver maybe
 * required...both config files (primary and redundant) can have the same keys
 * and syntax, but will be stored in separate config subtrees.
 * 
 * <PRE> // Read in app server settings and place under "APPSERVER2" Paramfile
 * asParams = new Paramfile("appserver2.cf"); OHSConfig cfg =
 * OHSConfig.getInstance(); cfg.addParamData(dbParams, "APPSERVER2"); OHSConfig
 * asCfg1 = cfg.getSubConfig("APPSERVER"); OHSConfig asCfg2 =
 * cfg.getSubConfig("APPSERVER2"); String serverKey = "SERVERNAME"; String
 * portKey = "PORTNUMBER";
 * 
 * AppServerProxy as = new AppServerProxy(asCfg1.getValue(serverKey),
 * asCfg1.getValue(portKey)); if (as.isDown() && asCfg2 != null) { // we have a
 * second server configured...try it! as = new
 * AppServerProxy(asCfg2.getValue(serverKey), asCfg2.getValue(portKey)); }
 * 
 * </PRE>
 * 
 * 
 * @version $Revision: 1.21.4.1 $
 * @author Nick Kornweibel for ESO.
 * @author amchavan, 07-Jan-2002: added
 * @author cmavros, 29-Apr-02 1728/P2PP - Save configurations to PAF files. <br>
 *         ChangeLog: cmavros, 03JULY02 - Add header to writeConfigFile()
 */

public class OHSConfig {

    /**
     * Field rcsid
     */
    private final static String rcsid = "$Id: OHSConfig.java,v 1.21.4.1 2007/10/24 11:30:47 pnunes Exp $";

    /* Standard log channel */
    /**
     * Field stdlog_
     */
    private static Logger stdlog_ = Logger.getLogger(OHSConfig.class);

    /* singleton instance */
    /**
     * Field instance_
     */
    protected static OHSConfig instance_ = null;

    /**
     * Field tree_
     */
    protected DefaultGTreeModel tree_ = null;

    /**
     * Field maxParamFileLines_
     */
    protected static int maxParamFileLines_ = 10000;

    /**
     * Field FILE_HDR
     */
    protected static String FILE_HDR = "#!--- USER SELECTION OPTION FILE\n"
            + "#!--- This is an automatically generated file.\n"
            + "#!--- It will be read and overwritten.\n"
            + "#!--- Do not edit this file.\n";

    /**
     * @return instance of singleton OHSConfig, creating and setting one if not
     *         already set.
     */
    public static OHSConfig getInstance() {
        if (instance_ == null) {
            instance_ = new OHSConfig();
        }
        return instance_;
    }

    /**
     * set the singleton OHSConfig instance.
     * 
     * @param i
     *            OHSConfig
     */
    public static void setInstance(OHSConfig i) {
        instance_ = i;
    }

    /**
     * @return true if a singleton OHSConfig has been set.
     */
    public static boolean isSingletonSet() {
        return (instance_ != null);
    }

    /**
     * Build an empty OHSConfig - it is not set as the singleton!
     */
    public OHSConfig() {
        DefaultGTreeNode root = new DefaultGTreeNode("config");
        tree_ = new DefaultGTreeModel(root);
        stdlog_.info("*******PLATFORM: " + getSystemProperty("os.name") +"*******");

    }

    /**
     * Build an OHSConfig from the node given.
     * 
     * @param root
     *            GTreeNode
     */
    public OHSConfig(GTreeNode root) {
        tree_ = new DefaultGTreeModel(root);
    }

    /**
     * Build an OHSConfig from the tree model given.
     * 
     * @param tree
     *            DefaultGTreeModel
     */
    public OHSConfig(DefaultGTreeModel tree) {
        tree_ = tree;
    }

    /**
     * Add data from a Paramfile formatted file to this tree.<BR>
     * The data is placed in the tree under the specified path, or directly
     * under the root if the path is empty or null.<BR>
     * The path is expected to be a valid ParamfileTreePath path.
     * 
     * @param paramfile
     *            Pathname of a Paramfile formatted file; pathames are relative
     *            to the installation directory.
     * 
     * @param path
     *            Is expected to be a valid ParamfileTreePath path.
     * 
     * @throws IOException
     *             If reading the file generated some I/O error.
     * @throws NotASCIIException
     *             If file contains non-ASCII data, including any 8-bit
     *             character.
     * @see org.eso.ohs.core.utilities.Paramfile
     * @see org.eso.ohs.core.utilities.ParamfileTreePath
     * 
     */
    public void addParamfile(String paramfile, String path)
            throws IOException, NotASCIIException {
        addParamData(new Paramfile(paramfile, maxParamFileLines_),
                new ParamfileTreePath(path));
    }

    /**
     * Add data from a Paramfile formatted file to this tree.<BR>
     * The data is placed in the tree directly under the root.
     * 
     * @param paramfile
     *            Pathname of a Paramfile formatted file; non-absolute pathames
     *            are relative to the directory where the JVM is running.
     * 
     * @throws IOException
     *             If reading the file generated some I/O error.
     * @throws NotASCIIException
     *             If file contains non-ASCII data, including any 8-bit
     *             character.
     * @see org.eso.ohs.core.utilities.Paramfile
     * 
     */
    public void addParamfile(String paramfile)
            throws IOException, NotASCIIException {
        stdlog_.info("loading param file: " + paramfile);
        addParamData(new Paramfile(paramfile, maxParamFileLines_),
                (GTreePath) null);

    }

    /**
     * override whatever is in private configuration file in $HOME dir
     * 
     * @param configFile
     *            String
     * @throws IOException
     * @throws NotASCIIException
     */
    public void retrievePrivateConfResource(String configFile)
            throws IOException, NotASCIIException {
        File configFileBase = new File(configFile);
        String fileName = configFileBase.getName();
        File privateConfigFile = new File(System.getProperty("user.home")
                + File.separator + fileName);
        stdlog_.debug("Retrieving private config file "
                + privateConfigFile.getAbsolutePath());
        if (privateConfigFile.exists()) {
            addParamfile(privateConfigFile.getAbsolutePath());
        } else {
            stdlog_.warn("No private config file "
                    + privateConfigFile.getAbsolutePath());
        }
    }

    /**
     * Method loadConfigFile Uses the OHSApplication algorithm for retrieving
     * the config file. &lt;configFile&gt; is the
     * &lt;configFilePath&gt;.getName()
     * <OL>
     * <li> Try to load &lt;configFilePath&gt; from classloader</li>
     * <li> Try to load &lt;configFilePath&gt; from current working directory<BR>
     * The properties already loaded in the previous step are overridden</li>
     * <li> Try to load OHSApplication.getOhsUserConfigDir()/&lt;configFile&gt;
     * from user home directory<BR>
     * The properties already loaded in the previous steps are overridden</li>
     * </OL>
     * 
     * @param configFilePath
     *            String
     * @param ohsDir
     *            String directory in which the data are saved
     * @return
     * @throws IOException
     * @throws NotASCIIException
     */
    public boolean loadConfigFile(String configFilePath, String ohsDir) {

        File configFile = new File(configFilePath);
        String confFileName = configFile.getName();
        boolean configFileLoaded = false;
        File privateConfigFile = new File(ohsDir + File.separator
                + confFileName);
        try {
            addParamfileAsResourceStream(configFilePath);
            stdlog_.info("load param file: " + configFilePath);
            configFileLoaded = true;
        } catch (Exception e) {
            stdlog_.debug("Cannot load param file: " + configFilePath + ": \n "
                    + e.getMessage());
        }
        try {
            addParamfile(configFilePath);
            stdlog_.info("load param file: " + configFilePath);
            configFileLoaded = true;
        } catch (Exception e) {
            stdlog_.debug("Cannot load param file: " + configFilePath + ": \n "
                    + e.getMessage());

        }
        try {
            addParamfile(privateConfigFile.getAbsolutePath());
            stdlog_.info("load param file: "
                    + privateConfigFile.getAbsolutePath());
            configFileLoaded = true;
        } catch (Exception e) {
            stdlog_.debug("Cannot load param file: "
                    + privateConfigFile.getAbsolutePath() + ": \n "
                    + e.getMessage());
        }
        return configFileLoaded;
    }

    /**
     * Method getFileContent Uses the OHSApplication algorithm for retrieving
     * the file.
     * 
     * @see loadConfigFile
     * @param configFilePath
     *            String
     * @param ohsDir
     *            String directory in which the data are saved
     * @throws IOException
     * @throws NotASCIIException
     */
    public String getFileContent(String keyFileName, String ohsDir) {

        File privateConfigFile = getFileLocal(keyFileName, ohsDir);
        String filePath = getValue(keyFileName);
        File configFile = new File(filePath);
        String confFileName = configFile.getName();

        String fileContent = null;

        try {
            InputStream s = FileIOUtils.getResourceStream(filePath);
            fileContent = FileIOUtils.fileToString(s);
            s.close();
            stdlog_.info("load file as ResourceStream: " + filePath);

        } catch (Exception e) {
            stdlog_.debug("Cannot load  file: " + filePath + ": \n "
                    + e.getMessage());
        }
        try {
            if (configFile.canRead()) {
                fileContent = FileIOUtils.fileToString(configFile);
                stdlog_.info("load file from file system: " + filePath);
            }
        } catch (Exception e) {
            stdlog_.debug("Cannot load  file: " + filePath + ": \n "
                    + e.getMessage());
        }
        try {
            if (privateConfigFile.canRead()) {
                fileContent = FileIOUtils.fileToString(privateConfigFile
                        .getAbsolutePath());
                stdlog_.info("load file: "
                        + privateConfigFile.getAbsolutePath());
            }
        } catch (Exception e) {
            stdlog_.debug("Cannot load file: "
                    + privateConfigFile.getAbsolutePath() + ": \n "
                    + e.getMessage());
        }
        return fileContent;
    }

    /**
     * @param ohsDir
     * @param filePath
     * @return File
     */
    public File getFileLocal(String keyFileName, String ohsDir) {
        String filePath = getValue(keyFileName);
        File privateConfigFile = new File(ohsDir + File.separator + filePath);
        return privateConfigFile;
    }

    /**
     * Add data from a Paramfile formatted file to this tree. The file is
     * retrieved from the classpath.<BR>
     * The data is placed in the tree directly under the root.
     * 
     * @param configFile
     *            Pathname of a Paramfile formatted file; non-absolute pathames
     *            are relative to the directory where the JVM is running. The
     *            file can be placed into a jar file.
     * @throws IOException
     *             If reading the file generated some I/O error.
     * @throws NotASCIIException
     *             If file contains non-ASCII data, including any 8-bit
     *             character.
     * 
     * @see org.eso.ohs.core.utilities.Paramfile
     * 
     */
    public void addParamfileAsResourceStream(String configFile)
            throws IOException, NotASCIIException {

        InputStream s = FileIOUtils.getResourceStream(configFile);
        addParamfile(s);
    }

    /**
     * Add data from a Paramfile formatted file to this tree.<BR>
     * The data is placed in the tree directly under the root.
     * 
     * @param paramfile
     *            Pathname of a Paramfile formatted file; non-absolute pathames
     *            are relative to the directory where the JVM is running.
     * 
     * @throws IOException
     *             If reading the file generated some I/O error.
     * @throws NotASCIIException
     *             If file contains non-ASCII data, including any 8-bit
     *             character.
     * @see org.eso.ohs.core.utilities.Paramfile
     * 
     */
    public void addParamfile(InputStream paramfile)
            throws IOException, NotASCIIException {
        InputStreamReader r = new InputStreamReader(paramfile);
        addParamData(new Paramfile(r, maxParamFileLines_), (GTreePath) null);
    }

    /**
     * Save configuration to a PAF file.<BR>
     * 
     * @param fileName
     *            Name of file for saving output;
     * @param data
     *            Data from a Paramfile object.
     * 
     * @throws IOException
     *             If reading the file generated some I/O error.
     * @throws NotASCIIException
     *             If file contains non-ASCII data, including any 8-bit
     *             character.
     * @see org.eso.ohs.core.utilities.Paramfile
     * 
     */
    public void saveParamDataToPAF(String fileName, Paramfile data)
            throws IOException, NotASCIIException {
        saveParamDataToPAF(fileName, data, (GTreePath) null);
    }

    /**
     * Save configuration to a PAF file.<BR>
     * 
     * @param fileName
     *            Name of file for saving output;
     * @param data
     *            Data from a Paramfile object.
     * @param path
     *            A valid ParamfileTreePath path.
     * 
     * @throws IOException
     *             If reading the file generated some I/O error.
     * @throws NotASCIIException
     *             If file contains non-ASCII data, including any 8-bit
     *             character.
     * @see org.eso.ohs.core.utilities.Paramfile
     * @see org.eso.ohs.core.utilities.ParamfileTreePath
     * 
     */
    public void saveParamDataToPAF(String fileName, Paramfile data, String path)
            throws IOException, NotASCIIException {
        saveParamDataToPAF(fileName, data, new ParamfileTreePath(path));
    }

    /**
     * Save configuration to a PAF file.<BR>
     * 
     * @param fileName
     *            Name of file for saving output;
     * @param data
     *            Data from a Paramfile object.
     * @param path
     *            A valid ParamfileTreePath path.
     * 
     * @throws IOException
     *             If reading the file generated some I/O error.
     * @throws NotASCIIException
     *             If file contains non-ASCII data, including any 8-bit
     *             character.
     * @see org.eso.ohs.core.utilities.Paramfile
     * @see org.eso.ohs.core.utilities.GTreePath
     * 
     */
    protected void saveParamDataToPAF(String fileName, Paramfile data,
            GTreePath path) throws IOException, NotASCIIException {
        Paramfile paramfile = new Paramfile();
        String header = ("CONFIGURATION " + TextUtils.currentISODateTime() + "(UT)\n");
        DefaultGTreeModel baseTree = null;
        String[] keys = data.getKeywords();
        for (int j = 0; j < keys.length; j++) {
            String key = (String) keys[j];
            String value = (String) data.getValue(key);
            paramfile.put(key, value);
        }
        try {
            if (fileName != null) {
                paramfile.writeTo(new FileWriter(fileName), header);
            }
        } catch (IOException e) {
            stdlog_.error("Error writing to file: \n" + e.getMessage());
        }
    }

    /**
     * Add data from a Paramfile object to this tree. The data is placed in the
     * tree directly under the root.
     * 
     * @param data
     *            Paramfile
     */
    public void addParamData(Paramfile data) {
        // cast null to specify method signature.
        addParamData(data, (GTreePath) null);
    }

    /**
     * Add data from a Paramfile object to this tree. The data is placed in the
     * tree under the specified path, or directly under the root if the path is
     * empty or null. The path is expected to be a valid ParamfileTreePath path.
     * 
     * @param data
     *            Paramfile
     * @param path
     *            String
     */
    public void addParamData(Paramfile data, String path) {
        addParamData(data, new ParamfileTreePath(path));
    }

    /**
     * Add data from a Paramfile object to this tree. The data is placed in the
     * tree under the specified path, or directly under the root if the path is
     * empty or null.
     * <P>
     * The "invisible" root node will be prepended to the path and so does need
     * to contain the root node.
     * 
     * @param data
     *            Paramfile
     * @param path
     *            GTreePath
     */
    protected void addParamData(Paramfile data, GTreePath path) {
        DefaultGTreeModel baseTree = null;
        if (path != null && path.getPath().length > 0) {
            GTreePath p = prependNode(tree_.getRoot(), path);
            tree_.addPath(p);
            GTreeNode base = tree_.getNode(p);
            baseTree = new DefaultGTreeModel(base);
        } else {
            // add to the root.
            baseTree = tree_;
        }
        String[] keys = data.getKeywords();
        for (int j = 0; j < keys.length; j++) {
            String k = (String) keys[j];
            String v = (String) data.getValue(k);
            // note the return code is ignored...it is assumed the keywords
            // are correct syntax because they were parsed from a Paramfile
            // data source.
            setValue(baseTree, k, v);
        }
    }

    /**
     * Add data from a PafFile object to this tree. The data is placed in the
     * tree directly under the root.
     * 
     * @param data
     *            PafFile
     */
    public void addPafData(PafFile data) {
        // cast null to specify method signature.
        addPafData(data, (GTreePath) null);
    }

    /**
     * Add data from a PafFile object to this tree. The data is placed in the
     * tree under the specified path, or directly under the root if the path is
     * empty or null.
     * <P>
     * The "invisible" root node will be prepended to the path and so does need
     * to contain the root node.
     * 
     * @param data
     *            PafFile
     * @param path
     *            String
     */
    protected void addPafData(PafFile data, String path) {
        addPafData(data, new ParamfileTreePath(path));
    }

    /**
     * Method addPafData
     * 
     * @param data
     *            PafFile
     * @param path
     *            GTreePath
     */
    protected void addPafData(PafFile data, GTreePath path) {
        DefaultGTreeModel baseTree = null;
        if (path != null && path.getPath().length > 0) {
            GTreePath p = prependNode(tree_.getRoot(), path);
            tree_.addPath(p);
            GTreeNode base = tree_.getNode(p);
            baseTree = new DefaultGTreeModel(base);
        } else {
            // add to the root.
            baseTree = tree_;
        }
        PafData[] keys = data.getIdentifierList();
        for (int j = 0; j < keys.length; j++) {
            String k = (String) keys[j].getIdentifier();
            String v = (String) keys[j].getValue();
            // note the return code is ignored...it is assumed the keywords
            // are correct syntax because they were parsed from a Paramfile
            // data source.
            setValue(baseTree, k, v);
        }
    }

    /**
     * Prepend a node to a path and return the new path. This is a utility
     * function to handle the OHSConfig's "invisible root".
     * 
     * @param node
     *            GTreeNode
     * @param path
     *            GTreePath
     * @return GTreePath
     */
    protected static GTreePath prependNode(GTreeNode node, GTreePath path) {
        GTreePath rc;
        GTreeNode[] pathNodes = new GTreeNode[path.getPath().length + 1];
        pathNodes[0] = node;
        System
                .arraycopy(path.getPath(), 0, pathNodes, 1,
                        path.getPath().length);
        rc = new DefaultGTreePath(pathNodes);
        return rc;
    }

    /**
     * Add/overwrite a keyword and value pair to the tree.
     * 
     * @param keyword
     *            String
     * @param value
     *            String
     * @return false if the keyword syntax is invalid and the path could not be
     *         added, else true. Note that the path will have the "invisible"
     *         root prepended before searching.
     */
    public boolean setValue(String keyword, String value) {
        return setValue(tree_, keyword, value);
    }

    /**
     * Add a keyword and value pair to the root of the given tree. The tree is
     * assumed to be a subtree of the OHSConfig data tree. If the tree is null,
     * the OHSConfig data tree is used.
     * 
     * @param subtree
     *            DefaultGTreeModel
     * @param keyword
     *            String
     * @param value
     *            Object
     * @return false if the keyword syntax is invalid and the path could not be
     *         added, else true. Note that the path will have the "invisible"
     *         root prepended before searching.
     */
    public boolean setValue(DefaultGTreeModel subtree, String keyword,
            Object value) {
        boolean rc = true;
        ParamfileTreePath path = new ParamfileTreePath(keyword);
        StringBuffer sb = new StringBuffer();

        if (!path.isValidPath()) {
            rc = false;
        } else {
            GTreeNode[] pathNode = path.getPath();
            int numOfNodes = pathNode.length;
            int i = 0;

            if (pathNode[i].getValue().equalsIgnoreCase(OHSKeys.platform)) {
                i += 1;
                if (i < numOfNodes) {
                    if (pathNode[i].getValue().equals(
                            getSystemProperty("os.name"))) {
                        i += 1;
                        if (i < numOfNodes) {
                            while (i < numOfNodes) {
                                path = new ParamfileTreePath(pathNode[i]
                                        .getValue());
                                sb.append("." + pathNode[i].getValue());
                                i++;
                            }
                            path = new ParamfileTreePath(sb.toString());
                            setValue(subtree, path, value);
                            rc = true;
                        }
                    } else {
                        rc = false;
                    }
                }
            } else {
                setValue(subtree, path, value);
                rc = true;
            }
        }
        return rc;
    }

    /**
     * Get System Property *
     * 
     * @param property
     *            String
     * @return String
     */
    public String getSystemProperty(String property) {
        return System.getProperty(property).replaceAll(" ", "");
    }

    /**
     * Add/overwrite an integer keyword and value pair to the tree.
     * 
     * @param keyword
     *            String
     * @param value
     *            int
     * @return false if the keyword syntax is invalid and the path could not be
     *         added, else true. Note that the path will have the "invisible"
     *         root prepended before searching.
     */
    public boolean setValue(String keyword, int value) {
        return setValue(keyword, "" + value);
    }

    /**
     * Add/overwrite an integer keyword and value pair to the tree.
     * 
     * @param keyword
     *            String
     * @param value
     *            boolean
     * @return false if the keyword syntax is invalid and the path could not be
     *         added, else true. Note that the path will have the "invisible"
     *         root prepended before searching.
     */
    public boolean setValue(String keyword, boolean value) {
        String val = value ? "true" : "false";
        return setValue(keyword, val);
    }

    /**
     * Add/overwrite an integer keyword and value pair to the tree.
     * 
     * @param keyword
     *            String
     * @param value
     *            Object
     * @return false if the keyword syntax is invalid and the path could not be
     *         added, else true. Note that the path will have the "invisible"
     *         root prepended before searching.
     */
    public boolean set(String keyword, Object value) {
        return setValue(tree_, keyword, value);
    }

    /**
     * Add a keyword and value pair to the root of the given tree. The tree is
     * assumed to be a subtree of the OHSConfig data tree. If the tree is null,
     * the OHSConfig data tree is used.
     * 
     * @param subtree
     *            DefaultGTreeModel
     * @param path
     *            GTreePath
     * @param value
     *            Object
     */

    public void setValue(DefaultGTreeModel subtree, GTreePath path, Object value) {
        GTreePath pth = prependNode(subtree.getRoot(), path);
        subtree.addPath(pth);
        GTreeNode node = subtree.getNode(pth);
        node.set(value);
    }

    /**
     * Remove a keyword from the tree. If the keyword is a node with children,
     * the node will not be removed, but the nodes value will be set to null. A
     * tree node will only be removed if all of its children have been removed.
     * Note that the path will have the "invisible" root prepended before
     * searching.
     * 
     * @param keyword
     *            String
     */
    public void removeKeyword(String keyword) {
        ParamfileTreePath path = new ParamfileTreePath(keyword);
        GTreeNode node = getNode(path);
        if (node != null) {
            node.set(null);
            GTreeNode parent = node.getParent();
            if (parent != null) {
                parent.removeChild(node);
            }
        }
    }

    /**
     * Return the map containig the {keyword,value} pair If the starting node is
     * null all the OHSConfig will be mapped
     * 
     * If the node doesn't exist, the map is empty
     * 
     * @param startSubTreeNode
     *            node from which retrieve the data
     * @return the map
     */
    public Map<String, Object> subTreeToMap(String startSubTreeNode) {
        GTreeNode node = null;
        if (startSubTreeNode == null) {
            node = tree_.getRoot();
        } else {
            ParamfileTreePath path = new ParamfileTreePath(startSubTreeNode);
            node = getNode(path);
        }
        Map<String, Object> map = new HashMap<String, Object>();
        if (node != null) {
            map = saveSubTreeToMap(map, node);
        }
        return map;
    }

    /**
     * 
     * Return the String array containig the children values. If the node
     * doesn't exist or is null, the array has length 0
     * 
     * @param startSubTreeNode
     * @return String[]
     */
    public String[] subTreeToStringArray(String startSubTreeNode) {
        String[] stringArray = new String[0];
        if (startSubTreeNode != null) {
            Map<String, Object> stringMap = subTreeToMap(startSubTreeNode);
            Collection<Object> stringList = stringMap.values();
            Iterator<Object> cursor = stringList.iterator();
            stringArray = new String[stringList.size()];
            int i = 0;
            while (cursor.hasNext()) {
                String errString = (String) cursor.next();
                stringArray[i] = errString;
                i++;
            }
        }
        return stringArray;
    }

    /**
     * Save OHSConfig subtree to PafFile If the starting node is null all the
     * OHSConfig will be saved to file. If the node doesn't exist, nothing is
     * saved.
     * 
     * @param fileName
     * @param startSubTreeNode
     *            node from which save the data
     * @return the node from which it starts to save
     */
    public GTreeNode saveSubTreeToPafFile(String fileName,
            String startSubTreeNode) {
        GTreeNode node = null;
        if (startSubTreeNode == null) {
            node = tree_.getRoot();
        } else {
            ParamfileTreePath path = new ParamfileTreePath(startSubTreeNode);
            node = getNode(path);
        }
        StringBuffer dataToBeSaved = new StringBuffer();
        if (node != null) {
            dataToBeSaved = saveSubTreeToString(dataToBeSaved, node);
        }
        writeConfigFile(fileName, dataToBeSaved);
        return node;
    }

    /**
     * Writes the modified buffer (after the removing of the option) to the
     * helper config file.
     * 
     * @param fileName
     *            String
     * @param newFileConfigBuffer
     *            buffer to be writed. If is null the file is not touch.
     */
    private void writeConfigFile(String fileName,
            StringBuffer newFileConfigBuffer) {
        if (fileName == null) {
            stdlog_.error("Config file name is null");
        } else {
            try {
                if (newFileConfigBuffer != null) {
                    PrintWriter out = new PrintWriter(new FileWriter(fileName));
                    String body = FILE_HDR + "#!--- "
                            + TextUtils.currentISODateTime() + " (UT)\n\n"
                            + newFileConfigBuffer.toString();
                    out.println(body);
                    out.close();
                }
            } catch (IOException e) {
                stdlog_.error("File \"" + fileName + "\" not writed", e);
            }
        }
    }

    /**
     * Save OHSConfig subtree to String in recursive way
     * 
     * @param map
     *            the pair key, value are appended
     * @param node
     *            starting node
     * @return the buffer of data
     */
    protected Map<String, Object> saveSubTreeToMap(Map<String, Object> map, GTreeNode node) {
        if (node != null) {
            Object value = node.getValue();
            // the node is only an intermidiate node (no values, it isn't an
            // entry)
            if (value != null && node.isLeaf()) {
                String nodePath = node.getPathName();
                map.put(nodePath, value);
            }
            for (Enumeration<GTreeNode> e = node.getChildren(); e.hasMoreElements();) {
                GTreeNode nextChild = e.nextElement();
                map = saveSubTreeToMap(map, nextChild);
            }
        }
        return map;
    }

    /**
     * Save OHSConfig subtree to String in recursive way
     * 
     * @param dataToBeSaved
     *            the pair key, value are appended
     * @param node
     *            starting node
     * @return the buffer of data
     */
    public StringBuffer saveSubTreeToString(StringBuffer dataToBeSaved,
            GTreeNode node) {
        if (dataToBeSaved == null) {
            dataToBeSaved = new StringBuffer();
        }
        if (node != null) {

            Object value = node.getValue();
            // the node is only an intermidiate node (no values, it isn't an
            // entry)
            if (value != null && node.isLeaf()) {
                String nodePath = node.getPathName();
                dataToBeSaved.append(nodePath);
                dataToBeSaved.append("  \"");
                dataToBeSaved.append(value);
                dataToBeSaved.append("\"\n");
            }

            for (Enumeration<GTreeNode> e = node.getChildren(); e.hasMoreElements();) {
                GTreeNode nextChild = e.nextElement();
                dataToBeSaved = saveSubTreeToString(dataToBeSaved, nextChild);
            }

        }
        return dataToBeSaved;
    }

    /**
     * Remove a subtree from the tree node. If the keyword is a node with
     * children, the node will be removed with all the children (as rm -rf)
     * 
     * Note that the path will have the "invisible" root prepended before
     * searching.
     * 
     * @param keyword
     *            String
     */
    public void removeSubTree(String keyword) {
        ParamfileTreePath path = new ParamfileTreePath(keyword);
        GTreeNode node = getNode(path);
        if (node != null) {
            removeNodeChildren(node);
        }
    }

    /**
     * Remove recursively a subtree
     * 
     * @param node
     *            GTreeNode
     */
    public void removeNodeChildren(GTreeNode node) {
        if (node != null) {
            for (Enumeration<GTreeNode> e = node.getChildren(); e.hasMoreElements();) {
                GTreeNode nextChild = e.nextElement();
                removeNodeChildren(nextChild);
            }
            node.set(null);
            GTreeNode parent = node.getParent();
            parent.removeChild(node);
        }
    }

    /**
     * @param keyword
     *            String
     * @return the tree node addressed by the keyword. Or null if not found or
     *         not a valid keyword. Note that the path will have the "invisible"
     *         root prepended before searching.
     */
    public GTreeNode getNode(String keyword) {
        ParamfileTreePath path = new ParamfileTreePath(keyword);
        GTreeNode rc = null;
        if (path.isValidPath()) {
            rc = getNode(path);
        } else {
            rc = null;
        }
        return rc;
    }

    /**
     * @param path
     *            GTreePath
     * @return the tree node addressed by the path. Or null if not found. Note
     *         that the path will have the "invisible" root prepended before
     *         searching.
     */
    public GTreeNode getNode(GTreePath path) {
        path = prependNode(tree_.getRoot(), path);
        return tree_.getNode(path);
    }

    /**
     * @param keyword
     *            String
     * @return the value (as a String) corresponding to the keyword, or "" if
     *         not found or invalid keyword.
     */
    public String getValue(String keyword) {
        String rc = "";
        GTreeNode node = getNode(keyword);
        if (node != null && node.getValue() != null) {
            rc = node.getValue();
        }
        return rc;
    }

    /**
     * @param keyword
     *            String
     * @param separator
     *            the
     * @return the value (as a String) corresponding to the keyword, or "" if
     *         not found or invalid keyword.
     */
    /**
     * @param keyword
     * @param separator
     * @return String[]
     */
    public String[] getValueStringArray(String keyword, String separator) {
        String rc = "";
        String[] valueArray = new String[0];
        GTreeNode node = getNode(keyword);
        if (node != null && node.getValue() != null) {
            rc = node.getValue();
        }
        if (rc != null) {
            valueArray = TextUtils.stringToStringArray(rc, separator, true);
        }
        return valueArray;
    }

    /**
     * @param keyword
     *            String
     * @return the value as an integer. throw NumberFormatException if not found
     *         or not an int.
     * @throws NumberFormatException
     */
    public int getInteger(String keyword) throws NumberFormatException {
        String val = getValue(keyword);
        return Integer.parseInt(val);
    }

    /**
     * @param keyword
     *            String
     * @param defaultVal
     *            int
     * @return the value as an integer, if not found or not an int, returns the
     *         default provided..
     */
    public int getInteger(String keyword, int defaultVal) {
        int rc = defaultVal;
        try {
            rc = getInteger(keyword);
        } catch (NumberFormatException e) {
            rc = defaultVal;
        }
        return rc;
    }

    /**
     * @param keyword
     *            String
     * @return true if the value is "T", "Y", "1", or "true" else false
     *         Comparison is NOT case-sensitive.
     */
    public boolean getBoolean(String keyword) {
        String val = getValue(keyword);
        return (val.equalsIgnoreCase("T") || val.equalsIgnoreCase("Y")
                || val.equalsIgnoreCase("TRUE") || val.equals("1"));
    }

    /**
     * @param keyword
     *            String
     * @return the value object corresponding to the keyword, or null if not
     *         found or invalid keyword.
     */
    public Object get(String keyword) {
        Object rc = null;
        GTreeNode node = getNode(keyword);
        if (node != null) {
            rc = node.get();
        }
        return rc;
    }

    /**
     * @param keyword
     *            String
     * @return the subtree addressed by the keyword as an OHSConfig, or null if
     *         not found. This method does not change the singleton instance
     *         setting of OHSConfig.
     */
    public OHSConfig getSubConfig(String keyword) {
        OHSConfig rc = null;
        GTreeNode node = getNode(keyword);
        if (node != null) {
            rc = new OHSConfig(node);
        }
        return rc;
    }

    /**
     * @return the super-tree of this OHSConfig, as an OHSConfig or null if
     *         there is no super-tree. This method does not change the singleton
     *         instance setting of OHSConfig.
     */
    public OHSConfig getSuperConfig() {
        OHSConfig rc = null;
        GTreeNode node = tree_.getRoot().getParent();
        if (node != null) {
            rc = new OHSConfig(node);
        }
        return rc;
    }

    /**
     * @return the tree data model used in this OHSConfig/
     */
    public GTreeModel getTree() {
        return tree_;
    }

    /**
     * Return the tree toString().
     * 
     * @return String
     */
    public String toString() {
        return tree_.toString();
    }

    // A M Chavan, 28-Mar-2003: moved up from subclass.
    /**
     * Method getBaseDir
     * 
     * @return String
     */
    public String getBaseDir() {
        return getValue(OHSKeys.BaseDirKey);
    }

    // A M Chavan, 28-Mar-2003: moved up from subclass.
    /**
     * Method getHomeDir
     * 
     * @return String
     */
    public String getHomeDir() {
        return getValue(OHSKeys.HomeDirKey);
    }

    // A M Chavan, 28-Mar-2003: moved up from subclass.
    /**
     * Method getConfigDir
     * 
     * @return String
     */
    public String getConfigDir() {
        return getValue(OHSKeys.ConfigDirKey);
    }

    /**
     * Process a keyword and compute the absolute pathname of its value.
     * 
     * @param key
     *            Keyword whose value we must process.
     * @return Full pathname of the file specified by the keyword. If the file
     *         is not an absolute path the a path is prepended to the file name.
     *         The prepended path will be the BaseDir setting, unless this key
     *         has been over ridden in the UserOverrideKey subconfig, in which
     *         case the HomeDir setting will be prepended.<BR>
     *         Note that Windows-style pathnames must be specified with double
     *         backslashes; e. g. <BR>
     *         <code>INSTRUMENTS.FOLDER&nbsp;&nbsp;&nbsp;"Z:\\tmp\\instruments"</code>
     * 
     */

    // A M Chavan, 28-Mar-2003: changed Javadoc, added support for "~" (tilde)
    // char representing the user's home directory;
    // commented, reformatted.
    // Moved up from subclass.
    protected String prependBase(String key) {

        String pathname = getValue(key);
        String overrideFile = getValue(OHSKeys.UserOverrideKey + "." + key);
        boolean overridden = (overrideFile.length() > 0);
        if (overridden)
            pathname = overrideFile;

        String ret = null; // this is what we return

        if (pathname != null && pathname.length() > 0) {

            // If the file pathname begins with "~/" we substitute
            // the first char (that is, the tilde) with the user's
            // home directory.
            if (pathname.indexOf("~/") == 0) {

                pathname = pathname.substring(1);
                pathname = getHomeDir() + pathname;
            }

            File f = new File(pathname);
            if (!f.isAbsolute()) {

                // The pathname is not absolute.
                // We prefix it with the base directory or
                // the user's home directory,
                // depending on whether the keyword was
                // defined in the installation's config
                // files or in the user's config file.
                String dir = null;
                if (overridden)
                    dir = getHomeDir();
                else
                    dir = getBaseDir();

                f = new File(dir + File.separator + pathname);
            }
            ret = f.getAbsolutePath();
        }

        return ret;
    }

    /**
     * Get the value of key keyName from this config data, and locate that value
     * of that property in the given paramfile. For example, if keyName is
     * TARGET.RA.PARAM, get that value of TARGET.RA.PARAM, and if a value is
     * found, use the returned value as the name of a key to locate in the
     * paramfile. The keyName is a base name which is search is evaluated
     * follows: 1. eval: keyName (e.g. TARGET.RA.PARAM) 2. if not evauated, or
     * no value, eval: keyName+".1" (e.g. TARGET. RA.PARAM.1) 3. if step 2
     * evaluated, but no value in paramfile, eval: keyName+". 2" (e.g. TARGET.
     * RA. PARAM.2) etc. 4. if step 3 evaluated, but no value in paramfile,
     * eval: keyName+". 3" (e.g. TARGET. RA. PARAM.3) 5. etc
     * 
     * @param keyName
     *            the base name of the key to use to get the value from the
     *            paramfile.
     * @param pf
     *            the Paramfile of data
     * @return String the value found, or "" if not found or the key could not
     *         be evaluated.
     */
    public String evalPropertyFromParamfile(String keyName, Paramfile pf) {
        String rc = "";
        String keyValue = getValue(keyName);
        if (keyValue.length() > 0) {
            rc = pf.getValue(keyValue);
        }
        if (rc.length() == 0) {
            int i = 1;
            // key did not evaluate or value not found, try using
            // the key as a base with number suffixes.
            do {
                keyValue = getValue(keyName + "." + i);
                i++;
                if (keyValue.length() > 0) {
                    rc = pf.getValue(keyValue);
                }
            } while (rc.length() == 0 && keyValue.length() > 0);
        }
        return rc;
    }

}
