package org.eso.ohs.core.utilities;

/**
 * The PAFTable class is a PAF data file in memory.
 * It may be build from a PAF format data string.
 * Currently contains no means to output PAF format strings.
 * This class is currently an "Adapter" to org.eso.ohs.core.utilities.PafFile.
 * @author Nick Kornweibel
 */

import java.io.IOException;
import java.io.StringReader;
import java.util.Vector;

import org.apache.log4j.Logger;

public class PAFTable {

    /**
     * Construct an empty PAF table
     */

    private Paramfile paramFile_ = null;
    private boolean dataValid_ = false;
    private String pafData_;

    /**
     * Standard log channel
     */
    private static Logger stdlog_ = Logger.getLogger(PAFTable.class);

    public PAFTable() {
    }

    /**
     * Initialize this PAF table from the PAF format string provided.
     * @param String PAF format string containing the data
     * @return boolean true if successfully parsed, else false
     */

    public boolean loadFromString(String pafData) {
        try {
            paramFile_ = new Paramfile(new StringReader(pafData), 100000);
            dataValid_ = true;
            pafData_ = pafData;
        } catch (IOException ex) {
            stdlog_.debug("Paramfile load from string error: " + ex.toString());
            dataValid_ = false;
        } catch (NotASCIIException ex) {
            stdlog_.debug("Paramfile load from string error: " + ex.toString());
            // @author amchavan, 23-Aug-2001
            dataValid_ = false;
        }
        return dataValid_;
    }
    public boolean checkCRC() {
        String keywordChksum = "PAF.CHCK.CHECKSUM";
        String pafChksum = paramFile_.getValue(keywordChksum);
        long readChkSum = CRC32PAFFile.getChecksum(new StringBuffer(pafData_));
        if (readChkSum == new Long(pafChksum).longValue())
            return true;
        return false;
    }

    /**
     * Get a StringSet of the parameters in this PAF table
     * if not PAF data was found, the array will be empty
     * @return StringSet of Strings of parameter names
     */

    public StringSet getParameters() {
        Vector<String> rc = new Vector<String>();
        if (dataValid_) {
            String[] params = paramFile_.getKeywords();

            for (int i = 0; i < params.length; i++) {
                rc.addElement(params[i]);
            }
        }
        return new StringSet(rc);
    }

    /**
     * Get a StringSet of Strings of the values in this PAF table
     * if not PAF data was found, the array will be empty
     * @return StringSet of parameter values
     */

    public StringSet getValues() {
        Vector<String> rc = new Vector<String>();
        if (dataValid_) {
            String[] params = paramFile_.getKeywords();

            for (int i = 0; i < params.length; i++) {
                rc.addElement(paramFile_.getValue(params[i]));
            }
        }
        return new StringSet(rc);
    }

    /**
     * Get the values matching this parameter - parameters do not
     * need to be unique in PAF-ese.
     * @param the name of the parameter
     * @return the values, or an empty StringSet if none located
     */

    public StringSet getValuesForParam(String parameter) {
        Vector<String> rc = new Vector<String>();
        String string = null;
        if (dataValid_) {
            string = paramFile_.getValue(parameter);
        } 
        if (string!= null) {
            rc.addElement(string);
        }
        return new StringSet(rc);
    }

    /**
     * Get the number of parameters stored in this PAF table
     * @return the number of parameters, or 0 if not PAF data
     */

    public int getParameterCount() {
        int rc = 0;
        if (dataValid_) {
            rc = paramFile_.getKeywords().length;
        }
        return rc;
    }

    /**
     * Add a parameter/value pair to the PAF table
     * @param param the name of the parameter - must be non-null.
     * @param val the value of the parameter - must be non-null.
     */

    public void addParameter(String param, String val) {
        throw new RuntimeException("PAFTable.addParameter() not supported");
    }
}
