package org.eso.ohs.core.utilities;

import gnu.regexp.RE;
import gnu.regexp.REException;
import gnu.regexp.REMatch;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.text.CharacterIterator;
import java.text.StringCharacterIterator;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Set;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;

/**

 This class parses a keyword/value ASCII file, whose syntax is similar
 to that of Parameter (PAF) files.
 <P>

 Parsing of non-ASCII characters is not currently supported: if a
 non-ASCII character is found, parsing terminates and an exception is
 thrown (note that for performance reasons the 'non-ASCII' test is
 applied only to the first few hundred chars).
 <P>

 Keywords found in the file are associated to their value,
 and made available through an API.
 <P>

 Note that: (a) there are no mandatory keywords and (b) lines whose
 syntax is not recognized are silently ignored (i.e., no
 complaints). Also, the input file gets read only once (by the class'
 constructor) and the input text is not kept.
 <P>

 Parsing of the input file is performed according to the following rules:

	<UL>
	  <LI>
	    parsing is line based, and ends after the first <em>n</em>
	    lines (where <em>n</em> is the value of
	    <code>defaultMaxNumLines_</code>);
	  <LI>
	    valid lines include a keyword and a value separated by
	    whitespace;
	  <LI>
	    a line's leading and trailing whitespace is ignored;
	  <LI>
	    the keyword is a sequence of words separated
	    by dots (period charachters);
	    a word is any sequence of alphanumeric characters
	    (mixed case allowed);
	  <LI>
	    a value may be:
	    <OL>
	      <LI>
		a single word (i.e., a sequence of ASCII chars with no
		intervening whitespace nor semicolons), possibly
		enclosed in double-quotes or terminated by a
		semicolon character;
	      <LI>
	      <UL>
	         <LI>
		   a sequence of words enclosed in double-quotes,
		   in which case
		   anything after the closing double-quote char is
		   ignored;
		 <LI>
		   if only the leading  double-quotes are found,
		   an exception is thrown;
		 <LI>
		   if no double-quotes are found,
		   the sequence of words
		   extends to the first semicolon character, or to the
		   end of the line if no semicolon is found;
         </UL>
	    </OL>

	  <LI>
	    everything else on the line (after the value) is ignored;
	  <LI>
	    non valid lines are ignored.
	</UL>
	<P>

	According to these rules, all of the following lines are
	valid, and associate the keyword INS.MASK.ID to the value
	1234:
	<P>
	<code>
	  INS.MASK.ID&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;1234<BR>
	  INS.MASK.ID 1234;<BR>
	  INS.MASK.ID "1234"<BR>
	  INS.MASK.ID "1234"; # comment<BR>
	</code>
	<P>

	The following lines are also valid, and associate the keyword
	TEL.TARG.NAME to the value "Eta Carinae":
	<P>
	<code>
	  TEL.TARG.NAME Eta Carinae<BR>
	  TEL.TARG.NAME "Eta Carinae"<BR>
	  TEL.TARG.NAME "Eta Carinae"; # comment<BR>
	  TEL.TARG.NAME "Eta Carinae" ignored junk<BR>
	  TEL.TARG.NAME  Eta Carinae; ignored junk<BR>
	</code>
	<P>

	The following lines are not valid and an exception will be thrown:
	<P>
	<code>
	    TEL.TARG.NAME "Eta Carinae<BR>
	</code>

	<P>

	The following lines are not valid and will be ignored:
	<P>
	<code>
	    PAF.HDR.START;<BR>
	    %ADD25R,11.575X0.526*%<BR>
	    G04_______________SLIT_CUTTING*<BR>
	</code>
	<P>

	<strong>Note</strong>: when parsing the a paramfile line, the backslash char
    '<code>\</code>' is interpreted according to the usual Unix conventions. If
    a value contains backslashes, as in the case of Windows-style
    pathnames, those backslash chars must be escaped (with another backslash);
    e.g.
    <p>
    <code>
        CACHE.FOLDER "C:\\tmp\\p2pp-cache"<BR>
    </code>

 * @version $Revision: 1.12.4.1 $
 * @author A M Chavan, 03-Aug-2001: creation.
 */

// $Id: Paramfile.java,v 1.12.4.1 2007/07/20 11:57:39 pnunes Exp $

public class Paramfile {

    /** New-line */
    private static final String NL = "\n";

    /** Empty string */
    private static final String EMPTY = "";

    /** Compiled regular expression matching the keyword of a
    paramfile line.
     */
    protected static RE reKeyword_ = null;

    /** Default max number of paramfile lines to be parsed. */
    protected static final int defaultMaxNumLines_ = 100;

    /**
     * We scan the first asciiTestLength_ chars of the input text: if
     * any non-ASCII char is found among those, an exception is
     * thrown.
     *
     * <br>ChangeLog: amchavan, 22-Aug-2001: created.
     */
    private static final int asciiTestLength_ = 256;

    /** Max number of paramfile lines to be parsed. */
    protected int maxNumLines_;

    /** Registry of keyword/value pairs */
    protected HashMap<String, String> kvPairs_;

    /**
     * Original pathname of the paramfile
     * <br>ChangeLog: A M Chavan, 17-Aug-2001: creation
     */
    protected String pathname_ = "<unknown>";

    /** Standard log channel */
    private static Logger stdlog_ = Logger.getLogger(Paramfile.class);

    /**
     * Create an empty param file object.
     */
    public Paramfile() {
        kvPairs_ = new LinkedHashMap<String, String>();
        if (reKeyword_ == null)
            initRegExps();
    }

    /**
     * Parse the first lines of an input file and extract all the
     * keyword/value pairs we can recognize.
     * Will use the default max number of paramfile lines.
     *
     * @param filename    Pathname of the file to parse
     * @see #defaultMaxNumLines_
     *
     * @throws IOException       If reading the file generated some I/O error.
     * @throws NotASCIIException If file contains non-ASCII data,
     *                           including any 8-bit character, or another
     *                           syntax error.
     */
    public Paramfile(String filename) throws IOException, NotASCIIException {
        this(filename, defaultMaxNumLines_);
    }

    /**
     * Parse the first lines of an input file and extract all the
     * keyword/value pairs we can recognize.
     * Will use the default max number of paramfile lines.
     *
     * @param Reader  text reader with input to be parsed
     * @see #defaultMaxNumLines_
     *
     * @throws IOException       If reading the file generated some I/O error.
     * @throws NotASCIIException If file contains non-ASCII data,
     *                           including any 8-bit character, or another
     *                           syntax error.
     */
    public Paramfile(Reader rdr) throws IOException, NotASCIIException {
        this(rdr, defaultMaxNumLines_);
    }

    /**
     * Parse the first lines of an input file and extract all the
     * keyword/value pairs we can recognize.
     *
     * @param rdr         A reader for the file to parse
     * @param maxNumLines Max number of paramfile lines to be parsed.
     *
     * @throws IOException       If reading the file generated some I/O error.
     * @throws NotASCIIException If file contains non-ASCII data,
     *                           including any 8-bit character, or another
     *                           syntax error.
     */
    public Paramfile(Reader rdr, int maxlines)
        throws IOException, NotASCIIException {

        this();

        // Initialize this class' fields
        maxNumLines_ = maxlines;

        // Read in and parse input stream
        String inputtext = FileIOUtils.fileToString( rdr );
        matchLines(inputtext);
    }

    /**
     * Parse the first lines of an input file and extract all the
     * keyword/value pairs we can recognize.
     *
     * @param filename    Pathname of the file to parse
     * @param maxNumLines Max number of paramfile lines to be parsed.
     *
     * @throws IOException       If reading the file generated some I/O error.
     * @throws NotASCIIException If file contains non-ASCII data,
     *                           including any 8-bit character, or another
     *                           syntax error.
     */
    public Paramfile(String filename, int maxNumLines)
        throws IOException, NotASCIIException {

        this();

        // Initialize this class' fields
        maxNumLines_ = maxNumLines;
        pathname_ = filename;

        // Read in and parse input file
        String inputtext = FileIOUtils.fileToString(filename);
        matchLines(inputtext);
    }

    /** Build all needed regexps */
    protected void initRegExps() {

        // The following patterns assume
        // that the input line is
        // trimmed of all leading white space
        //----------------------------------------

        String pWord = "\\w+";
        String pKeyword = "(" + pWord + "(\\." + pWord + ")*)";
        String pWhite = "\\s";
        String pNoWhite = "\\S";
        String pSomeWhite = pWhite + "+";
        String pAnything = ".*";

        String pKwOnly = pKeyword + pSomeWhite + pAnything;

        // Build all needed regexps
        try {
            reKeyword_ = new RE(pKwOnly);
        } catch (REException e) {
            DebugUtils.stackTrace("Invalid regexp");
        }
    }

    /** @return Max number of paramfile lines to be parsed */
    public int getMaxNumLines() {
        return maxNumLines_;
    }

    /** @return Default max number of paramfile lines to be parsed */
    public int getDefaultMaxNumLines() {
        return defaultMaxNumLines_;
    }

    /**
     *  Return the value associated to a keyword in the input file.
     *  <BR>
     *
     *  <strong>Note</strong>: in the case of multiple identical
     *  keywords, the last value found is returned (later values
     *  override previos values).
     *
     * @param keyword Keyword to look for.
     * @return Value of the input keyword, if found;
     *         the empty string ("") otherwise.
     */
    public String getValue(String keyword) {
        String t = kvPairs_.get(keyword);
        if (t == null)
            return ("");
        return (t);
    }

    /**
     * Read the value associated with a particular ID
     *
     * @param s unique identifier
     *
     * @deprecated This is introduced for backwards compatibility with
     *             the PafFile class.
     *             Use getValue( String keyword ) instead.
     */
    public String readLine(String s) {
        return (getValue(s));
    }

    /**
     * Return the value(s) associated with a given keyword
     * @param s unique identifier.
     * @return String array of keyword values
     *
     * @deprecated This is introduced for backwards compatibility with
     *             the PafFile class.
     *             Use getValue( String keyword ) instead.
     */
    public String[] readKeywordValues(String s) {
        String[] ret = null;
        String val = getValue(s);
        if (val != null) {
            ret = new String[1];
            ret[0] = val;
        } else {
            ret = new String[0];
        }
        return ret;
    }

    /**
     * @return  The list of keywords found in the paramfile
     */
    public String[] getKeywords() {

        Set<String> s = kvPairs_.keySet();
        String[] ret = new String[ s.size() ];
        Iterator<String> iter = s.iterator();
        int i = 0;
        while( iter.hasNext() ) {
            String key = iter.next();
            ret[ i++ ] = key;
        }
        return ret;
    }

    /**
     * @return  The list of keywords ending with suffix found in the paramfile
     */
    public String[] getKeywords(String suffix) {
    	String[] result;
        Set<String> s = kvPairs_.keySet();
        ArrayList<String> tmpList = new ArrayList<String>();
        Iterator<String> iter = s.iterator();
        int i = 0;
        while( iter.hasNext() ) {
            String key = iter.next();
            if (key.endsWith(suffix))
            	tmpList.add(key);
        }
        result = (String[]) tmpList.toArray(new String[tmpList.size()]);
        return result;
    }

    /**
     * @return  The number of keywords found in the paramfile
     */
    public int getKeywordNumber() {

        return (kvPairs_.size());
    }

    /**
     * @return  the list of keywords found in the paramfile
     * @deprecated This is introduced for backwards compatibility with
     *             the PafFile class.
     *             Use getKeywords( ) instead.
     */
    public String[] getIdentifierList() {
        return getKeywords();
    }

    /**
     * If the input parameter is a valid paramfile line, the
     * corresponding keyword/value pair will be added to the internal
     * hashtable
     *
     *  @return True if the input parameter is a
     *          valid paramfile line; false otherwise.
     *
     * @param line    Any string.
     * @param linenum Line counter.
     *
     * @throws NotASCIIException If file contains a syntax error.
     */
    private boolean matchLine(String line, int linenum)
        throws NotASCIIException {

        RE matchingRE = null;
        String encodedLine = null;
        String trimmedLine = line.trim(); // drop leading whitespace
        String keyword = null;
        String value = null;

        // Discard comment lines and empty lines
        //----------------------------------------
        if (trimmedLine.length() == 0 || trimmedLine.charAt(0) == '#')
            return false;

        // Discard all lines which do not start
        // with a keyword
        //----------------------------------------
        if (!reKeyword_.isMatch(trimmedLine))
            return false;

        // Recover the keyord, then drop it
        // with all following whitespace
        //----------------------------------------

        REMatch match = reKeyword_.getMatch(trimmedLine);
        keyword = match.toString(1); // keyword
        int endKeyw = keyword.length();
        trimmedLine = trimmedLine.substring(endKeyw).trim();

        // If the rest of the string does not
        // begin with double-quotes, then that
        // is our value.
        // Otherwise, look for the closing
        // double-quotes, and take whatever
        // is in between as the value.
        //----------------------------------------
        int l = trimmedLine.length();
        int endValue = -1;

        if (trimmedLine.charAt(0) != '\"') {

            // Look for the first non-escaped semicolon
            for (int i = 0; i < l; i++) {
                if (trimmedLine.charAt(i) == ';'
                    && trimmedLine.charAt(i - 1) != '\\') {
                    endValue = i;
                    break;
                }
            }

            // Did we find it?
            if (endValue < 0)
                value = trimmedLine; // NO, we didn't
            else
                value = trimmedLine.substring(0, endValue); // YES, we did
        } else {

            // Skip first double-quotes, then look
            // for the next non-escaped double-quotes
            boolean escapeNext = false;
            for (int i = 1; i < l; i++) {
                /*
                 * nkornwei: The following code is perhaps more correct.
                 */
                char c = trimmedLine.charAt(i);
                if (escapeNext) {
                    escapeNext = false;
                    continue;
                }
                if (c == '\\') {
                    escapeNext = true;
                } else if (c == '\"') {
                    endValue = i;
                    break;
                }

                /*
                 * Original code - replaced with above. (Left here for reference but
                 * can be removed after OHS 2.6).
                 * The code above does not handle escaped escaped characters e.g. \\
                 *
                if (trimmedLine.charAt(i) == '\"'
                    && trimmedLine.charAt(i - 1) != '\\') {
                    endValue = i;
                    break;

                }
                */

            }

            // Did we find it?
            if (endValue < 0) {
                // NO, we didn't: throw an exception
                String errmsg =
                    "No closing double-quote char at line " + linenum;
                throw new NotASCIIException(errmsg);
            }

            value = trimmedLine.substring(1, endValue);
        }

        //kvPairs_.put(keyword, TextUtils.revertEscapedString(value));
        putNotRepeat(keyword,TextUtils.revertEscapedString(value));
        return true;
    }

    /**
     * Populates the internal table of keyword/value pairs, each
     * pair corresponding to a valid keyword/value line in the input
     * string.
     *
     * @param line This string will be broken up in lines, and each
     *             line will then be matched individually.
     *
     * @throws NotASCIIException We scan the first asciiTestLength_
     *                           chars of the input text: if any
     *                           non-ASCII char is found among those,
     *                           this exception is thrown.
     *                           Also in the case of syntax errors.
     *
     * <br>ChangeLog: amchavan, 22-Aug-2001: added.
     *
     * @see #asciiTestLength_
     */
    protected void matchLines(String lines) throws NotASCIIException {

        // Check for non-ASCII chars in the text
        int i = TextUtils.isASCIIPrintable(lines, maxNumLines_);
        if (i >= 0)
            throw new NotASCIIException("at char " + i);

        int numLines = 0;

        // Need to do some magic here, because
        // StringTokenizer does not return empty
        // lines. So, we set the returnTokens
        // flag of the constructor to true, and
        // handle the newline chars ourselves.
        // ----------------------------------------

        StringTokenizer st = new StringTokenizer(lines, "\n", true);
        String line = EMPTY;
        while (st.hasMoreTokens()) {

            String token = st.nextToken();

            if( token.equals( NL )) {       // is it end-of-line?
                numLines++;
                matchLine( line, numLines );
                if (numLines >= maxNumLines_)
                    break;
                line = EMPTY;               //      ...and reset it
                continue;
            }
            line = token;
        }
        if( line != EMPTY )                 // any leftover lines?
            matchLine( line, numLines );    // YES, match that
    }

    /**
     * Utility method for checking if a String is a valid keyword.
     * This method is not used in the parsing of paramfiles in this class,
     * but is included in this class to encapsulate the param file definition.
     *
     * A keyword is a sequence of words separated by dots (".").
     * A word is any sequence of alphanumeric characters (mixed case allowed).
     *
     * <br>ChangeLog: Nick Kornweibel for ESO.
     */

    public static boolean isValidKeyword(String key) {
        boolean rc = true;
        if (key != null && key.length() != 0) {
            for (int i = 0; rc && i < key.length(); i++) {
                char c = key.charAt(i);
                rc = (Character.isLetterOrDigit(c) || c == '.');
            }
        } else {
            rc = false;
        }
        return rc;
    }

    /**
     * Main program, for testing only.<P>
     * Usage: java org.eso.ohs.core.utilities.Paramfile file1 [file2 ...]
     */
    public static void main(String[] args) {

        Paramfile p = null;

        for (int i = 0; i < args.length; i++) {

            try {
                p = new Paramfile(args[i], 100000);
            } catch (Exception e) {
                e.printStackTrace();
            }

            // Dump keyword/value pairs
            String[] keys = p.getKeywords();

            for (int j = 0; j < keys.length; j++) {
                String k = (String) keys[j];
                String v = (String) p.getValue(k);
                System.out.println(
                    j + "/" + p.getKeywordNumber() + " " + k + " " + v);
            }
        }
        p.reportToString();
    }

    public void report() {
        stdlog_.debug(reportToString());
    }

    public String reportToString() {
        StringBuffer rc = new StringBuffer();
        rc.append("=================Identifier List============\n");
        String data[] = getKeywords();
        for (int i = 0; i < data.length; i++) {
            String d = data[i];
            String val = getValue(d);
            rc.append("" + i + "\t<" + d + ">\t\t<" + val + ">\n");
        }
        rc.append("============================================\n");
        return rc.toString();
    }

    public static void report(String filePath) {
        Paramfile f = null;
        try {
            f = new Paramfile(filePath);
        } catch (Exception e) {
            System.out.println("Error reading param file: " + e.toString());
        }
        System.out.println(f.reportToString());
    }

    /**
     * @return Original pathname of the paramfile.
     * <br>ChangeLog: A M Chavan, 17-Aug-2001: creation
     */
    public String getPathname() {
        return pathname_;
    }

    /**
     * @return True if the string s includes whitespace characters,
     *         false otherwise.
     *
     * <br>ChangeLog: amchavan, 13-Sep-2001: creation
     */
    private boolean includesWhitespace(String s) {
        StringCharacterIterator it = new StringCharacterIterator(s);

        for (char c = it.first(); c != CharacterIterator.DONE; c = it.next())
            if (Character.isWhitespace(c))
                return true;

        return false;
    }

    /**
     * Two paramfiles are <em>equal</em> if they have the same set of
     * keywords, and if the corresponding values are also equal.
     *
     * @return True if this paramfile is equal to p; false otherwise.
     *
     * @param  other   Paramfile to compare to this one.
     *
     * <br>ChangeLog: amchavan, 13-Sep-2001: Creation.
     */
    public boolean equals(Paramfile other) {

        String[] ourKeywords = getKeywords();
        String[] otherKeywords = other.getKeywords();

        // Do we have the same number of keywords?
        if (ourKeywords.length != otherKeywords.length)
            return false; // NO, the paramfiles are different

        // Foreach keyword we have, look it up
        // in the other paramfile: if we find it,
        // check if the values are the same
        //----------------------------------------
        for (int i = 0; i < ourKeywords.length; i++) {

            String ourK = ourKeywords[i];
            boolean found = false;

            for (int j = 0; j < otherKeywords.length; j++)
                if (ourK.equals(otherKeywords[j])) {
                    found = true;
                    break;
                }

            // Did we find the i-th keyword in the other
            // paramfile?
            if (!found)
                return false; // NO, the paramfiles are different

            // Are the values equal?
            if (!getValue(ourK).equals(other.getValue(ourK)))
                return false; // NO, the paramfiles are different
        }
        return true;
    }

    /**
     * Generate an ASCII representation of the paramfile object, in a
     * format suitable for further parsing by this class.
     *
     * @return A textual report of the keyword/value pairs of the
     *         <BR>
     *         Note that:
     *         <UL>
     *           <LI> keyword/value pairs are placed on individual
     *                lines;
     *           <LI> a TAB character is used to separate the keyword
     *                from the value;
     *           <LI> values including whitespace characters are
     *                enclosed within double-quotes;
     *           <LI> the last character of the returned string is
     *                a newline char.
     *         </UL>
     *         For instance:
     *         <P>
     *         <code>
     *         PHASE1.DATABASE          opc63<BR>
     *         PHASE2.DATABASE          obrep2<BR>
     *         VALUE.WITH.BLANKS        "A string with blanks"<BR>
     *         </code>
     *
     * <br>ChangeLog: amchavan, 13-Sep-2001: Creation.
     */
    public String toString() {

        StringBuffer b = new StringBuffer();
        String[] keys = getKeywords();

        for (int j = 0; j < keys.length; j++) {

            String k = (String) keys[j];
            String v = getValue(k);
            boolean w = includesWhitespace(v);

            b.append(k);
            b.append('\t');
            if (w)
                b.append('"');
            b.append(v);
            if (w)
                b.append('"');
            b.append('\n');
        }

        return b.toString();
    }

    /**
     * Generate an ASCII representation of the paramfile object, in a
     * format suitable for further parsing by this class.
     *
     * @return The same text returned by <code>toString()</code>,
     *         prepended by the header. Each line of the header is
     *         in turn prepended with a comment (<code>'#'</code>)
     *         character and a space.
     *         For instance:
     *         <P>
     *         <code>
     *         # First line of the header<BR>
     *         # Second line<BR>
     *         PHASE1.DATABASE          opc63<BR>
     *         PHASE2.DATABASE          obrep2<BR>
     *         VALUE.WITH.BLANKS        "A string with blanks"<BR>
     *         </code>
     *
     * @param header  Text of the header. For instance:
     *                <BR>
     *                <code>First line of the header\nSecond line</code>
     *
     * @see #toString()
     *
     * <br>ChangeLog: amchavan, 13-Sep-2001: creation.
     */

    public String toString(String header) {

        if (header == null)
            return toString();

        StringBuffer b = new StringBuffer();
        StringTokenizer st = new StringTokenizer(header, "\n", false);

        // Process and prepend header
        //----------------------------------------
        while (st.hasMoreTokens()) {
            b.append("# ");
            b.append(st.nextToken());
            b.append('\n');
        }

        // Now append the actual paramfile,
        // and return
        //----------------------------------------
        b.append(toString());
        return b.toString();
    }

    /**
     * Generate an ASCII representation of the paramfile object, in a
     * format suitable for further parsing by this class, and write
     * it to a disk file.
     * <P>
     * Writes the same text returned by <code>toString()</code>.
     *
     * @param pathname  Pathname of the output file.
     *
     * @see #toString()
     *
     * <br>ChangeLog: amchavan, 13-Sep-2001: creation.
     */

    public void toFile(String pathname) throws IOException {

        toFile(pathname, null);
    }

    /**
     * Generate an ASCII representation of the paramfile object, in a
     * format suitable for further parsing by this class, and write
     * it to a disk file.
     * <P>
     * Writes the same text returned by the
     * <code>toString(String header)</code> method
     *
     * @param pathname  Pathname of the output file.
     * @param header  Text of the header:
     *
     * @see #toString(String)
     *
     * <br>ChangeLog: amchavan, 13-Sep-2001: creation.
     */

    public void toFile(String pathname, String header) throws IOException {

        FileOutputStream outs = new FileOutputStream(pathname);
        String text = toString(header);

        outs.write(text.getBytes());
        outs.close();
    }

    /**
     * Generate an ASCII representation of the paramfile object, in a
     * format suitable for further parsing by this class, and write
     * it to a a Writer object.
     *
     * Writes the same text returned by the toString() method.
     */

    public void writeTo(Writer out) throws IOException {
        writeTo(out, null);
    }

    public void writeTo(Writer out, String header) throws IOException {
        String text = toString(header);
        out.write(text);
        out.flush();
    }


    /**
     * Add a keyword, value pair to the param set
     */

    public void putNotRepeat(String key, String val) {
    	if (!kvPairs_.containsKey(key))
    		put(key,val);
    }


    /**
     * Add a keyword, value pair to the param set
     */

    public void put(String key, String val) {
        if (key != null
            && key.length() > 0
            && val != null
            && val.length() > 0) {
        		kvPairs_.put(key, val);
        }
    }
}
