package org.eso.ohs.core.utilities;

import java.util.Random;

/**
 * Generates a random String using a cryptographically
 * secure random number generator.
 * <p>
 * The alphabet (characters used in the passwords generated)
 * may be specified, and the random number generator can be
 * externally supplied.
 * <p>
 * Care should be taken when using methods that limit the types
 * of passwords may be generated.  Using an alphabet that is too
 * small, using passwords that are too short, requiring too many
 * of a certain type of character, or not allowing repetition,
 * may decrease security.
 * <p>
 * More information about this class is available from <a target="_top" href=
 * "http://ostermiller.org/utils/RandPass.html">ostermiller.org</a>.
 *
 * @author Stephen Ostermiller http://ostermiller.org/contact.pl?regarding=Java+Utilities
 * @since ostermillerutils 1.00.00
 */
public class RandomStringGenerator {

	/**
	 * Default length for passwords
	 *
	 */
	public static final int DEFAULT_PASSWORD_LENGTH = 6;

	/**
	 * Alphabet consisting of upper and lowercase letters A-Z and
	 * the digits 0-9 but with characters that are often mistaken
	 * for each other when typed removed. (I,L,O,U,V,i,l,o,u,v,0,1)
	 */
	public static final char[] NONCONFUSING_ALPHABET =
		{
			'A',
			'B',
			'C',
			'D',
			'E',
			'F',
			'G',
			'H',
			'J',
			'K',
			'M',
			'N',
			'P',
			'Q',
			'R',
			'S',
			'T',
			'W',
			'X',
			'Y',
			'Z',
			'a',
			'b',
			'c',
			'd',
			'e',
			'f',
			'g',
			'h',
			'j',
			'k',
			'm',
			'n',
			'p',
			'q',
			'r',
			's',
			't',
			'w',
			'x',
			'y',
			'z',
			'2',
			'3',
			'4',
			'5',
			'6',
			'7',
			'8',
			'9',
			};

	/**
	 * Random number generator used.
	 */
	protected Random rand_;


	/**
	 * Create a new random password generator
	 * with the default secure random number generator
	 * and default NONCONFUSING alphabet for all characters.
		 */
	public RandomStringGenerator() {
		this(new Random());
	}

	/**
	 * Create a new random password generator
	 * with the given secure random number generator
	 * and default NONCONFUSING alphabet for all characters.
	 *
	 * @param rand Secure random number generator to use when generating passwords.
	 */
	public RandomStringGenerator(Random rand) {
	   rand_ = rand;
	}


	/**
	 * Set the random number generator used by this random password generator.
	 *
	 * @param rand Secure random number generator to use when generating passwords.
	 */
	public void setRandomGenerator(Random rand) {
		rand_ = rand;
	}


	/**
	 * Fill the given buffer with random characters.
	 * <p>
	 * Using this method, the password character array can easily
	 * be reused for efficiency, or overwritten with new random
	 * characters for security.
	 * <p>
	 * NOTE: If it is possible for a hacker to examine memory to find passwords,
	 * the password should be overwritten in memory as soon as possible after i
	 * is no longer in use.
	 *
	 * @param pass buffer that will hold the password.
	 * @return the buffer, filled with random characters.
	 *
	 * @since ostermillerutils 1.00.00
	 */
	public char[] getPassChars(char[] pass) {
	
			int length = pass.length;
			for (int i = 0; i < length; i++) {
				char[] useAlph = NONCONFUSING_ALPHABET;				
				pass[i] = useAlph[rand_.nextInt(NONCONFUSING_ALPHABET.length)];
			}
		return (pass);
	}

	


	/**
	 * Generate a random password of the given length.
	 * <p>
	 * NOTE: Strings can not be modified.  If it is possible
	 * for a hacker to examine memory to find passwords, getPassChars()
	 * should be used so that the password can be zeroed out of memory
	 * when no longer in use.
	 *
	 * @param length The desired length of the generated password.
	 * @return a random password
	 *
	 */
	public String getPass(int length) {
		return (new String(getPassChars(new char[length])));
	}

	/**
	 * Generate a random password of the default length (6).
	 * <p>
	 * NOTE: Strings can not be modified.  If it is possible
	 * for a hacker to examine memory to find passwords, getPassChars()
	 * should be used so that the password can be zeroed out of memory
	 * when no longer in use.
	 *
	 * @return a random password

	 */
	public String getPass() {
		return (getPass(DEFAULT_PASSWORD_LENGTH));
	}
}
