package org.eso.ohs.dfs;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;

import org.apache.log4j.Logger;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentList;
import org.eso.ohs.instruments.ReadmeAttribute;
import org.eso.ohs.instruments.ReadmeAttributesFactory;

/**
 *
 *
 * <p>
 * $Id: ObservingRun.java,v 1.33.4.1 2007/08/01 12:06:10 pnunes Exp $<br>
 *
 * @author Bob Kemp
 */

public class ObservingRun extends BusinessObject {

    private static Logger stdlog_ = Logger.getLogger(ObservingRun.class);

    /** Version id for object serialization */
    static final long serialVersionUID = -7826695820902491018L;

    // Constants for "Type" property.
    public final static String TYPE_OBSERVING = "observing";

    public final static String TYPE_LARGE = "large";

    public final static String TYPE_DDT = "ddt";

    public final static String TYPE_ENGINEERING = "engineering";

    /**
     * Constants for "ObsMode" property.
     * <p>
     * <b>NOTE: </b> Comparisons MUST ignore case, i.e.&nbsp; use
     * "equalsIgnoreCase()"
     *
     * @see String#equalsIgnoreCase
     */
    public final static String OBSMODE_STANDARD = "S";

    public final static String OBSMODE_VISITOR = "V";

    public final static String OBSMODE_ENGINEERING = "E";

    /**
     * Instrument list. Example: {SUSI} or {SUSI EMMI}
     */
    private String[] instrumentList_;

    /**
     * Mode of the observing run: may be (L)ocal, (V)isitor, (S)ervice or
     * (R)emote.
     */
    private String mode_;

    /**
     * Number of hours assigned by the OPC to the observing run
     */
    private float decimalHours_;

    /**
     * Instrument code. Example: "SUSI"
     */
    private String instCode_;

    /**
     * Full name of the PI
     */

    private String piName_;

    /**
     * GTO Flag
     */

    private int gto_;

    /**
     * The IP version to be used in this run. Example: 72.1 Any value <= 0 is an
     * invalid version number.
     */
    private float ipversion_ = -1;

    /**
     * Observing period. Example: 58
     */
    private int period_;

    /**
     * Rank assigned by the OPC to the observing run
     */
    private int rank_;

    /**
     * Rank Class assigned by the OPC to the observing run
     */
    private String rankClass_;

    /**
     * Telescope code. Example: "NTT"
     */
    private String telescope_;

    /**
     * Title of the observing proposal
     */
    private String title_;

    /**
     * User name
     */
    private String userName_;

    /**
     * Database Id of the observing run
     */
    private long runId_;

    /**
     * Id of Principal Investigator (PI)
     */
    private int userId_;

    /**
     * Programme Id of the observing run
     * <p>
     * NOTE: this is different to the ProgId !!
     */
    private long programmeId_;

    /**
     * ProgId of the observing run.
     * <p>
     * NOTE: this is different to the Programme Id !!
     */
    private String progId_;

    /**
     * Run Description for the observing run ('A', 'B', etc) For example, the
     * 'B' in "265.A-5719(B)".
     */
    private String runDescription_;

    /**
     * The readme file for a given Observing Rum
     */

    private Readme readme;


    private float decimalNights;


    /**
     * A list of checksums that represent finding charts and the number of times
     * each finding chart is used in the system This is needed so we can see
     * when it is appropriate to delete a finding chart.
     */

    //private HashMap fcCountList = new HashMap();
    /**
     * null constructor
     */
    public ObservingRun() {
    }

    /**
     * Is a service mode run.
     *
     * @return Is in service mode
     */
    public boolean isServiceMode() {
        if (getObsMode().equalsIgnoreCase(OBSMODE_STANDARD))
            return true;
        return false;
    }

    /**
     * Get the type of the observing proposal
     *
     * @return the current value of the property
     */
    public String getType() {
        String name = getName();

        if (name.charAt(0) == ' ') {
            int start = name.indexOf('-') + 1;
            int end = name.indexOf('(');
            String propIdStr = name.substring(start, end);
            int proposalId = new Integer(propIdStr).intValue();

            if (proposalId > 9000) {
                return TYPE_ENGINEERING;
            } else {
                return TYPE_OBSERVING;
            }
        } else if (name.charAt(0) == '1') {
            return TYPE_LARGE;
        } else if (name.charAt(0) == '2') {
            return TYPE_DDT;
        } else {
            return null;
        }

    }

    /**
     * Get value for mode_ property
     *
     * @return the current value of the property
     */
    public String getObsMode() {
        return mode_;
    }

    /**
     * Get value for mode_ property
     *
     * @return the current value of the property
     */
    public void setObsMode(String mode) {
        mode_ = mode;
    }




    /**
     * Get value for instCode_ property
     *
     * @return the current value of the property
     */
    public String getInstCode() {
        return instCode_;
    }

    /**
     * Set new value for instCode_ property
     *
     * @param newValue
     *            the new value of the property
     */
    public void setInstCode(String newValue) {
        instCode_ = newValue;
    }

    /**
     * @return the ip version to be used with this observing run. or -1 if not
     *         set.
     */
    public float getIPVersion() {
        return ipversion_;
    }

    /**
     * Set the IP version to be used with this obs run
     */
    public void setIPVersion(float ipversion) {
        ipversion_ = ipversion;
    }

    /**
     * Get value for period_ property
     *
     * @return the current value of the property
     */
    public int getPeriod() {
        return period_;
    }

    /**
     * Get value for period_ property
     *
     * @return the current value of the property
     */
    public void setPeriod(int period) {
        period_ = period;
    }

    /**
     * Get value for rank_ property
     *
     * @return the current value of the property
     */
    public int getRank() {
        return rank_;
    }

    /**
     * Set value for rank_ property
     *
     * @param the
     *            new value of the property
     */
    public void setRank(int rank) {
        rank_ = rank;
    }

    /**
     * Get value for rankClass_ property
     *
     * @return the current value of the property
     */

    public String getRankClass() {
        return rankClass_;
    }

    /**
     * Set value for rankClass_ property
     *
     * @param the
     *            new value of the property
     */
    public void setRankClass(String rankClass) {
        rankClass_ = rankClass;
    }

    /**
     * Get value for runId_ property
     *
     * @return the current value of the property
     */
    public long getRunId() {
        return runId_;
    }

    /**
     * Set value for runId_ property
     *
     * @param the
     *            new value of the property
     */
    public void setRunId(long runId) {
        runId_ = runId;
    }

    /**
     * Get value for programmeId_ property
     *
     * @return the current value of the property
     */
    public long getProgrammeId() {
        return programmeId_;
    }

    /**
     * Set value for programmeId_ property
     *
     * @param the
     *            new value of the property
     */
    public void setProgrammeId(long programmeId) {
        programmeId_ = programmeId;
    }

    /**
     * Get value for progId_ property
     *
     * @return the current value of the property
     */
    public String getProgId() {
        return progId_;
    }

    /**
     * Set value for progId_ property
     *
     * @param the
     *            new value of the property
     */
    public void setProgId(String progId) {
        progId_ = progId;
    }

    /**
     * Get value for runDescription_ property
     *
     * @return the current value of the property
     */
    public String getRunDescription() {
        return runDescription_;
    }

    /**
     * Set value for runDescription_ property
     *
     * @param the
     *            new value of the property
     */
    public void setRunDescription(String runDescription) {
        runDescription_ = runDescription;
    }

    /**
     * Get value for telescope_ property
     *
     * @return the current value of the property
     */
    public String getTelescope() {
        return telescope_;
    }

    /**
     * Get value for telescope_ property
     *
     * @return the current value of the property
     */
    public void setTelescope(String telescope) {
        telescope_ = telescope;
    }

    /**
     * Get value for title_ property
     *
     * @return the current value of the property
     */
    public String getTitle() {
        return title_;
    }

    /**
     * Get value for title_ property
     *
     * @return the current value of the property
     */
    public void setTitle(String title) {
        title_ = title;
    }

    /**
     * Get value for userId_ property
     *
     * @return the current value of the property
     */
    public int getUserId() {
        return userId_;
    }

    /**
     * Set value for userId_ property
     *
     * @param the
     *            new value of the property
     */
    public void setUserId(int userid) {
        userId_ = userid;
    }

    /**
     * Get value for userName_ property
     *
     * @return the current value of the property
     */
    public String getUserName() {
        return userName_;
    }

    /**
     * Set value for userName_ property
     *
     * @param the
     *            new value of the property
     */
    public void setFirstName(String userName) {
        userName_ = userName;
        firePropertyChangeEvent();
    }

    /**
     * Set the pi associated with the run.
     *
     * @param the
     *            new value of the property
     */
    public void setPIName(String piName) {
        piName_ = piName;
        firePropertyChangeEvent();
    }

    /**
     * Set the pi name associated with the run.
     *
     * @param the
     *            new value of the property
     */
    public String getPIName() {
        return piName_;
    }

    /**
     * Returns the string used to refer to this type of BusinessObject in
     * messages to the user.
     *
     * @return user-oriented name
     */
    public static String getDisplayName() {
        return "Observing Run";
    }

    /**
     * Returns the filename suffix for files containing this type of object.
     *
     * @return suffix for filename when stored in local filesystem
     */
    public static String getSuffix() {
        return "or";
    }

    /**
     * Generic hook for "visitor" Design Pattern. See the BusinessVisitor
     * interface for more details.
     *
     * @param visitor
     *            the class whose visit() method we wish to invoke.
     *
     * @see BusinessVisitor
     */
    public Object accept(BusinessVisitor visitor, Object arg) {
        return visitor.visit(this, arg);
    }

    /**
     *
     * /**
     *
     * @see org.eso.ohs.dfs.StorableObject#postProecess()
     */
    public void postProecess() {
    }

    private void readObject(java.io.ObjectInputStream in) throws IOException,
            ClassNotFoundException {
        in.defaultReadObject();
        if (readme != null) {
            ReadmeAttribute[] readmeAttributes = readme.getRedmeAttributes();
            stdlog_.debug("Readme " + readme.getIpVersion());
            if (readmeAttributes != null) {
                Instrument inst = InstrumentList.getInstance().getInstrument(
                        getInstCode(),readme.getIpVersion());
                readmeAttributes = ReadmeAttributesFactory
                        .refreshReadmeAttributes(inst, readmeAttributes);
                readme.setRedmeAttributes(readmeAttributes);
                //readme.setIpVersion(getIPVersion());
            }

        }
    }

    /**
     * Get the readme file defintion for this run
     *
     * @return Returns the readme.
     */
    public Readme getReadme() {
        return readme;
    }

    /**
     * Set the readme file defintion for a run
     *
     * @param readme
     *            The readme to set.
     */
    public void setReadme(Readme readme) {
        this.readme = readme;
        firePropertyChangeEvent();
    }

    /**
     * @return Returns the gto_.
     */
    public int getGto() {
        return gto_;
    }

    /**
     * @param gto_
     *            The gto_ to set.
     */
    public void setGto(int gto) {
        this.gto_ = gto;
        firePropertyChangeEvent();
    }
    public static final Integer TOO_INT = new Integer(3);
    public static final Integer DDT_INT = new Integer(2);
    public static final Integer GTO_INT = new Integer(1);
    public static final Integer NORMAL_INT = new Integer(0);

    public String getProgType() {
        switch (gto_) {
        case 0:
            return "NORMAL";
        case 1:
            return "GTO";
        case 2:
            return "DDT";
        case 3:
            return "TOO";
        case 4:
            return "LARGE";
        case 5:
            return "SHORT";
        default:
            return "UNKNOWN";
        }
    }



    /**
     * Copies the value of the specified property of this object to
     * the given BusinessObject. Special processing for reamdme data
     * it should not be reinirialised from the database on downloasd
     * refresh obsering runs.
     * @param propName      property name to copy
     * @param destBO        destination for the copy operation
     */
    public void copyProperty(String propName, BusinessObject destBO)
        throws
            NoSuchMethodException,
            IllegalAccessException,
            InvocationTargetException {


        if (propName.equals("Readme")) {
            return;
        }
        super.copyProperty(propName, destBO);
    }





    /**
     * @return Returns the nights.
     */
    public float getDecimalNights() {
        return decimalNights;
    }

    /**
     * @param nights
     *            The nights to set.
     */
    public void setDecimalNights(float nights) {
        this.decimalNights = nights;
        firePropertyChangeEvent();
    }
    /**
     * @return Returns the decimalhours_.
     */
    public float getDecimalHours() {
        return decimalHours_;
    }
    /**
     * @param decimalhours_ The decimalhours_ to set.
     */
    public void setDecimalHours(float decimalHours) {
        this.decimalHours_ = decimalHours;
        firePropertyChangeEvent();
    }
}

