package org.eso.ohs.dfs;

import java.util.Hashtable;

import org.eso.ohs.core.docview.datatrans.AdaptorFactory;
import org.eso.ohs.core.docview.datatrans.ObjectAdaptor;
import org.eso.ohs.instruments.TemplateSignature;

/**
 * This class contains the summary information for a single BusinessObject
 * needed by the GUI so that the contents of a folder can be displayed
 * quickly.
 *
 * <p>
 * $Id: Summary.java,v 1.25 2006/04/26 09:26:49 tcanavan Exp $
 * <br>
 * @author Bob Kemp
 */

public class Summary extends SimpleStorableObject {

    /** Version id for object serialization */
    static final long serialVersionUID = -3509021718789997956L;

    /** Does the summarized object pass the verify() tests? */
    private boolean isValid_ = true;

    /** Type of object represented by this summary */
    private Class objectClass_;

    /** Is the summarized object already checked into the database? */
    private boolean isCheckedIn_;

    /**
     * A record of the links from other BusinessObjects to this one.
     */
    //    private Hashtable links = new Hashtable();

    /**
     * Property names and values for the object that is summarized.
     */
    private Hashtable<Object, Object> properties = new Hashtable<Object, Object>();

    private long[] childIds_ = new long[] {
    };

    public Summary() {
    }

    /**
     * Set the type of object that this instance summarizes.
     * 
     * @param objType	    type of object that this summarizes.
     */
    public void setObjectClass(Class cls) {
        objectClass_ = cls;
    }

    /**
     * Get the type of object that this instance summarizes.
     * 
     * @return	type of object that this summarizes.
     */
    public Class getObjectClass() {
        return objectClass_;
    }

    /**
     * Set value for isCheckedIn_ property
     *
     * @param newValue		new value of the property
     */
    public void setCheckedIn(boolean newValue) {
        isCheckedIn_ = newValue;
    }

    /**
     * Get value for isCheckedIn_ property
     *
     * @return the current value of the property
     */
    public boolean isCheckedIn() {
        return isCheckedIn_;
    }

    /**
     * Get value for isValid_ property
     *
     * @return the current value of the property
     */
    public boolean isValid() {
        return isValid_;
    }

    /**
     * Set value for isValid_ property
     *
     * @param newValue		new value of the property
     */
    public void setValid(boolean isValid) {
        isValid_ = isValid;
    }

    /**
     * Returns the string used to refer to this type of BusinessObject
     * in messages to the user.
     *
     * @return user-oriented name
     */
    public static String getDisplayName() {
        return "Summary";
    }

    /**
     * Returns the filename suffix for files containing this type of object.
     *
     * @return suffix for filename when stored in local filesystem
     */
    public static String getSuffix() {
        return "ERROR";
    }

    /**
     * Sets the value of a named property.
     * 
     * @param name	    name of the property
     * @param newValue	    new value of the named property
     */
    public void setPropertyValue(Object key, Object newValue) {
        if (newValue == null) {
            properties.remove(key);
        } else {
            properties.put(key, newValue);
        }
    }

    /**
     * Gets the value of a named property.
     * 
     * @return	    current value of the named property
     */
    public Object getPropertyValue(Object key) {
        return properties.get(key);
    }

    /**
     * Extracts the named property values from the BusinessObject
     * and stores them in the <code>properties</code> hashtable.
     * 
     * @param	bo  	    business object with values to be summarized.
     * @param	propNames   array of property names to remember.
     */
    public void extractProperties(BusinessObject bo, String[] propNames) {

        Class<? extends BusinessObject> cls = bo.getClass();

        properties.clear();

        // Handle Id and Name separately
        // We retain getX/setX for Id and Name as convenience methods,
        //   since it is so common to scan for a summary based on id.
        setId(bo.getId());
        setName(bo.getName());
        StringBuffer fcChartBuf = new StringBuffer();
        for (int j = 0; j < propNames.length; j++) {
            ObjectAdaptor oa =
                AdaptorFactory.getAdaptor(bo, propNames[j], false);
            Object obj = oa.getRawValue();
            if (obj instanceof BusinessObject)
                setPropertyValue(
                    propNames[j],
                    ((BusinessObject) obj).getName());
            else if (obj instanceof TemplateSignature)
                setPropertyValue(
                    propNames[j],
                    ((TemplateSignature) obj).getTemplateName());

            else
                setPropertyValue(propNames[j], obj);
        }

        updateChildIds(bo);
    }

    public void updateChildIds(BusinessObject bo) {
        BusinessObject[] children = bo.getBusObjChildren();
        long[] ids = new long[children.length];

        for (int j = 0; j < children.length; j++) {
            ids[j] = children[j].getId();
        }

        childIds_ = ids;
    }

    public long[] getChildIds() {
        return childIds_;
    }

    public String paramString() {
        Class cls = getObjectClass();
        String clName;

        if (cls != null) {
            clName = cls.getName();
            clName = clName.substring(clName.lastIndexOf('.') + 1);
        } else {
            clName = "";
        }

        return super.paramString() + ",class=" + clName;
    }
    /**
     * @see org.eso.ohs.dfs.StorableObject#postProecess()
     */
    public void postProecess() {
    }

}
