package org.eso.ohs.instruments;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ExtFileList;

/**
 * This class instanciates a Singleton Reference to a List of Instrument
 * objects. In order to obtain an object reference to the class you must call
 * the static method getInstance. If the object has already been instanciated it
 * will return a reference to this object in memory.
 * 
 * If you wish to extend this class so that your own InstrumentList class is
 * used, you need to do the following: 1. Create you class, extending this one.
 * 2. Create a factory for your class, implementing InstrumentListFactory. 3.
 * Assign an instance of your factory as the factory to be used:
 * InstrumentList.setInstrumentListFactory(new MyFactory(...));
 * 
 * @version $Revision: 1.42 $
 * @author Tim Canavan
 */

public class InstrumentList {

    private static final String rcsid = "$Id: InstrumentList.java,v 1.42 2006/04/27 10:37:43 tcanavan Exp $";

    /**
     * Factory for creating InstrumentList objects. If not set, the
     * DefaultInstrumentListFactory is used.
     */
    private static InstrumentListFactory ilFactory_ = null;

    /**
     * The name of the isf file used when an IP has versioning.
     */
    public static final String defaultISF = "default.isf";

    /**
     * Standard log channel
     */
    private static Logger stdlog_ = Logger.getLogger(InstrumentList.class);

    /**
     * The sole instance of the instrument list class
     */
    private static InstrumentList instance_ = null;

    /**
     * The actual list of instruments - this is a Hashtable of Hashtables. Each
     * entry in this Hashtable keyed under each instrument name. Eash hashtable
     * contains a set of Instrument objects, each object representing an IP
     * version for that instrument, keyed under the version value.
     */
    private Hashtable<String, Hashtable<String, Instrument>> instrumentList_ = new Hashtable<String, Hashtable<String, Instrument>>();
    
    
    /**
     * Ignored Instruments.
     * 
     */

    private List<String> ignoredInstruments_ = new ArrayList<String>();


    /**
     * This memeber determines if the InstrumentList is running in single IP
     * mode or not. It only has meaning if IPVersioning is enabled. Setting this
     * member to true will result in the IP with the highest version being
     * returned for the requested instrument regardless of the version
     * requested.
     * 
     * The member is set in the default constructor from AppConfig data.
     */
    private static boolean singleIPMode_ = false;

   
    /**
     * This function is the default connstrutor for the class It is disallowd
     * from being called because the class is a Singleton and the static
     * function getInstance should be called
     */
    protected InstrumentList() {
    }

    /**
     * @return an instance of an InstrumentList class which contains a handle to
     *         a list of instruments. This call assumes that the instance has
     *         already been setup with setupInstrumentList().
     */
    public static InstrumentList getInstance() {
        return instance_;
    }

    /**
     * Return an instance of an InstrumentList class which contains a handle to
     * a list of instruments Uses the standard .isf extension for ISF files and
     * obtains the instrument directory from the AppConfig.
     * 
     * @return List of innstruments that are insatanciated
     */
    public static synchronized InstrumentList setupInstrumentList(
            boolean singleIPMode, String instDir, String ipCacheDir)
            throws InstrumentException {
        if (instance_ == null) {
            //stdlog_.debug("MIPS InstrumentList getInstance()");
       
            singleIPMode_ = singleIPMode;
            
            Vector<String> instDirVector = new Vector<String>();

            //String instDir = AppConfig.getAppConfig().getInstrumDir();
            if (instDir != null && instDir.length() > 0) {
                instDirVector.addElement(instDir);
            }
            //String ipCacheDir =
            // AppConfig.getAppConfig().getIPVersionIPCache();
            if (ipCacheDir != null && ipCacheDir.length() > 0) {
                instDirVector.addElement(ipCacheDir);
            }
            String instDirs[] = new String[instDirVector.size()];
            instDirVector.copyInto(instDirs);
            setupInstrumentList(instDirs, ".isf");
        }
        return instance_;
    }

    /**
     * This function resets and reloads all instruments in the search
     * dirctories. It is typically executed after the instruments in the
     * directories have been updated.
     */
    public static void reloadInstrumentList(boolean singleIPMode,
            String instDir, String ipCacheDir) throws InstrumentException {
        resetInstrumentList();
        setupInstrumentList(singleIPMode, instDir, ipCacheDir);
    }

    /**
     * This is a utility function intended mainly for use in the unit tests. It
     * will reset the instance_ and the useIPVersion member.
     */

    public static void resetInstrumentList() {
        instance_ = null;
    }

    /**
     * Return an instance of an InstrumentList class which contains a handle to
     * a list of instruments
     * 
     * @param instrumentDirectory[]
     *            array of directory locations of instrument files
     * @param instrumentExtension
     *            extension of instrument file .isf
     * @return List of innstruments that are insatanciated
     */
    public static synchronized InstrumentList setupInstrumentList(
            String instrumentDirectory, String instrumentExtension)
            throws InstrumentException {
        if (instance_ == null) {
            String instDirs[] = new String[1];
            instDirs[0] = instrumentDirectory;
            setupInstrumentList(instDirs, instrumentExtension);
        }
        return instance_;
    }

    private static synchronized InstrumentList setupInstrumentList(
            String instDirs[], String instrumentExtension)
            throws InstrumentException {

        try {
            if (instance_ == null) {
                stdlog_.debug("MIPS InstrumentList getInstance(dirs, ext)");
                /*
                 * IP Version specifies that an IP be contained in a directory
                 * under which is a default.isf file. If this default.isf file
                 * does not exist, or the directory is not of the correct name
                 * format, it is ignored and assumed NOT to be an IP.
                 * 
                 * Example:
                 * 
                 * The ISAAC version 65.1 IP would be contained under the
                 * instruments directory as follows: instruments/ISAAC-065.1
                 * Under the above directory must be located a file of the name:
                 * 
                 * instruments/ISAAC-065.1/default.isf
                 */
                instance_ = create();
                ExtFileList filter;
                stdlog_.debug("Instrument directories listed: "
                        + instDirs.length);
                String mode = singleIPMode_ ? "Single IP Mode"
                        : "Multiple IP Mode";
                Hashtable<String, File> dirMap = new Hashtable<String, File>();
                for (int idx = 0; idx < instDirs.length; idx++) {
                    String instrumentDirectory = instDirs[idx];
                    File instDir = new File(instrumentDirectory);
                    String fullPath = instDir.getCanonicalPath();
                    Object exists = dirMap.get(fullPath);
                    if (exists != null) {
                        // we have already read this directory.
                        continue;
                    } else {
                        // save it so we know that we have read it.
                        dirMap.put(fullPath, instDir);
                    }
                    stdlog_.debug("Instrument directory[" + idx + "]: "
                            + instrumentDirectory + ", Using IP Versioning, "
                            + mode);

                    String[] fileList = (new File(instrumentDirectory)).list();

                    for (int i = 0; i < fileList.length; i++) {
                        File instrDir = new File(instrumentDirectory
                                + File.separator + fileList[i]);
                        if (instrDir.isDirectory()) {
                            File isfFile = new File(instrDir, defaultISF);
                            if (isfFile.exists()) {
                                Instrument inst = null;
                                if (fileList[i].indexOf("VIMOS") != -1) {
                                    //stdlog_.debug(
                                    //    "!!!!!!!!!!!! Constructing VIMOS IP");
                                    inst = new MaskInstrument(instrDir
                                            .getPath(), isfFile.getPath());
                                } else {
                                    inst = new Instrument(instrDir.getPath(),
                                            isfFile.getPath());
                                }
                                if (inst.hasValidVersion()) {
                                    stdlog_.debug("Adding Instrument "
                                            + inst.getInstrumentName()
                                            + " Version " + inst.getVersion());
                                    //instrumentList_.put(inst.getInstrumentName(),inst);
                                    instance_.putInstrument(inst);
                                    inst.initIndexFile();
                                } else {
                                	instance_.ignoredInstruments_.add(inst.getInstrumentName());
                                    stdlog_.debug("Instrument Package "
                                            + inst.getInstrumentName()
                                            + " has an illegal version ("
                                            + inst.getVersion()
                                            + ") and will be ignored.");
                                }
                            }
                        }
                    }
                }
            }
        } catch (IllegalArgumentException e) {
            throw new InstrumentException(e);
        } catch (IOException e1) {
            throw new InstrumentException(e1);
        }
        return instance_;
    }

    /**
     * Add an instrument to the hashtable of instruments. If IP versioning is
     * enabled: if the instrument and version already exist, replace it. If IP
     * versioning is no enabled: if the instrument already exists, replace it.
     */
    public void putInstrument(Instrument inst) throws InstrumentException {
        Hashtable<String, Instrument> ipSet = instrumentList_.get(inst
                .getInstrumentName());
        if (ipSet == null) {
            ipSet = new Hashtable<String, Instrument>();
            instrumentList_.put(inst.getInstrumentName(), ipSet);
        }
        Object exists = ipSet.get((new Float(inst.getVersionNumber()))
                .toString());
        if (exists != null) {
            Instrument currentIP = getInstrument(inst.getInstrumentName(), inst
                    .getVersionNumber());
            throw new InstrumentException("Instrument Package "
                    + inst.getInstrumentName() + " version "
                    + inst.getVersion()
                    + " exists multiple times in the Instrument List\n"
                    + "     " + inst.getTSFDirectory() + "\n" + "     "
                    + currentIP.getTSFDirectory());
        }
        ipSet.put((new Float(inst.getVersionNumber())).toString(), inst);
    }

    /**
     * @return the instrument requested, with the version requested, or null if
     *         not found If the version string is < 0.0 and ip versioning is
     *         enabled, the latest version of an Instrument will be returned.
     */
    public Instrument getInstrument(String inst, float version) {
        stdlog_.debug("MIPS GetInstrument: " + inst + " version: " + version);
        Instrument rc = null;
        Hashtable ipSet = instrumentList_.get(inst);
        if (ipSet != null) {
            // If version is less than 0, latest is returned.
            if (version < 0.0 || isInSingleIPMode()) {
                // latest version requested
                for (Enumeration e = ipSet.elements(); e.hasMoreElements();) {
                    Instrument i = (Instrument) e.nextElement();
                    if (rc == null
     
                    		|| rc.getVersionNumber() < i.getVersionNumber()) {
                        rc = i;
                    }
                }
            } else {
                rc = (Instrument) ipSet.get(new Float(version).toString());
            }
        }
        if (rc == null) {
            stdlog_.debug("MIPS GetInstrument: rc == null");
            // (new Exception()).printStackTrace();
        } else {
            stdlog_.debug("MIPS GetInstrument: rc != null");
        }
        return rc;
    }

    /**
     * @return an Instrument located by a directory path (instead of name and
     *         version), or null if no Instrument for that path exists in the
     *         cache.
     */

    public Instrument findInstrumentByDirectory(File dirPath) {
        // stdlog_.debug("MIPS FindInstrumentByDirectory: " + dirPath);
        Instrument[] insts = getInstrumentList();
        Instrument rc = null;
        if (insts != null) {
            for (int i = 0; i < insts.length; i++) {
                File tsfDir = new File(insts[i].getTSFDirectory());
                File cannonicalPath = null;
                try {
                    cannonicalPath = new File(tsfDir.getCanonicalPath());
                } catch (Exception e) {
                    // failed to get cannonical path - print a message
                    e.printStackTrace();
                }
                if (tsfDir.equals(dirPath)
                        || (cannonicalPath != null && cannonicalPath
                                .equals(dirPath))) {
                    rc = insts[i];
                    break;
                }
            }
        }
        return rc;
    }

    /**
     * @return the instrument requested AND remove it from the list, or null if
     *         not found.
     */
    public Instrument removeInstrument(String inst, float version) {
        Instrument rc = null;
        Hashtable ipSet = instrumentList_.get(inst);
        if (ipSet != null) {
            rc = (Instrument) ipSet.remove(new Float(version).toString());
            if (ipSet.size() == 0) {
                // remove the ipSet as well, as it is empty.
                instrumentList_.remove(inst);
            }
        }
        return rc;
    }

    /**
     * @return the named instument, or null if not found. If ip versioning is
     *         used the latest version of the instument package will be
     *         returned.
     */
    public Instrument getInstrument(String inst) {
        return getInstrument(inst, -1);
    }


    /**
     * @return true if the instrument list is in Single IP Mode.
     */
    public static boolean isInSingleIPMode() {
        return singleIPMode_;
    }

    /**
     * Returns a list of handles to instrument objects.
     * 
     * @return List of ALL instrument objects in the collection. This includes
     *         all versions of all instruments.
     */
    
    
    //private Hashtable<String, Hashtable<String, Instrument>> instrumentList_ 
    //= new Hashtable<String, Hashtable<String, Instrument>>();

    
    public Instrument[] getInstrumentList() {
        //stdlog_.debug("MIPS GetInstrumentList");
        Instrument[] rc = null;
        if (instance_ != null && instrumentList_ != null) {
            Vector<Instrument> instruments = new Vector<Instrument>();
            for (Enumeration <Hashtable<String, Instrument>> e = instrumentList_.elements(); e
                    .hasMoreElements();) {
                Hashtable<String,Instrument> ipSet = e.nextElement();
                for (Enumeration <Instrument> i = ipSet.elements(); i.hasMoreElements();) {
                    instruments.addElement(i.nextElement());
                }
            }
            rc = new Instrument[instruments.size()];
            for (int i = 0; i < instruments.size(); i++) {
                rc[i] = (Instrument) instruments.elementAt(i);
            }
        }
        return rc;
    }

    /**
     * Check if an instrument exists within the instrument list
     * 
     * @param instrumentName
     *            name of instrument to be located
     * @return exists or not exists
     */
    public boolean instrumentExists(String instrumentName) {
        return instrumentExists(instrumentName, -1);
    }

    public boolean instrumentExists(String instrumentName, float version) {
        return getInstrument(instrumentName, version) != null;
    }

    /**
     * Returns a list of instrument names
     * 
     * @return List of instrument names
     */

    public String[] getInstrumentNames() {
        String[] instNames = new String[instrumentList_.size()];
        Enumeration<String> keys = instrumentList_.keys();
        for (int i = 0; i < instrumentList_.size(); i++) {
            instNames[i] = keys.nextElement();
        }
        return instNames;
    }

    /**
     * Returns a list of telescope names
     * 
     * @return List of instrument names
     */

    public String[] getTelescopeNames() {
        Enumeration<String> keys = instrumentList_.keys();
        Vector<String> telVec = new Vector<String>();
        for (int i = 0; i < instrumentList_.size(); i++) {
            String name = keys.nextElement();
            String telescopeName = (getInstrument(name)).getTelescopeName();
            if (!telVec.contains(telescopeName))
                telVec.addElement(telescopeName);
        }
        String[] telescopeNames = new String[telVec.size()];
        telVec.copyInto(telescopeNames);
        return telescopeNames;
    }

    /**
     * Retuns a handle to an Instrument class
     * 
     * @return Instrument handle to an instrument class based on name.
     */
    /*
     * 
     * public static Instrument getInstrument(String name) throws
     * NullPointerException { Instrument inst =
     * (Instrument)instrumentList_.get(name); return inst; }
     */

    /**
     * Insert a new instrument into the instrument list
     * 
     * @param Reference
     *            to an instrument class
     */
    /*
     * public static void insertInstrument(Instrument inst) throws IOException {
     * String name = inst.getInstrumentName(); inst.initIndexFile();
     * if(instrumentList_.contains(name)) instrumentList_.remove(name);
     * instrumentList_.put(name,inst); }
     */

    public void insertInstrument(Instrument inst) throws IOException,
            InstrumentException {
        removeInstrument(inst.getInstrumentName(), inst.getVersionNumber());
        inst.initIndexFile();
        putInstrument(inst);
    }

    /**
     * Create an InstrumentList object - according to the factory assigned.
     */
    private static InstrumentList create() {
        InstrumentList rc;
        if (ilFactory_ != null) {
            stdlog_.debug("Creating InstrumentList from Custom Factory.");
            rc = ilFactory_.getInstrumentList();
        } else {
            stdlog_.debug("Creating InstrumentList from Default Factory.");
            rc = (new DefaultInstrumentListFactory()).getInstrumentList();
        }
        return rc;
    }

    /**
     * Set the InstrumentListFactory to be used to create InstrumentList
     * objects. The DefaultInstrumentList factory is used if this is not set.
     */
    public static void setInstrumentListFactory(InstrumentListFactory ilf) {
        ilFactory_ = ilf;
    }

    /**
     * @return the InstrumentListFactory used to create InstrumentList objects,
     *         or null if none is assigned.
     */
    public static InstrumentListFactory getInstrumentListFactory() {
        return ilFactory_;
    }

	public List<String> getIgnoredInstruments() {
		return ignoredInstruments_;
	}

	public void setIgnoredInstruments_(List<String> ignoredInstruments) {
		this.ignoredInstruments_ = ignoredInstruments;
	}
}