package org.eso.ohs.instruments;

import java.io.File;
import java.io.IOException;
import java.util.Hashtable;

import org.apache.log4j.Logger;


/**
 * This class is the interface to the instruments package from the
 * outside world. If caches the TemplateSignatureMetaData classes in
 * memory so this static read only data can be referenced by many 
 * TemplateSignature classes. This data includes Ranges for parameters etc..
 * @see TemplateSignature
 * @see TemplateSignatureMetaData
 * <p>
 * $Id: TemplateSignatureFactory.java,v 1.51 2006/04/26 09:26:49 tcanavan Exp $
 * @author Tim Canavan 
 * <br>ChangeLog: cmavros 15-10-01 Call getCriticalStatus() to 
 *   collect errors for Verification Report.
 */

public class TemplateSignatureFactory  {

   /** Standard log channel */
   private static Logger stdlog_ =
       Logger.getLogger( TemplateSignatureFactory.class );


    /**
     * Cache of previously created Template objects.
     * This saves a <b>lot</b> of memory but should be changed to
     * using Java reference objects once we're running under JDK1.2
     */
    private final static Hashtable<Instrument, Hashtable<String, TemplateSignatureMetaData>> cache_ = new Hashtable<Instrument, Hashtable<String, TemplateSignatureMetaData>>();



   /**
     * Returns a (possibly cached) Template object for the given
     * instrument and tsf file.  Repeated requests for a template
     * of the same instrument and tsf file will return the same
     * Template object -- this saves significant amounts of time
     * and memory.
     * 
     * @param instrument    the instrument object
     * @param tsfName       the name of the .tsf file
     * @return a Template object for the given instrument and .tsf file.
     * @throws IOException  if the template was not previously cached
     *              and an I/O occurred reading it from disk.
     */
    // public static TemplateSignature getTemplate(String instDir, String instrument, 
 
    public static TemplateSignature getTemplate(Instrument instrument,
            String tsfName) throws IOException {
        
        stdlog_.debug("Get Template: "+instrument.getInstrumentName()
                +" v: "+instrument.getVersionNumber()+" Template: "+tsfName);
        TemplateSignatureMetaData tsMet = getTemplateMetaData(instrument,
           tsfName);
        
        if(tsMet != null) {
            stdlog_.debug("GetTemplate: Meta Data Found");
            return new TemplateSignature(tsMet, instrument.getVersionNumber());
        } else {
            stdlog_.debug("GetTemplate: Meta Data Not Found.");
        }
        return null;            
    }

    /**
     * Returns a (possibly cached) Template object for the given
     * instrument and tsf file.  Repeated requests for a template
     * of the same instrument and tsf file will return the same
     * Template object -- this saves significant amounts of time
     * and memory.
     * 
     * @param instrument    the name of the instrument
     * @param tsfName       the name of the .tsf file
     * @return a Template object for the given instrument and .tsf file.
     * @throws IOException  if the template was not previously cached
     *              and an I/O occurred reading it from disk.
     */
    // public static TemplateSignatureMetaData getTemplateMetaData(String instDir,String instrument, 
    
    public static TemplateSignatureMetaData getTemplateMetaData(Instrument instrument, 
            String tsfName) throws IOException {
        // File tsfFile = new File(getAbsPath(instDir,instrument,tsfName));
        stdlog_.debug("MIPS GetTemplateMetaData file: "+instrument.getTSFDirectory()+File.separator+tsfName+".tsf");
        File tsfFile = new File(instrument.getTSFDirectory()+File.separator+tsfName+".tsf");
        if(!tsfFile.exists()) {
            return null;
        }
        Hashtable<String, TemplateSignatureMetaData> instrumentCache = cache_.get(instrument);
    
        if (instrumentCache == null) {  // Not seen instrument before?
            instrumentCache = new Hashtable<String, TemplateSignatureMetaData>();
            cache_.put(instrument, instrumentCache);
        }
    
        TemplateSignatureMetaData tmpl = instrumentCache.get(tsfName);   
        if (tmpl == null || isOutOfDate(tsfFile,tmpl)) {    // Not seen instrument before?
            tmpl = new TemplateSignatureMetaData(
            // getAbsPath(instDir,instrument,tsfName),instrument);
            instrument.getTSFDirectory()+File.separator+tsfName+".tsf", instrument);
            if(tmpl != null && !tmpl.getCriticalStatus()) {
                instrumentCache.put(tsfName, tmpl);
            } else {
                /** 
                 * ChangeLog: cmavros 15-10-01 
                 * Call getCriticalStatus() to collect 
                 * errors for Verification Report.
                 */
                if(tmpl.getCriticalStatus()){                                           
                    tmpl.toEngString();
                } 
                return null;
            }
	}
        else {
	    // stdlog_.debug("@@@@@@@@@ Template Meta Data from cache " + tsfName);
        }     
        return tmpl;
    }  
    
    /**
     * remove all the templates for an instrument from the cache.
     * if the instrument is not found in the cache, no action is taken
     * This call is typically made when an ISF file has changed and
     * all TSF files must be read in (i.e. in refreshTemplate()).
     */
    public static void removeInstrumentTemplates(Instrument inst) {
        cache_.remove(inst);
    }

    /**
     * Checks to see if a file is out of Date by comparing
     * the date stored in a files metadata with the physical file on disk
     * @param tsfFile File handle to tsf file on disk
     * @param tsMet MetaData for a given template
     * @return boolean true if .tsf file is newer than the original stored
     *
     */

    private static boolean isOutOfDate(File tsfFile,TemplateSignatureMetaData tsMet) {
        return (tsfFile.lastModified() > tsMet.getLastModified());
               
    }


    /**
     * Create Template Meta data class for a give .tsf file
     * @param instrument The Specific instrument object.
     * @param tsfName Template Signature file name used to create meta data from
     *    
     * @return TemplateSignatureMetaData refence to TemplateSignatureMetaData class
     */
    public static TemplateSignatureMetaData createTemplateMetaData(Instrument instrument,String tsfName) throws IOException {
        TemplateSignatureMetaData tmpl = new TemplateSignatureMetaData(
                instrument.getTSFDirectory()+File.separator+tsfName+".tsf",instrument);
        return tmpl;         
    }

    /**
     * Create Template Meta data class for a give .tsf file using default 
     * instrument directory from .cfg file
     * @param instrument Specific instrument FORS1 etc..
     * @param tsfName Template Signature file name used to create meta data from
     *    
     * @return TemplateSignatureMetaData refence to TemplateSignatureMetaData class
     */
/*
    public static TemplateSignatureMetaData createTemplateMetaData(String instrument,String tsfName) throws IOException {
        return createTemplateMetaData(AppConfig.getAppConfig().getInstrumDir(),
            instrument,tsfName);
    }
*/
    /**
     * Refresh a template with new MetaData from a .tsf file that has 
     * been updated. This is an extremely important method as it guarantees
     * the integerity of a users cache as .tsf files get updated
     * 
     * All values that are still valid in accordance with the new .tsf file
     * MetaData will be caried accross. 
     *
     * @param ts TemplateSignature to be updated from  
     * @param instDir Instrument directory
     *
     * @return boolean False if it cannot refresh a template else True
     */
    public static boolean refreshTemplate(TemplateSignature ts) throws IOException {
        return refreshTemplate(ts, false);
    }
    
    public static boolean refreshTemplate(TemplateSignature ts, boolean fullRefresh) throws IOException {
        TemplateSignatureMetaData tsMet=null;
        // refresh the ISF file if require
        Instrument inst = ts.getInstrument();
        if (inst != null && inst.refreshRequired()) {
            //stdlog_.debug("Instrument Refresh Required.........");
            // remove all templates for that instrument from the cache
            removeInstrumentTemplates(inst);
            // refresh the template
            inst.refreshInstrument();
            // retreive the instrument again from the InstrumentList.
            inst = ts.getInstrument();
            // force a template refresh as the ISF file has changed
            fullRefresh = true;
        }
        if (inst == null) {
            stdlog_.debug("REFRESH TEMPLATE: NO INST");
            // could not locate the instrument in the InstrumentList.
            return false;
        }
        File f = new File(inst.getTSFDirectory()+File.separator+ts.getTemplateName()+".tsf");
        if(!f.exists()) {
            // TSF file no longer exists.
            stdlog_.debug("REFRESH TEMPLATE: NO TSF");
            return false;
        }
        tsMet = getTemplateMetaData(inst,ts.getTemplateName());          
        if(tsMet == null) {
            stdlog_.debug("REFRESH TEMPLATE: NO META DATA");
            return false;
        }
        /*
         * Refresh the template meta data if the modified times have changed, or a refresh
         * is forced, or the template IP version number differs from the instrument IP version.
         * The Instrument IP version will be the latest and correct - the TSF IP version may not
         * be correct as the TSF data is cached in the local cache - and therefore may be
         * out of date.
         */
        /*
         * for debugging only.
        stdlog_.debug("TESTING REFRESH!: fr:"+fullRefresh+" lm1: "+f.lastModified()+" lm2: "+ts.getLastModified()
            +" Inst Ver:"+inst.getVersionNumber()+" TSF Ver:"+ts.getVersion());
        */
        if(fullRefresh || (inst.getVersionNumber() != ts.getVersion()) || (f.lastModified() != ts.getLastModified())) {       
            refreshParameters(ts,tsMet);         
        } else {
            ts.setMetaData(tsMet);
            ts.recreateParamMetaData();
            ts.createDisplayParamList();           
        }
        return true;
    }


    
    /**
     * Refresh parameters for a template. 
     * All values that are still valid in accordance with the new .tsf file
     * MetaData will be caried accross. 
     *
     * @param ts TemplateSignature to be updated from  
     * @param instDir Absolute Path for instruments directory
     *
     * @return boolean False if successfully refreshed parameters  else True
     */
    private static void refreshParameters(TemplateSignature ts,TemplateSignatureMetaData tsMet) {
        Parameter [] oldPlist = ts.getParamList();                                
        ts.setMetaData(tsMet);              
        ParamMetaData [] pMetaList = tsMet.getParamList();
        Parameter [] newlist = new Parameter[pMetaList.length];
        for(int j=0;j<pMetaList.length;j++) {
            newlist[j] = CreateParam.Create(pMetaList[j]);
            for(int k=0;k<oldPlist.length;k++) {
                if(pMetaList[j].getParameterName().equals(oldPlist[k].getParameterName())
                        && pMetaList[j].getType().equals(oldPlist[k].getType())
                        && !pMetaList[j].isFixed()) {                          
                    String value  = oldPlist[k].getValue();
                    newlist[j].setValue(value);                          
                }                        
            }
        }
                                  
                /*
                 * For debugging only.
                stdlog_.debug(
                        "============= PARAMETER LISTS =========::"
                        +ts.getTemplateName());
                for (int x=0;x<newlist.length;x++) {
                    stdlog_.debug(x+" newlist  :<"
                        +newlist[x].getParameterName()+">");
                    stdlog_.debug("pMetaList:<"
                        +pMetaList[x].getParameterName()+">");
                }
                for (int y=0;y<oldPlist.length;y++) {
                    stdlog_.debug(y+" oldList  :<"
                        +oldPlist[y].getParameterName()+">");
                }
                stdlog_.debug(
                        "============= END PARAMETER LISTS ===============");
                 */

        ts.setParamList(newlist);               
    }

    /**
     * Refresh a template given the default instrument directory
     * @param Reference to ts object to be refreshed
     * @return boolean False if it cannot refresh a template else True     
     */

    /*
    public static boolean refreshTemplate(TemplateSignature ts) throws IOException {
        return refreshTemplate(ts, AppConfig.getAppConfig().getInstrumDir());
    }
    */

    /**
     * Get absolute path for a .tsf file
     * @param instDir instruments directory
     * @param instrument instrument name  
     * @param tsfName template signature name
     * @return absolute path for a .tsf file     
     */

    //public static String getAbsPath(String instDir,String instrument,String tsfName) {
    //     return instDir + File.separator + instrument + File.separator + tsfName + ".tsf";
    //}


    /**
     * Get the absolute path for a .tsf file given with a 
     * default instrument directory
     * @param instrument instruments directory
     * @param tsfName instrument name
     * @return absolute path for a .tsf file     
     */

/*
    public static String getAbsPath(String instrument,String tsfName) {
         return AppConfig.getAppConfig().getInstrumDir()
             + File.separator + instrument + File.separator + tsfName + ".tsf";
    } 
*/             
}
