package org.eso.ohs.p2pp;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.NotActiveException;
import java.io.ObjectInputStream;
import java.util.Vector;

import javax.swing.event.EventListenerList;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.BusinessObject;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.SimpleStorableObject;
import org.eso.ohs.dfs.StorableObject;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.persistence.AppConfig;
import org.eso.ohs.persistence.Config;
import org.eso.ohs.persistence.DirectoryNode;
import org.eso.ohs.persistence.Media;
import org.eso.ohs.persistence.ObjectManager;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.persistence.ObjectTableEvent;
import org.eso.ohs.persistence.ObjectTableListener;

/**
 * This class contains the summary informaton needed by the GUI
 * so that the contents of a folder can be displayed quickly.
 * Without this, each OB, OD, Target, etc would have to be read
 * before the contents of a folder could be shown, which would
 * be too time-consuming for all except the smallest folders.
 * 
 * <p>
 *
 * $Id: Folder.java,v 1.2 2006/04/26 09:36:38 ddorigo Exp $
 *
 * <br>
 * @author Bob Kemp
 */

public class Folder
    extends SimpleStorableObject
    implements PropertyChangeListener, StorableObject {

    /** Standard log channel */
    private static Logger stdlog_ = Logger.getLogger(Folder.class);

    /** Version id for object serialization */
    static final long serialVersionUID = 7735096446635272018L;

    /** user-visible name of this folder */
    private String folderName_ = "Untitled";

    /** array of object summaries for user-visible objects in this folder */
    private Summary[] summaryInformation_ = {
    };

    /** medium on which the directory exists. */
    private transient Media device_;

    /** reference to the ObservingRun for this folder, if any. */
    private transient ObservingRun obsRun_;

    /** path of this directory on the given medium. */
    private transient DirectoryNode dirNode_;

    /** support for handling an object events */
    private transient EventListenerList objectTableListeners;

    /**
     * Default constructor for a Folder (summary information) object.
     * 
     * <p> A caller that uses this constructor, must also use
     * the setMediaNode() method to fully initialise the class.
     * This constructor is normally only called by the de-serialization
     * classes when an object of this class is read from disk.
     */
    public Folder() {
        initialise();
    }

    /**
     * Constructor for a Folder (summary information) object.
     * Builds the summary information from scratch.
     * 
     * @param node    the directory that this class summarizes.
     * @param device  the medium of the directory this class summarizes.
     * 
     * <p> XXX the setMediaNode() method should be removed/designed out!
     * 
     * @exception ObjectIOException
     *			if there was an I/O error reading the object.
     * @exception ObjectNotFoundException
     *			if there is no ObservationBlock with the given id
     * 			on this medium.
     */
    public Folder(DirectoryNode node, Media device)
        throws ObjectIOException, ObjectNotFoundException {
        this();
        setMediaNode(device, node);

        reconstructSummaryInformation();
    }

    /**
     * Set new value for obsRun_ property
     *
     * @param newValue the new value for the property
     */
    public void setObsRun(ObservingRun obr) {
        obsRun_ = obr;
    }

    /**
     * Get value for obsRun_ property
     *
     * @return the current value of the property
     */
    public ObservingRun getObsRun() {
        return obsRun_;
    }

    /**
     * Returns whether all business objects are valid for a given
     * CB/OB
     * @return boolean All business objects valid.
     */
    public boolean allValid(Summary summary) {
        boolean allValid = true;
        for (int i = 0; i < summaryInformation_.length; i++) {
            if (Config.getCfg().uniqueToTableId(summaryInformation_[i].getId())
                == Config.getCfg().uniqueToTableId(summary.getId())) {
                if (!summaryInformation_[i].isValid()) {
                    allValid = false;
                    break;
                }
            }
        }
        return allValid;
    }

    /**
     * Finds the observing run object for this folder.
     * Each directory below the root of the local storage
     * corresponds to an observing run.  This method returns the
     * observing run corresponding to this folder or an ancestor.
     * 
     * @return	observing run for this folder hierarchy, or null if not found
     */
    public ObservingRun findObsRun() {
        for (DirectoryNode node = getNode();
            node != null;
            node = node.getParent()) {
            StorableObject sobj = node.getObject();

            if (sobj instanceof Folder) {
                Folder folder = (Folder) sobj;

                if (folder.getObsRun() != null) {
                    return folder.getObsRun();
                }
            }
        }
        //	System.out.println("findObsRun: couldn't find an ObservingRun!");
        return null;
    }

    public Folder findRootFolder() {
        for (DirectoryNode node = getNode();
            node != null;
            node = node.getParent()) {
            StorableObject sobj = node.getObject();

            if (sobj instanceof Folder) {
                Folder folder = (Folder) sobj;

                if (folder.getObsRun() != null) {
                    return folder;
                }
            }
        }
        return null;
    }

    /**
     * Returns the directory node for this folder, ie its path in the hierarchy
     *
     * @return directory node
     */
    public DirectoryNode getDirNode() {
        return dirNode_;
    }

    public String getDescription() {
        return dirNode_.getDescription();
    }

    /**
     * Sets the medium and directory associated with this folder.
     * 
     * <p> XXX This method should be removed/designed out!!!
     * 
     * @param node    the directory that this class summarizes.
     * @param device  the medium of the directory this class summarizes.
     */
    public void setMediaNode(Media device, DirectoryNode node) {
        dirNode_ = node;
        device_ = device;
        firePropertyChangeEvent();
    }

    /**
     * Hook method for Java serialization interface.
     * 
     * @param in        the input stream from which the serialized objects
     *			will be read.
     * @exception IOException
     *		        an I/O error occurred while reading the object
     *			from external media.
     * @exception ClassNotFoundException
     *			The class for a serialized business
     *			object taken from disk cannot be found.
     *			This should never happen.
     * @exception NotActiveException
     *		      	thrown if the ObjectInputStream is used when
     *			not active.
     */
    private void readObject(ObjectInputStream in)
        throws IOException, ClassNotFoundException, NotActiveException {

        initialise();
        in.defaultReadObject();
    }

    /**
     * Initialise/reinitialise the transient variables of the class.
     */
    private void initialise() {
        ObjectManager objMgr = ObjectManager.getObjectManager();

        objectTableListeners = new EventListenerList();
        summaryInformation_ = new Summary[0];
    }

    /**
     * Catch PropertyChange events from the objects in the folder
     * and update the summary info, if appropriate.
     *
     * @param ev the event
     */
    public void propertyChange(PropertyChangeEvent ev) {
        //System.out.println("Folder " + getNode().getDescription()
        //		    + " saw change in "
        //		    + ((StorableObject)ev.getSource()).getId());
        try {
            updateSummaryEntry((BusinessObject) ev.getSource());
        } catch (ObjectNotFoundException e) {
            // XXX should do something more sensible with this
            stdlog_.error(e);
        } catch (ObjectIOException e) {
            // XXX should do something more sensible with this
            stdlog_.error(e);
        }
    }

    /**
     * Returns the DirectoryNode for this folder's path.
     * 
     * @return path for this folder.
     */
    public DirectoryNode getNode() {
        return dirNode_;
    }

    /**
     * Returns the list of summaries for objects in this folder.
     * 
     * @return array of object summaries.
     */
    public Summary[] getSummaryInformation() {
        return summaryInformation_;
    }

    /**
     * Returns a list of summaries in this folder for a given object type.
     * 
     * @return array of object summaries.
     */
    public Summary[] getSummaryInformationByType(Class cls) {
        Vector<Summary> summaryVec = new Vector<Summary>();

        for (int i = 0; i < summaryInformation_.length; i++) {
            if (summaryInformation_[i].getObjectClass().equals(cls))
                summaryVec.addElement(summaryInformation_[i]);
        }
        Summary[] summary = new Summary[summaryVec.size()];
        summaryVec.copyInto(summary);
        return summary;
    }

    /**
     * Returns true if an object with the given name exists in this folder.
     * 
     * @return true if it was found, false otherwise.
     */
    public boolean exists(String name) {
        if (summaryInformation_ != null) {
            for (int i = 0; i < summaryInformation_.length; i++) {
                if (summaryInformation_[i].getName().equals(name))
                    return true;
            }
        }
        return false;
    }

    /**
     * Returns the id of an object with the given name in this folder, or -1.
     * 
     * @return id if it was found, -1 otherwise.
     */
    public long getId(String name) {
        if (summaryInformation_ != null) {
            for (int i = 0; i < summaryInformation_.length; i++) {
                if (summaryInformation_[i].getName().equals(name))
                    return summaryInformation_[i].getId();
            }
        }
        return -1;
    }

    /**
     * Returns the medium on which this folder in stored.
     * 
     * @return a media type.
     */
    public Media getMedia() {
        return device_;
    }

    public void insertSummary(Summary summ) {
        Summary[] newBOList = new Summary[summaryInformation_.length + 1];

        System.arraycopy(
            summaryInformation_,
            0,
            newBOList,
            1,
            summaryInformation_.length);
        summaryInformation_ = newBOList;

        summaryInformation_[0] = summ;
    }

    /**
     * Updates the summary entry for the specified business object.
     * This method is usually called because a property of the
     * object changed or the summary is being recreated.
     * 
     * @param bObject	the business object with the summary to be updated.
     * 
     * @exception ObjectIOException
     *			An I/O error occurred while reading an object.
     * @exception ObjectNotFoundException
     *			The object was not found.
     *			Should never happen.
     */
    private void updateSummaryInternal(BusinessObject bObject)
        throws ObjectIOException, ObjectNotFoundException {
        Summary summary;

        summary = findSummary(bObject.getId());
        if (summary == null) {
            summary = new Summary();
        }

        summary.extractProperties(
            bObject,
            AppConfig.getAppConfig().folderColumnProperties(
                bObject.getClass()));

        String errs = bObject.verifyBO();

        if (errs.length() > 0) {
            summary.setValid(false);
        } else {
            summary.setValid(true);
        }

        summary.setCheckedIn(bObject.isCheckedIn());
        summary.setObjectClass(bObject.getClass());

        if (findSummary(bObject.getId()) == null) {
            insertSummary(summary);
        }

        //updateAllLinks(summary);
    }

    /**
     * Removes the summary entry for the specified business object.
     * This method is usually called because the object has been deleted
     * or moved.
     * 
     * @param bObject	the business object with the summary to be updated.
     */
    private Summary removeSummaryInternal(long id) {
        Summary deletedSummary = null;
        Class mo;
        String fname;
        int i;

        for (i = 0; i < summaryInformation_.length; i++) {
            if (summaryInformation_[i].getId() == id) {
                break;
            }
        }

        if (i < summaryInformation_.length) {
            Summary[] newBL = new Summary[summaryInformation_.length - 1];

            System.arraycopy(summaryInformation_, 0, newBL, 0, i);
            System.arraycopy(
                summaryInformation_,
                i + 1,
                newBL,
                i,
                newBL.length - i);

            deletedSummary = summaryInformation_[i];
            summaryInformation_ = newBL;

            //removeAllLinks(deletedSummary);
        }

        return deletedSummary;
    }

    /**
     * Add summary info for a new BusinessObject.
     * 
     * @param bObject	the business object with the summary to be added.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error reading the object.
     * @exception ObjectNotFoundException
     *			if there is no ObservationBlock with the given id
     * 			on this medium.
     */
    public void addSummaryEntry(BusinessObject bObject, Summary summ)
        throws ObjectNotFoundException, ObjectIOException {
        if (summ == null) {
            updateSummaryInternal(bObject);
        } else {
            insertSummary(summ);
        }
        fireObjectTableEvent(
            ObjectTableEvent.OBJECT_ADDED,
            device_,
            dirNode_,
            bObject.getId(),
            bObject.getClass());
    }

    /**
     * Updates the summary entry for an existing BusinessObject.
     * 
     * @param bObject	the business object with the summary to be updated.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error reading the object.
     * @exception ObjectNotFoundException
     *			if there is no ObservationBlock with the given id
     * 			on this medium.
     */
    public void updateSummaryEntry(BusinessObject bObject)
        throws ObjectNotFoundException, ObjectIOException {
        updateSummaryInternal(bObject);
        fireObjectTableEvent(
            ObjectTableEvent.OBJECT_CHANGED,
            device_,
            dirNode_,
            bObject.getId(),
            bObject.getClass());
    }

    /**
     * Removes the summary entry for a BusinessObject.
     * 
     * @param bObject	the business object with the summary to be removed.
     */
    public Summary removeSummaryEntry(long id, Class objType) {
        Summary summ = removeSummaryInternal(id);
        fireObjectTableEvent(
            ObjectTableEvent.OBJECT_REMOVED,
            device_,
            dirNode_,
            id,
            objType);

        return summ;
    }

    /**
     * Recreates the summary information for the current folder.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error reading the object.
     *                  authentication.
     */
    public void reconstructSummaryInformation() throws ObjectIOException {
        ObjectManager objMgr = ObjectManager.getObjectManager();

        summaryInformation_ = objMgr.listObjects(device_, dirNode_);
        if (device_ == Media.LOCAL) {
            for (int j = 0; j < summaryInformation_.length; j++) {
                try {
                    // As a side-effect of reading, the summary
                    // info will be updated in updateSummaryEntry()
                    // Thus we don't care about the exceptions.
                    StorableObject sobj =
                        objMgr.getBusObj(
                            device_,
                            summaryInformation_[j].getId(),
                            summaryInformation_[j].getObjectClass());

                    if (sobj instanceof BusinessObject) {
                        updateSummaryEntry((BusinessObject) sobj);
                    }

                    if (sobj instanceof ObservingRun) {
                        setObsRun((ObservingRun) sobj);
                        setName(((ObservingRun) sobj).getName());
                    }
                } catch (ObjectIOException e) {
                    // do nothing
                } catch (ObjectNotFoundException e) {
                    // do nothing
                }
            }
        }
        firePropertyChangeEvent();
    }

    /**
     * Returns the summary in this folder with the given id, or null.
     * 
     * @return Summary if it was found, null otherwise.
     */
    public Summary findSummary(long id) {
        Summary[] summaries = getSummaryInformation();

        for (int j = 0; j < summaries.length; j++) {
            if (summaries[j].getId() == id) {
                return summaries[j];
            }
        }
        return null;
    }

    public String getDisplayName() {
        return "Folder";
    }

    public static String getSuffix() {
        return "sum";
    }

    /**
     * Returns a string representation of the object.
     * Mainly used for debugging.
     *
     * @return the string representation
     */
    public String toString() {
        return "Folder["
            + (dirNode_ == null ? "node=<none>" : "node=" + dirNode_.getId())
            + ",media="
            + device_
            + "]";
    }

    /**
     * Adds a new listener to be notified of created/read/deleted objects
     *
     * @param lnr the listener
     */
    public void addObjectTableListener(ObjectTableListener lnr) {
        objectTableListeners.add(ObjectTableListener.class, lnr);
    }

    /**
     * Removes a listener, which will no longer be notified of changes.
     *
     * @param lnr the listener
     */
    public void removeObjectTableListener(ObjectTableListener lnr) {
        objectTableListeners.remove(ObjectTableListener.class, lnr);
    }

    /**
     * Inform listeners of a change to the StorableObjects in this Folder.
     */
    private void fireObjectTableEvent(
        int eventType,
        Media device,
        DirectoryNode dir,
        long id,
        Class objType) {
        ObjectTableEvent objEvent =
            new ObjectTableEvent(eventType, device, dir, id, objType);
        Object[] listeners = objectTableListeners.getListenerList();

        for (int i = listeners.length - 2; i >= 0; i -= 2) {
            Object lnr = listeners[i + 1];
            if (listeners[i] == ObjectTableListener.class) {
                if (eventType == ObjectTableEvent.OBJECT_ADDED) {
                    ((ObjectTableListener) lnr).tableObjectAdded(objEvent);
                } else if (eventType == ObjectTableEvent.OBJECT_CHANGED) {
                    ((ObjectTableListener) lnr).tableObjectChanged(objEvent);
                } else if (eventType == ObjectTableEvent.OBJECT_REMOVED) {
                    ((ObjectTableListener) lnr).tableObjectRemoved(objEvent);
                }
            }
        }
        firePropertyChangeEvent();
    }
    /**
     * @see org.eso.ohs.dfs.StorableObject#postProecess()
     */
    public void postProecess() {
    }

}
