package org.eso.ohs.persistence;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.swing.event.EventListenerList;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.BusinessObject;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.StorableObject;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.persistence.StorageManager.UnpackedStorableObject;


/**
 * This class manages the in-memory copies of all StorableObjects.
 * When a reference to some StorableObject is requested by a client,
 * it first looks it up in a table to see if the object is already
 * in memory and, if so, returns it.  This ensures that only one
 * copy of any StorableObject is in memory at any one time, which will
 * eventually be written to the external media.  Thus an object cannot
 * appear to have several values at once.
 * 
 * <p>
 * If the request object is not in memory, it is retrieved from the
 * associated external media, added to the inMemory table and returned.
 * As part of the retrieval process, any StorableObjects referred to by
 * this object are also read into memory and referred to correctly.
 * 
 * <p>
 * For correct operation, there should be at most one instance of this
 * object.
 * <p>
 * The DirectoryNode concept is kept in this abstract for compatibility
 * and because could be implemented in subclasses. If a child of this class
 * doesn't use it, could pass always null and doesn't harm.
 * <p>
 * $Id: AbstractObjectManager.java,v 1.15.4.1 2007/10/05 11:03:37 pnunes Exp $
 * <br>* @author Bob Kemp
 */
public abstract class AbstractObjectManager {

    private static Logger stdlog_ = 
	Logger.getLogger( AbstractObjectManager.class );
 
   /** Invalid object id.  Used to indicate that there is no such object */
    public static final long NO_OBJECT = 0;

    /** Utility constant for invoking getter methods, using reflection */
    protected static final Object[] NO_ARGS = new Object[]{};

    /** A reference to the singleton AbstractObjectManager instance. */
    protected static AbstractObjectManager objMgr_ = null;

    protected Registry registry_;

    /**
     * References to StorageManager objects for supported Media.
     * The table is keyed by the Media value.  Supported media can be
     * added at any time using the setStorageManager() method.
     * 
     * @see #setStorageManager
     */
    protected Hashtable<Media, StorageManager> mediaMgr_ = new Hashtable<Media, StorageManager>();

    /* Support for handling object events */
    protected EventListenerList objectListeners;


    /**
     * Construct an instance of the AbstractObjectManager.
     * The constructor is protected, so that we can ensure there is
     * only one instance of the AbstractObjectManager, i.e. it's a singleton.
     */
    protected AbstractObjectManager() {
	objectListeners  = new EventListenerList();
	createRegistry();
    }



    /**
     * Install a new instance of the AbstractObjectManager.
     * 
     * @param objMgr	new AbstractObjectManager instance.
     */
    public static void setObjectManager(AbstractObjectManager objMgr) {
	objMgr_ = objMgr;
    }


    /**
     * Returns the appropiate Storage Manager for a particular Media type.
     * 
     * @param device The selected media that is being used
     * 
     * @return Storage manager for a particular device
     */
    public StorageManager getStorageManager(Media device) {
	StorageManager stMgr = mediaMgr_.get(device);

	if (stMgr == null) {
	    throw new IllegalArgumentException(
			    "unsupported media: " + device.toString()
			);
	}
	return stMgr;
    }

    /**
     * Set the StorageManager object for a given Media object.
     * 
     * @param device	the storage medium to be managed
     * @param stMgr	the storage manager object
     * 
     * @exception ObjectIOException
     * 			thrown if there is an I/O error while initialising
     * 			the StorageManager
     */
    public void setStorageManager(Media device, StorageManager stMgr)
			    throws ObjectIOException {
	mediaMgr_.put(device, stMgr);
	stMgr.initialise();
    }


    /**
     * Creates and installs a new instance of the Registry.
     * Hook for subclasses.
     */
    protected void createRegistry() {
	registry_ = new Registry();
    }




    /**
     * Startup hook for subclasses.
     * 
     * @see #restart
     * @see #shutdown
     */
    public abstract void  startup();


    /**
     * Shutdown hook for subclasses.
     * 
     * @see #startup
     * @see #restart
     */
    public abstract  void  shutdown();


    /**
     * Shuts down the current ObjectManager and AutoSaver and then
     * creates a new instance with the same StorageManager objects.
     * 
     * @exception ObjectIOException
     *			if an I/O error occurred during the operation
     * 
     * @see #startup
     * @see #shutdown
     */
    public void restart()
			throws ObjectIOException {
	Media device;

	shutdown();
	startup();

	Enumeration<Media> enum0 = mediaMgr_.keys();

	while (enum0.hasMoreElements()) {
	    device = enum0.nextElement();
	    objMgr_.setStorageManager(
		device,
		mediaMgr_.get(device)
	    );
	}
    }


    /**
     * Returns a deep copy of the given tree of business objects
     * after adding them to the Registry.
     * 
     * @param device	the storage medium to be managed
     * @param dir	directory to contain the copy objects
     * @param bo	top-level BusinessObject of 'tree' to copy
     * 
     * @exception ObjectIOException
     * 			thrown if there is an I/O error while registering
     * 
     * @see BusinessObject#deepCopy
     */
    public BusinessObject deepCopy(Media device,
				   DirectoryNode dir,
				   BusinessObject bo)
	                           throws ObjectIOException {
	return deepCopy(device, dir, bo, 0);
    }


    /**
     * Returns a deep copy of the given tree of business objects
     * after adding them to the Registry.
     * 
     * @param device	the storage medium to be managed
     * @param dir	directory to contain the copy objects
     * @param bo	top-level BusinessObject of 'tree' to copy
     * @param baseId	Id on which to base the new object id,
     * 			(StorageManager-dependent)
     * 
     * @exception ObjectIOException
     * 			thrown if there is an I/O error while registering
     * 
     * @see BusinessObject#deepCopy
     */
    public BusinessObject deepCopy(Media device, DirectoryNode dir,
					    BusinessObject bo, long baseId)
					    throws ObjectIOException {
	BusinessObject boCopy = bo.deepCopy();

	recursiveRegister(device, dir, boCopy, baseId);

	return boCopy;
    }


    /**
     * Adds each element of the given tree of business objects
     * to the Registry, if it's not already there.
     * 
     * @param device	the storage medium containing the objects
     * @param dir	directory containing the objects
     * @param bo	top-level BusinessObject of 'tree' to register
     * 
     * @exception ObjectIOException
     * 			thrown if there is an I/O error while registering
     * 
     * @see #deepCopy
     */
    public void recursiveRegister(Media device, DirectoryNode dir,
					    BusinessObject bo)
					    throws ObjectIOException {
	recursiveRegister(device, dir, bo, 0);
    }

    /**
     * Adds each element of the given tree of business objects
     * to the Registry, if it's not already there.
     * 
     * @param device	the storage medium containing the objects
     * @param dir	directory containing the objects
     * @param bo	top-level BusinessObject of 'tree' to register
     * @param baseId	Id on which to base the new object id,
     * 			(StorageManager-dependent)
     * 
     * @exception ObjectIOException
     * 			thrown if there is an I/O error while registering
     * 
     * @see #deepCopy
     */
    public void recursiveRegister(Media device, DirectoryNode dir,
					    BusinessObject bo, long baseId)
					    throws ObjectIOException {
	long id;	
	setBusObjId(device, bo, baseId);
	id = bo.getId();
        setBusObjIdRecursive(device,bo,id);
        recursiveRegisterInternal(device,dir,bo);

    }

    private void recursiveRegisterInternal(Media device,DirectoryNode dir,BusinessObject bo) throws ObjectIOException {
         BusinessObject[] children = bo.getBusObjChildren();
	 for (int j=0; j < children.length; j++) {
	     recursiveRegisterInternal(device,dir,children[j]);
	 }
         register(device,dir,bo);
    }


    private void setBusObjIdRecursive(Media device,BusinessObject bo,long baseId) throws ObjectIOException {	
        BusinessObject[] children = bo.getBusObjChildren();
	for (int j=0; j < children.length; j++) {
	    setBusObjIdRecursive(device,children[j],baseId);
	}
        setBusObjId(device, bo, baseId);
    }

    /**
     * Get the root DirectoryNode object for a given Media object.
     * If used the directory node should be implemented in the subclasses.
     * 
     * @param device The storage medium Local Store, RDBMS
     * 
     * @return Root Directory (null)
     * 
     * @exception ObjectIOException
     * 			thrown if the StorageManager for the media
     * 			could not return the root directory
     */
    public DirectoryNode getRoot(Media device)
			    throws ObjectIOException {
	return null;
    }

    /**
     * Set the id property of a (new) BusinessObject.
     * 
     * @param device	the medium on which the object is stored.
     * @param bo	the BusinessObject with the id to be initialised.
     * @param baseId	Id on which to base the new object id,
     * 			(StorageManager-dependent)
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */
    protected void setBusObjId(Media device, StorableObject bo, long baseId)
				throws ObjectIOException {
	if (bo.getId() == 0) {
	    bo.setId(
		getNewId(
		    device,
		    bo.getClass(),
		    baseId
		)
	    );
	}
    }

    /**
     * Asks the StorageManager to calculate a unique id.
     *
     * @param device	Media containing object for which id is needed.
     * @param objType	type of the object to be given an id
     * @return Unique id
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */
    protected long getNewId(Media device, Class objType)
				throws ObjectIOException {
	return getNewId(device, objType, 0);
    }


    /**
     * Asks the StorageManager to calculate a unique id.
     *
     * @param device	Media containing object for which id is needed.
     * @param objType	type of the object to be given an id
     * @param baseId	Id on which to base the new object id,
     * 			(StorageManager-dependent)
     * @return Unique id
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */
    protected long getNewId(Media device, Class objType, long baseId)
				throws ObjectIOException {
	return getStorageManager(device).getNewId(objType, baseId);
    }


    /**
     * Tests whether the specified object is in the Registry.
     *
     * @param id	id of a StorableObject
     * 
     * @return 		true if the specified object is in the Registry,
     * 			false otherwise.
     */
    public boolean isInRegistry(long id) {
	return (registry_.lookup(id) != null);
    }

    /**
     * Add a link to this StorableObject into the Registry
     * so that it can be found quickly when needed.  This is used to
     * ensure that we only have one copy of an object in memory at
     * any one time!
     * 
     * @param device	medium on which the object is stored.
     * @param dir	directory in which the object is/should be stored.
     * @param bo	StorableObject with the id to be initialised.
     */
    protected StorableObject linkBusObj(Media device, DirectoryNode dir,
			    boolean isNew, StorableObject bo) {
	RegistryEntry boReg = null;
	StorageManager stMgr = getStorageManager(device);
	RegistryEntry oi = new RegistryEntry(bo, device, stMgr, dir, isNew);

	// oi.obj    = bo;
	// oi.stMgr  = stMgr;
	// oi.device = device;
	// oi.dir    = dir;
	// oi.isNew_  = isNew;

	boReg = registry_.add(bo.getId(), oi);
	return boReg.getObject();
    }


    public RegistryEntry lookup(long id) {
       return registry_.lookup(id);        
    }








    /**
     * Delet a Business Object. Remove the object from the Registry.
     * 
     * @param device	the medium on which the object is deleted
     * @param dir	the directory in which the object is/should be deleted.
     * @param bo	the StorableObject with the id to be initialised.
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */     
    public void deleteBusObj(Media device, long id, Class objType)
				    throws ObjectNotFoundException,
					    ObjectIOException {
	StorageManager stMgr = getStorageManager(device);
	RegistryEntry oi = registry_.lookup(id);
	StorableObject sobj = null;
	DirectoryNode node = null;

	if (oi != null) {
	    node = oi.getDirectory();
	    sobj = oi.getObject();
	}
	else {
	    node = stMgr.find(id, objType);
	}

	stMgr.delete(node, id, objType);
	registry_.remove(id);

	fireObjectEvent(ObjectEvent.OBJECT_REMOVED,
			device, null, node, sobj, id, objType);
    }

  

    /**
     * Add the Business Object to the Registry.         
     * 
     * @param device	the medium on which the object is to be stored
     * @param sobj	the BusinessObject to be added.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */     
    public StorableObject register(Media device, StorableObject sobj)
				    throws ObjectIOException {
	return register(device, getRoot(device), sobj, 0);
    }



   /**
     * Add the Business Object to the Registry.   The object is already in persistence storage      
     * 
     * @param device	the medium on which the object is to be stored
     * @param sobj	the BusinessObject to be added.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */     
    public StorableObject registerPersistence(Media device, StorableObject sobj)
				    throws ObjectIOException {
	return registerPersistence(device, getRoot(device), sobj, 0);
    }




    /**
     * Add the Business Object to the Registry.         
     * 
     * @param device	the medium on which the object is to be stored
     * @param sobj	the BusinessObject to be added.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */     
    public StorableObject register(Media device, StorableObject sobj, long baseId)
				    throws ObjectIOException {
	return register(device, getRoot(device), sobj, baseId);
    }

    /**
     * Add a Business Object. Add the Business Object to the Registry.         
     * 
     * @param device	the medium on which the object is to be stored
     * @param dir	the directory in which the object is/should be stored.
     * @param sobj	the BusinessObject to be added.
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */     
    public StorableObject register(Media device, DirectoryNode dir,
						    StorableObject sobj)
				    throws ObjectIOException {
	return register(device, dir, sobj, 0);
    }

    /**
     * Add a New Storable Object.
     * 
     * @param device	the medium on which the object is to be stored
     * @param dir	the directory in which the object is/should be stored.
     * @param sobj	the StorableObject to be added.
     * @param baseId	Id on which to base the new object id,
     * 			(StorageManager-dependent)
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */     
    public StorableObject register(Media device, DirectoryNode dir,
				StorableObject sobj, long baseId)
						throws ObjectIOException {
	setBusObjId(device, sobj, baseId);
	if (sobj instanceof BusinessObject) {
	    ((BusinessObject)sobj).setOwnerId(Config.getCfg().getAccountId());
	}
	StorableObject sobjReg = linkBusObj(device, dir, true, sobj);
	fireObjectEvent(ObjectEvent.OBJECT_CREATED,
			device, null, dir, sobjReg,
			sobjReg.getId(), sobjReg.getClass());
	return sobjReg;
    }
  /**
     * Add a persistence Storable Object.
     * 
     * @param device	the medium on which the object is to be stored
     * @param dir	the directory in which the object is/should be stored.
     * @param sobj	the StorableObject to be added.
     * @param baseId	Id on which to base the new object id,
     * 			(StorageManager-dependent)
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */     
    public StorableObject registerPersistence(Media device, DirectoryNode dir,
				StorableObject sobj, long baseId)
						throws ObjectIOException {
	setBusObjId(device, sobj, baseId);
	if (sobj instanceof BusinessObject) {
	    ((BusinessObject)sobj).setOwnerId(Config.getCfg().getAccountId());
	}
	StorableObject sobjReg = linkBusObj(device, dir, false, sobj);
	fireObjectEvent(ObjectEvent.OBJECT_CREATED,
			device, null, dir, sobjReg,
			sobjReg.getId(), sobjReg.getClass());
	return sobjReg;
    }




    /**
     * Add a New Storable Object array.
     * 
     * @param device	the medium on which the object is to be stored
     * @param sobjArray	the StorableObject array to be added.
     * 			(StorageManager-dependent)
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */     
    public  StorableObject[] register(Media device,
			 StorableObject[] sobjArray)
	throws ObjectIOException {
	Vector<StorableObject> sobjVect = new Vector<StorableObject>();
	StorableObject[] currSobjArray = null;
	if (sobjArray != null) {

	    for (int i = 0; i < sobjArray.length; i++) {
		sobjVect.addElement(register( device,  sobjArray[i]));
	    }
	}
	currSobjArray = new StorableObject[sobjVect.size()];
	sobjVect.copyInto(currSobjArray);
	return currSobjArray;
    }


    /**
     * Add a persistence Storable Object array.
     * 
     * @param device	the medium on which the object is to be stored
     * @param sobjArray	the StorableObject array to be added.
     * 			(StorageManager-dependent)
     * 
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */     
    public  StorableObject[] registerPersistence(Media device,
			 StorableObject[] sobjArray)
	throws ObjectIOException {
	Vector<StorableObject> sobjVect = new Vector<StorableObject>();
	StorableObject[] currSobjArray = null;
	if (sobjArray != null) {

	    for (int i = 0; i < sobjArray.length; i++) {
		sobjVect.addElement(registerPersistence( device,  sobjArray[i]));
	    }
	}
	currSobjArray = new StorableObject[sobjVect.size()];
	sobjVect.copyInto(currSobjArray);
	return currSobjArray;
    }






    /**
     * Write a StorableObject to its permanent media.  The in-memory
     * copy remains available.
     * 
     * @param sobj	Object to be stored.
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */
    public void putBusObj(StorableObject sobj)
				throws ObjectNotFoundException,
					ObjectIOException {
	RegistryEntry oi = registry_.lookup(sobj.getId());
	UnpackedStorableObject objSpec;

	if (oi == null) {
	    throw new ObjectNotFoundException(
			    "object not in memory: "
			    + sobj.getId()
			);
	}

	objSpec = disassembleReferences(oi.getObject());
	StorageManager stMgr = oi.getStorageManager();

	if (oi.isNew()) {
	    stMgr.write(oi.getDirectory(), sobj.getId(), sobj.getClass(), objSpec);
	}
	else {
	    stMgr.update(oi.getDirectory(), sobj.getId(), sobj.getClass(), objSpec);
	}
	oi.setNew(false);
    }


    /**
     * Write a StorableObject to its permanent media.  The in-memory
     * copy remains available. It doesn't analyse all the nested business objects
     * 
     * @param sobj	Object to be stored.
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */
    public void putBusObjNoRecursive(StorableObject sobj)
				throws ObjectNotFoundException,
					ObjectIOException {
	stdlog_.debug("putBusObjNoRecursive");
	RegistryEntry oi = registry_.lookup(sobj.getId());
	// it contains the object but it doesn't disassemble the references
	UnpackedStorableObject objSpec = new UnpackedStorableObject(sobj);

	if (oi == null) {
	    throw new ObjectNotFoundException(
			    "object not in memory: "
			    + sobj.getId()
			);
	}

	StorageManager stMgr = oi.getStorageManager();

	if (oi.isNew()) {
	    stMgr.write(oi.getDirectory(), sobj.getId(), sobj.getClass(), objSpec);
	}
	else {
	    stMgr.update(oi.getDirectory(), sobj.getId(), sobj.getClass(), objSpec);
	}
	oi.setNew(false);
    }


    /**
     * Forget a BusinessObject, ie remove it from the Registry.
     * When using this method, be careful that there no classes
     * that are holding a reference to it.  If you try (eg) to write
     * the BusinessObject after removing it from the Registry, an
     * exception wil be thrown.
     * 
     * <p>
     * If the object is not currently in the Registry, the method
     * just returns.
     * 
     * @param id	Id of object to be forgotten
     * 
     */
    public void forgetBusObj(Media device,long id,Class objType) {
	RegistryEntry oi = registry_.lookup(id);

	if (oi != null) {
	    DirectoryNode node = oi.getDirectory();
	    StorableObject sobj = oi.getObject();
	    registry_.remove(id);
            fireObjectEvent(ObjectEvent.OBJECT_FORGOTTEN,
			device,null,node, sobj, id, objType);
	}
    }


    /**
     * Refresh a BusinessObject instance from external media.
     * This is useful in case the version in a database has changed
     * and we want the updated version, rather that the local cached copy.
     *<br>
     *Only simple properties are updated, 
     * and they must have both get and set methods.
     * fro tretrieving an obj from scratch, is better to 
     * use forgetBusObj
     * 
     * @param device	External medium on which the object is stored
     * @param id	Id of object to be refreshed
     * @param objType	Type of object to be refreshed
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * @exception ObjectIOException
     *			if there was an I/O error refreshing the object.
     */
    public BusinessObject refreshBusObj(Media device, long id, Class objType)
				throws ObjectNotFoundException,
					ObjectIOException {
	if ( ! BusinessObject.class.isAssignableFrom(objType)) {
	    throw new ObjectIOException(
			"refreshBusObj: cannot refresh this type of object: "
			+ objType.getName()
		    );
	}

	if ( ! isInRegistry(id) ) {
	    stdlog_.debug ("nothing to refresh. Not found id "+ id);
	    return null;	// Not in Registry -- nothing to refresh
	}

	StorageManager stMgr	    = getStorageManager(device);
	DirectoryNode node   	    = stMgr.find(id, objType);
	UnpackedStorableObject uso  = stMgr.read(node, id, objType);
	BusinessObject dbBO	    = (BusinessObject)uso.getObject();
	BusinessObject localBO	    = (BusinessObject)getBusObj(
					    device,
					    id,
					    objType
					);

	try {
	    String[] props = dbBO.simpleProperties();
	    dbBO.copyProperties(props, localBO);
	}
	catch (NoSuchMethodException e) {
	    throw new ObjectIOException(
		"refreshBusObj: NoSuchMethod: " + e.getMessage()
	    );
	}
	catch (InvocationTargetException e) {
	    throw new ObjectIOException(
		    "refreshBusObj: InvocationTargetException:"
		    + e.getTargetException().getMessage()
		);
	}
	catch (IllegalAccessException e) {
	    throw new ObjectIOException(
		    "refreshBusObj: IllegalAccess: " + e.getMessage()
		);
	}
    return localBO;
    }


    /**
     * Read a StorableObject from its permanent media, link it into the
     * StorableObject infrastructure, etc and return it.
     * 
     * @param device	the medium on which the StorableObject is stored.
     * @param id	the id of the StorableObject to store.
     * @param objType	the type of the object to store.
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     * 
     * @return StorableObject
     */
    public StorableObject getBusObjNoRecursive(Media device, long id, Class objType)
				throws ObjectNotFoundException,
					ObjectIOException {
	stdlog_.debug("getBusObjNoRecursive");
	RegistryEntry oi = registry_.lookup(id);   
	stdlog_.debug("id " +id);

	if (oi != null) {
	    return oi.getObject();
	}

	StorageManager stMgr		= getStorageManager(device);
	DirectoryNode node   		= stMgr.find(id, objType);
	UnpackedStorableObject objSpec	= stMgr.read(node, id, objType);

	if (objSpec == null) {
org.eso.ohs.core.utilities.DebugUtils.stackTrace("SHOULDN'T HAPPEN");
	    return null;
	}

	StorableObject sobj = (StorableObject)objSpec.getObject(); 

	linkBusObj(device, node, false, sobj);

	stdlog_.debug("linkBusObj");

	fireObjectEvent(ObjectEvent.OBJECT_READ,
			device, null, node, sobj,
			sobj.getId(), sobj.getClass());
	return sobj;
    }




    /**
     * Read a StorableObject from its permanent media, link it into the
     * StorableObject infrastructure, etc and return it.
     * 
     * @param device	the medium on which the StorableObject is stored.
     * @param id	the id of the StorableObject to store.
     * @param objType	the type of the object to store.
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     * 
     * @return StorableObject
     */
    public StorableObject getBusObj(Media device, long id, Class objType)
				throws ObjectNotFoundException,
					ObjectIOException {
	RegistryEntry oi = registry_.lookup(id);   

	if (oi != null) {
	    return oi.getObject();
	}

	StorageManager stMgr		= getStorageManager(device);
	DirectoryNode node   		= stMgr.find(id, objType);
	UnpackedStorableObject objSpec	= stMgr.read(node, id, objType);

	if (objSpec == null) {
org.eso.ohs.core.utilities.DebugUtils.stackTrace("SHOULDN'T HAPPEN");
	    return null;
	}

	StorableObject sobj = (StorableObject)objSpec.getObject(); 

	reassembleReferences(device, objSpec);

	linkBusObj(device, node, false, sobj);

	fireObjectEvent(ObjectEvent.OBJECT_READ,
			device, null, node, sobj,
			sobj.getId(), sobj.getClass());
	return sobj;
    }

    /**
     * Restore the saved the links in a StorableObject that has been read
     * from external media.
     * 
     * <p>
     * The objSpec parameter must follow the format as results from
     * StorageManager.read().
     * 
     * <p>
     * <b>Note:</b> XXX after migration to JDK 1.2, this method should
     *              ignore the setX/getX methods and directly set
     *              the transient fields.
     * 
     * @param device	the medium on which the StorableObject is stored.
     * @param objSpec	the array of objects read from the storage medium.
     * @param objType	the type of the object to store.
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     * 
     * @see StorageManager#read
     */
    protected void reassembleReferences(Media device, UnpackedStorableObject objSpec)
				    throws ObjectIOException,
					    ObjectNotFoundException {
	StorableObject sobj = objSpec.getObject();
	Enumeration enum0 = objSpec.getProperties();
	long lastMod = sobj instanceof BusinessObject
			    ? ((BusinessObject)sobj).getLastModifiedDate()
			    : -1;

	// ASSERT objSpec.length is odd

	while (enum0.hasMoreElements()) {
	    String propName = (String)enum0.nextElement();
	    long nestedId   = objSpec.getNestedObjId(propName);
	    Object nestedObj = null;

	    try {
		Class cl = sobj.getClass();
		Method getMeth = cl.getMethod("get" + propName,
				    new Class[]{}
				);
		Method setMeth = cl.getMethod("set" + propName,
				    new Class[]{getMeth.getReturnType()}
				);
	    
		if (nestedId != 0L) {
			    // XXX It should which dbase contains an objType
		    Class objType = getMeth.getReturnType();
		    Media dev = device;

		    if (objType == ObservingRun.class
				&& device.equals(Media.PHASE2_DB)) {
			dev = Media.PHASE1_DB;
		    }

		    nestedObj = getBusObj(
					dev,
					nestedId,
					objType
				    );	        
		}

		setMeth.invoke(sobj, new Object[]{ nestedObj });
	    }
	    catch (IllegalAccessException e) {
		stdlog_.error(e);
	    }
	    catch (InvocationTargetException e) {		
		stdlog_.error(e);
	    }
	    catch (NoSuchMethodException e) {
		stdlog_.error(e);
		throw new IllegalArgumentException("corrupt data: " + propName);
	    }
	}

	if (lastMod >= 0) {
	    ((BusinessObject)sobj).setLastModifiedDate(lastMod);
	}
    }


    /**
     * Save the links in a StorableObject that is to be written
     * to external media.
     * 
     * <p>
     * <b>Note:</b> XXX after migration to JDK 1.2, this method should
     *              ignore the setX/getX methods and directly set
     *              the transient fields.
     *
     * @param bo Use the StorableObject to create an array of Objects 
     *           that contains the field names and method names of the
     *           Objects that link to these objects.
     * @return Object[] A generic list of objects that contain the methods
     *                  and fields from the linked objects.
     *
     */
    protected UnpackedStorableObject disassembleReferences(StorableObject sobj) {
	UnpackedStorableObject uso = new UnpackedStorableObject(sobj);

	try {
	    Class cl = sobj.getClass();
	    Method[] meths = cl.getMethods();
	    Class<BusinessObject> boClass = BusinessObject.class;

	    for (int j=0; j < meths.length; j++) {
		Method m = meths[j];
		Class retType = m.getReturnType();

		    // Look for a getX() method returning a subclass
		    // of StorableObject.
		if (m.getName().startsWith("get")
			&& boClass.isAssignableFrom(retType)
			&& m.getParameterTypes().length == 0) {

		    BusinessObject field =
			(BusinessObject) m.invoke(sobj, NO_ARGS);

		    uso.addNestedObjId(
			m.getName().substring(3),
			field == null ? 0 : field.getId()
		    );
		}
	    }
	}
	catch (IllegalAccessException e) {
	    stdlog_.error(e);
	}
	catch (InvocationTargetException e) {
	    stdlog_.error(e);
	}

	return uso;
    }



  



    /**
     * Gets the lists of different object types(targets,ob's,od's) for a given directory
     * 
     *
     * @param device A reference to the selected media. 
     * @param node The directory that is used to select the appropiate
     *             directory
     * @return Returns a list of ObjectFileInfo for a given directory
     *
     * @exception ObjectIOException
     *			if there was an I/O error writing the object.
     */
  
    public Summary[] listObjects(Media device, DirectoryNode node)
				    throws ObjectIOException {
	return getStorageManager(device).listObjects(node);
    }



   

    /**
     * Adds a new listener to be notified of created/read/deleted objects
     * 
     * @param lnr The listener
     */
    public void addObjectListener(ObjectListener lnr) {
	objectListeners.add(ObjectListener.class, lnr);
    }

    /**
     * Removes a listener, which will no longer be notified of changes.
     * 
     * @param lnr The listener
     */
    public void removeObjectListener(ObjectListener lnr) {
	objectListeners.remove(ObjectListener.class, lnr);
    }

    /**
     * Inform listeners of a change to the Registry of StorableObjects
     *
     * @param evtType types of ObjecyEvent
     * @param dir directory that the event will take place on.
     * @param sobj reference to storable object
     */
    protected void fireObjectEvent(int evtType, Media device,
					DirectoryNode fromDir,
					DirectoryNode toDir,
					StorableObject sobj,
					long id,
					Class objType) {
	ObjectEvent objEvent = new ObjectEvent(
			    evtType, device, fromDir, toDir, sobj, id, objType);
	Object[] listeners = objectListeners.getListenerList();

	for (int i = listeners.length-2; i>=0; i-=2) {
	    Object lnr = listeners[i+1];
	    if (listeners[i] == ObjectListener.class) {
		if (evtType == ObjectEvent.OBJECT_READ
			|| evtType == ObjectEvent.OBJECT_CREATED) {
		    ((ObjectListener)lnr).objectAdded(objEvent);
		}
		else if (evtType == ObjectEvent.OBJECT_MOVED) {
		    ((ObjectListener)lnr).objectMoved(objEvent);
		}
		else if(evtType ==  ObjectEvent.OBJECT_REMOVED) {
		    ((ObjectListener)lnr).objectRemoved(objEvent);
		}
	        else if(evtType ==  ObjectEvent.OBJECT_FORGOTTEN){
		    ((ObjectListener)lnr).objectRemovedFromRegistry(objEvent);
		}
	    }              
	}
    }
}


