package org.eso.ohs.persistence;

import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.AccountInfo;
import org.eso.ohs.dfs.FindingChart;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.Readme;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.instruments.TemplateVersionInfo;

/**
 * The DecoratorStorageManager provides a Decorator pattern around
 * a StorageManager.
 *
 * The class its self does nothing, being nothing more than an
 * empty (transparent) decorator.
 * 
 * This class should be extended with the relevant methods overridden,
 * when doing so, they must remember to call the method in super (this class)
 * to ensure that the original (decorated) StorageManager is called.
 *
 * @see AuthenticatingStorageMgr
 * @see StorageManager
 */
public class DecoratorStorageManager implements StorageManager {

    private StorageManager stMgr_;

    public DecoratorStorageManager(StorageManager stMgr) {
        stMgr_ = stMgr;
    }

    public void setAccountId(int userId)
                throws ObjectIOException {
        stMgr_.setAccountId(userId);
    }

    public int getAccountId() {
        return stMgr_.getAccountId();
    }

    public DirectoryNode getRoot()
                throws ObjectIOException {
        return stMgr_.getRoot();
    }

    public String[] verifyTemplateSignatures(TemplateVersionInfo[] tsList)
                throws ObjectIOException {
        return stMgr_.verifyTemplateSignatures(tsList);
    }

    public FindingChart []  populateFindingCharts(long boId)
	    throws ObjectIOException {
	return stMgr_.populateFindingCharts(boId);
    }
    
    
    /**
     * Checkin the readme file for a given observing run
     * @param Readme file business object
     */
    
    public void checkInReadme(ObservingRun obsRun)
    throws ObjectIOException {
         stMgr_.checkInReadme(obsRun);
    }
    
    /**
     * Checkout the readme file for a given observing run.
     * @param boId Observing run definition. 
     */
    
    public String checkOutReadme(ObservingRun obsRun)
    throws ObjectIOException {
         return stMgr_.checkOutReadme(obsRun);
    }


    public byte[] downloadInstrumentPackage(String instrument, float version)
                throws ObjectIOException {
       return stMgr_.downloadInstrumentPackage(instrument, version);
    }

    public Media getMedia() {
        return stMgr_.getMedia();
    }


    public long getNewId(Class mo, long baseId)
                throws ObjectIOException {
        return stMgr_.getNewId(mo, baseId);
    }

    public void initialise()
                throws ObjectIOException {
        stMgr_.initialise();
    }

    public boolean loginUser(String password)
            throws ObjectIOException {
        return stMgr_.loginUser(password);
    }

    public DirectoryNode find(long id, Class objType)
                throws ObjectNotFoundException,
                    ObjectIOException {
        return stMgr_.find(id, objType);
    }

    public StorageManager.UnpackedStorableObject read(DirectoryNode node, long id,
                        Class objType)
                throws ObjectNotFoundException, ObjectIOException {
        return stMgr_.read(node, id, objType);
    }

    public void write(DirectoryNode node, long id, Class objType,
                    StorageManager.UnpackedStorableObject obj) throws ObjectIOException {
        stMgr_.write(node, id, objType, obj);
    }

    public void multiStore(StorageManager.MultiObjSpec[] mos)
                throws ObjectIOException {
        stMgr_.multiStore(mos);
    }

    public void update(DirectoryNode node, long id,
                Class objType, StorageManager.UnpackedStorableObject obj)
                throws ObjectIOException {
        stMgr_.update(node, id, objType, obj);
    }


    public void delete(DirectoryNode node, long id, Class objType)
                throws ObjectIOException {
        stMgr_.delete(node, id, objType);
    }

    public void move(long id, Class objType,
                DirectoryNode fromNode,
                DirectoryNode toNode)
                throws ObjectNotFoundException,
                    ObjectIOException {
        stMgr_.move(id, objType, fromNode, toNode);
    }

    public boolean createDirectory(DirectoryNode node, long id)
                throws ObjectIOException {
        return stMgr_.createDirectory(node, id);
    }


    public boolean moveDirectory(DirectoryNode fromNode,
                        DirectoryNode toNode)
                throws ObjectIOException,
                    ObjectNotFoundException {
        return stMgr_.moveDirectory(fromNode, toNode);
    }

    public boolean deleteDirectory(DirectoryNode node)
                throws ObjectIOException {
        return stMgr_.deleteDirectory(node);
    }


    public Summary[] listObjects(DirectoryNode node)
            throws ObjectIOException {

        return stMgr_.listObjects(node);
    }

    public void checkIn(StorageManager.MultiObjSpec[] moi)
                throws ObjectIOException {
        stMgr_.checkIn(moi);
    }

    public boolean verifyObject(StorageManager.MultiObjSpec[] moi)
                throws ObjectIOException {
        return stMgr_.verifyObject(moi);
    }

    public void checkOut(StorageManager.IdType[] idTypes)
                throws ObjectNotFoundException,
                    ObjectIOException {
        stMgr_.checkOut(idTypes);
    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#usdSubmit(org.eso.ohs.dfs.ObservingRun)
     */
    public void usdSubmit(ObservingRun obsRun) throws ObjectIOException {
        stMgr_.usdSubmit(obsRun);
    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#populateReadmeFile(org.eso.ohs.dfs.ObservingRun)
     */
    public Readme populateReadmeFile(ObservingRun obsRun) throws ObjectIOException {
        return stMgr_.populateReadmeFile(obsRun);
    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#setReadmeFileStatus(java.lang.String, long)
     */
    public void setReadmeFileStatus(String status, long runId) throws ObjectIOException {
        stMgr_.setReadmeFileStatus(status,runId);

    }

    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#populateReadmeFileReplicate(org.eso.ohs.dfs.ObservingRun)
     */
    public Readme populateReadmeFileReplicate(ObservingRun obsRun) throws ObjectIOException {
        return stMgr_.populateReadmeFileReplicate(obsRun);

    }


	/* (non-Javadoc)
	 * @see org.eso.ohs.persistence.StorageManager#getAccountInfo(java.lang.String, java.lang.String)
	 */
	public AccountInfo getAccountInfo(String username, String password)
			throws ObjectIOException {		
		return stMgr_.getAccountInfo(username, password);
	}

	/* (non-Javadoc)
	 * @see org.eso.ohs.persistence.StorageManager#getAccountInfo(int, java.lang.String)
	 */
	public AccountInfo getAccountInfo(int accountId, String password)
			throws ObjectIOException {
		return stMgr_.getAccountInfo(accountId, password);
	}
}
