package org.eso.ohs.persistence;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eso.ohs.core.utilities.Convert;
import org.eso.ohs.dfs.phase1.Cycles;
import org.eso.ohs.dfs.phase1.GtoInstruments;
import org.eso.ohs.dfs.phase1.GtoManagers;
import org.eso.ohs.dfs.phase1.GtoObservations;
import org.eso.ohs.dfs.phase1.InstrumentConfig;
import org.eso.ohs.dfs.phase1.Offered;
import org.eso.ohs.persistence.dao.GtoInstrumentsDAO;
import org.eso.ohs.persistence.dao.InstrumentConfigDAO;
import org.springframework.context.ApplicationContext;
import org.springframework.dao.DataAccessException;

/**
 * This class imports a target list from a text file.
 * 
 * @author pnunes 02.11.2005
 * 
 */
public class ImportGtoTargets {

	// regular expressions for the target parameters
	public static final String SEPARATOR = ",";

	private static final String RA_HOURS = "((0*[0-1]?[0-9]|0*2[0-3])";

	private static final String DEC_DEGREES = "((\\+|-)?(0*[0-8]?[0-9]|0*90)";

	private static final String MINUTES = "(0*[0-5]?[0-9])";

	private static final String SECONDS = "((0*[0-5]?[0-9])(\\.[0-9]*)?))";

	public static final String RA_EXPR = RA_HOURS + ":" + MINUTES + ":"
			+ SECONDS;

	public static final String RA_EXPR2 = RA_HOURS + " " + MINUTES + " "
			+ SECONDS;

	public static final String RA_EXPR3 = RA_HOURS + "h " + MINUTES + "m "
			+ SECONDS + "s";

	public static final String DEC_EXPR = DEC_DEGREES + ":" + MINUTES + ":"
			+ SECONDS;

	public static final String DEC_EXPR2 = DEC_DEGREES + " " + MINUTES + " "
			+ SECONDS;

	public static final String DEC_EXPR3 = DEC_DEGREES + "d " + MINUTES + "m "
			+ SECONDS + "s";;

	public static final String EQUINOX_EXPR = "((J|B)?([1-9]([0-9]{3})))?";

	public static final String PI_NAME_EXPR = "(([a-zA-Z0-9(\\-)*( *)(\\.)(\\')*])*)";

	public static final String TARGET_NAME_EXPR = "(([a-zA-Z0-9(\\+)*(\\-)*( *)(\\.)*])*)";

	// public static final String PROPER_MOTION_EXPR =
	// "(\\+|-)?([0-9]+)(\\.[0-9]*)?";
	// public static final String EPOCH_EXPR = "([0-9]{4}(\\.[0-9]?))?";

	// error messages
	public static final String INVALID_PROPER_MOTION_FORMAT_ERROR_MSG = "Invalid proper motion format";

	public static final String INVALID_PI_NAME_ERROR_MSG = "Invalid characters in the PI Surname";

	public static final String INVALID_RA_FORMAT_ERROR_MSG = "Invalid RA format or value out of range (00:00:00.000 - 23:59:59.999): ";

	public static final String INVALID_TARGET_NAME_ERROR_MSG = "Invalid Target name";

	public static final String INVALID_DEC_FORMAT_ERROR_MSG = "Invalid DEC format or value out of range (-90:00:00.000 - +90:00:00.000): ";

	public static final String INVALID_LINE_FORMAT_ERROR_MSG = "Invalid line format. A valid line must have no less than 10 and no more than 11, comma separated, parameters (don't use commas in the comments field) ";

	public static final String INVALID_INSTRUMENT_ERROR_MSG = "Invalid instrument ";

	public static final String INVALID_MODE_ERROR_MSG = "Invalid mode ";
	public static final String INVALID_PARAMETER_ERROR_MSG = "Invalid parameter";
	public static final String INVALID_EXPTIME_ERROR_MSG = "Invalid exptime, it must be a positive integer";
	public static final String INVALID_MOVING_TARGET_ERROR_MSG = "In the moving target field, valid values are \"yes\" or \"no\"";
	public static final String INVALID_GTO_LP_ERROR_MSG = "In the GTOLP field, valid values are \"yes\" or \"no\"";

	public static final String INVALID_EQUINOX_FORMAT_ERROR_MSG = "Invalid equinox format";

	// default values
	public static final String DEFAULT_EQUINOX = "J2000";

	public static final String DEFAULT_EPOCH = "2000";
	
	private int cycleId_ = -1;
	
	private int userId_ = -1;
	

	/**
	 * Imports a target list file
	 * 
	 * @param inputFile
	 * @return Target array
	 * @throws IOException
	 */
	public GtoObservations[] importGtoTargets(File inputFile, GtoManagers gtoManager, ApplicationContext context) throws IOException,
			GtoTargetImportException {
		
		GtoObservations[] res = null;
		String cycleId = gtoManager.getCycleId().toString();
		InstrumentConfigDAO instrumentConfigDAO = (InstrumentConfigDAO) context.getBean("instrumentConfigDAO");
		GtoInstrumentsDAO gtoInstrumentDAO = (GtoInstrumentsDAO) context.getBean("gtoInstrumentsDAO");
		Hashtable validInstrumentModes = getInstrumentModeTable(gtoManager, gtoInstrumentDAO, instrumentConfigDAO);
		//setUserId(userId);
		//retrieveValidInstrumentsModes(gtoManager);
		
		LinkedList<GtoObservations> targetList = new LinkedList<GtoObservations>();
		LineNumberReader reader = new LineNumberReader(
				new FileReader(inputFile));

		String line = reader.readLine();
		int lineNumber = reader.getLineNumber();
		Vector<String> gtoErrorList = new Vector<String>();
		
		while (line != null) {

			try {
				String[] values = parseLine(line, validInstrumentModes);

				if (values != null && values.length > 0) {
					GtoObservations gtoTarget = createGtoTarget(values, gtoManager, instrumentConfigDAO, cycleId);
					targetList.add(gtoTarget);
				}

				

			} catch (GtoTargetImportException e) {
				
				//throw new GtoTargetImportException("Error at line: "+ lineNumber + " : " + e.getMessage());
				System.out.println("Error at line: "+ lineNumber + " : " + e.getMessage());
				gtoErrorList.add("Error at line: "+ lineNumber + " : " + e.getMessage());
			}
			line = reader.readLine();
			lineNumber = reader.getLineNumber();
		}

		if (gtoErrorList.size() > 0) {
			System.out.println("There have been errors");
			GtoTargetImportException gtoExceptions = new GtoTargetImportException("Errors --- Errors --- Errors");
			gtoExceptions.setErrorList(gtoErrorList);
			System.out.println("Throw the exception list");
			throw gtoExceptions;
		}
		
		res = (GtoObservations[]) targetList.toArray(new GtoObservations[targetList.size()]);
		return (res);
	}

	
	private void setCycleId(String cycleIdString) {
		
		Integer cycleIdInteger = new Integer(cycleIdString);
		int cycleId_ = cycleIdInteger.intValue();
		
	}
	
	
	
	

	/**
	 * Parses a line
	 * 
	 * Assumes the following sequence, with tab separators Target name
	 * (mandatory), RA (mandatory), DEC (mandatory) , equinox (optional) ,
	 * propRA (optional), propDEC (optional) , Epoch (optional)
	 * 
	 * Other assumptions: - if equinox is not there, it uses the default value -
	 * propDEC and propRA should always come together - epoch is only to be
	 * there if there the two previous were defined, it also has a default when
	 * missing
	 * 
	 * 
	 * @param line
	 * @return a string array containing the values
	 * @throws GtoTargetImportException
	 */
	protected String[] parseLine(String line, Hashtable validInstrumentModes) throws GtoTargetImportException {
		
		
		String[] values = null;
		String trimmedLine = line.trim();
		if (trimmedLine.startsWith("#")) {
			return values;
		}
		
		LinkedList<String> valueList = new LinkedList<String>();

		StringTokenizer tok = new StringTokenizer(line, SEPARATOR);
		Pattern[] pattern;

		if (tok.countTokens() == 0)
			return (new String[0]);
		else if (tok.countTokens() > 11 || tok.countTokens() < 10) {
			throw new GtoTargetImportException(INVALID_LINE_FORMAT_ERROR_MSG);
		} else {

			// gets the pi name - mandatory
			if (tok.hasMoreTokens()) {
				String piName = tok.nextToken();
				pattern = new Pattern[1];
				pattern[0] = Pattern.compile(PI_NAME_EXPR);
				
					matchPatterns(valueList, pattern, piName,
							INVALID_PI_NAME_ERROR_MSG);
				

			}

//			 gets the target name - mandatory
			if (tok.hasMoreTokens()) {
				String targetName = tok.nextToken();
				pattern = new Pattern[1];
				pattern[0] = Pattern.compile(TARGET_NAME_EXPR);
				matchPatterns(valueList, pattern, targetName,
							INVALID_TARGET_NAME_ERROR_MSG);		
			}
			
			if (tok.hasMoreTokens()) {
				// gets the RA - mandatory, try the 3 different formats
				String raString = tok.nextToken();
				if(!Convert.isValidRA(raString)) {
					throw new GtoTargetImportException(INVALID_RA_FORMAT_ERROR_MSG + raString + " has an invalid format.");
				} else {
					valueList.add(raString);
				}
				
				
				//pattern = getRAPatterns();
				//matchPatterns(valueList, pattern, ra, INVALID_RA_FORMAT_ERROR_MSG);
			}

			if (tok.hasMoreTokens()) {
				// gets the DEC - mandatory
				String decString = tok.nextToken();
				if(!Convert.isValidDec(decString)) {
					throw new GtoTargetImportException(INVALID_DEC_FORMAT_ERROR_MSG + decString + " has an invalid format.");
				} else {
					valueList.add(decString);
					
				}
				
				//pattern = getDECPatterns();
				//matchPatterns(valueList, pattern, dec,INVALID_DEC_FORMAT_ERROR_MSG);
				
			}

			String instrument = "";
			if (tok.hasMoreTokens()) {
				// gets the INSTRUMENT - mandatory
				instrument = tok.nextToken();
				if (validInstrumentModes.containsKey(instrument)) {
					valueList.add(instrument);
				}else {
					String validInstruments = getValidInstruments(validInstrumentModes);
					throw new GtoTargetImportException(
							INVALID_INSTRUMENT_ERROR_MSG + instrument + " Valid instruments are: " + validInstruments);
				}
				
				
//				Vector instruments = new Vector();
//				instruments.add("FORS1");
//				instruments.add("AMBER");
//
//				if (instruments.contains(instrument)) {
//					valueList.add(instrument);
//				} else {
//					throw new GtoTargetImportException(
//							INVALID_INSTRUMENT_ERROR_MSG);
//				}
			}

			if (tok.hasMoreTokens()) {
				// gets the MODE - mandatory
				String mode = tok.nextToken();
				List modeList = (List) validInstrumentModes.get(instrument);
				if(modeList.contains(mode)) {
					valueList.add(mode);
				} else {
					String validModes = getValidModes(modeList);
					throw new GtoTargetImportException(INVALID_MODE_ERROR_MSG + mode + " Valid modes are: " + validModes);
				}
				
//				Vector modes = new Vector();
//				modes.add("IPOL");
//				modes.add("IMG");
//				modes.add("PRE-IMG");
//				modes.add("LR-HK");
//
//				if (modes.contains(mode)) {
//					valueList.add(mode);
//				} else {
//					throw new GtoTargetImportException(INVALID_MODE_ERROR_MSG);
//				}
			}

			if (tok.hasMoreTokens()) {
				// gets the PARAMETER - mandatory
				String parameter = tok.nextToken();

				
				if (!parameter.equals("")) {
					valueList.add(parameter);
				} else {
					throw new GtoTargetImportException(INVALID_PARAMETER_ERROR_MSG);
				}
			}
			
			if (tok.hasMoreTokens()) {
				// gets the EXPTIME - mandatory
				String exptime = tok.nextToken();

				Integer exposureTime = null;
				try {
					exposureTime = new Integer(exptime);
					
				}
				catch(NumberFormatException e) {
					throw new GtoTargetImportException(INVALID_EXPTIME_ERROR_MSG);
				}
				
				if (exposureTime.intValue() <= 0) {
					
					throw new GtoTargetImportException(INVALID_EXPTIME_ERROR_MSG);
					
				} else {
				  valueList.add(exptime);
				}
			}
				
			if (tok.hasMoreTokens()) {
				// gets the Moving Target flag - mandatory
				String moving = tok.nextToken();

				
				if (moving.toLowerCase().equals("yes") || moving.toLowerCase().equals("no")) {
					valueList.add(moving.toLowerCase());
				} else {
					throw new GtoTargetImportException(INVALID_MOVING_TARGET_ERROR_MSG);
				}
				
			}
			
			
			if (tok.hasMoreTokens()) {
				// gets the gto lp flag - mandatory
				String gtolp = tok.nextToken();

				
				if (gtolp.toLowerCase().equals("yes") || gtolp.toLowerCase().equals("no")) {
					valueList.add(gtolp.toLowerCase());
				} else {
					throw new GtoTargetImportException(INVALID_GTO_LP_ERROR_MSG);
				}
				
			}
			
			if (tok.hasMoreTokens()) {
				// gets the comments 
				String comments = tok.nextToken();
					valueList.add(comments);
				
			}
			
		}

		Object[] obj = valueList.toArray();
		values = new String[obj.length];
		for (int i = 0; i < values.length; i++) {
			values[i] = (String) obj[i];
		}

		return (values);

	}

	public String getValidModes(List modeList) {
		
		StringBuffer stringBuffer = new StringBuffer();
		for (int i =0; i< modeList.size(); i++) {
			stringBuffer.append(modeList.get(i));
			stringBuffer.append(" ");
		}
		
		String validModes = stringBuffer.toString();
		return validModes;
	}
	
public String getValidInstruments(Hashtable instrumentModes) {
		
		StringBuffer stringBuffer = new StringBuffer();
		for (Enumeration e = instrumentModes.keys() ; e.hasMoreElements() ;) {
			String instrument = (String) e.nextElement();
	         stringBuffer.append(instrument);
	         stringBuffer.append(" ");

	     }
		String validInstruments = stringBuffer.toString();
		return validInstruments;
	}

	/**
	 * Matches the value parameter against one of the patterns in the array If
	 * it doens't match any pattern, throws an exception Otherwise adds the
	 * value to the list
	 * 
	 * @param valueList
	 * @param patterns
	 * @param value
	 * @throws GtoTargetImportException
	 */
	private void matchPatterns(LinkedList<String> valueList, Pattern[] patterns,
			String value, String errorMessage) throws GtoTargetImportException {
		boolean matched = false;

		for (int i = 0; i < patterns.length; i++) {
			Matcher matcher = patterns[i].matcher(value);
			if (matcher.matches()) {
				matched = true;
				break;
			}

		}

		if (matched) {
			valueList.add(value);
		} else {
			throw new GtoTargetImportException(errorMessage);
		}
	}

	/**
	 * Creates a Target object, assuming the following order on the array:
	 * 
	 * Target name, RA, DEC, Equinox, PropRA, PropDEC, Epoch the last three are
	 * optional
	 * 
	 * @param values
	 * @return
	 */
	private GtoObservations createGtoTarget(String[] values, GtoManagers gtoManager,InstrumentConfigDAO instrumentConfigDAO, String cycleId) {
		GtoObservations target = new GtoObservations();

		//set gtoManager
		target.setGtoManagers(gtoManager);
		
		//set current cycle id
		//int cycleId = getCycleId();
		Cycles cycle = new Cycles();
		cycle.setId(new Integer(cycleId));
		target.setCycle(cycle);
		
		String name = values[0];
		target.setPiSurname(name);

		String targetName = values[1];
		target.setTargetName(targetName);
		
		String raString = values[2];
		//int ra = parseRA(raString);
		int ra = Convert.HHMMSSToMilliarcsec(raString);
		int ramillisec = Convert.HHMMSSToMillisec(raString);
		//int ra = 123456;
		target.setRa(new Integer(ra));
		target.setStandardFormatRa(raString);
		
		String decString = values[3];
		//int dec = parseDEC(decString);
		//int dec = 123456;
		int dec = Convert.DDMMSSToMilliarcsec(decString);
		target.setDec(new Integer(dec));
		target.setStandardFormatDec(decString);
		
		/** TODO
		 * Retrieve the instrument configuration.
		 */
		String instrument = values[4];
		String mode = values[5];
//		InstrumentConfig instrumentConfig = new InstrumentConfig();
//		instrumentConfig.getId().
//		
		Offered offeredInst = new Offered();
    	Offered.Id offeredId = new Offered.Id();
    	offeredId.setCycleId(new Integer(cycleId));
    	offeredId.setInstrumentCode(instrument);
    	offeredInst.setId(offeredId);
    	InstrumentConfig instConfig = new InstrumentConfig();
    	instConfig.setMode(mode);
    	instConfig.setOffered(offeredInst);
    	//instConfig.setInstrumentField(new Integer("12345"));
		InstrumentConfig instrumentConfigFromDb = instrumentConfigDAO.find(instConfig);
		target.setInstrumentConfig(instrumentConfigFromDb);
		
		String parameter = values[6];
		target.setParameter(parameter);
		
		String exptimeString = values[7];
		Integer exptimeInteger = new Integer(exptimeString);
		int exptime = exptimeInteger.intValue();
		target.setExposureTime(new Integer(exptime));
		
		String movingTargetString = values[8];
		byte movingTarget = 0;
		if (movingTargetString.equals("yes")) {
			movingTarget = 1;
		}
		target.setMovingFlag(new Byte(movingTarget));
		
		String gtoLpString = values[9];
		byte gtoLp = 0;
		if (gtoLpString.equals("yes")) {
			gtoLp = 1;
		}
		target.setGtolpFlag(new Byte(gtoLp));
		
		String comments = values[10];
		target.setComments(comments);
		
		return (target);
	}

	private int getCycleId() {
		System.out.println("return the cycle id " + cycleId_);
		return cycleId_;
	}

	/**
	 * @param ra
	 * @return
	 */
	protected int parseRA(String ra) {
		Pattern[] patterns = getRAPatterns();

		Matcher matcher = null;
		for (int i = 0; i < patterns.length; i++) {
			matcher = patterns[i].matcher(ra);
			if (matcher.matches()) {
				break;
			}
		}

		String hours = matcher.group(2);
		String minutes = matcher.group(3);
		String seconds = matcher.group(4);
		String s = hours + ":" + minutes + ":" + seconds;
		int result = Convert.HHMMSSToMillisec(s);

		return (result);

	}

	/**
	 * @param dec
	 * @return
	 */
	protected int parseDEC(String dec) {

		Pattern[] patterns = getDECPatterns();

		Matcher matcher = null;
		for (int i = 0; i < patterns.length; i++) {
			matcher = patterns[i].matcher(dec);
			if (matcher.matches()) {
				break;
			}
		}

		String hours = null;
		if (matcher.group(2) != null)
			hours = matcher.group(2) + matcher.group(3);
		else
			hours = matcher.group(3);

		String minutes = matcher.group(4);
		String seconds = matcher.group(5);

		String s = hours + ":" + minutes + ":" + seconds;
		int result = Convert.DDMMSSToMilliarcsec(s);

		return (result);

	}

	private Pattern[] getRAPatterns() {
		Pattern[] patterns = new Pattern[3];
		patterns = new Pattern[3];
		patterns[0] = Pattern.compile(RA_EXPR);
		patterns[1] = Pattern.compile(RA_EXPR2);
		patterns[2] = Pattern.compile(RA_EXPR3);
		return (patterns);

	}

	private Pattern[] getDECPatterns() {
		Pattern[] patterns = new Pattern[3];
		patterns = new Pattern[3];
		patterns[0] = Pattern.compile(DEC_EXPR);
		patterns[1] = Pattern.compile(DEC_EXPR2);
		patterns[2] = Pattern.compile(DEC_EXPR3);
		return (patterns);

	}

	/**
	 * @return Returns the userId.
	 */
	public int getUserId() {
		return userId_;
	}

	/**
	 * @param userId The userId to set.
	 */
	public void setUserId(String userIdString) {
		Integer userIdInteger = new Integer(userIdString);
		int userId_ = userIdInteger.intValue();
	}
	
private Hashtable getInstrumentModeTable(GtoManagers gtoManager, GtoInstrumentsDAO gtoInstrumentDAO, InstrumentConfigDAO instrumentConfigDAO) {
		
		Hashtable<String,List<String>> instrumentModeTable = new Hashtable<String,List<String>>();
		
		Integer cycleId = gtoManager.getCycleId();
		List gtoInstrumentList = null;
		try {
			System.out.println("Retrieving instruments for the manager");
			gtoInstrumentList = gtoInstrumentDAO.find(gtoManager);
		} catch (DataAccessException e1) {
			System.out.println("Database error");
		} catch (IllegalArgumentException e2) {	
			System.out.println("Instruments not found");
			return instrumentModeTable;
		}
		
		for (int i = 0; i<gtoInstrumentList.size(); i++) {
			
			GtoInstruments instrumentObj = (GtoInstruments) gtoInstrumentList.get(i);
			String instrument = instrumentObj.getId().getInstrument();
			//String instrument = (String)gtoInstrumentList.get(i);
			List<String> modeList;
			try {
				List instrumentConfigList = instrumentConfigDAO.findByCycleIdInstrument(cycleId, instrument.trim());
				modeList = getModeList(instrumentConfigList);
				instrumentModeTable.put(instrument.trim(), modeList);
			} catch (DataAccessException e) {
			}
			
		}
		
		return instrumentModeTable;
	}
	
	
private List<String> getModeList(List instrumentConfigList) {
	
	Vector<String> modeVector = new Vector<String>();
	for (int i = 0; i<instrumentConfigList.size(); i++) {
		InstrumentConfig instrumentConfig = (InstrumentConfig) instrumentConfigList.get(i);
		String mode = instrumentConfig.getMode();
		modeVector.add(mode);
	}
	
	return modeVector;
	
}
	
	
}
