package org.eso.ohs.persistence;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.Convert;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.instruments.TemplateSignature;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;


/**
 * This class imports an XML file produced by the Survey Area Definition Tool (SADT)
 * and creates ObservationBlock business objects based on the file and an initial OB
 * 
 * @author pnunes 14.02.2006
 *
 */
public class ImportSurvey {
	

	/** Standard log channel */
    private static Logger stdlog_ = Logger.getLogger(ImportSurvey.class);
    
    
    public static String TILE_TEMPLATE_NAME = "VIRCAM_img_obs_tile";
    public static String TILE_PATTERN_PARAM = "SEQ.TILE.ID";
    public static String PAW_PRINT_OFFSETX_PARAM = "OFFSETX";
    public static String PAW_PRINT_OFFSETY_PARAM = "OFFSETY";
    public static String PAW_PRINT_ROTATION_PARAM = "OFFSETY";
    
	
    public static String TILE_PATTERN_XML_ATTR = "pattern";
    public static String PAW_PRINT_OFFSETX_XML_ATTR ="offsetx";
    public static String PAW_PRINT_OFFSETY_XML_ATTR ="offsety";
    
						
	/**
	 * Imports a survey definition File and returns an array of ObservationBlock
	 * objects based on an parent OB
	 * 
	 * @param sadtFile
	 * @param parentOB
	 * @return
	 * @throws ImportSurveyException
	 * @throws IOException
	 */
	public ObservationBlock [] importSADTFile(File sadtFile, ObservationBlock parentOB) throws ImportSurveyException, IOException {
		
		if (sadtFile == null) {
			throw new IllegalArgumentException("Parameter sadtFile is null");
		}
		
		if (sadtFile == null) {
			throw new IllegalArgumentException("Parameter parentOB is null");
		}
		
		LinkedList<ObservationBlock> blocks = new LinkedList<ObservationBlock>();
		
		// Code to parse the XML
		SAXBuilder builder = new SAXBuilder();
		try {
			Document doc = builder.build(sadtFile);
			
			Element root = doc.getRootElement();
	        List elements = root.getChildren();

	        stdlog_.debug("Importing " + sadtFile.toString());	
	        
	        
	        for (Iterator iter = elements.iterator(); iter.hasNext();) {
	            Element node = (Element) iter.next();	            	           
	            	            	
				// get all the tiles and create an OB per tile
				List allChildren = node.getChildren();
							
				for (Iterator iterator = allChildren.iterator(); iterator
						.hasNext();) {
					Element tile = (Element) iterator.next();
					
					stdlog_.debug("Reading tile " + tile.toString());	
					
					ObservationBlock ob = (ObservationBlock)parentOB.deepCopy();
			
					setTarget(tile, ob);
					//setTemplate(tile,ob);
					
					blocks.add(ob);					
				}
	        }  
	        

		} catch (JDOMException e) {
			e.printStackTrace();
			throw new ImportSurveyException(e);
		} 
		
		

		// end while loop xml file
		
		
		
		ObservationBlock [] result = new ObservationBlock[blocks.size()];
		for (int i = 0; i < blocks.size(); i++) {
			result[i] = blocks.get(i);
		}
		
		return(result);
		
	}

	
	/*
	 * SEQ.REF.FILES           (none)          List of reference setup files containing
                                        telescope SETUP keywords defining
                                        AG and AO stars for each pawprint.
	 */
	protected void setTemplate(Element tile, ObservationBlock ob) {
						
		// <PAWPRINT angle="0.0" dec="0.6149350222222223" id="0_4:1" offsetx="-0.475" offsety="0.0" ra="11.836696347576138">
		
		TemplateSignature [] templates = ob.getOd().getSignatures();
		
		for (int i = 0; i < templates.length; i++) {
			TemplateSignature temp = templates[i];
			
			String name = temp.getTemplateName();
			if (name.equals(TILE_TEMPLATE_NAME)) {
				
				
				// sets the pattern
				String pattern = tile.getAttribute(TILE_PATTERN_XML_ATTR).toString();				
				temp.setParameterValue(TILE_PATTERN_PARAM,pattern);
				
				// sets the tile offsets 				
				
				//	get all the pawprints 
				List allChildren = tile.getChildren();
							
				for (Iterator iterator = allChildren.iterator(); iterator
						.hasNext();) {
					Element pawPrint = (Element) iterator.next();
					
					stdlog_.debug("Reading paw print" + pawPrint.toString());			
				}
			}
		}
	}
	
	
	

	/**
	 * @param tile
	 * @param ob
	 */
	private void setTarget(Element tile, ObservationBlock ob) {
		// get RA and DEC and (eventually) other stuff from the file
		String raStr = tile.getAttribute("ra").getValue();
		String decStr = tile.getAttribute("dec").getValue();
		
		// in the SDT file, RA comes in degrees
		double raDouble = new Double(raStr).doubleValue() / 15;
		double decDouble = new Double(decStr).doubleValue();
			
		
		int RA = Convert.degreesToMilliarcsecs(raDouble);
		int DEC = Convert.degreesToMilliarcsecs(decDouble);

		stdlog_.debug("RA: " + raStr + " Milliarcsecs:" + RA );
		stdlog_.debug("DEC: " + decStr + " Milliarcsecs:" + DEC );
		
		ob.getTarget().setRA(RA);
		ob.getTarget().setDeclination(DEC);
	}
	
	/**
	 * Copies the relevant values from obSource to obDest
	 * 
	 * @param obSource
	 * @param obDest
	 */
	public void copyImportedValues(ObservationBlock obSource , ObservationBlock obDest) {
		
		if (obSource == null) {
			throw new IllegalArgumentException("Parameter obSource is null");
		}
		
		if (obDest == null) {
			throw new IllegalArgumentException("Parameter obDest is null");
		}
		
		
		obDest.getTarget().setRA(  obSource.getTarget().getRA() );
		obDest.getTarget().setDeclination(  obSource.getTarget().getDeclination() );
		
	}
	
}
