package org.eso.ohs.persistence.dbase;

import java.io.IOException;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import javax.mail.MessagingException;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ConnectionPool;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.core.utilities.Sybase;
import org.eso.ohs.core.utilities.javamail.MailManagement;
import org.eso.ohs.dfs.AccountInfo;
import org.eso.ohs.dfs.BusinessObject;
import org.eso.ohs.dfs.CalibrationBlock;
import org.eso.ohs.dfs.FindingChart;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.Readme;
import org.eso.ohs.dfs.StorableObject;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.instruments.TemplateVersionInfo;
import org.eso.ohs.persistence.AppConfig;
import org.eso.ohs.persistence.Config;
import org.eso.ohs.persistence.DirectoryNode;
import org.eso.ohs.persistence.Media;
import org.eso.ohs.persistence.OHSApplication;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.persistence.StorageManager;
import org.eso.ohs.persistence.dbase.phase2.FindingChartsDBIO;
import org.eso.ohs.persistence.dbase.phase2.OBHistory;
import org.eso.ohs.persistence.dbase.phase2.ReadmeFileDBIO;
import org.jdom.JDOMException;

/**
 * Abstract superclass for database storage manager. Contains generic methods
 * for read, write, update and delete, for acquiring connections, etc.
 * 
 * @version $Revision: 1.8.4.7 $
 * 
 * @author Bob Kemp
 * 
 * @see StorageManager
 * @see org.eso.ohs.persistence.ObjectManager
 */
public abstract class DbaseStorageMgr implements StorageManager {

    /**
     * <br>
     * ChangeLog: amchavan, 15-Jan-2002: added
     */
    private final static String rcsid =
        "$Id: DbaseStorageMgr.java,v 1.8.4.7 2007/11/06 14:19:44 pnunes Exp $";

    /**
     * <br>
     * ChangeLog: amchavan, 15-Jan-2002: added
     */
    private static Logger stdlog_ = Logger.getLogger(DbaseStorageMgr.class);

    /** DirectoryNode object representing the subtreeDir_ directory */
    private DirectoryNode rootNode_;

    /** UserId of the currently logged-in user */
    private int userId_;

    /** Media to which we are interfacing. Possibly reset in constructor */
    private Media device_ = Media.DBASE;

    /**
     * Field registry_
     */
    protected HandlerRegistry registry_;

    /**
     * Construct a storage manager for the given directory subtree. This subtree
     * should be in an area such that the user will not alter the contents of
     * the hierarchy.
     * 
     * @param device
     *            Media to which we are interfacing.
     */
    public DbaseStorageMgr(Media device) {
        this();

        device_ = device;
    }

    /**
     * Construct a storage manager for the given directory subtree. This subtree
     * should be in an area such that the user will not alter the contents of
     * the hierarchy.
     */
    public DbaseStorageMgr() {
        createRegistry();
        rootNode_ = new DirectoryNode(-2);
    }

    /**
     * Hook to allow subclasses to create specialised versions of their
     * DbaseIOHandler registries.
     */
    protected abstract void createRegistry();

    /**
     * Sets the id of the current user.
     * 
     * @param userId
     *            the id of the end-user
     * @see org.eso.ohs.persistence.StorageManager#setAccountId(int)
     */
    public void setAccountId(int userId) {
        userId_ = userId;
        registry_.setUserId(userId);
    }

    /**
     * Gets the id of the current user.
     * 
     * @return the id of the end-user
     * @see org.eso.ohs.persistence.StorageManager#getAccountId()
     */
    public int getAccountId() {
        return userId_;
    }

    /**
     * Connects to the database where objects of some class are stored.
     * 
     * @param objType
     *            The class we're interested in.
     * 
     * @return A JDBC Connection object, pointing to the database where objects
     *         of the <tt>objType</tt>> class are stored.
     * 
     * @exception ObjectIOException
     *                an error occured during an access to a business object
     * 
     * <br>
     * ChangeLog: amchavan, 15-Jan-2002: fixed bug, reformatted, fixed Javadoc.
     * @exception SQLException
     *                an error occurred during the dbase operation
     */
    protected Connection acquireConnection(Class objType)
        throws ObjectIOException, SQLException {
        Config cfg = Config.getCfg();
        String suffix = BusinessObject.getSuffix(objType);
        String url = cfg.getUrl(suffix);
        String user = cfg.getDbUserName();
        String pwd = cfg.getDbPassword();

        stdlog_.debug("The url: \"" + url + "\" the user \"" + user + "\" ");
        return ConnectionPool.getPool().acquireConnection(url, user, pwd);
    }

    /**
     * Returns the cached JDBC database Connection object creating it first, if
     * necessary.
     * 
     * @return JDBC Connection object
     * 
     * @exception ObjectIOException
     *                an error occured during an access to a business object
     * @exception SQLException
     *                an error occurred during the dbase operation
     */
    protected abstract Connection acquireConnection()
        throws ObjectIOException, SQLException;

    /**
     * Returns the cached JDBC database Connection object creating it first, if
     * necessary.
     * 
     * @param con
     *            Connection
     * @exception ObjectIOException
     *                an error occured during an access to a business object
     */
    protected void releaseConnection(Connection con) throws ObjectIOException {
        ConnectionPool.getPool().releaseConnection(con);
    }

    /**
     * The root node of the directory tree.
     * 
     * @return a representation of the root of the directory tree.
     * @see org.eso.ohs.persistence.StorageManager#getRoot()
     */
    public DirectoryNode getRoot() {
        return rootNode_;
    }

    /**
     * Verify that the TemplateSignature files used by the StorageManager are
     * the same version as those specified (and used by the client)
     * 
     * @param tsList
     *            array of version info objects for each template used in an OB
     * 
     * @return a list of the names of TemplateSignatures that are out of date.
     * 
     * @exception ObjectIOException
     *                if there was an I/O error while checking
     * @see org.eso.ohs.persistence.StorageManager#verifyTemplateSignatures(TemplateVersionInfo[])
     */
    public String[] verifyTemplateSignatures(TemplateVersionInfo[] tsList)
        throws ObjectIOException {
        // We're local, so by definition the templates are
        //  sufficiently uptodate.
        return new String[] {
        };
    }

    /**
     * Method populateFindingCharts
     * 
     * @param boId
     *            long
     * @return FindingChart[]
     * @throws ObjectIOException
     * @see org.eso.ohs.persistence.StorageManager#populateFindingCharts(long)
     */
    public FindingChart[] populateFindingCharts(long boId)
        throws ObjectIOException {
        Connection con = null;
        SQLException error;
        try {
            con = acquireConnection(ObservationBlock.class);
            FindingChart fcs[] =
                FindingChartsDBIO.read(
                    con,
                    Config.getCfg().uniqueToTableId(boId),
                    true);
            return fcs;
        } catch (SQLException e) {
            error = null;
        } catch (IOException e) {
            throw new ObjectIOException(e);
        } catch (ObjectNotFoundException e) {
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                releaseConnection(con);
            }
        }
        throw new ObjectIOException(error.getMessage());
    }

    /**
     * Method to checkin a readme file into the repository.
     * 
     * @param boId
     *            long
     * @see org.eso.ohs.persistence.StorageManager#populateFindingCharts(long)
     */
    public void checkInReadme(ObservingRun obsRun) throws ObjectIOException {
        Connection con = null;
        ObjectIOException error = null;
        String emailaddress = null;
        try {

            con = acquireConnection(ObservationBlock.class);
            emailaddress = ReadmeFileDBIO.write(con, obsRun);
            con.commit();
        } catch (SQLException e) {
            e.printStackTrace();
            try {
                con.rollback();
            } catch (SQLException e1) {
                e1.printStackTrace();
            }
            error = new ObjectIOException(e.getMessage());
        } catch (IOException e) {
            e.printStackTrace();
            error = new ObjectIOException(e.getMessage());
        } catch (ObjectNotFoundException e) {
            e.printStackTrace();
            error = new ObjectIOException(e.getMessage());
        } catch (ObjectIOException e) {
            error = e;
        }
        if (con != null) {
            try {
                con.setAutoCommit(true);
            } catch (SQLException e1) {
                // TODO Auto-generated catch block
                e1.printStackTrace();
            }
            releaseConnection(con);
        }
        if (error != null)
            throw error;

        if (error == null) {
            if (emailaddress != null && emailaddress.length() > 0) {
                try {
                    sendEmail(
                        emailaddress,
                        obsRun.getProgId() + " CHECKED IN ",
                        "Readme file checked in please run OT for details");
                } catch (MessagingException e1) {
                    stdlog_.error(e1.getMessage());
                }
            }
        }
    }
    
    


    /**
     * @param obsRun
     * @param emailaddress
     */
    private synchronized void sendEmail  (
        String emailaddress,
        String subject,
        String body)  throws MessagingException {
        MailManagement mgmt = new MailManagement();
        mgmt.setSmtpHost(AppConfig.getAppConfig().getSmtpName());
            stdlog_.debug("Seding Mail " + emailaddress);
            mgmt.postTextPlainMail(emailaddress, "", subject, body);
         
    }

    public void usdSubmit(ObservingRun obsRun) throws ObjectIOException {
        Connection con = null;
        SQLException error;
        String emailaddress;
        try {
            con = acquireConnection(ObservationBlock.class);
            int countReadme =
                ReadmeFileDBIO.getReadmeExists(
                    con,
                    Config.getCfg().uniqueToTableId(obsRun.getId()));
            int countOBS =
                ReadmeFileDBIO.getHasObs(
                    con,
                    Config.getCfg().uniqueToTableId(obsRun.getId()));
            if (countOBS <= 0 && countReadme <= 0) {
                throw new ObjectIOException(" No OBs or Readme have been submiited for this run. p2pp-submit failed");
            }
            if (countOBS <= 0) {
                throw new ObjectIOException(" No OBs have been submitted for this run. p2pp-submit failed");
            }
            if (countReadme <= 0) {
                throw new ObjectIOException(" No Readme has been submitted for this run. p2pp-submit failed");
            }
            boolean notAcceptedAndCheckedOut =
                ReadmeFileDBIO.notAcceptedAndCheckedOut(
                    con,
                    Config.getCfg().uniqueToTableId(obsRun.getId()));
            if(notAcceptedAndCheckedOut) {
                throw new ObjectIOException(" Please checkin your readme file for this run");
            }
        } catch (SQLException e) {
            e.printStackTrace();
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                releaseConnection(con);
            }
        }
        try {
            sendEmail(
                AppConfig.getAppConfig().getP2PPSubmit(),
                obsRun.getProgId(),
                "");
        } catch (MessagingException e1) {
            stdlog_.error(e1.getMessage());
            throw new ObjectIOException("ESO mail server is down\nPlease try again later");
        }
    }

    /**
     * 
     * @param boId
     *            long
     * @see org.eso.ohs.persistence.StorageManager#populateFindingCharts(long)
     */
    public String checkOutReadme(ObservingRun obsRun)
        throws ObjectIOException {
        Connection con = null;
        SQLException error;
        String readme = null;
        String emailaddress = null;
        try {
            con = acquireConnection(ObservationBlock.class);
            readme =
                ReadmeFileDBIO.checkOut(
                    con,
                    Config.getCfg().uniqueToTableId(obsRun.getId()));
            emailaddress =
                ReadmeFileDBIO.getUSDEmailAddress(
                    con,
                    Config.getCfg().uniqueToTableId(obsRun.getId()));
        } catch (SQLException e) {
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                releaseConnection(con);
            }
        }
        if (emailaddress != null && emailaddress.length() > 0) {
            try {
                sendEmail(
                    emailaddress,
                    obsRun.getProgId() + " Readme Checked Out ",
                    "");
            } catch (MessagingException e1) {
                stdlog_.error(e1.getMessage());
            }
        }
        return readme;
    }

    /**
     * @param instrument
     *            String
     * @param version
     *            float
     * @return the requested IP as a byte array, or null if the IP does not
     *         exist on the server.
     * @exception ObjectIOException
     *                when there is an I/O exception.
     * @see org.eso.ohs.persistence.StorageManager#downloadInstrumentPackage(String,
     *      float)
     */
    public byte[] downloadInstrumentPackage(String instrument, float version)
        throws ObjectIOException {
        return null;
    }

    /**
     * Get the type of media maintained by this StorageManager instance.
     * 
     * @return an instance of the Media class.
     * @see org.eso.ohs.persistence.StorageManager#getMedia()
     */
    public Media getMedia() {
        return device_;
    }

    /**
     * Calculate a unique id for local storage.
     * 
     * @param objType
     *            Class
     * @param baseId
     *            long
     * @return Unique id
     * 
     * @exception ObjectIOException
     *                if there was an I/O error writing the object.
     * @see org.eso.ohs.persistence.StorageManager#getNewId(Class, long)
     */
    public synchronized long getNewId(Class objType, long baseId)
        throws ObjectIOException {
        Connection con = null;
        CallableStatement cstmt = null;
        ResultSet rs = null;
        Config cfg = Config.getCfg();

        if (baseId != 0) {
            return cfg.tableToUniqueId(cfg.uniqueToTableId(baseId), objType);
        }

        try {
            String obrepDbName =
                Config.getCfg().getDbName(ObservationBlock.getSuffix());
            con = acquireConnection(ObservationBlock.class);
            cstmt =
                con.prepareCall(
                    "{? = call " + obrepDbName + "..opc_get_key(?)}");

            cstmt.registerOutParameter(1, java.sql.Types.INTEGER);

            cstmt.setString(2, "BUS_OBJ_ID");

            // The following calls match the code for
            //   the stored procedure 'opc_get_key'.
            // If you change one, you _must_ change the other.
            stdlog_.debug(">>>>>>>>>>>>> C Statement " + cstmt);
            cstmt.execute();
            cstmt.getMoreResults();
            cstmt.getMoreResults();
            rs = cstmt.getResultSet();
            if (rs != null) {
                rs.next();
                stdlog_.debug(
                    " !!!!!!!!! Config "
                        + Config.getCfg().tableToUniqueId(
                            rs.getLong(1),
                            objType));
                return Config.getCfg().tableToUniqueId(rs.getLong(1), objType);
            }
        } catch (SQLException e) {
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                try {
                    if (cstmt != null) {
                        cstmt.close();
                    }
                } catch (SQLException e) {
                    e.printStackTrace();
                } finally {
                    releaseConnection(con);
                }
            }
        }

        throw new ObjectIOException("DbaseStorageMgr: could not generate a new database id");
    }

    /**
     * Returns a DirectoryNode representing the path within the storage medium
     * of an object file with the given id and type.
     * 
     * @param id
     *            id of the StorableObject's file.
     * @param objType
     *            type of stored StorableObject.
     * 
     * @return the directory in which to find the object.
     * @see org.eso.ohs.persistence.StorageManager#find(long, Class)
     */
    public DirectoryNode find(long id, Class objType) {
        return getRoot();
    }

    /**
     * Log user into the database
     * 
     * @param password
     * @return boolean true on successful login else false
     * 
     * @exception ObjectIOException
     *                Wraps any exception that occurred during execution,
     *                usually an SQLexception from accessing the database.
     * 
     * <br>
     * ChangeLog: amchavan, 28-Aug-2001: fixed bug: was looping forever
     * @see org.eso.ohs.persistence.StorageManager#loginUser(String)
     */
    public boolean loginUser(String password) throws ObjectIOException {

        Connection con = null;
        PreparedStatement stmt = null;

        try {
            con = acquireConnection(ObservingRun.class);
            if (con == null)
                stdlog_.debug("?? Connection is NULL");
            stmt =
                con.prepareStatement(
                    "SELECT rtrim(passwd) from accounts "
                        + "WHERE account_id = ? "
                        + "AND passwd = ?");
            stmt.setInt(1, getAccountId());
            stmt.setString(2, password);
            stmt.execute();

            ResultSet rs = stmt.getResultSet();

            if (!rs.next()) { // Should give one result set
                return false;
            }

            return true;
        } catch (SQLException e) {

        	handleSQLException(e);
        	throw new ObjectIOException(e);

        } catch (Exception e) {
            e.printStackTrace();
            stdlog_.error("!! GET CLASS " + e.getClass());
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                try {
                    if (stmt != null) {
                        stmt.close();
                    }
                } catch (SQLException e) {
                    throw new ObjectIOException(e);
                } finally {
                    releaseConnection(con);
                }
            }
        }
    }
    
    
    /* (non-Javadoc)
     * @see org.eso.ohs.persistence.StorageManager#getAccountInfo(int, java.lang.String)
     */
    public AccountInfo getAccountInfo(int id, String password) throws ObjectIOException  {	
    	
		Connection con = null;
		PreparedStatement stmt = null;
		AccountInfo accountInfo = null;
		int accountId = -1;
		boolean activated = false;
		String username = null;
		
		
		try {
			// acquire connection
			con = acquireConnection(ObservingRun.class);						
			if (con == null) {
				stdlog_.debug("?? Connection is NULL");
				throw new ObjectIOException("Couldn't acquire connection!");
			}
			
			
			// find the account
			stmt = con.prepareStatement("SELECT accounts.account_id, rtrim(username) username , accounts.activated from accounts "
					+ " WHERE account_id = ? " + "AND passwd = ? ");
			stmt.setInt(1, id);
			stmt.setString(2, password);
								
			stmt.execute();

			ResultSet rs = stmt.getResultSet();

			if (!rs.next()) { // Should give one result set
				accountId = -1;
			} else {
				accountId = rs.getInt("account_id");	
				activated = rs.getBoolean("activated");
				username = rs.getString("username");
			}
						
			
			if (accountId != -1) {
				
				// great, we've found the account											
				// find the roles for the account				
				List<String> roles = findRoles(con, accountId);
				
				
				boolean engineering = false;
				boolean operations = false;
				
				// get the the engineering and operations role names from config
				String engineeringRoleName = AppConfig.getAppConfig().getEngineeringRoleName().toUpperCase();
				String operationsRoleName = AppConfig.getAppConfig().getOperationsRoleName().toUpperCase();
												
								
				// does this user have these roles?
				if (roles.size() > 0) {
					if (roles.contains(engineeringRoleName))
						engineering = true;
					if (roles.contains(operationsRoleName))
						operations = true;
				}
				
				
				// create AccountInfo object to return
				accountInfo = createAccountInfo(username, password, accountId,
						activated, engineering, operations);
												
			}
			
			return(accountInfo);

		} catch (SQLException e) {
			  handleSQLException(e);
			  throw new ObjectIOException (e);
		} finally {
            if (con != null) {
                try {
                    if (stmt != null) {
                        stmt.close();
                    }
                } catch (SQLException e) {
                    throw new ObjectIOException(e);
                } finally {
                    releaseConnection(con);
                }
            }
        }
		
		
    }

	private void handleSQLException(SQLException e) throws ObjectIOException {
		// @author amchavan, 29-May-2002: better error message.
		// In fact, one gets an INVALID_LOGIN error code both
		// if the user ID is wrong, or the password.
		stdlog_.error("SQLException", e);
		String msg0 = "Cannot connect to the database server:\n";
		String msg2 = " in configuration file.";
		String msgLogin = msg0 + "invalid user ID or password" + msg2;
		String msgPasswd = msg0 + "password" + msg2;
		SQLException e2 = e;

		while (e2 != null) {

		    // All database operations for the OHS go through this method
		    // If there is a Remote Login Error from Sybase we want
		    // to output our own less cryptic message + Sybases message

		    // @author amchavan, 28-Aug-2001: separated handling
		    // of the two conditions
		    if (e2.getErrorCode() == Sybase.INVALID_LOGIN)
		        throw new ObjectIOException(e, msgLogin);

		    if (e2.getErrorCode() == Sybase.INVALID_PASSWORD)
		        throw new ObjectIOException(e, msgPasswd);

		    // @author amchavan, 28-Aug-2001: fixed bug:
		    // was looping forever
		    e2 = e2.getNextException();
		}

		stdlog_.error(e);
		
	}
    
	

	
	/* (non-Javadoc)
	 * @see org.eso.ohs.persistence.StorageManager#getAccountInfo(java.lang.String, java.lang.String)
	 */
	public AccountInfo getAccountInfo(String username, String password) throws ObjectIOException  {
		Connection con = null;
		PreparedStatement stmt = null;
		AccountInfo accountInfo = null;
		int accountId = -1;
		boolean activated = false;
		
		
		try {
			// acquire connection
			con = acquireConnection(ObservingRun.class);						
			if (con == null) {
				stdlog_.debug("?? Connection is NULL");
				throw new ObjectIOException("Couldn't acquire connection!");
			}
			
			
			// find the account
			stmt = con.prepareStatement("SELECT account_id, activated from accounts "
					+ " WHERE username = ? " + "AND passwd = ? ");
			stmt.setString(1, username);
			stmt.setString(2, password);
								
			stmt.execute();

			ResultSet rs = stmt.getResultSet();

			if (!rs.next()) { // Should give one result set
				accountId = -1;
			} else {
				accountId = rs.getInt("account_id");	
				activated = rs.getBoolean("activated");
			}
						
			
			if (accountId != -1) {
				
				// great, we've found the account											
				// find the roles for the account				
				List<String> roles = findRoles(con, accountId);
				
				
				boolean engineering = false;
				boolean operations = false;
				
				// get the the engineering and operations role names from config
				String engineeringRoleName = AppConfig.getAppConfig().getEngineeringRoleName().toUpperCase();
				String operationsRoleName = AppConfig.getAppConfig().getOperationsRoleName().toUpperCase();
												
								
				// does this user have these roles?
				if (roles.size() > 0) {
					if (roles.contains(engineeringRoleName))
						engineering = true;
					if (roles.contains(operationsRoleName))
						operations = true;
				}
				
				
				// create AccountInfo object to return
				accountInfo = createAccountInfo(username, password, accountId,
						activated, engineering, operations);
												
			}
			
			return(accountInfo);

		} catch (SQLException e) {
			  handleSQLException(e);
			  throw new ObjectIOException (e);
		} finally {
            if (con != null) {
                try {
                    if (stmt != null) {
                        stmt.close();
                    }
                } catch (SQLException e) {
                    throw new ObjectIOException(e);
                } finally {
                    releaseConnection(con);
                }
            }
        }
		
		
		
	}

	private AccountInfo createAccountInfo(String username, String password,
			int accountId, boolean activated, boolean engineering,
			boolean operations) {
		AccountInfo accountInfo;
		accountInfo = new AccountInfo();
		accountInfo.setAccountId(accountId);
		accountInfo.setUsername(username);
		accountInfo.setPasswd(password);
		accountInfo.setActivated(activated);
		accountInfo.setEngineering(engineering);
		accountInfo.setOperations(operations);
		return accountInfo;
	}
	
	
	/**
	 * Finds a list of role names for an accountId
	 * 
	 * @param con
	 * @param accountId
	 * @return
	 * @throws SQLException
	 */
	private List<String> findRoles(Connection con,Integer accountId ) throws SQLException {
		PreparedStatement stmt = null;
		List<String> roles = new ArrayList<String>();
		stmt = con.prepareStatement("SELECT rtrim(upper(role_name)) as role_name FROM accounts,accounts_roles,roles  "
				+ " WHERE accounts.account_id = accounts_roles.account_id AND  " 
				+ " accounts_roles.role_id = roles.role_id AND"
				+ " accounts.account_id = ?"
		);
		stmt.setInt(1, accountId);				
		stmt.execute();

		ResultSet rs  = stmt.getResultSet();

		while (rs.next()) {					
			roles.add(rs.getString("role_name").toUpperCase());	
		}
		
		return(roles);
	}
	

    /**
	 * Returns an array of objects representing the contents of an object file
	 * with the given id and type from the chosen directory.
	 * 
	 * @param node
	 *            directory containing the file.
	 * @param id
	 *            id of the object file.
	 * @param objType
	 *            type of stored object.
	 * 
	 * @return the array of objects and object references, as for
	 *         StorageManager.read().
	 * 
	 * @exception ObjectNotFoundException
	 *                if the object was not found.
	 * @exception ObjectIOException
	 *                if there was an I/O error reading an object.
	 * 
	 * @see #write
	 * @see #update
	 */
    public UnpackedStorableObject read(
        DirectoryNode node,
        long id,
        Class objType)
        throws ObjectNotFoundException, ObjectIOException {
        stdlog_.debug("read DirectoryNode");
        return read(id, objType);
    }

    /**
     * Returns an array of objects representing the contents of an object file
     * with the given id and type from the chosen directory.
     * 
     * @param id
     *            id of the object file.
     * @param objType
     *            type of stored object.
     * 
     * @return the array of objects and object references, as for
     *         StorageManager.read().
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found.
     * @exception ObjectIOException
     *                if there was an I/O error reading an object.
     * 
     * @see #write
     * @see #update
     */
    public UnpackedStorableObject read(long id, Class objType)
        throws ObjectNotFoundException, ObjectIOException {
        stdlog_.debug("read");
        stdlog_.debug("Reading handler " + objType);
        DbaseIOHandler handler = registry_.getHandler(objType);
        stdlog_.debug(" handler is " + handler.getClass().getName());
        UnpackedStorableObject uso = null;
        Connection con = null;
        Statement stmt = null;

        id = Config.getCfg().uniqueToTableId(id);
        stdlog_.debug(">>>>>>>>>>> Id " + id);
        try {
            stdlog_.debug(" Id " + id);
            con = acquireConnection();
            stmt = con.createStatement();
            stdlog_.debug(" Id 2" + id);
            uso = handler.read(con, id);

            return uso;
        } catch (IOException e) {
            stdlog_.error("", e);
            throw new ObjectIOException(e);
        } catch (SQLException e) {
            stdlog_.error("", e);
            throw new ObjectIOException(e);
        } catch (IllegalArgumentException e) {
            stdlog_.error("", e);
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                try {
                    if (stmt != null) {
                        stmt.close();
                    }
                } catch (SQLException e) {
                    stdlog_.error("", e);
                    throw new ObjectIOException(e);
                } finally {
                    releaseConnection(con);
                }
            }
        }
    }

    /**
     * Returns an array of objects representing the contents of an object file
     * with the given id and type from the chosen directory.
     * 
     * @param node
     *            directory containing the file.
     * @param otherTables
     *            String
     * @param whereClause
     *            retriever of a list of objects.
     * @param objType
     *            type of stored object.
     * 
     * @return the array of objects and object references, as for
     *         StorageManager.read().
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found.
     * @exception ObjectIOException
     *                if there was an I/O error reading an object.
     * 
     * @see #write
     * @see #update
     */
    public UnpackedStorableObject[] read(
        DirectoryNode node,
        String otherTables,
        String whereClause,
        Class objType)
        throws ObjectNotFoundException, ObjectIOException {
        return read(otherTables, whereClause, objType);
    }

    /**
     * Returns an array of objects representing the contents of an object file
     * with the given id and type from the chosen directory.
     * 
     * @param otherTables
     *            String
     * @param whereClause
     *            retriever of a list of objects.
     * @param objType
     *            type of stored object.
     * 
     * @return the array of objects and object references, as for
     *         StorageManager.read().
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found.
     * @exception ObjectIOException
     *                if there was an I/O error reading an object.
     * 
     * @see #write
     * @see #update
     */
    public UnpackedStorableObject[] read(
        String otherTables,
        String whereClause,
        Class objType)
        throws ObjectNotFoundException, ObjectIOException {

        stdlog_.debug("WhereSql:" + whereClause + "; type=" + objType);

        DbaseIOHandler handler = registry_.getHandler(objType);
        UnpackedStorableObject[] usoArray = null;
        Connection con = null;
        Statement stmt = null;

        try {
            con = acquireConnection();
            stmt = con.createStatement();
            usoArray = handler.read(stmt, whereClause, otherTables);

            return usoArray;
        } catch (IOException e) {
            e.printStackTrace();
            throw new ObjectIOException(e);
        } catch (SQLException e) {
            e.printStackTrace();
            throw new ObjectIOException(e);
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                try {
                    if (stmt != null) {
                        stmt.close();
                    }
                } catch (SQLException e) {
                    e.printStackTrace();
                    throw new ObjectIOException(e);
                } finally {
                    releaseConnection(con);
                }
            }
        }
    }

    /**
     * Writes an array of objects with the given id and type to this storage
     * medium, in the chosen directory.
     * 
     * @param node
     *            directory in which the file should be saved.
     * @param id
     *            id under which to store the StorableObject.
     * @param objType
     *            type of StorableObject.
     * @param obj
     *            array of objects and object references to store, as described
     *            in the section for the read method. if an I/O error occurred
     *            while reading the file
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while writing the file
     * 
     * @see #read
     * @see #update
     */
    public void write(
        DirectoryNode node,
        long id,
        Class objType,
        UnpackedStorableObject obj)
        throws ObjectIOException {
        Connection con = null;

        stdlog_.debug(
            "WRITE: id="
                + id
                + "("
                + Config.getCfg().uniqueToTableId(id)
                + "); type="
                + objType);

        try {
            con = acquireConnection();
            con.setAutoCommit(false);
            store(con, true, id, objType, obj);
            con.commit();
        } catch (SQLException e) {
            try {
                con.rollback();
            } catch (SQLException exc) {
                exc.printStackTrace();
            }
            e.printStackTrace();
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                try {
                    con.setAutoCommit(true);
                } catch (SQLException e) {
                    e.printStackTrace();
                }
                releaseConnection(con);
            }
        }
    }

    /**
     * Writes an array of objects with the given id and type to this storage
     * medium, in the chosen directory.
     * 
     * @param con
     *            directory in which the file should be saved.
     * @param isNew
     *            boolean
     * @param id
     *            id under which to store the StorableObject.
     * @param objType
     *            type of StorableObject.
     * @param obj
     *            array of objects and object references to store, as described
     *            in the section for the read method. if an I/O error occurred
     *            while reading the file
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while writing the file
     * 
     * @see #read
     * @see #update
     */
    private void store(
        Connection con,
        boolean isNew,
        long id,
        Class objType,
        UnpackedStorableObject obj)
        throws ObjectIOException {
        DbaseIOHandler handler = registry_.getHandler(objType);
        System.out.println(handler.getManagedClass());

        Statement stmt = null;
        stdlog_.debug(
            "id in store method ------------------------------> " + id);
        id = Config.getCfg().uniqueToTableId(id);

        try {
            stdlog_.debug("IS NEW " + isNew);
            if (isNew) {
                handler.write(con, id, obj);
            } else {
                handler.update(con, id, obj);
            }
        } catch (SQLException e) {
            stdlog_.debug("URL: " + Config.getCfg().getUrl("OR")+" user: "+ Config.getCfg().getDbUserName());
            try {
                con.rollback();
            } catch (SQLException exc) {
                throw new ObjectIOException(exc);
            }
            throw new ObjectIOException(e);
        } catch (RuntimeException e) {
            e.printStackTrace();
            throw e;
        } finally {
            try {
                if (stmt != null) {
                    stmt.close();
                }
            } catch (SQLException e) {
                throw new ObjectIOException(e);
            }
        }
    }

    /**
     * Write several object to the external media in an atomic operation.
     * 
     * @param mos
     *            Array of object descriptions.
     * @exception ObjectIOException
     *                if an I/O error occurred while writing the object
     * 
     * @see #read
     * @see #write
     * @see #update
     */
    public void multiStore(MultiObjSpec[] mos) throws ObjectIOException {
        Connection con = null;

        stdlog_.debug("MULTISTORE: #objs = " + mos.length);

        try {
            con = acquireConnection();
            con.setAutoCommit(false);

            for (int j = 0; j < mos.length; j++) {
                stdlog_.debug("**************** mos index *********** " + j);
                MultiObjSpec m = mos[j];
                if (m.obj == null) {
                    System.out.println(
                        "***************** There is a problem in multi *************");
                }
                if (m.obj.getObject() == null) {
                    System.out.println(
                        "***************** There is a problem in multi *************");
                }
                long id = m.obj.getObject().getId();
                stdlog_.debug("multistore: obj " + j + " = " + id);

                store(con, m.isNew(), id, m.getCls(), m.obj);
            }

            con.commit();
        } catch (SQLException e) {
            try {
                con.rollback();
            } catch (SQLException exc) {
                exc.printStackTrace();
            }
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                try {
                    con.setAutoCommit(true);
                } catch (SQLException e) {
                    e.printStackTrace();
                }
                releaseConnection(con);
            }
        }
        CalibrationBlock cb = getCalBlk(mos);
        try {
            OBHistory.addEvent(
                Config.getCfg().uniqueToTableId(cb.getId()),
                null,
                OHSApplication.getShortDesc(),
                new Integer(cb.getOwnerId()).toString(),
                cb.getStatus());
        } catch (SQLException e) {
            throw new ObjectIOException(e);
        }
    }

    /**
     * @param mos
     *            MultiObjSpec[]
     * @return the dbase id for the calibration block in the idType array
     *         passed. This assumes there is exactly ONE Calib or Obs Block in
     *         the array.
     */
    private CalibrationBlock getCalBlk(MultiObjSpec[] mos) {
        StringBuffer rc = new StringBuffer();

        for (int j = 0; j < mos.length; j++) {
            if (mos[j].getCls().equals(CalibrationBlock.class)
                || mos[j].getCls().equals(ObservationBlock.class)) {
                CalibrationBlock cb = (CalibrationBlock) mos[j].obj.getObject();
                return cb;
            }
        }

        return null;
    }

    /**
     * Updates an array of objects with the given id and type on this storage
     * medium, in the chosen directory.
     * 
     * @param node
     *            directory in which the file should be saved.
     * @param id
     *            id under which to store the StorableObject.
     * @param objType
     *            type of StorableObject.
     * @param obj
     *            array of objects and object references to store, as described
     *            in the section for the read method. if an I/O error occurred
     *            while reading the file
     * 
     * @exception ObjectIOException
     *                if there was an I/O error while writing the object.
     * 
     * @see #read
     * @see #write
     */
    public void update(
        DirectoryNode node,
        long id,
        Class objType,
        UnpackedStorableObject obj)
        throws ObjectIOException {
        Connection con = null;

        stdlog_.debug(
            "UPDATE: id="
                + id
                + "("
                + Config.getCfg().uniqueToTableId(id)
                + "); type="
                + objType);

        try {
            con = acquireConnection();
            con.setAutoCommit(false);
            store(con, false, id, objType, obj);
            con.commit();
        } catch (SQLException e) {
            try {
                con.rollback();
            } catch (SQLException exc) {
                throw new ObjectIOException(exc);
            }
            throw new ObjectIOException(e);
        } catch (RuntimeException e) {
            e.printStackTrace();
            throw e;
        } finally {
            if (con != null) {
                try {
                    con.setAutoCommit(true);
                } catch (SQLException e) {
                    e.printStackTrace();
                }
                releaseConnection(con);
            }
        }
    }

    /**
     * Deletes the object with the given id and type on this storage medium, in
     * the chosen directory.
     * 
     * @param node
     *            directory in which the file should be saved.
     * @param id
     *            id under which to store the StorableObject.
     * @param objType
     *            type of StorableObject.
     * 
     * @exception ObjectIOException
     *                if there was an I/O error deleting the object.
     * 
     * @see #multiDelete
     */
    public void delete(DirectoryNode node, long id, Class objType)
        throws ObjectIOException {
        IdType idt = new IdType(id, objType);

        stdlog_.debug(
            "DELETE: id="
                + id
                + "("
                + Config.getCfg().uniqueToTableId(id)
                + "); type="
                + objType);

        multiDelete(new IdType[] { idt });
    }

    /**
     * Deletes an array of objects with the given id and type on this storage
     * medium, in the chosen directory.
     * 
     * @param idTypes
     *            array of id and object type pais
     * 
     * @exception ObjectIOException
     *                if there was an I/O error deleting the object.
     * 
     * @see #delete
     */
    protected void delete(StorableObject obj, Class objType)
        throws ObjectIOException {
        Config cfg = Config.getCfg();
        DbaseIOHandler handler;
        Connection con = null;
        Statement stmt = null;
        long id;

        try {
            con = acquireConnection();
            stmt = con.createStatement();
            id = cfg.uniqueToTableId(obj.getId());
            handler = registry_.getHandler(objType);
            handler.delete(stmt, obj);
            con.commit();
        } catch (SQLException e) {
            try {
                con.rollback();
            } catch (SQLException exc) {
                exc.printStackTrace();
            }
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                try {
                    if (stmt != null) {
                        stmt.close();
                    }
                } catch (SQLException e) {
                    e.printStackTrace();
                } finally {
                    releaseConnection(con);
                }
            }
        }
        try {

            if (objType.equals(CalibrationBlock.class)
                || objType.equals(ObservationBlock.class)) {
                OBHistory.addEvent(
                    Config.getCfg().uniqueToTableId(obj.getId()),
                    null,
                    OHSApplication.getShortDesc(),
                    new Integer(getAccountId()).toString(),
                    CalibrationBlock.PARTIALLY_DEFINED);
            }

        } catch (SQLException e) {
            throw new ObjectIOException(e);
        }

    }

    /**
     * Deletes an array of objects with the given id and type on this storage
     * medium, in the chosen directory.
     * 
     * @param idTypes
     *            array of id and object type pais
     * 
     * @exception ObjectIOException
     *                if there was an I/O error deleting the object.
     * 
     * @see #delete
     */
    protected void multiDelete(IdType[] idTypes) throws ObjectIOException {
        Config cfg = Config.getCfg();
        DbaseIOHandler handler;
        Connection con = null;
        Statement stmt = null;
        long id;

        try {
            con = acquireConnection();
            con.setAutoCommit(false);
            stmt = con.createStatement();

            for (int j = 0; j < idTypes.length; j++) {
                id = cfg.uniqueToTableId(idTypes[j].id);
                handler = registry_.getHandler(idTypes[j].objType);

                handler.delete(stmt, id);
            }

            con.commit();
        } catch (SQLException e) {
            try {
                con.rollback();
            } catch (SQLException exc) {
                exc.printStackTrace();
            }
            throw new ObjectIOException(e);
        } finally {
            if (con != null) {
                try {
                    con.setAutoCommit(true);
                    if (stmt != null) {
                        stmt.close();
                    }
                } catch (SQLException e) {
                    e.printStackTrace();
                } finally {
                    releaseConnection(con);
                }
            }
        }
        try {
            for (int i = 0; i < idTypes.length; i++) {
                if (idTypes[i].objType.equals(CalibrationBlock.class)
                    || idTypes[i].objType.equals(ObservationBlock.class)) {
                    OBHistory.addEvent(
                        Config.getCfg().uniqueToTableId(idTypes[i].id),
                        null,
                        OHSApplication.getShortDesc(),
                        new Integer(getAccountId()).toString(),
                        CalibrationBlock.PARTIALLY_DEFINED);
                }
            }
        } catch (SQLException e) {
            throw new ObjectIOException(e);
        }

    }

    /**
     * Moves an object between directories on this storage medium.
     * 
     * @param id
     *            id of the StorableObject.
     * @param objType
     *            type of StorableObject.
     * @param fromNode
     *            directory from which the object should be moved.
     * @param toNode
     *            directory to which the object should be moved.
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found in fromNode.
     * @exception ObjectIOException
     *                if an I/O error occurred while during the operation
     * 
     * @see org.eso.ohs.persistence.StorageManager#move(long, Class,
     *      DirectoryNode, DirectoryNode)
     */
    public void move(
        long id,
        Class objType,
        DirectoryNode fromNode,
        DirectoryNode toNode)
        throws ObjectNotFoundException, ObjectIOException {
        // Do nothing -- we don't support directories
    }

    /**
     * Null operation -- no separate initialisation required.
     * 
     * @see org.eso.ohs.persistence.StorageManager#initialise()
     */
    public void initialise() {
    }

    /**
     * Creates the directory on the file system.
     * 
     * @param node
     *            directory where the directory should be placed
     * @param id
     *            name of the directory to be created
     * @return True if directory is created successfully else false
     * @see org.eso.ohs.persistence.StorageManager#createDirectory(DirectoryNode,
     *      long)
     */

    public boolean createDirectory(DirectoryNode node, long id) {
        return false;
    }

    /**
     * Moves a directory within the file system.
     * 
     * @param fromNode
     *            directory to be moved
     * @param toNode
     *            new parent directory
     * 
     * @return True on success false on error
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found.
     * @see org.eso.ohs.persistence.StorageManager#moveDirectory(DirectoryNode,
     *      DirectoryNode)
     */
    public boolean moveDirectory(DirectoryNode fromNode, DirectoryNode toNode)
        throws ObjectNotFoundException {
        return false; // No hierarchical directories in the dbase.
    }

    /**
     * Deletes the directory from the file system.
     * 
     * @param node
     *            The directory to be deleted.
     * @return True if the directory is deleted successfully otherwise false
     * @see org.eso.ohs.persistence.StorageManager#deleteDirectory(DirectoryNode)
     */
    public boolean deleteDirectory(DirectoryNode node) {
        return false;
    }

    /**
     * Builds a list of Summary objects for a given directory. These Summary
     * objects contain the unique Id of the object and the Class which contains
     * details about the type of Object that is being referenced be it an
     * Observation Description,Observation Block etc..
     * 
     * @param node
     *            Directory from which you can create a list of objects
     * @return Array of ObjectFileInfo
     * 
     * @exception ObjectIOException
     *                if there was an I/O error listing an object.
     * @see org.eso.ohs.persistence.StorageManager#listObjects(DirectoryNode)
     */
    public Summary[] listObjects(DirectoryNode node) throws ObjectIOException {
        Enumeration<Class> enumeration = registry_.objectTypes();
        Vector<Summary> vec = new Vector<Summary>();
        DbaseIOHandler handler;
        Connection con = null;
        Statement stmt = null;
        Class objType;

        while (enumeration.hasMoreElements()) {
            objType = enumeration.nextElement();

            try {
                con = acquireConnection();
                stmt = con.createStatement();

                handler = registry_.getHandler(objType);

                Summary[] summ = handler.listObjects(stmt);
                for (int j = 0; j < summ.length; j++) {
                    vec.addElement(summ[j]);
                }

                stmt.close();
                stmt = null;
            } catch (SQLException e) {
                e.printStackTrace();
                throw new ObjectIOException(e);
            } finally {
                if (con != null) {
                    try {
                        if (stmt != null) {
                            stmt.close();
                        }
                    } catch (SQLException e) {
                        e.printStackTrace();
                    } finally {
                        releaseConnection(con);
                    }
                }
            }
        }

        Summary[] sums = new Summary[vec.size()];
        vec.copyInto(sums);
        return sums;
    }

    /**
     * Returns a string representation of this object and its fields.
     * 
     * @return a string representation.
     */
    public String toString() {
        return "DbaseStorageMgr[" + "rootNode=" + getRoot() + "]";
    }

    /**
     * Registry of DbaseIOHandler classes for object types known to this
     * manager.
     * 
     * @author ddorigo
     */
    public static class HandlerRegistry {

        /** UserId of the currently logged-in user */
        private int userId_;

        /** IO Handlers for all the business objects */
        private Hashtable<Class, DbaseIOHandler> ioHandlers_ = new Hashtable<Class, DbaseIOHandler>();

        /**
         * Constructor.
         */
        public HandlerRegistry() {
            initialiseRegistry();
        }

        /**
         * Method setUserId
         * 
         * @param userId
         *            int
         */
        public void setUserId(int userId) {
            userId_ = userId;
        }

        /**
         * Method getUserId
         * 
         * @return int
         */
        public int getUserId() {
            return userId_;
        }

        /**
         * Initialises the registry with handler classes for every object type.
         */
        protected void initialiseRegistry() {
            stdlog_.debug("dbase.DbaseStorageMgr initialiseRegistry");
            register(new DbaseHandlerUnknown());
        }

        /**
         * Method objectTypes
         * 
         * @return Enumeration
         */
        public Enumeration<Class> objectTypes() {
            return ioHandlers_.keys();
        }

        /**
         * Method register
         * 
         * @param handler
         *            DbaseIOHandler
         */
        protected void register(DbaseIOHandler handler) {
            stdlog_.debug(
                " register handler : " + handler.getClass().getName());
            stdlog_.debug("Registering class " + handler.getManagedClass());
            ioHandlers_.put(handler.getManagedClass(), handler);
            stdlog_.debug(" done ");

        }

        /**
         * Return the appropriate Handler for a particular class
         * 
         * @param objType
         *            Class
         * @return Appropriate dbase handler
         */
        private DbaseIOHandler lookup(Class objType) {
            return ioHandlers_.get(objType);
        }

        /**
         * Returns an instance of the DbaseIOHandler for the given object type.
         * 
         * @param objType
         *            object type to be read from/written to the dbase
         * @return a DbaseIOHandler object to do the I/O.
         */
        public DbaseIOHandler getHandler(Class objType) {
            stdlog_.debug(
                "!!!!!!!!!!!!!!!!!!!!!! Lookup Class !!!!!!!!!!!!!!!! "
                    + objType);
            DbaseIOHandler handler = lookup(objType);

            if (handler == null) {
                handler = lookup(Object.class);
            }

            handler.setUserId(getUserId());

            return handler;
        }
    }

    /**
     * Updates an array of objects with the given id and type on this storage
     * medium, in the chosen directory.
     * 
     * @param obj
     *            array of objects and object references to store, as described
     *            in the section for the read method. if an I/O error occurred
     *            while reading the file
     * 
     * @param objType
     *            type of StorableObject.
     * @exception ObjectIOException
     *                if there was an I/O error while writing the object.
     * 
     * @see #read
     * @see #write
     */
    public void update(StorableObject obj, Class objType)
        throws ObjectIOException {
        CalibrationBlock cb = null;
        String oldStatus = null;
        if (obj instanceof CalibrationBlock) {
            oldStatus = cb.getStatus();
            cb = (CalibrationBlock) obj;
        }

        DbaseIOHandler handler = registry_.getHandler(objType);
        Connection con = null;
        Statement stmt = null;
        try {
            con = acquireConnection();
            con.setAutoCommit(false);

            stmt = con.createStatement();
            handler.update(stmt, obj);
            con.commit();
        } catch (SQLException e) {
            try {
                con.rollback();
            
            } catch (SQLException exc) {
                throw new ObjectIOException(exc);
            }
            throw new ObjectIOException(e);
        } catch (RuntimeException e) {
            e.printStackTrace();
            throw e;
        } finally {
            try {
                if (stmt != null) {
                    stmt.close();
                }
            } catch (SQLException e) {
                e.printStackTrace();
            }
            if (con != null) {
                try {
                    con.setAutoCommit(true);
                } catch (SQLException e) {
                    e.printStackTrace();
                }
                releaseConnection(con);
            }
        }
        if (cb != null) {
            try {
                if (!oldStatus.equals(cb.getStatus())) {
                    OBHistory.addEvent(
                        Config.getCfg().uniqueToTableId(cb.getId()),
                        null,
                        OHSApplication.getShortDesc(),
                        new Integer(Config.getCfg().getAccountId()).toString(),
                        cb.getStatus());
                }
            } catch (SQLException e) {
                throw new ObjectIOException(e);
            }
        }
    }

    /**
     * Populate Readme file with all data from .usg
     * file and inst specific comments along with the readme
     */

    public Readme populateReadmeFile(ObservingRun obsRun)
        throws ObjectIOException {
        Connection con = null;
        Readme readme = null;
        try {
            con = acquireConnection(ObservationBlock.class);
            return ReadmeFileDBIO.read(con, obsRun);
        } catch (SQLException e) {
            throw new ObjectIOException(e);
        } catch (JDOMException e1) {
            throw new ObjectIOException(e1);
        } catch (IOException e1) {
            throw new ObjectIOException(e1);
        } finally {
            if (con != null) {
                releaseConnection(con);
            }
        }
    }
    
    /**
     * Populate Readme file with all data from .usg
     * file and inst specific comments along with the readme
     */

    public Readme populateReadmeFileReplicate(ObservingRun obsRun)
        throws ObjectIOException {
        Connection con = null;
        Readme readme = null;
        try {
            con = acquireConnection(ObservationBlock.class);
            return ReadmeFileDBIO.readReadmeReplicate(con, obsRun);
        } catch (SQLException e) {
            throw new ObjectIOException(e);
        } catch (JDOMException e1) {
            throw new ObjectIOException(e1);
        } catch (IOException e1) {
            throw new ObjectIOException(e1);
        } finally {
            if (con != null) {
                releaseConnection(con);
            }
        }
    }

    /**
     * Change the status of a readme file giben certain rules.
     */

    public void setReadmeFileStatus(String status, long runId)
        throws ObjectIOException {
        Connection con = null;
        Readme readme = null;
        ObjectIOException error = null;
        try {
            con = acquireConnection(ObservationBlock.class);
            ReadmeFileDBIO.setReadmeFileStatus(con, status, runId);
            con.commit();
        } catch (SQLException e) {
            e.printStackTrace();
            try {
                con.rollback();
            } catch (SQLException e1) {
                e1.printStackTrace();
            }
            error = new ObjectIOException(e.getMessage());
        }  catch (ObjectIOException e) {
            error = e;
        }
        if (con != null) {
            try {
                con.setAutoCommit(true);
            } catch (SQLException e1) {
                e1.printStackTrace();
            }
            releaseConnection(con);
        }
        if (error != null)
            throw error;
    }
}