package org.eso.ohs.persistence.dbase.phase2;

import java.io.IOException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.core.utilities.SqlUtils;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.dfs.ObservationDescription;
import org.eso.ohs.instruments.Instrument;
import org.eso.ohs.instruments.InstrumentList;
import org.eso.ohs.instruments.TemplateSignature;
import org.eso.ohs.persistence.Config;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.persistence.StorageManager.UnpackedStorableObject;
import org.eso.ohs.persistence.dbase.DbaseIOHandler;

/**
 * I/O class for ODs stored in the database.
 *
 * <p>
 * $Id: DbaseHandlerOD.java,v 1.1.8.1 2007/10/11 08:27:31 pnunes Exp $
 * <br>
 * @author Bob Kemp
 * 
 * @see org.eso.ohs.persistence.dbase.phase2.DbaseMgr
 * @see org.eso.ohs.persistence.ObjectManager
 */

public class DbaseHandlerOD extends DbaseIOHandler {

    /** Standard log channel */
    private static Logger stdlog_ =
        Logger.getLogger(DbaseHandlerOD.class);

    private static String obrepDbName =
        Config.getCfg().getDbName(ObservationBlock.getSuffix());

    /**
     * Refers to the appropriate DFS class
     * @return ObservationDescription class
     */

    public Class getManagedClass() {
        return ObservationDescription.class;
    }

    /**
     * Reads the object from the database, partially reassembling it.
     * References to other business objects are stored in other fields
     * of the array
     * 
     * @param stmt  JDBC Statement for contacting the dbase
     * @param id    per-table id of the object to be retrieved.
     * @return      array in the format required by reassembleReferences()
     * 
     * @exception SQLException
     *          an error occurred during the dbase operation
     * @exception ObjectNotFoundException
     *          the requested object was not found in the database
     * @exception IOException
     *              if an I/O error occurred while reading the object
     * @exception ObjectIOException IO Eror occured while reading from the database
     * 
     * @see org.eso.ohs.persistence.ObjectManager#disassembleReferences
     * @see org.eso.ohs.persistence.ObjectManager#reassembleReferences
     */
    public UnpackedStorableObject read(Statement stmt, long id)
        throws ObjectNotFoundException, ObjectIOException, IOException, SQLException {
        ResultSet rs = null;
        ObservationDescription od = (ObservationDescription) newInstance();
        UnpackedStorableObject obj = new UnpackedStorableObject(od);

        String sql =
            "SELECT * FROM "
                + obrepDbName
                + "..obs_descriptions"
                + " WHERE od_id ="
                + id;

        if (!Config.getCfg().isInOperationsMode()) {
            sql =
                sql
                    + " AND "
                    + obrepDbName
                    + "..obs_descriptions.user_id="
                    + userId_;
        }
        // stdlog_.debug("sql = " + sql);

        rs = stmt.executeQuery(sql);

        if (!rs.next()) { // Only one with this id !!
            throw new ObjectNotFoundException("Object not found in database");
        }

        od.setId(
            Config.getCfg().tableToUniqueId(
                rs.getLong("od_id"),
                ObservationDescription.class));

        od.setOwnerId(rs.getInt("user_id"));

        od.setName(rs.getString("item_name"));
        float ipv = rs.getFloat("ipversion");
        if (ipv < 0) {
            // a NULL value in the database is set to 0.0.
            ipv = 0;
        }
        //od.setIPVersion(ipv);
        od.setInstrument(rs.getString("instrument"), ipv);

        od.setMode(rs.getString("mode"));
        od.setUserComments(rs.getString("item_description").trim());
        od.setCloneParentId(rs.getLong("local_storage_id"));

        rs.close();

        Instrument inst =
            InstrumentList.getInstance().getInstrument(
                od.getInstrument(),
                od.getIPVersion());
        /*
        if (inst == null) {
            // we do not currently have the instrument package downloaded.
            // download the instrument package and reload the instrument list.
            String ipCacheDir = InstrumentList.getIPCacheDirectory();
            File f = new File(ipCacheDir);
            InstrumentPackageDBIO.getIP(od.getInstrument(), od.getIPVersion(), f.getAbsolutePath(), true);
            InstrumentList.reloadInstrumentList();
        }
        */
        // Try and find the instrument again.
        inst =
            InstrumentList.getInstance().getInstrument(
                od.getInstrument(),
                od.getIPVersion());
        if (inst == null) {
            stdlog_.debug(
                "IP Template Not Found: "
                    + od.getInstrument()
                    + " version: "
                    + od.getIPVersion());

            String versionMessage =
                InstrumentList.isInSingleIPMode()
                    ? ""
                    : " version " + od.getIPVersion();
            throw new IOException(
                "Instrument Package for "
                    + od.getInstrument()
                    + versionMessage
                    + " could not be found.");
        }

        TemplateSignature[] tsList = TemplateSigDBIO.read(stmt, inst, id);

        // Loop starts at 1 to skip acquisition template
        for (int j = 1; j < tsList.length; j++) {
            if (tsList[j] == null) {
                throw new ObjectIOException("Missing TemplateSignature: " + j);
            }
            if (inst != null && inst.hasMasks() && j == 1)
                MaskDBIO.read(stmt, tsList[1], 1, id);
            od.appendSignature(tsList[j]);
        }

        return obj;
    }
    
    
     

    /**
     * Writes the object to the database, for the first time.
     * 
     * @param stmt  JDBC Statement for contacting the dbase
     * @param id    per-table id of the object to be written.
     * @param obj   an array: the first element is the BO to be stored.
     * @exception SQLException  an error occurred during the dbase operation
     * 
     * @see org.eso.ohs.persistence.ObjectManager#disassembleReferences
     * @see org.eso.ohs.persistence.ObjectManager#reassembleReferences
     */
    public void write(Statement stmt, long id, UnpackedStorableObject obj)
        throws SQLException {
        ObservationDescription od = (ObservationDescription) obj.getObject();
        StringBuffer sb = new StringBuffer();

        sb.append(id);
        sb.append(",");
        sb.append(od.getOwnerId());
        sb.append(",");
        sb.append(SqlUtils.sqlEscapedString(od.getName()));
        sb.append(",");
        sb.append(SqlUtils.sqlEscapedString(od.getInstrument()));
        sb.append(",");
        sb.append(SqlUtils.sqlEscapedString(od.getMode()));
        sb.append(",");
        sb.append(SqlUtils.sqlEscapedString(od.getUserComments()));
        sb.append(",");
        sb.append("'");
        sb.append(od.getCloneParentId());
        sb.append("'");
        sb.append(",");
        sb.append("" + od.getIPVersion());

        String sqlText =
            "INSERT INTO obs_descriptions"
                + " (od_id, user_id, item_name, instrument,"
                + "  mode, item_description, local_storage_id, ipversion)"
                + " VALUES ("
                + sb
                + ")";
        stdlog_.debug(sqlText);
        stmt.executeUpdate(sqlText);

        TemplateSigDBIO.write(stmt, id, od.getSignatures(), false);
        Instrument inst =
            InstrumentList.getInstance().getInstrument(od.getInstrument());

        TemplateSignature[] tsList = od.getSignatures();
        if (inst.hasMasks()) {
            try {
                if (tsList != null && tsList.length > 0)
                    MaskDBIO.write(stmt, tsList[0], 1, id);
            } catch (ObjectIOException e) {
                e.printStackTrace();
            }
        }
    }

    /** Writes the object to the database, for the first time.
        * 
        * @param id    per-table id of the object to be written.
        * @param obj   an array: the first element is the BO to be stored.
        * @exception SQLException  an error occurred during the dbase operation
        * 
        * @see org.eso.ohs.persistence.ObjectManager#disassembleReferences
        * @see org.eso.ohs.persistence.ObjectManager#reassembleReferences
        */
    public void write(Connection conn, long id, UnpackedStorableObject obj)
        throws SQLException {
        Statement stmt = conn.createStatement();
        write(stmt, id, obj);
    }

    /**
     * Updates an object that is already stored in the database.
     * 
     * @param stmt  JDBC Statement for contacting the dbase
     * @param id    per-table id of the object to be written.
     * @param obj   an array: the first element is the BO to be stored.
     * @exception SQLException  an error occurred during the dbase operation
     * 
     * @see org.eso.ohs.persistence.ObjectManager#disassembleReferences
     * @see org.eso.ohs.persistence.ObjectManager#reassembleReferences
     */
    public void update(Statement stmt, long id, UnpackedStorableObject obj)
        throws SQLException {
        delete(stmt, id);
        write(stmt, id, obj);
    }

    /**
     * Deletes an object from the database.
     * 
     * @param stmt  JDBC Statement for contacting the dbase
     * @param id    per-table id of the object to be written.
     * @exception SQLException  an error occurred during the dbase operation
     */
    public void delete(Statement stmt, long id) throws SQLException {

        stmt.executeUpdate(
            "DELETE FROM obs_descriptions" + " WHERE od_id = " + id);

        TemplateSigDBIO.delete(stmt, id, false);
    }
}
