package org.eso.ohs.persistence.dbase.phase2;

import java.io.IOException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Vector;

import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.TimeInterval;
import org.eso.ohs.persistence.Config;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.persistence.StorageManager.UnpackedStorableObject;
import org.eso.ohs.persistence.dbase.DbaseIOHandler;

/**
 * I/O class for Masks stored in the database.
 * 
 * <p>
 * $Id: DbaseHandlerTimeInterval.java,v 1.4 2006/04/26 09:36:38 ddorigo Exp $
 * <br>
 * 
 * @author Bob Kemp
 * 
 * @see org.eso.ohs.persistence.dbase.phase2.DbaseMgr
 */

public class DbaseHandlerTimeInterval extends DbaseIOHandler {

    private static final int SHIFT_TI_INDEX = 100000000;

    private static final long SHIFT_TI_ID = 1000;

    private static final String TI_ID_COL = "ti_id";

    private static final String TI_INDEX_COL = "ti_index";

    private static final String TI_TYPE_COL = "ti_type";

    /**
     * Constructor for CS I/O class.
     */
    public DbaseHandlerTimeInterval() {
    }

    /**
     * Managed class
     * 
     * @return Mask class
     */

    public Class getManagedClass() {
        return TimeInterval.class;
    }

    /**
     * Reads the object from the database, partially reassembling it. References
     * to other business objects are stored in other fields of the array
     * 
     * @param stmt
     *            JDBC Statement for contacting the dbase
     * @param id
     *            per-table id of the object to be retrieved.
     * @return array in the format required by reassembleReferences()
     * 
     * @exception SQLException
     *                an error occurred during the dbase operation
     * @exception ObjectNotFoundException
     *                the requested object was not found in the database
     * @exception IOException
     *                if an I/O error occurred while reading the object
     * 
     * @see org.eso.ohs.persistence.ObjectManager#disassembleReferences
     * @see org.eso.ohs.persistence.ObjectManager#reassembleReferences
     */
    public UnpackedStorableObject read(Statement stmt, long id)
            throws ObjectNotFoundException, IOException, SQLException {
        throw new IOException("Operation not supported for this object");
    }

    public static long idForTimeInterval(long tiId, int tiIndex, char tiType) {
        int tiTypeInt = Character.getNumericValue(tiType);
        long id = tiTypeInt + SHIFT_TI_ID * (SHIFT_TI_INDEX * tiIndex + tiId);
        return id;
    }

    /**
     * Read data
     * 
     * @param sql
     *            statement
     * @param whereClause
     *            where condition for retreiving the object
     * @exception SQLException
     *                an error occurred during the dbase operation
     * @exception ObjectNotFoundException
     *                the requested object was not found in the database
     * @exception ObjectIOException
     *                IO Eror occured while reading from the database
     */

    public UnpackedStorableObject[] read(Statement stmt, String whereClause,
            String otherTables)
            throws ObjectIOException, ObjectNotFoundException, SQLException {

        if (otherTables == null) {
            otherTables = "";
        }
        UnpackedStorableObject[] ob_array = null;
        TimeInterval t_i = null;
        String sql = buildSqlQuery(whereClause, otherTables);
        stdlog_.debug(sql);
        ResultSet rs = stmt.executeQuery(sql);

        try {
            Vector<TimeInterval> ti_vector = new Vector<TimeInterval>();

            while (rs.next()) {

                t_i = new TimeInterval();
                // New element
                ti_vector.addElement(t_i);

                long tiId = idForTimeInterval(rs.getLong(TI_ID_COL), rs
                        .getInt(TI_INDEX_COL), rs
                        .getString(TI_TYPE_COL)
                        .charAt(0));

                t_i.setId(Config.getCfg().tableToUniqueId(tiId,
                        TimeInterval.class));
                setTiFields(t_i, rs);
            }
            rs.close();

            ob_array = new UnpackedStorableObject[ti_vector.size()];
            for (int i = 0; i < ti_vector.size(); i++) {
                t_i = ti_vector.elementAt(i);
                ob_array[i] = new UnpackedStorableObject(t_i);
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }

        return ob_array;
    }

    public static String buildSqlQuery(String whereClause, String otherTables) {
        String sql = getSelectSql() + otherTables + " WHERE " + whereClause
                + " ORDER BY " + Phase2SelectStmt.timeIntervalsTable.get()
                + "."+TI_ID_COL;
        return sql;
    }

    /**
     * Set the prog fields from the result set. Only for avoiding duplicated
     * code
     * 
     */
    protected void setTiFields(TimeInterval timeInterval, ResultSet rs)
            throws SQLException {
        timeInterval.setObId(rs.getInt(TI_ID_COL));
        timeInterval.setStartTime(rs.getInt("start_interval"));
        timeInterval.setEndTime(rs.getInt("end_interval"));
        timeInterval.setPreference(rs.getInt("preference"));
        timeInterval.setType(rs.getString(TI_TYPE_COL));
        timeInterval.setIndexId(rs.getInt(TI_INDEX_COL));
    }

    /**
     * @return Returns the run_sql.
     */
    private static String getSelectSql() {
        String run_sql = "SELECT DISTINCT "
                + Phase2SelectStmt.timeIntervalsTable.get() + "."+TI_INDEX_COL+",  "
                + Phase2SelectStmt.timeIntervalsTable.get() + "."+TI_ID_COL+", "
                + Phase2SelectStmt.timeIntervalsTable.get()
                + ".start_interval, "
                + Phase2SelectStmt.timeIntervalsTable.get() + ".end_interval, "
                + Phase2SelectStmt.timeIntervalsTable.get() + ".preference, "
                + Phase2SelectStmt.timeIntervalsTable.get() + "."+TI_TYPE_COL+" "
                + "FROM " + Phase2SelectStmt.timeIntervalsTable.get() + " ";
        return run_sql;
    }

    /**
     * Writes the object to the database, for the first time.
     * 
     * @param stmt
     *            JDBC Statement for contacting the dbase
     * @param id
     *            per-table id of the object to be written.
     * @param obj
     *            an array: the first element is the BO to be stored.
     * @exception SQLException
     *                an error occurred during the dbase operation
     * 
     * @see org.eso.ohs.persistence.ObjectManager#disassembleReferences
     * @see org.eso.ohs.persistence.ObjectManager#reassembleReferences
     */
    public void write(Statement stmt, long id, UnpackedStorableObject obj)
            throws SQLException {
        throw new SQLException("Operation not supported for this object");
    }

    /**
     * Updates an object that is already stored in the database.
     * 
     * @param stmt
     *            JDBC Statement for contacting the dbase
     * @param id
     *            per-table id of the object to be written.
     * @param obj
     *            an array: the first element is the BO to be stored.
     * @exception SQLException
     *                an error occurred during the dbase operation
     * 
     * @see org.eso.ohs.persistence.ObjectManager#disassembleReferences
     * @see org.eso.ohs.persistence.ObjectManager#reassembleReferences
     */
    public void update(Statement stmt, long id, UnpackedStorableObject obj)
            throws SQLException {
        throw new SQLException("Operation not supported for this object");
    }

    /**
     * Deletes an object from the database.
     * 
     * @param stmt
     *            JDBC Statement for contacting the dbase
     * @param id
     *            per-table id of the object to be written.
     * @exception SQLException
     *                an error occurred during the dbase operation
     */
    public void delete(Statement stmt, long id) throws ObjectIOException {
        throw new ObjectIOException("Operation not supported for this object");

    }

}
