package org.eso.ohs.persistence.dbase.phase2;

import java.io.IOException;
import java.io.StringReader;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.ConnectionPool;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.core.utilities.PafData;
import org.eso.ohs.core.utilities.PafFile;
import org.eso.ohs.core.utilities.SqlUtils;
import org.eso.ohs.dfs.CalibrationBlock;
import org.eso.ohs.dfs.Mask;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.instruments.Parameter;
import org.eso.ohs.instruments.TemplateSignature;
import org.eso.ohs.persistence.Config;
import org.eso.ohs.persistence.dbase.DbaseIOHandler;

/**
 * @author tcanavan
 *
 * This class handles the insertion of mask data into the databasse. This mask
 * data is linked to the ob via the obsblk_masks table. The masks table contains
 * the actual mask data   
 * 
 * */
public class MaskDBIO {

    private static Logger stdlog_ = Logger.getLogger(MaskDBIO.class);

    private static String obrepDbName =
        Config.getCfg().getDbName(ObservationBlock.getSuffix());

    /**    
     * Read template signature(s) from database
     * @param stmt Sql statement
     * @param instrumentName Instrument names associated with template
     * @param boID Id of ts(s) to be read
     * @exception SQLException
     *            an error occurred during the dbase operation   
     * @exception IOException
     *              if an I/O error occurred while reading the object
     * @exception ObjectIOException IO Eror occured while reading from the database
     *
     * @return array of template signatures
     */

    public static void write(
        Statement stmt,
        TemplateSignature ts,
        int tsNo,
        long boId)
        throws SQLException, ObjectIOException {
        Parameter[] params = ts.getParamList();
        String instrumentName = ts.getInstrumentName();
        Vector<AdpData> adpList = initAdpList(tsNo, boId, params);
        writeMaskData(stmt, adpList);
    }
    
    private static Vector<AdpData> initAdpList(
        int tsNo,
        long boId,
        Parameter[] params) {
        Vector<AdpData> adpList = new Vector<AdpData>();
        for (int i = 0; i < params.length; i++) {
            String paramName = params[i].getParameterName();
            if (paramName.indexOf("INS.ADP") != -1) {
                AdpData adpData =
                    new AdpData(
                        boId,
                        tsNo,
                        getKeywordIndex(paramName),
                        params[i].getValue());
                adpList.addElement(adpData);
            }
        }
        return adpList;
    }

    public static void read(
        Statement stmt,
        TemplateSignature ts,
        int tsNo,
        long boId)
        throws SQLException {
        Parameter[] params = ts.getParamList();
        Vector<AdpData> adpList = initAdpList(tsNo, boId, params);
        for (int j = 0; j < params.length; j++) {
            String paramName = params[j].getParameterName();
            if (paramName.indexOf("INS.ADM") != -1) {
                setADMData(stmt, adpList, paramName, params[j]);
            }
            if (paramName.indexOf("INS.MASK") != -1
                && paramName.indexOf(".NO") != -1) {
                setSlotNumber(stmt, adpList, paramName, params[j]);
            }
            if (paramName.indexOf("INS.MASK") != -1
                && paramName.indexOf(".ID") != -1) {
                setMaskId(stmt, adpList, paramName, params[j]);
            }
        }
    }

    public static void incExecCount(TemplateSignature ts)
        throws ObjectIOException {
        Connection con = null;
        Statement stmt = null;
        SQLException error = null;
        try {
            Config cfg = Config.getCfg();
            con =
                ConnectionPool.getPool().acquireConnection(
                    cfg.getUrl(ObservationBlock.getSuffix()),
                    cfg.getDbUserName(),
                    cfg.getDbPassword());

            stmt = con.createStatement();
            con.setAutoCommit(false);
            Parameter[] params = ts.getParamList();
            for (int j = 0; j < params.length; j++) {
                String paramName = params[j].getParameterName();
                if (paramName.indexOf("INS.MASK") != -1
                    && paramName.indexOf(".ID") != -1) {
                    setExecCount(stmt, params[j].getValue());
                }
            }
            con.commit();
        } catch (SQLException e) {
            e.printStackTrace();
            error = e;
            try {
                con.rollback();
            } catch (SQLException ex) {
                // nothing here...already have an error recorded!
            }
        } finally {
            if (stmt != null) {
                try {
                    stmt.close();
                } catch (SQLException e) {
                    if (error == null) {
                        error = e;
                    }
                }
            }
            if (con != null) {
                try {
                    con.setAutoCommit(true);
                } catch (SQLException e) {
                }
                ConnectionPool.getPool().releaseConnection(con);
            }
            // throw an exception if there was an error encountered.
            if (error != null) {
                throw new ObjectIOException(error);
            }
        }

    }
    private static void setExecCount(Statement stmt, String bar_code)
        throws SQLException {

        String sql =
            "UPDATE "
                + obrepDbName
                + "..masks SET "
                + obrepDbName
                + "..masks.exec_count = "
                + "(("
                + obrepDbName
                + "..masks.exec_count)+1)"
                + " where "
                + obrepDbName
                + "..masks.bar_code = "
                + SqlUtils.sqlEscapedString(bar_code)
                + " AND  "
                + obrepDbName
                + "..masks.bar_code != '-99'";
        System.out.println("Increment Exec count " + sql);
        stmt.executeUpdate(sql);
    }

    /**
         * Get the inetger value at the end of a keyword  
         * @param  keyword Kwyword to search for integer
         * @return Integer value  at the end of a keyword
         */

    private static int getKeywordIndex(String keyword) {
        int index = -1;
        int startIndex = -1;
        int endIndex = -1;

        for (int k = 0; k < keyword.length(); k++) {
            if (Character.isDigit(keyword.charAt(k))) {
                startIndex = k;
            }
        }
        //dIndex = keyword.lastIndexOf('.');
        String idxStr = keyword.substring(startIndex, startIndex + 1);
        try {
            index = new Integer(idxStr).intValue();
        } catch (NumberFormatException e) {
            //stdlog_.error( e );
        }
        return index;
    }

    private static void setADMData(
        Statement stmt,
        Vector<AdpData> adpList,
        String paramName,
        Parameter p)
        throws SQLException {
        int index = getKeywordIndex(paramName);
        stmt.execute("set TEXTSIZE " + 1000000);
        AdpData adpData = getAdpDataFromKeywordIndex(adpList, index);
        String sql =
            "SELECT "
                + obrepDbName
                + "..masks.adm_file from "
                + obrepDbName
                + "..masks"
                + ", "
                + obrepDbName
                + "..obsblk_masks"
                + " WHERE  "
                + obrepDbName
                + "..obsblk_masks.ts_index = "
                + adpData.tsIndex
                + " AND    "
                + obrepDbName
                + "..obsblk_masks.adp_index = "
                + adpData.adpIndex
                + " AND    "
                + obrepDbName
                + "..obsblk_masks.ob_id = "
                + adpData.obId
                + " AND    "
                + obrepDbName
                + "..obsblk_masks.mask_id = "
                + obrepDbName
                + "..masks.mask_id";
        System.out.println(sql);
        ResultSet rs = stmt.executeQuery(sql);
        String adm_file = null;
        while (rs.next()) {
            adm_file = rs.getString("adm_file");
        }
        if (adm_file != null)
            p.setValue(adm_file);
        else {
            System.out.println("ERROR adm files is corrupt");
        }
    }

    private static void setSlotNumber(
        Statement stmt,
        Vector<AdpData> adpList,
        String paramName,
        Parameter p)
        throws SQLException {
        int index = getKeywordIndex(paramName);
        AdpData adpData = getAdpDataFromKeywordIndex(adpList, index);
        String sql =
            "SELECT "
                + obrepDbName
                + "..masks.mask_slot_num from "
                + obrepDbName
                + "..masks"
                + ", "
                + obrepDbName
                + "..obsblk_masks"
                + " WHERE  "
                + obrepDbName
                + "..obsblk_masks.ts_index = "
                + adpData.tsIndex
                + " AND    "
                + obrepDbName
                + "..obsblk_masks.adp_index = "
                + adpData.adpIndex
                + " AND    "
                + obrepDbName
                + "..obsblk_masks.ob_id = "
                + adpData.obId
                + " AND "
                + obrepDbName
                + "..obsblk_masks.mask_id = "
                + obrepDbName
                + "..masks.mask_id";
        System.out.println(sql);
        ResultSet rs = stmt.executeQuery(sql);
        int slotNo = -1;
        while (rs.next()) {
            slotNo = rs.getInt("mask_slot_num");
        }
        if (slotNo != -1)
            p.setValue(new Integer(slotNo).toString());

    }

    private static void setMaskId(
        Statement stmt,
        Vector<AdpData> adpList,
        String paramName,
        Parameter p)
        throws SQLException {
        int index = getKeywordIndex(paramName);
        //if(index -1 >= adpList.size()) {
        //    return;
        //}
        AdpData adpData = getAdpDataFromKeywordIndex(adpList, index);
        //AdpData adpData = (AdpData)adpList.elementAt(index-1);
        String sql =
            "SELECT "
                + obrepDbName
                + "..masks.bar_code from "
                + obrepDbName
                + "..masks"
                + ", "
                + obrepDbName
                + "..obsblk_masks"
                + " WHERE  "
                + obrepDbName
                + "..obsblk_masks.ts_index = "
                + adpData.tsIndex
                + " AND    "
                + obrepDbName
                + "..obsblk_masks.adp_index = "
                + adpData.adpIndex
                + " AND    "
                + obrepDbName
                + "..obsblk_masks.ob_id = "
                + adpData.obId
                + " AND "
                + obrepDbName
                + "..obsblk_masks.mask_id = "
                + obrepDbName
                + "..masks.mask_id";
        System.out.println(sql);
        ResultSet rs = stmt.executeQuery(sql);
        String maskId = null;
        while (rs.next()) {
            maskId = rs.getString("bar_code");
        }
        if (maskId != null)
            p.setValue(maskId);
    }

    private static AdpData getAdpDataFromKeywordIndex(
        Vector<AdpData> adpList,
        int index) {
        for (int i = 0; i < adpList.size(); i++) {
            AdpData adpData = adpList.elementAt(i);
            if (adpData.adpIndex == index)
                return adpData;
        }
        return null;
    }

    private static class AdpData {

        private int tsIndex;
        private int adpIndex;
        private long obId;
        private int keywordIndex;
        private String adpData;

        public AdpData(long ob_id, int ts_index, int adpIndex) {
            this.tsIndex = ts_index;
            this.obId = ob_id;
            this.adpIndex = adpIndex;
        }

        public AdpData(
            long ob_id,
            int ts_index,
            int adpIndex,
            String adp_data) {
            this.tsIndex = ts_index;
            this.obId = ob_id;
            this.adpIndex = adpIndex;
            this.adpData = adp_data;
        }

        /**
         * Returns the adpData.
         * @return String
         */
        public String getAdpData() {
            return adpData;
        }

        /**
         * Returns the adpIndex.
         * @return int
         */
        public int getAdpIndex() {
            return adpIndex;
        }

        /**
         * Returns the keywordIndex.
         * @return int
         */
        public int getKeywordIndex() {
            return keywordIndex;
        }

        /**
         * Returns the obId.
         * @return long
         */
        public long getObId() {
            return obId;
        }

        /**
         * Returns the tsIndex.
         * @return int
         */
        public int getTsIndex() {
            return tsIndex;
        }

        /**
         * Sets the adpData.
         * @param adpData The adpData to set
         */
        public void setAdpData(String adpData) {
            this.adpData = adpData;
        }

        /**
         * Sets the adpIndex.
         * @param adpIndex The adpIndex to set
         */
        public void setAdpIndex(int adpIndex) {
            this.adpIndex = adpIndex;
        }

        /**
         * Sets the keywordIndex.
         * @param keywordIndex The keywordIndex to set
         */
        public void setKeywordIndex(int keywordIndex) {
            this.keywordIndex = keywordIndex;
        }

        /**
         * Sets the obId.
         * @param obId The obId to set
         */
        public void setObId(long obId) {
            this.obId = obId;
        }

        /**
         * Sets the tsIndex.
         * @param tsIndex The tsIndex to set
         */
        public void setTsIndex(int tsIndex) {
            this.tsIndex = tsIndex;
        }

    }
    private static void writeMaskData(Statement stmt, Vector<AdpData> adpList)
        throws SQLException, ObjectIOException {

        StringReader rdr;
        PafFile paf;
        for (int i = 0; i < adpList.size(); i++) {
            AdpData adpData = adpList.elementAt(i);
            try {

                rdr = new StringReader(adpData.adpData);

                paf = new PafFile(rdr);
            } catch (IOException e) {
                throw new ObjectIOException(e);
            }
            String sql =
                "INSERT INTO obsblk_masks (ob_id, ts_index, adp_index) "
                    + "VALUES ("
                    + adpData.obId
                    + ","
                    + adpData.tsIndex
                    + ","
                    + adpData.adpIndex
                    + ")";
            stmt.executeUpdate(sql);
            String naid = getMaskNAID(paf);
            boolean found = false;
            long mskDbId = -1;
            sql =
                "SELECT mask_id from "
                    + obrepDbName
                    + "..masks"
                    + "  where "
                    + obrepDbName
                    + "..masks.mask_naid = "
                    + SqlUtils.sqlEscapedString(naid);
            ResultSet rs = stmt.executeQuery(sql);
            while (rs.next()) {
                mskDbId = rs.getLong(1);
                found = true;
            }
            rs.close();
            if (!found) {
                mskDbId = DbaseIOHandler.getDbId();

                // Should be moved into adpData array
                sql =
                    "INSERT INTO "
                        + obrepDbName
                        + "..masks (mask_id,mask_naid,mask_name) "
                        + "VALUES ("
                        + mskDbId
                        + ","
                        + SqlUtils.sqlEscapedString(getMaskNAID(paf))
                        + ","
                        + SqlUtils.sqlEscapedString(paf.getPafName())
                        + ")";
                System.out.println(sql);
                stmt.executeUpdate(sql);
            }
            updateObsBlk(
                stmt,
                mskDbId,
                adpData.obId,
                adpData.tsIndex,
                adpData.adpIndex);
        }
    }

    private static void updateObsBlk(
        Statement stmt,
        long mskDbId,
        long ob_id,
        int ts_index,
        int adp_index)
        throws SQLException {
        String sql =
            "UPDATE "
                + obrepDbName
                + "..obsblk_masks "
                + " SET "
                + obrepDbName
                + "..obsblk_masks.mask_id = "
                + mskDbId
                + " WHERE "
                + obrepDbName
                + "..obsblk_masks.ob_id = "
                + ob_id
                + " AND "
                + obrepDbName
                + "..obsblk_masks.ts_index = "
                + ts_index
                + " AND "
                + obrepDbName
                + "..obsblk_masks.adp_index = "
                + adp_index;
        System.out.println(sql);
        stmt.executeUpdate(sql);
    }

    public static void delete(Statement stmt, long id) throws SQLException {
        String sql =
            "SELECT masks.mask_id from masks,obsblk_masks where obsblk_masks.ob_id = "
                + id
                + " AND obsblk_masks.mask_id = masks.mask_id";
        stdlog_.debug(sql);
        ResultSet rs = stmt.executeQuery(sql);
        Vector<String> mskIdVec = new Vector<String>();
        StringBuffer buf = new StringBuffer();
        while (rs.next()) {
            mskIdVec.addElement(new Integer(rs.getInt(1)).toString());
        }
        rs.close();
        if (mskIdVec.size() == 0)
            return;
        buf.append("(");
        for (int i = 0; i < mskIdVec.size(); i++) {
            buf.append(mskIdVec.elementAt(i));
            if (i < mskIdVec.size() - 1)
                buf.append(",");
        }
        buf.append(")");
        String sqlStatus =
            "SELECT  status from obs_blocks where obs_blocks.ob_id in "
                + "(select ob_id from obsblk_masks where mask_id in "
                + buf.toString()
                + ")";
        stdlog_.debug(sqlStatus);
        rs = stmt.executeQuery(sqlStatus);
        boolean alreadyAccepted = false;
        while (rs.next()) {
            String status = rs.getString("status");
            if (!status.equals(CalibrationBlock.DEFINED)
                && !status.equals(CalibrationBlock.REJECTED))
                alreadyAccepted = true;
        }
        rs.close();
        if (alreadyAccepted)
            throw new SQLException(
                "Cannot checkout OB "
                    + id
                    + "\n Other OB's that share the same masks as OB "
                    + id
                    + "\n have already had their masks manufactured ");
        sql = "DELETE FROM obsblk_masks where ob_id = " + id;
        stdlog_.debug(sql);
        stmt.executeUpdate(sql);
        Vector<String> maskDelVec = new Vector<String>();
        for (int i = 0; i < mskIdVec.size(); i++) {
            sql =
                "select count(*) from obsblk_masks WHERE obsblk_masks.mask_id = "
                    + mskIdVec.elementAt(i);
            stdlog_.debug(sql);
            rs = stmt.executeQuery(sql);
            while (rs.next()) {
                int count = rs.getInt(1);
                if (count == 0)
                    maskDelVec.addElement(mskIdVec.elementAt(i));
            }
            rs.close();
        }
        for (int i = 0; i < maskDelVec.size(); i++) {
            sql =
                "DELETE from masks where masks.mask_id = "
                    + maskDelVec.elementAt(i);
            stdlog_.debug(sql);
            stmt.executeUpdate(sql);

        }
    }
    public static IdInfo[] getMaskIdsForOB(long id) throws SQLException {
        ConnectionPool pool = ConnectionPool.getPool();
        Connection conn = null;
        Statement stmt = null;
        Vector<IdInfo> vec = new Vector<IdInfo>();
        id = Config.getCfg().uniqueToTableId(id);
        try {
            conn =
                pool.acquireConnection(
                    Config.getCfg().getUrl(ObservationBlock.getSuffix()),
                    Config.getCfg().getDbUserName(),
                    Config.getCfg().getDbPassword());
            stmt = conn.createStatement();
            String sql =
                "SELECT "
                    + obrepDbName
                    + "..masks.bar_code , "
                    + obrepDbName
                    + " ..obsblk_masks.adp_index, "
                    + obrepDbName
                    + "..masks.mask_status, "
                    + obrepDbName
                    + "..masks.mask_name"
                    + " FROM "
                    + obrepDbName
                    + "..masks ,"
                    + obrepDbName
                    + "..obsblk_masks"
                    + " WHERE "
                    + obrepDbName
                    + "..obsblk_masks.ob_id = "
                    + id
                    + " AND "
                    + obrepDbName
                    + "..obsblk_masks.mask_id = "
                    + obrepDbName
                    + "..masks.mask_id"
                    + " ORDER BY "
                    + obrepDbName
                    + "..obsblk_masks.adp_index";
            System.out.println("sql = " + sql);
            ResultSet rs = stmt.executeQuery(sql);
            MaskDBIO mkDBIO = new MaskDBIO();
            while (rs.next()) {
                IdInfo idInfo = mkDBIO.new IdInfo();
                idInfo.adpIndex = rs.getInt("adp_index");
                idInfo.maskId = rs.getString("bar_code");
                idInfo.maskStatus = rs.getString("mask_status");
                idInfo.maskName = rs.getString("mask_name");
                vec.addElement(idInfo);
            }
            IdInfo[] result = new IdInfo[vec.size()];
            vec.copyInto(result);
            return result;
        } finally {
            if (conn != null) {
                try {
                    stmt.close();
                } catch (SQLException e) {
                    // ignore
                }
                pool.releaseConnection(conn);
            }
        }
    }

    public static int getNumMasksForStatus(String status) throws SQLException {
        ConnectionPool pool = ConnectionPool.getPool();
        Connection conn = null;
        Statement stmt = null;
        Vector vec = new Vector();
        try {
            int count = 0;
            conn =
                pool.acquireConnection(
                    Config.getCfg().getUrl(ObservationBlock.getSuffix()),
                    Config.getCfg().getDbUserName(),
                    Config.getCfg().getDbPassword());
            stmt = conn.createStatement();
            String sql =
                "SELECT COUNT(*) FROM "
                    + obrepDbName
                    + "..masks "
                    + " WHERE "
                    + obrepDbName
                    + "..masks.mask_status = "
                    + SqlUtils.sqlEscapedString(status);
            System.out.println("sql = " + sql);
            ResultSet rs = stmt.executeQuery(sql);
            while (rs.next()) {
                count = rs.getInt(1);
            }
            return count;
        } finally {
            if (conn != null) {
                try {
                    stmt.close();
                } catch (SQLException e) {
                    // ignore
                }
                pool.releaseConnection(conn);
            }
        }
    }

    public static int getMaskStoredCount() throws SQLException {
        ConnectionPool pool = ConnectionPool.getPool();
        Connection conn = null;
        Statement stmt = null;
        Vector vec = new Vector();
        try {
            int count = 0;
            conn =
                pool.acquireConnection(
                    Config.getCfg().getUrl(ObservationBlock.getSuffix()),
                    Config.getCfg().getDbUserName(),
                    Config.getCfg().getDbPassword());
            stmt = conn.createStatement();
            String sql =
                "SELECT COUNT(*) FROM "
                    + obrepDbName
                    + "..masks "
                    + " WHERE "
                    + obrepDbName
                    + "..masks.mask_status != "
                    + SqlUtils.sqlEscapedString(Mask.MASK_DISCARDED)
                    + " AND "
                    + obrepDbName
                    + "..masks.mask_status != "
                    + SqlUtils.sqlEscapedString(Mask.MASK_REMANUFACTURE)
                    + " AND "
                    + obrepDbName
                    + "..masks.bar_code != "
                    + SqlUtils.sqlEscapedString("-99");
            System.out.println("sql = " + sql);
            ResultSet rs = stmt.executeQuery(sql);
            while (rs.next()) {
                count = rs.getInt(1);
            }
            return count;
        } finally {
            if (conn != null) {
                try {
                    stmt.close();
                } catch (SQLException e) {
                    // ignore
                }
                pool.releaseConnection(conn);
            }
        }
    }

    private static String getMaskNAID(PafFile paf) {
        PafData[] pafData = paf.getIdentifierList();
        for (int j = 0; j < pafData.length; j++) {
            if (pafData[j].getIdentifier().equals("INS.MASK.NAID")) {
                return pafData[j].getValue();
            }
        }
        return paf.getPafIdentifier();
    }

    public class IdInfo {
        private int adpIndex;
        private String maskId;
        private String maskStatus;

        private String maskName;

        /**
         * Returns the adpIndex.
         * @return int
         */
        public int getAdpIndex() {
            return adpIndex;
        }

        /**
         * Returns the maskId.
         * @return String
         */
        public String getMaskId() {
            return maskId;
        }

        /**
         * Returns the maskStatus.
         * @return String
         */
        public String getMaskStatus() {
            return maskStatus;
        }

        /**
         * Sets the adpIndex.
         * @param adpIndex The adpIndex to set
         */
        public void setAdpIndex(int adpIndex) {
            this.adpIndex = adpIndex;
        }

        /**
         * Sets the maskId.
         * @param maskId The maskId to set
         */
        public void setMaskId(String maskId) {
            this.maskId = maskId;
        }

        public String getMaskName() {
            return maskName;
        }

        public void setMaskName(String maskName) {
            this.maskName = maskName;
        }

        /**
         * Sets the maskStatus.
         * @param maskStatus The maskStatus to set
         */
        public void setMaskStatus(String maskStatus) {
            this.maskStatus = maskStatus;
        }

        public boolean equals(Object obj) {
            if (((IdInfo) obj).getMaskId().equals(getMaskId()))
                return true;
            return false;
        }

    }

}
