package org.eso.ohs.phase2.apps.p2ppAppServer;

import java.io.BufferedWriter;
import java.text.SimpleDateFormat;

import org.eso.ohs.core.utilities.Logger;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.FindingChart;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.Readme;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.instruments.TemplateVersionInfo;
import org.eso.ohs.persistence.DecoratorStorageManager;
import org.eso.ohs.persistence.DirectoryNode;
import org.eso.ohs.persistence.Media;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.persistence.StorageManager;

/**
 * @version $Id: LoggingStorageManager.java,v 1.6 2004/02/17 14:31:56 tcanavan
 *          Exp $
 * 
 * A StorageManager class to log all method calls. <BR>
 * Each method delegates to methods in an auxiliary StorageManager but their
 * start/success/failure is logged to an output stream.
 * 
 * <p>
 * The class is typically used as
 * 
 * <pre>
 * StorageManager stMgr = new LoggingStorageManager(new FileWriter(
 *         &quot;/tmp/someLogFile&quot;), someStorageMgr);
 * </pre>
 * 
 * @author Bob Kemp <br>
 *         ChangeLog: amchavan, 11-Oct-2001: factored out the Logger class to
 *         the <code>utilities</code> package. <br>
 *         ChangeLog: cmavros 22Apr02 1195/P2PP Validation of OBs through P2PP
 *         AppServer.
 * 
 * @see StorageManager
 */
public class LoggingStorageManager extends DecoratorStorageManager {

    /**
     * Logger object to which all operations are logged..
     */
    private Logger logger_;

    /** Format for encoding/decoding text for an ISO format timestamp */
    private SimpleDateFormat isoTimeFormat_;

    public LoggingStorageManager(BufferedWriter wtr, StorageManager stMgr) {
        super(stMgr);
        logger_ = new Logger(wtr);
    }

    public LoggingStorageManager(Logger logger, StorageManager stMgr) {
        super(stMgr);
        logger_ = logger;
    }

    /**
     * Sets the id of the current user.
     * 
     * @param userId
     *            the id of the end-user
     * 
     * @exception ObjectIOException
     *                if there was an I/O error while setting the userId
     */
    public void setAccountId(int userId) throws ObjectIOException {
        final String whoami = "setUserId";

        try {
            logger_.logStart(whoami, String.valueOf(userId));
            super.setAccountId(userId);
            logger_.setUserId(userId);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Gets the id of the current user.
     * 
     * @return the id of the end-user
     */
    public int getAccountId() {
        final String whoami = "getUserId";

        try {
            logger_.logStart(whoami, "");
            int result = super.getAccountId();
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Get the root node of the DirectoryNode tree representing the directories
     * found on this medium. For example, a representation of the root of the
     * local StorableObject storage subtree.
     * 
     * @return a representation of the root node
     * 
     * @exception ObjectIOException
     *                if there was an I/O error while determining the directory
     *                hierarchy.
     */
    public DirectoryNode getRoot() throws ObjectIOException {
        final String whoami = "getRoot";

        try {
            logger_.logStart(whoami, "");
            DirectoryNode result = super.getRoot();
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * @return the requested IP as a byte array, or an empty (size 0) array if
     *         the IP does not exist on the server.
     * @exception ObjectIOException
     *                when there is an I/O exception.
     */
    public byte[] downloadInstrumentPackage(String instrument, float version)
            throws ObjectIOException {
        final String whoami = "verifyTemplateSignatures";

        try {
            logger_.logStart(whoami, "download ip for " + instrument + " v"
                    + version);
            byte[] result = super
                    .downloadInstrumentPackage(instrument, version);
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Verify that the TemplateSignature files used by the StorageManager are
     * the same version as those specified (and used by the client)
     * 
     * @param tsList
     *            array of version info objects for each template used in an OB
     * 
     * @return a list of the names of TemplateSignatures that are out of date.
     * 
     * @exception ObjectIOException
     *                if there was an I/O error while checking
     */
    public String[] verifyTemplateSignatures(TemplateVersionInfo[] tsList)
            throws ObjectIOException {
        final String whoami = "verifyTemplateSignatures";

        try {
            logger_.logStart(whoami, tsList.length + " templates to verify");
            String[] result = super.verifyTemplateSignatures(tsList);
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    public FindingChart[] populateFindingCharts(long boId)
            throws ObjectIOException {
        final String whoami = "populateFindingCharts";
        for (;;) {
            try {
                logger_.logStart(whoami, boId + " finding chart binary data");
                FindingChart[] fcCharts = super.populateFindingCharts(boId);
                logger_.logSuccess(whoami);
                return fcCharts;
            } catch (RuntimeException e) {
                logger_.logException(whoami, e);
                throw e;
            } catch (ObjectIOException e) {
                logger_.logException(whoami, e);
                throw e;
            }
        }
    }

    /**
     * Get the type of media maintained by this StorageManager instance.
     * 
     * @return an instance of the Media class.
     */
    public Media getMedia() {
        final String whoami = "getMedia";

        try {
            logger_.logStart(whoami, "");
            Media result = super.getMedia();
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Calculates a unique id for a StorableObject.
     * 
     * @return Unique id
     * 
     * @exception ObjectIOException
     *                if there was an I/O error writing the object.
     */
    public long getNewId(Class cls, long baseId) throws ObjectIOException {
        final String whoami = "getNewId";

        try {
            logger_.logStart(whoami, cls.getName(), String.valueOf(baseId));
            long result = super.getNewId(cls, baseId);
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Initialise the data structures for the appropiate media. For local
     * storage this will be a directory structure.
     * 
     * @exception ObjectIOException
     *                if there was an I/O error writing the object.
     */
    public void initialise() throws ObjectIOException {
        final String whoami = "initialise";

        try {
            logger_.logStart(whoami);
            super.initialise();
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Authenticates the end-user.
     * 
     * @param password
     *            password for the current user
     * 
     * @return true if user was successfully authenticated, false otherwise
     * 
     * @exception ObjectIOException
     *                if there was an I/O error while authenticating
     */
    public boolean loginUser(String password) throws ObjectIOException {
        final String whoami = "loginUser";

        try {
            logger_.logStart(whoami, password);
            boolean result = super.loginUser(password);
            if (result)
                logger_.logSuccess(whoami);
            else
                logger_.logFailure(whoami);

            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Returns a DirectoryNode representing the path within the storage medium
     * of an object with the given id and type. For media that do not support
     * directories, the root directory may be returned.
     * 
     * @param id
     *            the id of the StorableObject.
     * @param cls
     *            the type of stored StorableObject.
     * 
     * @return Directory in which to find the object.
     * 
     * @exception ObjectIOException
     *                if there was an I/O error writing the object.
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found.
     */
    public DirectoryNode find(long id, Class objType)
            throws ObjectNotFoundException, ObjectIOException {
        final String whoami = "find";

        try {
            logger_.logStart(whoami, String.valueOf(id), objType.getName());
            DirectoryNode result = super.find(id, objType);
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectNotFoundException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Returns an array of objects representing the contents of an object with
     * the given id and type from the chosen directory. If the medium does not
     * support directories, the <code>node</code> parameter is ignored.
     * 
     * <p>
     * The format for the array of objects is documented in StorageManager.
     * 
     * @param node
     *            directory from which the object should be loaded.
     * @param id
     *            id of the StorableObject.
     * @param objType
     *            type of StorableObject.
     * 
     * @return the array of objects and object references, as described above.
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found.
     * @exception ObjectIOException
     *                if an I/O error occurred while reading the object
     * 
     * @see StorageManager#read
     * @see #write
     * @see #update
     */
    public UnpackedStorableObject read(DirectoryNode node, long id,
            Class objType) throws ObjectNotFoundException, ObjectIOException {
        final String whoami = "read";

        try {
            logger_.logStart(whoami, String.valueOf(id), objType.getName());
            UnpackedStorableObject result = super.read(node, id, objType);
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectNotFoundException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Writes an array of objects with the given id and type to this storage
     * medium, in the chosen directory. If the medium does not support
     * directories, the <code>node</code> parameter is ignored.
     * 
     * @param node
     *            the directory in which the object should be saved.
     * @param id
     *            the id under which to store the StorableObject.
     * @param objType
     *            the type of StorableObject.
     * @param obj
     *            the array of objects and object references to store, as
     *            described in the section for the read method.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while writing the object
     * 
     * @see #read
     * @see #update
     */
    public void write(DirectoryNode node, long id, Class objType,
            UnpackedStorableObject obj) throws ObjectIOException {
        final String whoami = "write";

        try {
            logger_.logStart(whoami, String.valueOf(id), objType.getName());
            super.write(node, id, objType, obj);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Write several object to the external media in an atomic operation.
     * 
     * @param moi
     *            Array of object descriptions.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while writing the object
     * 
     * @see #read
     * @see #write
     * @see #update
     */
    public void multiStore(MultiObjSpec[] mos) throws ObjectIOException {
        final String whoami = "multiStore";

        try {
            logger_.logStart(whoami, mos.length + " elements", mos[0].obj);
            super.multiStore(mos);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Updates an array of objects with the given id and type on this storage
     * medium, in the chosen directory. If the medium does not support
     * directories, the <code>node</code> parameter is ignored.
     * 
     * @param node
     *            the directory in which the object should be saved.
     * @param id
     *            the id under which to store the StorableObject.
     * @param objType
     *            the type of StorableObject.
     * @param obj
     *            the array of objects and object references to store, as
     *            described in the section for the read method.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while updating the object
     * 
     * @see #read
     * @see #write
     */
    public void update(DirectoryNode node, long id, Class objType,
            UnpackedStorableObject obj) throws ObjectIOException {
        final String whoami = "update";

        try {
            logger_.logStart(whoami, String.valueOf(id), objType.getName());
            super.update(node, id, objType, obj);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Deletes an object with the given directory, id and type from this storage
     * medium. If the medium does not support directories, the <code>node</code>
     * parameter is ignored.
     * 
     * @param node
     *            directory from which the object should be deleted.
     * @param id
     *            id of the StorableObject.
     * @param objType
     *            type of StorableObject.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while updating the object
     * 
     * @see #read
     * @see #update
     */
    public void delete(DirectoryNode node, long id, Class objType)
            throws ObjectIOException {
        final String whoami = "delete";

        try {
            logger_.logStart(whoami, String.valueOf(id), objType.getName());
            super.delete(node, id, objType);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Moves an object between directories on this storage medium.
     * 
     * @param id
     *            id of the StorableObject.
     * @param cls
     *            type of StorableObject.
     * @param fromNode
     *            directory from which the object should be moved.
     * @param toNode
     *            directory to which the object should be moved.
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred while updating the object
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found in fromNode.
     */
    public void move(long id, Class objType, DirectoryNode fromNode,
            DirectoryNode toNode) throws ObjectNotFoundException,
            ObjectIOException {
        final String whoami = "move";

        try {
            logger_.logStart(whoami, String.valueOf(id), toNode);
            super.move(id, objType, fromNode, toNode);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Create a directory in local storage
     * 
     * @param node
     *            The directory to be created
     * @param name
     *            The name of the directory to be created.
     * 
     * @return True on success false on error
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred during the operation
     */
    public boolean createDirectory(DirectoryNode node, long id)
            throws ObjectIOException {
        final String whoami = "createDirectory";

        try {
            logger_.logStart(whoami, node, String.valueOf(id));
            boolean result = super.createDirectory(node, id);
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Move a directory.
     * 
     * @param node
     *            directory to be moved
     * @param toNode
     *            the new parent directory
     * 
     * @return True on success false on error
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred during the operation
     * 
     * @exception ObjectNotFoundException
     *                if the object was not found in fromNode.
     */
    public boolean moveDirectory(DirectoryNode fromNode, DirectoryNode toNode)
            throws ObjectIOException, ObjectNotFoundException {
        final String whoami = "moveDirectory";

        try {
            logger_.logStart(whoami, fromNode, toNode);
            boolean result = super.moveDirectory(fromNode, toNode);
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectNotFoundException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Delete a directory from local storeage
     * 
     * @param node
     *            Directory to be deleted
     * 
     * @return True on success, false on failure
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred during the operation
     */
    public boolean deleteDirectory(DirectoryNode node) throws ObjectIOException {
        final String whoami = "deleteDirectory";

        try {
            logger_.logStart(whoami, node);
            boolean result = super.deleteDirectory(node);
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Returns a list of Summary objects associated with a directory.
     * 
     * @param node
     *            Directory to be used to build up list of objects
     * 
     * @return Array of Summary[] objects describing the directory contents
     * 
     * @exception ObjectIOException
     *                if an I/O error occurred during the operation
     */
    public Summary[] listObjects(DirectoryNode node) throws ObjectIOException {
        final String whoami = "listObjects";

        try {
            logger_.logStart(whoami, node);
            Summary[] result = super.listObjects(node);
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Do a EVM verification test for the specified objects in a single atomic
     * operation.
     * 
     * @param usos
     *            Descriptions of each StorableObject to be checked in.
     * 
     * @return boolean if object passed verification
     */
    public boolean verifyObject(MultiObjSpec[] moi) throws ObjectIOException {
        final String whoami = "verifyObject";
        try {
            logger_.logStart(whoami, String.valueOf(moi.length));
            boolean result = super.verifyObject(moi);
            logger_.logSuccess(whoami);
            return result;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Check in the specified objects in a single atomic operation.
     * 
     * @param usos
     *            Descriptions of each StorableObject to be checked in.
     * 
     * @return array of error strings. If length = 0, no errors.
     */
    public void checkIn(MultiObjSpec[] moi) throws ObjectIOException {
        final String whoami = "checkIn";

        try {
            logger_.logStart(whoami, String.valueOf(moi.length));
            super.checkIn(moi);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }

    /**
     * Checkin the readme file for a given observing run
     * 
     * @param Readme
     *            file business object
     */

    public void checkInReadme(ObservingRun obsRun) throws ObjectIOException {
        final String whoami = "checkInReadme";
        try {
            logger_.logStart(whoami,obsRun.getProgId());
            super.checkInReadme(obsRun);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }

    }

    /**
     * Checkout the readme file for a given observing run.
     * 
     * @param boId
     *            Observing run definition.
     */

    public String checkOutReadme(ObservingRun obsRun) throws ObjectIOException {
        final String whoami = "checkOutReadme";
        try {
            logger_.logStart(whoami,obsRun.getProgId());
            String xmlStr = super.checkOutReadme(obsRun);
            logger_.logSuccess(whoami);
            return xmlStr;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }
    
    /**
     * Wrapper for populating readme file
     * 
     * @param boId
     *            Observing run definition.
     */

    public Readme populateReadmeFile(ObservingRun obsRun) throws ObjectIOException {
        final String whoami = "populateReadmeFile";
        try {
            logger_.logStart(whoami,obsRun.getProgId());
            Readme readme = super.populateReadmeFile(obsRun);
            logger_.logSuccess(whoami);
            return readme;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }
    
    
    public Readme populateReadmeFileReplicate(ObservingRun obsRun) throws ObjectIOException {
        final String whoami = "populateReadmeFileReplicate";
        try {
            logger_.logStart(whoami,obsRun.getProgId());
            Readme readme = super.populateReadmeFileReplicate(obsRun);
            logger_.logSuccess(whoami);
            return readme;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }


    
    
    public void usdSubmit(ObservingRun obsRun) throws ObjectIOException {
        final String whoami = "usdSubmit";
        try {
            logger_.logStart(whoami,obsRun.getProgId());        
            super.usdSubmit(obsRun);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }
    
    
    public void setReadmeFileStatus(String status,long runId) throws ObjectIOException {
        final String whoami = "setReadmeFileStatus";
        try {
            logger_.logStart(whoami,status);        
            super.setReadmeFileStatus(status,runId);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }


    
    
    
    

    /**
     * Check out the ObsBlk or CalBlk with the specified id plus any attached
     * objects, ie Target, ObsDescr, etc.
     * 
     * @param id
     *            id of the ObsBlk or CalBlk to be checked out
     * 
     * @return CheckOutStatus object that contains a return status and (if
     *         successful) references to the checked-out objects.
     */
    public void checkOut(IdType[] idTypes) throws ObjectNotFoundException,
            ObjectIOException {
        final String whoami = "checkOut";

        try {
            logger_.logStart(whoami, String.valueOf(idTypes[0].id));
            super.checkOut(idTypes);
            logger_.logSuccess(whoami);
            return;
        } catch (RuntimeException e) {
            logger_.logException(whoami, e);
            throw e;
        } catch (ObjectIOException e) {
            logger_.logException(whoami, e);
            throw e;
        }
    }
}

