package org.eso.ohs.phase2.apps.p2ppAppServer;

import java.rmi.Remote;
import java.rmi.RemoteException;

import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.AccountInfo;
import org.eso.ohs.dfs.FindingChart;
import org.eso.ohs.dfs.ObservingRun;
import org.eso.ohs.dfs.Readme;
import org.eso.ohs.dfs.Summary;
import org.eso.ohs.instruments.TemplateVersionInfo;
import org.eso.ohs.persistence.DirectoryNode;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.persistence.StorageManager;
import org.eso.ohs.persistence.StorageManager.IdType;
import org.eso.ohs.persistence.StorageManager.MultiObjSpec;
import org.eso.ohs.persistence.StorageManager.UnpackedStorableObject;

/**
 * This interface is just a placeholder to merge the StorageManager
 * and java.rmi.Remote interfaces.
 * 
 * <p>
 * $Id: RmiManagerInterface.java,v 1.11.8.4 2007/11/07 14:05:51 pnunes Exp $
 * <br>
 * @author Bob Kemp
 * <br>ChangeLog: cmavros 22Apr02 1195/P2PP Validation of OBs through P2PP AppServer.
 */

public interface RmiManagerInterface extends Remote {
    // This interface is just a placeholder to merge the StorageManager
    // and java.rmi.Remote interfaces.


    /**
     * Calculates a unique id for a StorableObject.
     *
     * @return Unique id
     * 
     * @exception ObjectIOException
     *			if the method call generated an I/O error
     */
    public long getNewId(Class mo, long baseId)
			    throws ObjectIOException,
				    RemoteException;

    /**
     * Get the root node of the DirectoryNode tree representing
     * the directories found on this medium.  For example, a
     * representation of the root of the local StorableObject
     * storage subtree.
     * 
     * @return a representation of the root node
     * 
     * @exception ObjectIOException
     *			if there was an I/O error while determining the
     * 			directory hierarchy.
     */
    public DirectoryNode getRoot()
			    throws ObjectIOException,
				    RemoteException;

    /**
     * Sets the id of the current user.
     * 
     * @param userId	the id of the end-user
     * 
     * @exception ObjectIOException
     *			if there was an I/O error while setting the userId
     */
    public void setUserId(int userId)
			    throws ObjectIOException,
				    RemoteException;

    /**
     * Authenticates the end-user.
     * 
     * @param password	password for the current user
     * 
     * @return	true if user was successfully authenticated, false otherwise
     * 
     * @exception ObjectIOException
     *			if there was an I/O error while authenticating
     */
    public boolean loginUser(String password)
			    throws ObjectIOException,
				    RemoteException;

    /**
     * Returns a DirectoryNode representing the path within the
     * storage medium of an object with the given id and type.
     * For media that do not support directories, the root directory
     * may be returned.
     * 
     * @param id	the id of the StorableObject.
     * @param objType	the type of stored StorableObject.
     * 
     * @return Directory in which to find the object.
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * 
     * @exception ObjectIOException
     *			if the method call generated an I/O error
     */
    public DirectoryNode find(long id, Class objType)
			    throws ObjectNotFoundException,
				    ObjectIOException,
				    RemoteException;


    /**
     * Returns an array of objects representing the contents of an
     * object with the given id and type from the chosen directory.
     * If the medium does not support directories, the <code>node</code>
     * parameter is ignored.
     * 
     * <p>
     * The array of objects contains:
     * <ul>
     *     <li> as the first (0th) element, an instance of a
     *          StorableObject subclass; followed by
     *     <li> an alternating sequence of values:
     *          <ul>
     *              <li> at even indexes, a property name, ie a string X
     *                   such that setX() and getX() methods exist.
     *              <li> at odd indexes, either an instance of
     *                   java.lang.Long containing the the id of the
     *                   StorableObject subclass to which the property
     *                   should refer or a reference to the subclass itself.
     *          </ul>
     * </ul>
     * 
     * @param node	directory from which the object should be loaded.
     * @param id	id of the StorableObject.
     * @param objType	type of StorableObject.
     * 
     * @return the array of objects and object references,
     * 			as described above.
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * @exception ObjectIOException
     * 			if an I/O error occurred while reading the object
     * 
     * @see #write
     * @see #update
     */
    public UnpackedStorableObject read(DirectoryNode node, long id,
				Class objType)
				throws ObjectNotFoundException,
					ObjectIOException,
					RemoteException;

    /**
     * Writes an array of objects with the given id and type to
     * this storage medium, in the chosen directory.  If the medium
     * does not support directories, the <code>node</code> parameter
     * is ignored.
     * 
     * @param node	the directory in which the object should be saved.
     * @param id	the id under which to store the StorableObject.
     * @param objType	the type of StorableObject.
     * @param obj	the array of objects and object references to store,
     *			as described in the section for the read method.
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred while writing the object
     * 
     * @see #read
     * @see #update
     */
    public void write(DirectoryNode node, long id,
				Class objType, UnpackedStorableObject obj)
				throws ObjectIOException,
					RemoteException;

    /**
     * Updates an array of objects with the given id and type on
     * this storage medium, in the chosen directory.  If the medium
     * does not support directories, the <code>node</code> parameter
     * is ignored.
     * 
     * @param node	the directory in which the object should be saved.
     * @param id	the id under which to store the StorableObject.
     * @param objType	the type of StorableObject.
     * @param obj	the array of objects and object references to store,
     *			as described in the section for the read method.
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred while updating the object
     * 
     * @see #read
     * @see #write
     */
    public void update(DirectoryNode node, long id,
				Class objType, UnpackedStorableObject obj)
				throws ObjectIOException,
					RemoteException;


    /**
     * Deletes an object with the given directory, id and type from
     * this storage medium.  If the medium does not support
     * directories, the <code>node</code> parameter is ignored.
     * 
     * @param node	directory from which the object should be deleted.
     * @param id	id of the StorableObject.
     * @param objType	type of StorableObject.
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred while updating the object
     * 
     * @see #read
     * @see #update
     */
    public void delete(DirectoryNode node, long id,
				Class objType)
				throws ObjectIOException,
					RemoteException;

    /**
     * Deletes an object with the given directory, id and type from
     * this storage medium.  If the medium does not support
     * directories, the <code>node</code> parameter is ignored.
     * 
     * @param node	directory from which the object should be deleted.
     * @param id	id of the StorableObject.
     * @param objType	type of StorableObject.
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred while updating the object
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     * 
     * @see #read
     * @see #update
     */
    public void move(long id, Class objType,
				DirectoryNode fromNode,
				DirectoryNode toNode)
				throws ObjectNotFoundException,
					ObjectIOException,
					RemoteException;

    /**
     * Create a directory in local store.
     *
     * @param node The directory to be created
     * @param name The name of the directory to be created.
     * 
     * @return True on success false on error
     * 
     * @exception ObjectIOException
     *			if the method call generated an I/O error
     */ 
    public boolean createDirectory(DirectoryNode node, long id)
				throws RemoteException,
					ObjectIOException;

    /**
     * Move a directory.
     *
     * @param fromNode	directory to be moved
     * @param toNode	the new parent directory
     * 
     * @return True on success false on error
     * 
     * @exception ObjectIOException
     *			if the method call generated an I/O error
     * 
     * @exception ObjectNotFoundException
     *			if the object was not found.
     */ 
    public boolean moveDirectory(DirectoryNode fromNode,
					    DirectoryNode toNode)
				throws ObjectNotFoundException,
					ObjectIOException,
					RemoteException;

    /**
     * Delete a directory from local storeage
     *      
     * @param node Delete a directory from Local Storeage
     * 
     * @return True on success false on error
     * 
     * @exception ObjectIOException
     *			if the method call generated an I/O error
     */
    public boolean deleteDirectory(DirectoryNode node)
				throws ObjectIOException,
					RemoteException;


    /**
     * Returns a list of ObjectInfo objects associated with a directory.
     * 
     * @param node Directory to be used to build up list of objects
     * @return An array of ObjectFileInfo[] objects 
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred while updating the object
     */
    public Summary[] listObjects(DirectoryNode node)
			    throws ObjectIOException,
				    RemoteException;

    /**
     * Do a EVM verification test for the specified objects in a single atomic operation.
     *
     * @param   usos    Descriptions of each StorableObject to do EVM verification
     *
     * @return boolean if OB passed EVM test
     */
    public boolean verifyObject(MultiObjSpec[] moi)
				throws ObjectIOException,
					RemoteException;
					
			
			
    /**
     * @return the requested IP as a byte array, or an empty (size 0) array
     * if the IP does not exist on the server.
     * @exception ObjectIOException when there is an I/O exception.
     */
    public byte[] downloadInstrumentPackage(String instrument, float version)
            throws ObjectIOException, RemoteException;
 

    /**
     * Check in the specified objects in a single atomic operation.
     * 
     * @param	usos	Descriptions of each StorableObject to be checked in.
     * 
     * @return		array of error strings.
     *			If length = 0, no errors.
     */
    public void checkIn(MultiObjSpec[] moi)
				throws ObjectIOException,
					RemoteException;


    /**
     * Check out the ObsBlk or CalBlk with the specified id plus
     * any attached objects, ie Target, ObsDescr, etc.
     * 
     * @param	id	id of the ObsBlk or CalBlk to be checked out
     * 
     * @return		CheckOutStatus object that contains a return
     *			status and (if successful) references to the
     *			checked-out objects.
     */
    public void checkOut(IdType[] idTypes)
				throws ObjectNotFoundException,
					ObjectIOException,
					RemoteException;


    /**
     * Write several object to the external media in an atomic operation.
     *      
     * @param moi	Array of object descriptions.
     * 
     * @exception ObjectIOException
     * 			if an I/O error occurred while writing the object
     * 
     * @see #read
     * @see #write
     * @see #update
     */
    public void multiStore(StorageManager.MultiObjSpec[] moi)
				throws ObjectIOException,
					RemoteException;

    /**
     * Verify that the TemplateSignature files used by the StorageManager
     * are the same version as those specified (and used by the client)
     * 
     * @param tsList	array of version info objects for each template
     *			used in an OB
     * 
     * @return a list of the names of TemplateSignatures that are out of date.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error while checking
     */
    public FindingChart [] populateFindingCharts(long boId)
			    throws ObjectIOException,
				    RemoteException;


    /**
     * Verify that the TemplateSignature files used by the StorageManager
     * are the same version as those specified (and used by the client)
     * 
     * @param tsList	array of version info objects for each template
     *			used in an OB
     * 
     * @return a list of the names of TemplateSignatures that are out of date.
     * 
     * @exception ObjectIOException
     *			if there was an I/O error while checking
     */
    public String[] verifyTemplateSignatures(TemplateVersionInfo[] tslist)
			    throws ObjectIOException,
				    RemoteException;
    
    
    
    /**
     * Checkin the readme file for a given observing run
     * @param Readme file business object
     */
    
    public void checkInReadme(ObservingRun obsRun)
    throws ObjectIOException,RemoteException;    
    
    /**
     * Checkout the readme file for a given observing run.
     * @param boId Observing run definition. 
     */
    
    public String checkOutReadme(ObservingRun obsRun)
    throws ObjectIOException,RemoteException;    

    
    public void usdSubmit(ObservingRun obsRun)
    throws ObjectIOException,RemoteException;

    /**
     * @param obsRun
     */
    public Readme populateReadmeFile(ObservingRun obsRun) throws ObjectIOException,RemoteException;

//	/**
//	 * Returns a user Id, given a username and password
//	 * 
//	 * @param username
//	 * @param passwd
//	 * @return a user Id or -1 if not existing or invalid
//	 */
//	public int getUserIdFromUsername(String username, String passwd) throws ObjectIOException,RemoteException;
//
//	/**
//	 * Checks if an user account is activated
//	 * 
//	 * @param username
//	 * @return
//	 */
//	public boolean isAccountActivated(String username)throws ObjectIOException,RemoteException;  
	

	/**
	 * @param username
	 * @param password
	 * @return
	 * @throws ObjectIOException
	 * @throws RemoteException
	 */
	public AccountInfo getAccountInfo(String username, String password) throws ObjectIOException, RemoteException;
	
	/**
	 * @param accountId
	 * @param password
	 * @return
	 * @throws ObjectIOException
	 * @throws RemoteException
	 */
	public AccountInfo getAccountInfo(Integer accountId, String password) throws ObjectIOException, RemoteException;



}

