package org.eso.ohs.scripting;

/**
 * This interface defines the ObservationBlock. From this interface,
 * all OB level data is available, as well as the requesting of
 * Template numbers and TemplateAPI objects.
 *
 * IMPORTANT: Changes to this interface must result in a new version string
 * in the ScriptingBox class.
 *
 * @author Nick Kornweibel
 */

public interface ObservationBlockAPI {

    /**
     * Get the number of templates in this OB (NOT including ACQ template?).
     * @return the number of templates
     */

    public int getNumberOfODTemplates();


    /**
     * Get the number of finding charts attached to an OB
     * @return the number of finding charts attached to an OB
     */

    public int getNumberOfFindingCharts();

    /**
     * Get the number of finding charts attached to an OB
     * @return the number of finding charts attached to an OB
     */

    public int getNumberOfSiderealTimeIntervals();


    /**
     * Get the template at the specified index
     * @param template index (>=0)
     * @return TemaplateAPI at that index, or null if not found
     */

    public TemplateAPI getTemplateAt(int templateIndex);

    /**
     * Get the name of this OB
     * @return the name of the OB
     */

    public String getName();

    /**
     * Get the OBID of this OB (the database id)
     * @return string containing OBID
     */

    public String getId();

    /**
     * Get the run id for this OB
     * @return string containing the run-id of this OB
     */

    public String getRunID();

    /**
     * Get the mode of the observing run (S: service, E: engineering
     * V: visitior).
     * @return String
     */
    public String getRunMode();

    /**
     * Get the name of the instrument used in this OB
     * @return string containing the name of the instrument
     */

    public String getInstrumentName();

    /**
     * Get the name of the target for this OB
     * @return string containing the name of the target
     */

    public String getTargetName();

    /**
     * Get the string form of the RA value for this OB
     * @return string containing the RA e.g. "05:45:36.5"
     */

    public int getTargetRa();

    /**
     * Get the string form of the DEC value for this OB
     * @return string containing the DEC e.g. "-78:11:09.0"
     */

    public int getTargetDec();

    /**
     * Get the type of this OB:
     *     "O" for Observation
     *     "C" for Calibration
     * @return string containing type of this OB
     */

    public String getType();

    /**
     * Get the target for this OB
     * @return the TargetAPI for this OB
     */

    public TargetAPI getTarget();

    /**
     * Get the Constraint Set for this OB
     * @return ConstraintSet for this OB
     */

    public ConstraintSetAPI getConstraintSet();

    /**
     * Get the timing intervals for this OB
     * @return TimingIntervalSetAPI object for this OB
     */

    public TimeIntervalSetAPI getTimeIntervals();

    /**
     * Get the sidereal timing intervals for this OB
     * @return TimingIntervalSetAPI object for this OB
     */

    public TimeIntervalSetAPI getSTTimeIntervals();



    /**
     * Get the template that is the acquisition template for this OB
     * This is typically equivalent to getTemplateAt(0)
     * @return Acquisition Template for this OB, or null if none present
     */

    public TemplateAPI getAcquisitionTemplate();

    /**
     * @return the Exposure time, in seconds, or 0 if not calculated
     * or not known.
     */
    public int getExposureTime();

    /**
     * @return the Execution time, in seconds, or 0 if not calculated
     * or not known.
     */
    public int getExecutionTime();

    /**
     * Set the execution time property in the OB
     * @param execTime
     */

    public void setExecutionTime(int execTime);

    /**
     * Set the exposure time propert in the OB
     * @param expTime
     */

    public void setExposureTime(int expTime);


    /**
     * The instruments comments for a particular OB/CB
     * @return instrumentComments
     */

    public String getInstrumentComments();

    /**
     * The user comments for a particular OB/CB
     * @return userComments
     */

    public String getUserComments();


    /**
     * @return
     */
    public FindingChartSetAPI getFindingCharts();


}
