package org.eso.ohs.scripting;

/**
 * An adaptor class from ObservationBlockAPI to ObservationBlock
 * This class accepts an ObservationBlock and exposes the
 * ObservationBlockAPI.
 *
 * @author Nick Kornweibel
 * <br>ChangeLog: cmavros 03/07/02 getId() no longer get Media from ObjMgr
 *    but interpret from blk_.getId()
 *
 */

import org.eso.ohs.dfs.CalibrationBlock;
import org.eso.ohs.dfs.FindingChart;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.dfs.Target;
import org.eso.ohs.dfs.TimeInterval;
import org.eso.ohs.instruments.TemplateSignature;
import org.eso.ohs.persistence.Config;

public class ObservationBlockAdaptor implements ObservationBlockAPI {

    private CalibrationBlock blk_ = null;

    /**
     * Contructor - requires an CalibrationBlock to act as adaptor
     * @param the observation block object, assumes non null
     */

    public ObservationBlockAdaptor(CalibrationBlock blk) {
        blk_ = blk;
    }

    /**
     * Get the Cal block (not to be used by the EVM scripts directly).
     * If needed in Jacl, can be used as follows:
     *     set oba [java::cast org.eso.ohs.scripting.ObservationBlockAdaptor $ob]
     *     set blk [$oba getBlock]
     *     $blk getCalculateExecutionTime $blk
     *
     * @return CalibrationBlock
     */
    public CalibrationBlock getBlock() {
        return blk_;
    }

    /**
     * Get the number of templates in this OB
     * @return the number of templates (NOT including acq template)
     */

    public int getNumberOfODTemplates() {
        int rc = 0;
        TemplateSignature[] od = blk_.getOd().getSignatures();
        if (od != null) {
           rc = od.length;
        }
        return rc;
    }



    /**
     * Get the number of finding charts in an OB
     * @return the number of finding charts
     */

    public int getNumberOfFindingCharts() {
        int rc = 0;
        if(blk_ instanceof ObservationBlock) {
           FindingChart[] fcs = ((ObservationBlock)blk_).getFindingCharts();
           if (fcs != null) {
              rc = fcs.length;
           }
        }
        return rc;
    }

    /**
     * Get a Template for the specified index.
     * If 0, acq is return (or null if no acq).
     * @param the index of the required template (0 for acq)
     * @return the template object for that index, or null if not found
     */

    public TemplateAPI getTemplateAt(int templateIndex) {
        TemplateAPI rc = null;
        TemplateSignature tmpl = null;

        if (templateIndex == 0 && blk_ instanceof ObservationBlock) {
            TemplateSignature tpl =
                    ((ObservationBlock)blk_).getTemplateSignature();
            if (tpl != null) {
                rc = new TemplateAdaptor(tpl, 0);
            }
        } else if (templateIndex > 0) {
            templateIndex--;
            TemplateSignature[] od = blk_.getOd().getSignatures();
            if (od != null && od.length > templateIndex) {
               rc = new TemplateAdaptor(od[templateIndex], templateIndex+1);
            }
        }
        return rc;
    }

    /**
     * Get the name of this OB
     * @return the name of the OB
     */

    public String getName() {
        return ScriptingUtilities.nullToEmptyString(blk_.getName());
    }

    /**
     * Get the observation ID of this OB
     * @return the OB-ID String for this Observation Block
     */
/*
 *
 * OLD IMPLEMENTATION - repleaced by getId() below.
    public String getId() {
        String rc;
        Media device = ObjectManager.getObjectManager().lookup(blk_.getId()).getMedia();

       if (device == null) {
            // this observation block has no Media association. This occurs when it is
            // an OB received by the AppServer as a Multi Object Spec from a client.
            // we assign the media based on the id of the block....
            if (blk_.getId() < 0) {
                device = Media.LOCAL;
            } else {
                device = Media.DBASE;
            }
        }

        if(device.equals(Media.DBASE))
            rc = ""+blk_.getDbaseId();
        else {
            if (blk_.getDbaseId() != 0) {
                rc = ""+Config.getCfg().uniqueToTableId(blk_.getDbaseId());
            } else {
                rc = ""+blk_.getId();
           }
        }
        return rc;
    }
 */

    public String getId() {
        String rc;
        if(blk_.getId() == 0) // AppServer
            rc = "" + blk_.getId();
        else if(blk_.getId() < 0) { // Local Cacge
            if (blk_.getDbaseId() != 0) {
                rc = ""+Config.getCfg().uniqueToTableId(blk_.getDbaseId());
            } else {
                rc = ""+blk_.getId();
            }
        }
        else // Dbase Id
            rc = "" + blk_.getDbaseId();
       return rc;
    }


    /**
     * Get the RunID of this OB.
     * @return the run id (prog-id) of this OB
     */

    public String getRunID() {
        String rc = ScriptingUtilities.nullToEmptyString(
                blk_.getObsRun().getProgId());
        return rc;
    }

    /**
     * Get the mode of the observing run (S: service, E: engineering
     * V: visitior).
     * @return String
     */
    public String getRunMode() {
        String rc = ScriptingUtilities.nullToEmptyString(
                blk_.getObsRun().getObsMode());
        return rc;
    }

    /**
     * Get the name of the instrument
     * @return the name of the instrument used this OB
     */

    public String getInstrumentName() {
        String rc = blk_.getOd().getInstrument();
        return ScriptingUtilities.nullToEmptyString(rc);
    }

    /**
     * Get the target - returning null if this block is not
     * an ObservationBlock
     */

    private Target accessTarget() {
        Target rc = null;
        if (blk_ instanceof ObservationBlock) {
            rc = ((ObservationBlock)blk_).getTarget();
        }
        return rc;
    }

    /**
     * Get the name of the target
     * @return the name of the target of this OB
     */

    public String getTargetName() {
        String rc = null;
        Target t = accessTarget();
        if (t != null) {
            rc = t.getName();
        }
        return ScriptingUtilities.nullToEmptyString(rc);
    }

    /**
     * Get the target ra
     * @return the RA value for the target
     */

    public int getTargetRa() {
        int rc = 0;
        Target t = accessTarget();
        if (t != null) {
            rc = t.getRA();
        }
        return rc;
    }

    /**
     * Get the target dec
     * @return the Dec value for this target
     */

    public int getTargetDec() {
        int rc = 0;
        Target t = accessTarget();
        if (t != null) {
            rc = t.getDeclination();
        }
        return rc;
    }

    /**
     * Get the type of the OB
     * @return string containing type of this OB
     */

    public String getType() {
        return blk_.getType();
    }

    /**
     * Get the target for this block
     * @return the TargetAPI for this OB, or null if not target
     */

    public TargetAPI getTarget() {
        TargetAPI rc;
        Target t = accessTarget();
        if (t != null) {
            rc = new TargetAdaptor(t);
        } else {
            rc = null;
        }
        return rc;
    }

    /**
     * Get the Constraint Set for this OB
     * @return ConstraintSet for this OB
     */

    public ConstraintSetAPI getConstraintSet() {
        ConstraintSetAPI rc;
        if (blk_ instanceof ObservationBlock) {
            rc = new ConstraintSetAdaptor(
                    ((ObservationBlock)blk_).getConstraintSet());
        } else {
            rc = null;
        }
        return rc;
    }

    /**
     * Get the time intervals for this OB
     * @return TimeIntervalSetAPI object for this OB or null if not found
     */

    public TimeIntervalSetAPI getTimeIntervals() {
        TimeIntervalSetAPI rc;
        if ((blk_ instanceof ObservationBlock) &&
                ((ObservationBlock)blk_).getTimeIntervals() != null) {
            rc = new TimeIntervalSetAdaptor(
                    ((ObservationBlock)blk_).getTimeIntervals());
        } else {
            rc = null;
        }
        return rc;
    }

    /**
     * Get the time intervals for this OB
     * @return TimeIntervalSetAPI object for this OB or null if not found
     */

    public TimeIntervalSetAPI getSTTimeIntervals() {
        TimeIntervalSetAPI rc;
        if ((blk_ instanceof ObservationBlock) &&
                ((ObservationBlock)blk_).getSTTimeIntervals() != null) {
            rc = new TimeIntervalSetAdaptor(
                    ((ObservationBlock)blk_).getSTTimeIntervals());
        } else {
            rc = null;
        }
        return rc;
    }


    /**
     * Get the template that is the acquisition template for this OB
     * This is equivalent to getTemplateAt(0)
     * @return Acquisition Template for this OB, or null if none in OB
     */

    public TemplateAPI getAcquisitionTemplate() {
        return getTemplateAt(0);
    }

    /**
     * @return the Exposure time in seconds. Or 0 if not known.
     */
    public int getExposureTime() {
        return blk_.getExposureTime();
    }

    /**
     * @return the Execution time, in seconds, or 0 if not known.
     */
    public int getExecutionTime() {
        return blk_.getExecutionTime();
    }

    /**
     *
     * @see org.eso.ohs.scripting.ObservationBlockAPI#setExecutionTime(int)
     */
    public void setExecutionTime(int execTime) {
         blk_.setExecutionTime(execTime);
    }

    /**
     *
     * @see org.eso.ohs.scripting.ObservationBlockAPI#setExposureTime(int)
     */

    public void setExposureTime(int expTime) {
         blk_.setExposureTime(expTime);
    }

    /* (non-Javadoc)
     * @see org.eso.ohs.scripting.ObservationBlockAPI#getNumberOfSiderealTimes()
     */
    public int getNumberOfSiderealTimeIntervals() {
        // TODO Auto-generated method stub
        int rc = 0;
        if(blk_ instanceof ObservationBlock) {
           TimeInterval [] tis = ((ObservationBlock)blk_).getSTTimeIntervals();
           if (tis != null) {
              rc = tis.length;
           }
        }
        return rc;
    }

    /**
     * Implementation of the getInstrumentComments method
     */

	public String getInstrumentComments() {
		String instComments = blk_.getInstrumentComments();
		if(instComments == null)
			return "";
		return instComments;

	}

	/* (non-Javadoc)
	 * @see org.eso.ohs.scripting.ObservationBlockAPI#getUserComments()
	 */
	public String getUserComments() {
		String userComments = blk_.getUserComments();
		if(userComments == null)
			return "";
		return userComments;
	}

	/* (non-Javadoc)
	 * @see org.eso.ohs.scripting.ObservationBlockAPI#getFindingCharts()
	 */
	public FindingChartSetAPI getFindingCharts() {
		FindingChartSetAPI rc;
		if (blk_ instanceof ObservationBlock) {
			rc = new FindingChartSetAdaptor(((ObservationBlock) blk_)
					.getFindingCharts());
		} else {
			rc = null;
		}

		return (rc);
	}



}

