package org.eso.ohs.scripting;

/**
 * Implementation of EVMImpl that acts as the implementation object for
 * EVMExecutor objects. This class is used when the scripting package
 * is available and can be initialized. If initialization fails, this
 * class is replaced by the DummyEVMImpl as the implementing class in 
 * the EVMExecutor object.
 *
 * See EVMImpl and EVMExecutor for a description of how they form the
 * Bridge pattern.
 *
 * This class extends the ScriptingBox abstract class.
 *
 * Note: the timeout is not currently supported.
 *
 * @author Nick Kornweibel
 */

import java.io.File;
import java.util.Enumeration;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.ohs.dfs.CalibrationBlock;
import org.eso.ohs.evm.EVMException;
import org.eso.ohs.evm.EVMImpl;
import org.eso.ohs.evm.EVMMessageListener;
import org.eso.ohs.evm.EVMScriptFinder;

public class ScriptingEVMImpl extends ScriptingBox implements EVMImpl {

    /**
     * Logging instance.
     */
    protected static Logger stdlog_ =
        Logger.getLogger(ScriptingEVMImpl.class);

    /**
     * Set of OBs that are _currently_ exposed to the scripting env
     */
    private ObservationBlockAPI[] testObs_ = null;

    /**
     * The object that locates the scripts for each OB
     */
    private EVMScriptFinder scriptFinder_ = null;

    /**
     * Set of registered EVM message listeners
     */
    private Vector<EVMMessageListener> messageListeners_ = new Vector<EVMMessageListener>();

    /**
     * How long scripts should be allowed to run.
     * NOT IMPLEMENTED.
     */
    private long scriptTimeout_ = 0L;

    /**
     * Default constructor - this constructor must (and is implicitly)
     * call the default ScriptingBox constructor in order to set the
     * instance of ScriptingBox available to the scripting environment
     */

    public ScriptingEVMImpl() {
        setOBsForTesting(null, -1, -1);
    }

    /**
     * Initialize the object with the defined scripting drivers and
     * defined script finder object
     *
     * @param scriptDriver the list of supported ScriptInterpreter
     * classes to be loaded
     * @param script finder object
     * @throws EVMException if initialization fails
     */

    public void initImpl(String[] scriptDrivers, EVMScriptFinder sf)
        throws EVMException {

        if (scriptDrivers == null || scriptDrivers.length == 0 || sf == null) {
            throw new EVMException("Empty or null parameters");
        }
        scriptFinder_ = sf;

        ScriptInterpreterManager man = ScriptInterpreterManager.getInstance();
        try {
            for (int i = 0; i < scriptDrivers.length; i++) {
                if (!man.addScriptInterpreter(scriptDrivers[i])) {
                    throw new EVMException(
                        "Script driver not found: " + scriptDrivers[i]);
                }
            }
        } catch (Exception ex) {
            throw new EVMException(ex.getMessage());
        }
    }

    /**
     * Set OBs to be tested/verified (from start to end, not incl)
     * @param array of OBs
     * @param start index of OBs to be tested
     * @param end index of OBs to be tested
     */

    private void setOBsForTesting(CalibrationBlock[] obs, int start, int end) {
        if (obs == null) {
            testObs_ = new ObservationBlockAPI[0];
        } else {
            testObs_ = new ObservationBlockAPI[end - start];
            for (int i = start; i < end; i++) {
                testObs_[i - start] = new ObservationBlockAdaptor(obs[i]);
            }
        }
    }

    /**
     * Verify the OBs given. Loop through each OB, finding the scripts
     * to execute. "Expose" the OB to the ScriptingBox environment, and
     * verify the OB.
     */

    public void verifyImpl(CalibrationBlock[] obs) throws EVMException {
        for (int i = 0; i < obs.length; i++) {
            Vector<File> scriptList = scriptFinder_.findScriptsEVM(obs[i]);
            runScripts(obs, i, scriptList);
        }
    }

    public void runScripts(CalibrationBlock[] obs, int i, Vector<File> scriptList)
        throws EVMException {
        if (scriptList != null && scriptList.size() > 0) {
            stdlog_.debug("EVM: Performing EVM tests.");
            setOBsForTesting(obs, i, i + 1);
            runScript(scriptList);
        } else {
            stdlog_.debug(
                "EVM: Initialized but no "
                    + "EVM scripts located for this instrument.");
        }
    }

    public void runExecScripts(
        CalibrationBlock[] obs,
        int i,
        Vector<File> scriptList)
        throws EVMException {
        if (scriptList != null && scriptList.size() > 0) {
            stdlog_.debug("EVM: Performing EVM tests.");
            setOBsForTesting(obs, i, i + 1);
            runScript(scriptList);
        }
    }

    /**
     * Impelmentation of ETC calculations.
     * @see org.eso.ohs.evm.EVMImpl#calculateETCSImpl(CalibrationBlock[])
     */
    public void calculateETCSImpl(CalibrationBlock[] obs) throws EVMException {

stdlog_.debug(">>>>>>>>>IN ETCS IMPLEMENTATION");

        for (int i = 0; i < obs.length; i++) {
            Vector<File> scriptList = scriptFinder_.findScriptsExecTime(obs[i]);
            runExecScripts(obs, i, scriptList);
        }
    }

    /**
     * For each script, load the driver matching the file extension
     * of the script and execute the script
     * @param Vector of File objects pointing to the scripts
     */

    private void runScript(Vector<File> scriptList) throws EVMException {
        for (int i = 0; i < scriptList.size(); i++) {
            Object script = scriptList.elementAt(i);
            if (script instanceof File) {
                runScriptImpl((File) script);
            } else {
                throw new EVMException("Non-File object returned by EVMScriptFinder");
            }
        }
    }

    /**
     * Set the script timeout, 0L for no timeout
     * TIMEOUTS NOT IMPLEMENTED.
     * @param script timeout time in milliseconds
     */

    public void setScriptTimeoutImpl(long millis) {
        scriptTimeout_ = millis;
    }

    /**
     * Get the script timeout time 
     * TIMEOUTS NOT IMPLEMENTED.
     * @return the timeout value in milliseconds
     */

    public long getScriptTimeoutImpl() {
        return scriptTimeout_;
    }

    /**
     * Return the set of supported scripting file extensions
     * @return set of supported script file extensions
     */

    public String[] getSupportedScriptExtensionsImpl() {
        return ScriptInterpreterManager.getInstance().getFileExtensions();
    }

    /**
     * Run a script
     * @param the script File to be run
     * @throws if an error occurs executing the script
     */

    public void runScriptImpl(File script) throws EVMException {
        ScriptInterpreterManager man = ScriptInterpreterManager.getInstance();
        // extract the file extension
        String fileName = script.getName();
        stdlog_.debug("Filename " + fileName);
        int extDot = fileName.lastIndexOf('.');
        String ext = null;
        if (extDot < 0) {
            // force an index oob exception below
            extDot = fileName.length();
        }
        try {
            ext = fileName.substring(extDot + 1);
        } catch (StringIndexOutOfBoundsException ex) {
            throw new EVMException(
                "No script driver found for: " + script.getPath());
        }
        ScriptInterpreter si = man.selectInterpreterByFileExt(ext);
        if (si == null) {
            throw new EVMException(
                "No script driver found for: " + script.getPath());
        } else {
            try {
                stdlog_.debug("Executing script");
                si.executeFile(script);
            } catch (ScriptInterpreterException siex) {
                throw new EVMException(siex.getMessage());
            }
        }
    }

    /**
     * Add an EVMMessageListener to the set of listeners
     * @param the EVMessageListener object to be added
     */

    public void addEVMMessageListenerImpl(EVMMessageListener m) {
        if (m != null) {
            messageListeners_.addElement(m);
        }
    }

    /**
     * Remove an EVMMessageListener from the set of listeners
     * @param the EVMessageListener object to be removed
     */

    public void removeEVMMessageListenerImpl(EVMMessageListener m) {
        if (m != null) {
            messageListeners_.removeElement(m);
        }
    }

    /**
     * Required by ScriptingBox
     * Get the number of Observation Blocks exposed to the scripting
     * environment. This method must be implemented.
     * @return array of ObservationBlockAPI objects
     */

    public int getNumberOfObservationBlocks() {
        return testObs_.length;
    }

    /**
     * Required by ScriptingBox
     * Get the ObservationBlock at the specified index (>=0)
     * This method must be implemented.
     * @param the index of the OB
     * @return the ObservationBlockAPI at that index, or null
     */

    public ObservationBlockAPI getObservationBlockAt(int i) {
        return testObs_[i];
    }

    /**
     * Forward info message to listeners
     * @param String message
     */

    public void infoMessage(String message) {
        // call super for debugging.
        //super.infoMessage(message);
        for (Enumeration<EVMMessageListener> e = messageListeners_.elements();
            e.hasMoreElements();
            ) {
            EVMMessageListener l = e.nextElement();
            l.infoMessage(message);
        }

    }

    /**
     * Forward warning message to listeners
     * @param String message
     */

    public void warningMessage(String message) {
        // call super for debugging.
        //super.warningMessage(message);
        for (Enumeration<EVMMessageListener> e = messageListeners_.elements();
            e.hasMoreElements();
            ) {
            EVMMessageListener l = e.nextElement();
            l.warningMessage(message);
        }
    }

    /**
     * Forward error message to listeners
     * @param String message
     */

    public void errorMessage(String message) {
        // call super for debugging.
        //super.errorMessage(message);
        for (Enumeration<EVMMessageListener> e = messageListeners_.elements();
            e.hasMoreElements();
            ) {
            EVMMessageListener l = e.nextElement();
            l.errorMessage(message);
        }
    }

    /**
     * Print in the report header.
     * @param message to be printed in the report header.
     */

    public void headerMessage(String message) {
        // call super for debugging.
        //super.headerMessage(message);
        for (Enumeration<EVMMessageListener> e = messageListeners_.elements();
            e.hasMoreElements();
            ) {
            EVMMessageListener l = e.nextElement();
            l.headerMessage(message);
        }
    }

    /**
     * Print in the report footer.
     * @param message to be printed in the report footer.
     */

    public void footerMessage(String message) {
        // call super for debugging.
        //super.footerMessage(message);
        for (Enumeration<EVMMessageListener> e = messageListeners_.elements();
            e.hasMoreElements();
            ) {
            EVMMessageListener l = e.nextElement();
            l.footerMessage(message);
        }
    }

    /**
     * Forward clear warnings event to listeners
     */

    public void clearWarnings() {
        for (Enumeration<EVMMessageListener> e = messageListeners_.elements();
            e.hasMoreElements();
            ) {
            EVMMessageListener l = e.nextElement();
            l.warningsWereCleared();
        }
    }

    /**
     * Forward clear errors event to listeners
     */

    public void clearErrors() {
        for (Enumeration<EVMMessageListener> e = messageListeners_.elements();
            e.hasMoreElements();
            ) {
            EVMMessageListener l = e.nextElement();
            l.errorsWereCleared();
        }
    }
}
